/*
 * crypto/cryptoapi.c
 *
 * Written by Alexander Kjeldaas <astor@fast.no> 1998-11-15
 *
 * 2000-10-15 Harald Welte <laforge@gnumonks.org>
 * 		- ported to Linux 2.4 
 * 
 * This module is still using old 2.2 module_init/exit semantics for
 * backwards compatibility - HW
 *
 * Copyright 1998 by Alexander Kjeldaas. Redistribution of this file
 * is permitted under the GNU Public License.
 */

#include <linux/version.h>
#include <linux/module.h>

#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif 

#include <linux/init.h>
#include <linux/config.h>
#include <linux/string.h>
#include <linux/crypto.h>
#include <linux/proc_fs.h>


static struct proc_dir_entry *proc_crypto;

#ifdef CONFIG_PROC_FS
static int cipher_read_proc(char *page, char **start, off_t off,
			int count, int *eof, void *data);
static int digest_read_proc(char *page, char **start, off_t off,
			int count, int *eof, void *data);
#endif

LIST_HEAD(ciphers);
LIST_HEAD(digests);

static struct transform_group transforms[MAX_TRANSFORM] = {
	/* digest */
	{ TRANSFORM_DIGEST, "digest", SPIN_LOCK_UNLOCKED, &digests, 
#ifdef CONFIG_PROC_FS
	  NULL, &digest_read_proc 
#endif
	},
	/* cipher */
	{ TRANSFORM_CIPHER, "cipher", SPIN_LOCK_UNLOCKED, &ciphers, 
#ifdef CONFIG_PROC_FS
	  NULL, &cipher_read_proc 
#endif
	}
};

#ifdef CONFIG_PROC_FS
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,18)
/* We have to emulate the 2.4 function proc_mkdir, because it is not
 * present in 2.2 - HW */
static struct proc_dir_entry *proc_mkdir(const char *name, 
		struct proc_dir_entry *parent)
{
	struct proc_dir_entry *ent = NULL;
	const char *fn = name;
	int len;

	if (!parent)
		parent = &proc_root;
	len = strlen(fn);

	ent = kmalloc(sizeof(struct proc_dir_entry) + len + 1, GFP_KERNEL);
	if (!ent)
		goto out;
	memset(ent, 0, sizeof(struct proc_dir_entry));
	memcpy(((char *) ent) + sizeof(*ent), fn, len + 1);
	ent->name = ((char *) ent) + sizeof(*ent);
	ent->namelen = len;
	ent->ops = NULL;
	ent->nlink = 2;
	ent->mode = S_IFDIR | S_IRUGO | S_IXUGO;

	proc_register(parent, ent);

out:
	return ent;
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
static struct proc_dir_entry *create_proc_read_entry(const char *name,
	mode_t mode, struct proc_dir_entry *base, 
	read_proc_t *read_proc, void * data)
{
	struct proc_dir_entry *res=create_proc_entry(name,mode,base);
	if (res) {
		res->read_proc=read_proc;
		res->data=data;
	}
	return res;
}

#endif
#endif

/**
 * find_transform_by_id - Find transform implementation
 * @id: The identifier for the transform.
 * @tgroup: The identifier for the transform group the transform belongs to.
 *
 * Returns a ptr to the transform on success, NULL on failure.
 * Valid tgroup values are:
 *
 * %TRANSFORM_CIPHER - When looking for ciphers
 *
 * %TRANSFORM_DIGEST - When looking for digests
 *
 * You might want to use the wrapper-functions find_cipher_by_id(int
 * id), and find_digest_by_id(int id) instead of this one.
 */                                                                             
struct transform_implementation *
find_transform_by_id(int id, int tgroup)
{
	struct list_head *tmp;
	struct transform_group *tg;
#ifdef CONFIG_KMOD
	int mod_try = 0;
retry:
#endif
	if (tgroup >= MAX_TRANSFORM)
		return NULL;
	tg = &transforms[tgroup];
	read_lock(&tg->tg_lock);
	for (tmp = tg->tg_head->next; tmp != tg->tg_head; tmp = tmp->next) {
		struct transform_implementation *t;
		t = list_entry(tmp, struct transform_implementation, t_list);
		if (t->t_id == id) {
			read_unlock(&tg->tg_lock);
			return t;
		}
	}
	read_unlock(&tg->tg_lock);
#ifdef CONFIG_KMOD
	if (!mod_try) {
		char module_name[50];
		mod_try = 1;
		sprintf(module_name, "%s-%d", tg->tg_name, id & (~CIPHER_CBC));
		request_module(module_name);
		goto retry;
	}
#endif
	return NULL;
}


/**
 * find_transform_by_name - Find transform implementation
 * @name: The name of the transform.
 * @tgroup: The identifier for the transform group the transform belongs to.
 *
 * Returns a ptr to the transform on success, NULL on failure.
 * Valid tgroup values are:
 *
 * %TRANSFORM_CIPHER - When looking for ciphers
 *
 * %TRANSFORM_DIGEST - When looking for digests
 *
 * You might want to use the wrapper-functions
 * find_cipher_by_name(const char *name), and
 * find_digest_by_name(const char *name) instead of this one.
 */                                                                             
struct transform_implementation *
find_transform_by_name(const char *name, int tgroup)
{
	struct list_head *tmp;
	struct transform_group *tg;

	if (tgroup >= MAX_TRANSFORM)
		return NULL;
	tg = &transforms[tgroup];

	read_lock(&tg->tg_lock);
	for (tmp = tg->tg_head->next; tmp != tg->tg_head; tmp = tmp->next) {
		struct transform_implementation *t;
		t = list_entry(tmp, struct transform_implementation, t_list);
		if (strcmp(t->t_name, name) == 0) {
			read_unlock(&tg->tg_lock);
			return t;
		}
	}
	read_unlock(&tg->tg_lock);
	return NULL;
}

/**
 * register_transform - Register new transform.
 * @ti: Initialized transform implementation struct.
 * @tgroup: The identifier for the transform group the transform should belong to.
 *
 * Adds a transform from the crypto API. ti->t_group is set to point
 * to the correct transform group according to tgroup, the transform
 * is added to the group's transform-list, and a /proc files are
 * created if CONFIG_PROC_FS=y Returns 0 on success.  Valid tgroup
 * values are:
 *
 * %TRANSFORM_CIPHER - When adding ciphers
 *
 * %TRANSFORM_DIGEST - When adding digests
 *
 */                                                                             
int register_transform(struct transform_implementation *ti, int tgroup)
{
	int err = 0;
	struct transform_group *tg;
	
	if (tgroup >= MAX_TRANSFORM) {
		return -1;
	}
	INIT_LIST_HEAD(&ti->t_list);
	tg = ti->t_group = &transforms[tgroup];
	write_lock(&ti->t_group->tg_lock);
	list_add(&ti->t_list, ti->t_group->tg_head);
	write_unlock(&ti->t_group->tg_lock);
	if (!err) {
		MOD_INC_USE_COUNT;
		printk(KERN_INFO "cryptoapi: Registered %s (%d)\n", 
			ti->t_name, ti->t_id);
		ti->t_proc = create_proc_read_entry(ti->t_name,
						S_IFREG|S_IRUGO,
						tg->tg_proc_parent_dir,
						tg->read_proc, (void *)ti);
	}
	return err;
}

/**
 * unregister_transform - Unregister new transform.
 * @ti: Initialized transform implementation struct.
 *
 * Removes a transform from the crypto API.  Returns 0 on success,
 * non-zero on failure to remove /proc entry.
 *
 */                                                                             
int unregister_transform(struct transform_implementation *ti)
{
	int ret = 0;

	if (!list_empty(&ti->t_list)) {
		write_lock(&ti->t_group->tg_lock);
		list_del(&ti->t_list);
		write_unlock(&ti->t_group->tg_lock);
		ret = 0;
	}

#ifdef CONFIG_PROC_FS
	if (ti->t_proc) {
		ti->t_proc = NULL;
		remove_proc_entry(ti->t_name, ti->t_group->tg_proc_parent_dir);
	}
#endif

	if (!ret) 
		MOD_DEC_USE_COUNT;

	return ret;
}

#ifdef CONFIG_PROC_FS
static int cipher_read_proc(char *page, char **start, off_t off,
			    int count, int *eof, void *data)
{
	struct cipher_implementation *ci;
	int len = 0;
	
	ci = (struct cipher_implementation *)data;

	len = sprintf(page, "cipher_id:         %d\n"
		      "cipher_name:       %s\n"
		      "blocksize:         %d\n"
		      "keysize_mask:      0x%08x\n"
		      "ivsize:            %d\n"
		      "key_schedule_size: %d\n",
		      ci->trans.t_id, ci->trans.t_name, 
		      ci->blocksize, ci->key_size_mask,
		      ci->ivsize, ci->key_schedule_size);
	*eof=1;

	return len;
}

static int digest_read_proc(char *page, char **start, off_t off,
			int count, int *eof, void *data)
{
	struct digest_implementation *ci;
	int len = 0;
	
	ci = (struct digest_implementation *)data;

	len = sprintf(page, "digest_id:         %d\n"
		      "digest_name:       %s\n"
		      "blocksize:         %d\n"
		      "working_size:      %d\n",
		      ci->trans.t_id, ci->trans.t_name, 
		      ci->blocksize, ci->working_size);
	*eof=1;

	return len;
}
#endif


#ifdef MODULE
int __init init_module(void)
#else
int __init cryptoapi_init(void)
#endif
{
	int i;

#ifdef CONFIG_PROC_FS
	proc_crypto = proc_mkdir("crypto", NULL);

	for (i = 0; i<sizeof(transforms)/sizeof(struct transform_group); i++) {
		transforms[i].tg_proc_parent_dir = 
			proc_mkdir(transforms[i].tg_name, proc_crypto);
	}
#endif
	
#ifdef CONFIG_DIGEST_MD5
	init_md5();
#endif
#ifdef CONFIG_DIGEST_SHA1
        init_sha1();
#endif
#ifdef CONFIG_CIPHER_AES
	init_rijndael();
#endif
#ifdef CONFIG_CIPHER_TWOFISH
	init_twofish();
#endif
#ifdef CONFIG_CIPHER_SERPENT
	init_serpent();
#endif
#ifdef CONFIG_CIPHER_RC5
	init_rc5();
#endif
#ifdef CONFIG_CIPHER_RC6
	init_rc6();
#endif
#ifdef CONFIG_CIPHER_MARS
	init_mars();
#endif
#ifdef CONFIG_CIPHER_DFC
	init_dfc();
#endif
#ifdef CONFIG_CIPHER_BLOWFISH
	init_blowfish();
#endif
#ifdef CONFIG_CIPHER_IDEA
	init_idea();
#endif
#ifdef CONFIG_CIPHER_DES
	init_des();
#endif
#ifdef CONFIG_CIPHER_DES_EDE3
	init_des_ede3();
#endif

	return 0;
}

#ifdef MODULE
void cleanup_module(void) 
{
	int i;

#ifdef CONFIG_PROC_FS
	for (i = 0; i<sizeof(transforms)/sizeof(struct transform_group); i++) {
		struct proc_dir_entry *p = transforms[i].tg_proc_parent_dir;
		remove_proc_entry(transforms[i].tg_name, p);
	}
	remove_proc_entry("crypto", NULL);
#endif

}
#endif


EXPORT_SYMBOL(find_transform_by_id);
EXPORT_SYMBOL(find_transform_by_name);
EXPORT_SYMBOL(register_transform);
EXPORT_SYMBOL(unregister_transform);
