/*
 *    idea.c - C source code for IDEA block cipher.
 *      IDEA (International Data Encryption Algorithm), formerly known as
 *      IPES (Improved Proposed Encryption Standard).
 *      Algorithm developed by Xuejia Lai and James L. Massey, of ETH Zurich.
 *      This implementation modified and derived from original C code
 *      developed by Xuejia Lai.
 *      Zero-based indexing added, names changed from IPES to IDEA.
 *      CFB functions added.  Random number routines added.
 *
 *      Extensively optimized and restructured by Colin Plumb.
 *
 *      There are two adjustments that can be made to this code to
 *      speed it up.  Defaults may be used for PCs.  Only the -DIDEA32
 *      pays off significantly if selectively set or not set.
 *      Experiment to see what works best for your machine.
 *
 *      Multiplication: default is inline, -DAVOID_JUMPS uses a
 *              different version that does not do any conditional
 *              jumps (a few percent worse on a SPARC), while
 *              -DSMALL_CACHE takes it out of line to stay
 *              within a small on-chip code cache.
 *      Variables: normally, 16-bit variables are used, but some
 *              machines (notably RISCs) do not have 16-bit registers,
 *              so they do a great deal of masking.  -DIDEA32 uses "int"
 *              register variables and masks explicitly only where
 *              necessary.  On a SPARC, for example, this boosts
 *              performace by 30%.
 *
 *      The IDEA(tm) block cipher is covered by patents held by ETH and a
 *      Swiss company called Ascom-Tech AG.  The Swiss patent number is
 *      PCT/CH91/00117, the European patent number is EP 0 482 154 B1, and
 *      the U.S. patent number is US005214703.  IDEA(tm) is a trademark of
 *      Ascom-Tech AG.  There is no license fee required for noncommercial
 *      use.  Commercial users may obtain licensing details from Dieter
 *      Profos, Ascom Tech AG, Solothurn Lab, Postfach 151, 4502 Solothurn,
 *      Switzerland, Tel +41 65 242885, Fax +41 65 235761.
 *
 *      The IDEA block cipher uses a 64-bit block size, and a 128-bit key
 *      size.  It breaks the 64-bit cipher block into four 16-bit words
 *      because all of the primitive inner operations are done with 16-bit
 *      arithmetic.  It likewise breaks the 128-bit cipher key into eight
 *      16-bit words.
 *
 *      For further information on the IDEA cipher, see the book:
 *        Xuejia Lai, "On the Design and Security of Block Ciphers",
 *        ETH Series on Information Processing (ed. J.L. Massey) Vol 1,
 *        Hartung-Gorre Verlag, Konstanz, Switzerland, 1992.  ISBN
 *        3-89191-573-X.
 *
 *      This code runs on arrays of bytes by taking pairs in big-endian
 *      order to make the 16-bit words that IDEA uses internally.  This
 *      produces the same result regardless of the byte order of the
 *      native CPU.
 */

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <asm/byteorder.h>
#include <linux/crypto.h>
#include <linux/string.h>

#define IDEAKEYSIZE 16
#define IDEABLOCKSIZE 8

#define IDEAROUNDS 8
#define IDEAKEYLEN (6*IDEAROUNDS+4)

#define low16(x) ((x) & 0xFFFF)
typedef unsigned char byte;

typedef struct {
    u16 ek[IDEAKEYLEN];
    u16 dk[IDEAKEYLEN];
} IDEA_context;

#if defined(__mc68000__) || defined (__sparc__) || defined (__PPC__) \
    || (defined(__mips__) && (defined(MIPSEB) || defined (__MIPSEB__)) )
#define BIG_ENDIAN_HOST 1
#else
#define LITTLE_ENDIAN_HOST 1
#endif

/* 
 * Compute the multiplicative inverse of x, modulo 65537, using
 * Euclid's algorithm. It is unrolled twice to avoid swapping the
 * registers each iteration, and some subtracts of t have been changed
 * to adds.
 */
static u16 mulInv(u16 x)
{
    u16 t0, t1;
    u16 q, y;

    if (x <= 1)
        return x;               /* 0 and 1 are self-inverse */
    t1 = 0x10001L / x;          /* Since x >= 2, this fits into 16 bits */
    y = 0x10001L % x;
    if (y == 1)
        return low16(1 - t1);
    t0 = 1;
    do {
        q = x / y;
        x = x % y;
        t0 += q * t1;
        if (x == 1)
            return t0;
        q = y / x;
        y = y % x;
        t1 += q * t0;
    } while (y != 1);
    return low16(1 - t1);
}

/*
 * Expand a 128-bit user key to a working encryption key ek
 */
static void ideaExpandKey(byte const *userkey, u16 * ek)
{
    int i, j;

    for (j = 0; j < 8; j++) {
        ek[j] = (userkey[0] << 8) + userkey[1];
        userkey += 2;
    }
    for (i = 0; j < IDEAKEYLEN; j++) {
        i++;
        ek[i + 7] = ek[i & 7] << 9 | ek[(i + 1) & 7] >> 7;
        ek += i & 8;
        i &= 7;
    }
}

/* 
 * Compute IDEA decryption key dk from an expanded IDEA encryption key
 * ek Note that the input and output may be the same.  Thus, the key
 * is inverted into an internal buffer, and then copied to the output.
 */
static void ideaInvertKey(u16 const *ek, u16 dk[IDEAKEYLEN])
{
    int i;
    u16 t1, t2, t3;
    u16 temp[IDEAKEYLEN];
    u16 *p = temp + IDEAKEYLEN;

    t1 = mulInv(*ek++);
    t2 = -*ek++;
    t3 = -*ek++;
    *--p = mulInv(*ek++);
    *--p = t3;
    *--p = t2;
    *--p = t1;

    for (i = 0; i < IDEAROUNDS - 1; i++) {
        t1 = *ek++;
        *--p = *ek++;
        *--p = t1;

        t1 = mulInv(*ek++);
        t2 = -*ek++;
        t3 = -*ek++;
        *--p = mulInv(*ek++);
        *--p = t2;
        *--p = t3;
        *--p = t1;
    }
    t1 = *ek++;
    *--p = *ek++;
    *--p = t1;

    t1 = mulInv(*ek++);
    t2 = -*ek++;
    t3 = -*ek++;
    *--p = mulInv(*ek++);
    *--p = t3;
    *--p = t2;
    *--p = t1;
	/* Copy and destroy temp copy */
    memcpy(dk, temp, sizeof(temp));
	memset(&temp,0,sizeof(temp));
}

/* IDEA encryption/decryption algorithm */
/* Note that in and out can be the same buffer */
static void ideaCipher(u16 const inbuf[4], u16 outbuf[4],
                       u16 const *key)
{
    register u16 x1, x2, x3, x4, s2, s3;
    u16 *in, *out;
    register u16 t16;        /* Temporaries needed by MUL macro */
    register u32 t32;
    int r = IDEAROUNDS;
#define MUL(x,y) \
        ((t16 = (y)) ? \
                (x=low16(x)) ? \
                        t32 = (u32)x*t16, \
                        x = low16(t32), \
                        t16 = t32>>16, \
                        x = (x-t16)+(x<t16) \
                : \
                        (x = 1-t16) \
        : \
                (x = 1-x))

    in = (u16 *) inbuf;
    x1 = *in++;
    x2 = *in++;
    x3 = *in++;
    x4 = *in;
#ifndef LITTLE_ENDIAN_HOST
    x1 = (x1 >> 8) | (x1 << 8);
    x2 = (x2 >> 8) | (x2 << 8);
    x3 = (x3 >> 8) | (x3 << 8);
    x4 = (x4 >> 8) | (x4 << 8);
#endif
    do {
        MUL(x1, *key++);
        x2 += *key++;
        x3 += *key++;
        MUL(x4, *key++);

        s3 = x3;
        x3 ^= x1;
        MUL(x3, *key++);
        s2 = x2;
        x2 ^= x4;
        x2 += x3;
        MUL(x2, *key++);
        x3 += x2;

        x1 ^= x2;
        x4 ^= x3;

        x2 ^= s3;
        x3 ^= s2;
    } while (--r);
    MUL(x1, *key++);
    x3 += *key++;
    x2 += *key++;
    MUL(x4, *key);

    out = (u16 *) outbuf;
#ifdef LITTLE_ENDIAN_HOST
    *out++ = x1;
    *out++ = x3;
    *out++ = x2;
    *out = x4;
#else
    x1 = low16(x1);
    x2 = low16(x2);
    x3 = low16(x3);
    x4 = low16(x4);
    *out++ = (x1 >> 8) | (x1 << 8);
    *out++ = (x3 >> 8) | (x3 << 8);
    *out++ = (x2 >> 8) | (x2 << 8);
    *out = (x4 >> 8) | (x4 << 8);
#endif
}

int idea_encrypt(struct cipher_context *cx,
		 const u8 *in, u8 *out, int size)
{
	IDEA_context *c=(IDEA_context *)cx->keyinfo;

	u16 input_buffer[4];
	u16 output_buffer[4];
	u32 *in_blk = (u32 *)in;
	u32 *out_blk = (u32 *)out;
	u32 n;
	
	for (; size >= 8; size -= 8)
    {
		n=*in_blk++;
		input_buffer[0]=n & 0xffff;
		n=n>>16;
		input_buffer[1]=n & 0xffff;

		n=*in_blk++;
		input_buffer[2]=n & 0xffff;
		n=n>>16;
		input_buffer[3]=n & 0xffff;

		ideaCipher( input_buffer,output_buffer, c->ek );

		n=0;
		n|=output_buffer[1];
		n=n<<16;
		n|=output_buffer[0];

		*out_blk++=n;

		n=0;
		n|=output_buffer[3];
		n=n<<16;
		n|=output_buffer[2];

		*out_blk++=n;
	}
	return 0;
}

int idea_decrypt(struct cipher_context *cx, 
		 const u8 *in, u8 *out, int size)
{
	IDEA_context *c=(IDEA_context *)cx->keyinfo;

	u16 input_buffer[8];
	u16 output_buffer[8];
	u32 *in_blk = (u32 *)in;
	u32 *out_blk = (u32 *)out;
	u32 n;

	for (; size >= 8; size -= 8)
    {
		n=*in_blk++;
		input_buffer[0]=n & 0xffff;
		n=n>>16;
		input_buffer[1]=n & 0xffff;

		n=*in_blk++;
		input_buffer[2]=n & 0xffff;
		n=n>>16;
		input_buffer[3]=n & 0xffff;

		ideaCipher( input_buffer,output_buffer, c->dk );

		n=0;
		n|=output_buffer[1];
		n=n<<16;
		n|=output_buffer[0];

		*out_blk++=n;

		n=0;
		n|=output_buffer[3];
		n=n<<16;
		n|=output_buffer[2];

		*out_blk++=n;
	}
	return 0;
}

int idea_set_key(struct cipher_context *cx, 
		 unsigned char *key, int keybytes)
{
	IDEA_context *c=(IDEA_context *)cx->keyinfo;

	if(keybytes != 16)
		return -EINVAL;

    ideaExpandKey(key, c->ek);
    ideaInvertKey(c->ek, c->dk);
    return 0;
}

static void idea_lock(void)
{
	MOD_INC_USE_COUNT;
}

static void idea_unlock(void)
{
	MOD_DEC_USE_COUNT;
}   

#define CIPHER_BITS_64
#define CIPHER_NAME(x) idea##x
#include "gen-cbc.h"
#include "gen-ecb.h"

static struct cipher_implementation idea = {
	{{NULL,NULL},CIPHER_IDEA, "idea"},
	blocksize: 8,
	ivsize: 0,
	key_schedule_size: IDEA_KEY_SCHEDULE_SIZE,
	key_size_mask: CIPHER_KEYSIZE_128,
	INIT_CIPHER_BLKOPS(idea_ecb),
	INIT_CIPHER_OPS(idea)
};

static struct cipher_implementation idea_cbc = {
	{{NULL,NULL},CIPHER_IDEA | CIPHER_CBC, "idea-cbc"},
	blocksize: 8,
	ivsize: 8,
	key_schedule_size: IDEA_KEY_SCHEDULE_SIZE,
	key_size_mask: CIPHER_KEYSIZE_128,
	INIT_CIPHER_BLKOPS(idea_cbc),
	INIT_CIPHER_OPS(idea)
};

#ifdef MODULE
int __init init_module(void)
#else
int __init init_idea(void)
#endif
{
    if (register_cipher(&idea))
        printk(KERN_WARNING "Couldn't register idea encryption\n");
    if (register_cipher(&idea_cbc))
        printk(KERN_WARNING "Couldn't register idea-cbc encryption\n");
    return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
	if (unregister_cipher(&idea))
		printk(KERN_WARNING "Couldn't unregister idea encryption\n");
	if (unregister_cipher(&idea_cbc))
		printk(KERN_WARNING "Couldn't unregister idea-cbc encryption\n");
}
#endif
