/*
 * crypto/testing/speed.c
 *
 * Written by Alexander Kjeldaas <astor@guardian.no> 1998-11-23
 *
 * Copyright 1998 by Alexander Kjeldaas. Redistribution of this file
 * is permitted under the GNU Public License.
 */

#define __KERNEL__

#define printk printk_dummy

#include <linux/crypto.h>

#undef memcpy
#undef memset
#undef printk

#include "libcfuncs.h"
#include "misc.h"


int cipher_correct(struct cipher_implementation *ci);
int test_cipher(struct cipher_implementation *ci, int encrypt);
int test_cbc_cipher(struct cipher_implementation *ci, int encrypt);

int main()
{
	int cipher;
	/* Register all ciphers */
	cryptoapi_init();
	
	for (cipher = 1; cipher < MAX_CIPHER; cipher++) {
		struct cipher_implementation *ci;
		struct timeval tv1,tv2;
		int modes[] = {0, CIPHER_CBC};
		int test[] = {1, 0};
		int i;
		for (i = 0; i < sizeof(modes)/sizeof(int); i++) {
			ci = find_cipher_by_id(cipher | modes[i]);
			if (!ci)
				continue;
			if (test[i]) {
				PRINTF("Testing cipher %s, number %d..", 
				       ci->trans.t_name, ci->trans.t_id);
				if (!cipher_correct(ci)) {
					PRINTF("seems to work\n");
				} else {
					PRINTF("doesnt work!\n");
				}
			}
			gettimeofday(&tv1, NULL);
			if (test_cipher(ci, 1))
				return 1;
			gettimeofday(&tv2, NULL);
			PRINTF("encrypt %-15s = %8d usec/MB; %6.3f MB/s; "
			       "%6.3f Mb/s\n", 
			       ci->trans.t_name, (int)timediff(&tv2,&tv1), 
			       1000000.0/timediff(&tv2,&tv1), 
			       1000000.0*8.0/timediff(&tv2,&tv1));
			gettimeofday(&tv1, NULL);
			if (test_cipher(ci,0))
				return 1;
			gettimeofday(&tv2, NULL);
			PRINTF("decrypt %-15s = %8u usec/MB; %6.3f MB/s; "
			       "%6.3f Mb/s\n", 
			       ci->trans.t_name, (int)timediff(&tv2,&tv1), 
			       1000000.0/timediff(&tv2,&tv1), 
			       1000000.0*8.0/timediff(&tv2,&tv1));
		}
	}
	exit(0);
}

#if 0
static inline void rdtsc(long long *x)
{
   asm volatile (".byte 0x0f, 0x31\n\t"
		 "movl %%eax, %0\n\t"
		 "movl %%edx, %1"
		 : "=m" (((long *)x)[0]), "=m"(((long *)x)[1])
		 ::"ax","dx");
}
#endif

int test_cipher(struct cipher_implementation *ci, int encrypt)
{
	int fd,i,j,bs,size;
	char buf[512], buf_in[512], buf_out[512];
        static u8 key[] = {0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x80,
                           0x20, 0x02, 0x02, 0x40, 0x40, 0x04, 0x04, 0x20};
        struct cipher_context *cx;
	cx = (struct cipher_context *) malloc(sizeof(struct cipher_context));
	cx->ci = ci;
	cx->keyinfo = (u32 *) malloc(ci->key_schedule_size);
	errno=ci->set_key(cx, (char *)key, sizeof(key));
	if(errno) {
		perror("Can't set key");
		return 1;
	}
	fd = open("speed.c", O_RDONLY);
	if (!fd) {
		perror("Couldn't open speed.c");
		return 1;
	}
	if ((size = read(fd, buf, sizeof(buf))) < sizeof(buf)) {
		perror("Couldn't read 512 bytes from speed.c");
		return 1;
	}
	close(fd);
	memcpy(buf_in, buf, sizeof(buf));
	bs = ci->blocksize;
	if (bs == 0) {
		bs = 512;
	}
	for (i = 0; i < 1024; i++) {
		char *in =  &buf_in[0];
		char *out =  &buf_out[0];
		int (*func)(struct cipher_context *cx,
			    const u8 *in, u8 *out, int size);
		if (encrypt)
			func = ci->encrypt;
		else
			func = ci->decrypt;

		for (j = 512; j > 0; j-=bs) {
			errno=func(cx, (u8 *)in, (u8 *)out, bs);
			if(errno) {
				perror("Can't encrypt/decrypt in test_cipher()");
				return 1;
			}
			in += bs; out += bs;
		}
		in =  &buf_in[0]; 
		out = &buf_out[0];
		for (j = 512; j > 0; j-=bs) {
			func(cx, (u8 *)out, (u8 *)in, bs);
			in += bs; out += bs;
		}
	}
	return 0;
}

int cipher_correct(struct cipher_implementation *ci)
{
	int fd,j,bs,size,err;
	char buf[512], buf_enc[512],buf_dec[512];
	static u8 key[] = {0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x41, 0x80,
                           0x40, 0x60, 0x40, 0x40, 0x40, 0x04, 0x40, 0x00};  
	struct cipher_context *cx;
	cx = (struct cipher_context *) malloc(sizeof(struct cipher_context));
	cx->ci = ci;
	cx->keyinfo = (u32 *) malloc(ci->key_schedule_size);
	errno=ci->set_key(cx, (char *)key, sizeof(key));
	if(errno) {
		printf("errno=%d key=%d\n",errno,sizeof(key));
		perror("Can't set key");
		return 1;
	}
	fd = open("speed.c", O_RDONLY);
	if (!fd) {
		perror("Couldn't open speed.c");
		return 1;
	}
	if ((size = read(fd, buf, sizeof(buf))) < sizeof(buf)) {
		perror("Couldn't read 512 bytes from speed.c");
		return 1;
	}
	close(fd);
	bs = ci->blocksize;
	if (bs == 0) {
		return 0;
	}
	{
		char *in =  &buf[0];
		char *out =  &buf_enc[0];
		for (j = 512; j > 0; j-=bs) {
			errno=ci->encrypt(cx, (u8 *)in, (u8 *)out, bs);
			if(errno) {
				perror("Can't encryp in cipher_correct");
				return 1;
			}

			in += bs; out += bs;
		}
		in =  &buf_enc[0]; 
		out = &buf_dec[0];
		for (j = 512; j > 0; j-=bs) {
			errno=ci->decrypt(cx, (u8 *)in, (u8 *)out, bs);
			if(errno) {
				perror("Can't encrypt in cipher_correct");
				return 1;
			}

			in += bs; out += bs;
		}
		if ((err = my_memcmp(buf, buf_dec, sizeof(buf)))) {
			PRINTF("failed ecb check");
			PRINTF("\nPlaintext buffer\n");
			dump_u8_buffer(buf, 512);
			PRINTF("\nEncrypted plaintext\n");
			dump_u8_buffer(buf_enc, 512);
			PRINTF("\nDecrypted buffer\n");
			dump_u8_buffer(buf_dec, 512);
			exit(err);
		}
	}
	{
		struct cipher_implementation *ci_cbc;
		ci_cbc = find_cipher_by_id(ci->trans.t_id | CIPHER_CBC);
		if (!ci_cbc)
			return 0; /* not supporting cbc is ok */
		memset(cx->iv, 0, sizeof(cx->iv));
		ci_cbc->encrypt(cx, (u8 *)buf, (u8 *)buf_enc, 512);
		ci_cbc->decrypt(cx, (u8 *)buf_enc, (u8 *)buf_dec, 512);
		if ((err = my_memcmp(buf, buf_dec, sizeof(buf)))) {
			PRINTF("failed cbc check");
			PRINTF("\nPlaintext buffer\n");
			dump_u8_buffer(buf, 512);
			PRINTF("\nEncrypted plaintext\n");
			dump_u8_buffer(buf_enc, 512);
			PRINTF("\nDecrypted buffer\n");
			dump_u8_buffer(buf_dec, 512);
			exit(err);
		}
	}
	return 0;
}

