/*
 * crypto/testing/testapi.c
 *
 * Written by Alexander Kjeldaas <astor@guardian.no> 2000-10-08
 *
 * Copyright 2000 by Alexander Kjeldaas. Redistribution of this file
 * is permitted under the GNU Public License.
 */

/* Regression tests for the kernel crypto API */

#define __KERNEL__

#define printk printk_dummy

#include <linux/crypto.h>
#undef memcpy
#undef memset
#undef printk

#include "libcfuncs.h"
#include "misc.h"

u8 plaintext[32+64];
u8 ciphertext[32+64];
u8 testres[512];
u8 key[512];
int plaintextlen, ciphertextlen, keylen;

static int test_fence(struct cipher_context *cx);
static void show_usage();


int main(int argc, char **argv)
{
	const char *cipher = "aes-cbc";
	struct cipher_implementation *ci = NULL;
	struct cipher_context *cx;
	int arg, err;
	int i;

	/* Register all ciphers */
	cryptoapi_init();

	memset(plaintext, 0, sizeof(plaintext));
	memset(ciphertext, 0, sizeof(ciphertext));
	memset(testres, 0, sizeof(testres));
	memset(key, 0, sizeof(key));

	while((arg = getopt(argc,argv,"hc:k:")) != EOF) {
		switch(arg) {
		case 'h': 
			show_usage(); 
			break;
		case 'c': 
			cipher = strdup(optarg);
			if (!(ci = find_cipher_by_name(cipher))) {
				PRINTF("no cipher by that name (%s)\n", optarg);
				exit(1);
			}
			break;
		}
	}
	if (!ci && !(ci = find_cipher_by_name(cipher))) {
		PRINTF("no cipher by that name (%s)\n", cipher);
		exit(1);
	}
	cx = (struct cipher_context *) malloc(sizeof(struct cipher_context));
	cx->ci = ci;
	cx->keyinfo = (u32 *) malloc(ci->key_schedule_size);
	make_random_buf(key, sizeof(key));
	if ((err = ci->set_key(cx, (char *)key, 128/8))) {
		PRINTF("set_key returned error %d\n", err);
		exit(1);
	}
	err = 0;
	for (i = 0; i < 100; i++) {
		err = err | test_fence(cx);
	}
	PRINTF("[%s] Fencepost tests\n", err ? "FAIL" : " OK ");
	return 0;
}

static int test_fence(struct cipher_context *cx) 
{
	/* Test that a cipher fills whatever buffer we ask it to fill,
	   but that it doesn't touch anything else. We have a fence
	   consisting of 16 bytes at the beginning of the output
	   buffer, and 16 bytes at the end of the output buffer */
	int enclen = rand() % (sizeof(ciphertext) - 32);
	u8 fence[] = {0, 0xff};
	int i, j, err=0;
	make_random_buf(ciphertext, sizeof(ciphertext));
	/* Check a few times to avoid false positives */
	for (i = 0; i < sizeof(fence); i++) {
		memset(plaintext, fence[i], sizeof(plaintext));
		cx->ci->decrypt(cx, ciphertext, plaintext+16, enclen);
		for (j = 0; j < 16; j++) {
			if (plaintext[j] != fence[i]) {
				PRINTF("fencepost error at beginning of buf"
				       " idx: %u, cont:%x, not %x, encl:%d\n",
				       j, plaintext[j], fence[i], enclen);
				err++;
				break;
			}
			if (plaintext[16+enclen+j] != fence[i]) {
				PRINTF("fencepost error at end of buf"
				       " idx: %d, cont:%x, not %x, encl:%d\n",
				       j, plaintext[16+enclen+j], fence[i],
				       enclen);
				dump_u8_buffer(plaintext+16, enclen+16);
				err++;
				break;
			}
		}
	}
	if (err == 1) {
		PRINTF("fencepost errors reported are probably false "
		       "positives\n");
	}
	return err;
}

void show_usage() 
{
	printf("Usage: testapi [-h] -c <cipher>\n");
	exit(1);
}
	
