/*
 * crypto/testing/testcip.c
 *
 * Written by Alexander Kjeldaas <astor@guardian.no> 1998-12-17
 *
 * Copyright 1998 by Alexander Kjeldaas. Redistribution of this file
 * is permitted under the GNU Public License.
 */

#define __KERNEL__

#define printk printk_dummy

#include <linux/crypto.h>
#undef memcpy
#undef memset
#undef printk

#include "libcfuncs.h"
#include "misc.h"

void show_usage();

int plaintext[512];
int ciphertext[512];
int testres[512];
char key[512];
int plaintextlen, ciphertextlen, keylen;

int main(int argc, char **argv)
{
	struct cipher_implementation *ci = NULL;
	struct cipher_context *cx;
	int given_args = 0;
	int arg, err;
	int swap_to_processor = 0;

	/* Register all ciphers */
	cryptoapi_init();

	memset(plaintext, 0, sizeof(plaintext));
	memset(ciphertext, 0, sizeof(ciphertext));
	memset(testres, 0, sizeof(testres));
	memset(key, 0, sizeof(key));

	

	while((arg = getopt(argc,argv,"hp:e:c:k:s")) != EOF) {
		switch(arg) {
		case 'h': 
			show_usage(); 
			break;
		case 'p': 
			plaintextlen = parsehex(optarg, (char *)&plaintext); 
			given_args |= 2;
			break;
		case 'e': 
			ciphertextlen = parsehex(optarg, (char *)&ciphertext); 
			given_args |= 4;
			break;
		case 'k': 
			keylen = parsehex(optarg, key);
			given_args |= 8;
			break;
		case 'c': 
			if (!(ci = find_cipher_by_name(optarg))) {
				PRINTF("no cipher by that name (%s)\n", optarg);
				exit(1);
			}
			given_args |= 1; 
			break;
		case 's': 
			swap_to_processor = 1;
			break;
		}
	}
	if (given_args != 15) 
		show_usage();
	if (ciphertextlen != plaintextlen) {
		PRINTF("size of plaintext and ciphertext differ\n");
		exit(1);
	}

	if (swap_to_processor) {
		swab32_buf((char *)ciphertext, ciphertextlen);
	}

	cx = (struct cipher_context *) malloc(sizeof(struct cipher_context));
	cx->ci = ci;
	cx->keyinfo = (u32 *) malloc(ci->key_schedule_size);
	if ((err = ci->set_key(cx, (char *)key, keylen))) {
		PRINTF("set_key returned error %d\n", err);
		exit(1);
	}
	if ((err = ci->encrypt(cx, (u8 *)plaintext, (u8 *)testres, 
			       plaintextlen))) {
		PRINTF("encrypt returned error %d\n", err);
	}


	if (memcmp(testres, ciphertext, ciphertextlen)) {
		PRINTF("encrypted text and ciphertext differ\n");
		PRINTF("generated ciphertext (%d bytes):\n", ciphertextlen);
		dump_buffer((char *)testres, ciphertextlen);
		PRINTF("correct ciphertext (%d bytes):\n", ciphertextlen);
		dump_buffer((char *)ciphertext, ciphertextlen);
	}


	if ((err = ci->decrypt(cx, (u8 *)ciphertext, (u8 *)testres, 
			       ciphertextlen))) {
		PRINTF("decrypt returned error %d\n", err);
	}
	if (memcmp(testres, plaintext, plaintextlen)) {
		PRINTF("decrypted ciphertext and plaintext differ\n");
		PRINTF("generated plaintext (%d bytes):\n", plaintextlen);
		dump_buffer((char *)testres, plaintextlen);
		PRINTF("correct plaintext (%d bytes):\n", plaintextlen);
		dump_buffer((char *)plaintext, plaintextlen);
	}
	return 0;
}

void show_usage() 
{
	printf("Usage: testcip [-h] -c <cipher> -p <plaintext> -e <encryptedtext>\n");
	printf("       plaintext and encryptedtext are hex\n");
	exit(1);
}
	
