/*
 * linux/drivers/block/loop_cast.c
 *
 * PURPOSE
 *	CAST-128 encryption loop data transformation ("cast").
 *
 * DESCRIPTION
 *
 * HISTORY
 *	May 23, 1997 - Andrew E. Mileski
 *	Written, tested, and released.
 */
#include <linux/blkdev.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/locks.h>
#include <linux/loop.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/slab.h>

/* Optimize the CAST-128 code */
#if !defined(i386) && !defined(i486)
#if defined(CONFIG_M386)
#define i386
#endif
#if defined(CONFIG_M486) || defined(CONFIG_M586) || defined(CONFIG_M686)
#define i486
#endif
#endif /* !i386 && !i486 */

typedef __u32 U32;
typedef __u8 U8;

/* Internal Prototypes */
static void cast_keySchedule(U32 *, U32 *);
static U32 cast_decrypt(U32 *, U32 *, U32 *, U32);
static U32 cast_encrypt(U32 *, U32 *, U32 *, U32);
static int cast_release(struct loop_device *lo);
static int cast_init(struct loop_device *lo, struct loop_info *info);
static int cast_transfer(struct loop_device *lo, int cmd, char *raw_buf,
			 char *loop_buf, int size, int real_block);

/* Loop device specific data */
struct cast_public {
	U32 key[LO_KEY_SIZE >> 2];
};
struct cast_private {
	U32 skeys[32];
};
struct cast_data {
	struct cast_public public;
	struct cast_private private;
};


static void cast_lock(struct loop_device *lo)
{
	MOD_INC_USE_COUNT;
}

static void cast_unlock(struct loop_device *lo)
{
	MOD_DEC_USE_COUNT;
}   

/* Loop data transformation entry for "cast" */
static struct loop_func_table cast_funcs =
{ 
	number:   LO_CRYPT_CAST128,
	transfer: cast_transfer,
	init:     cast_init,
	release:  cast_release,
	lock:     cast_lock,
	unlock:   cast_unlock
};

/*
 * loop_cast_init / init_module
 *
 * PURPOSE
 *	Register the loop data transformation "cast".
 *
 * HISTORY
 *	May 3, 1997 - Andrew E. Mileski
 *	Written, tested, and released.
 */
#if defined(MODULE)
extern int init_module(void)
#else
int __init loop_cast_init(void)
#endif defined(MODULE)
{
	int err;

	if ((err=loop_register_transfer(&cast_funcs))) {
		printk(KERN_WARNING "Couldn't register cast-128 encryption\n");
		return err;
	}
  
	printk(KERN_INFO "loop: registered cast-128 encryption \n");
  
	return 0;
}

#if defined(MODULE)
/*
 * cleanup_module
 *
 * PURPOSE
 *	Unregister the loop data transformation "cast".
 *
 * HISTORY
 *	May 3, 1997 - Andrew E. Mileski
 *	Written, tested, and released.
 */
extern
void cleanup_module (void)
{
	loop_unregister_transfer(LO_CRYPT_CAST128);
}
#endif defined(MODULE)

static int cast_init(struct loop_device *lo, struct loop_info *info)
{ 
	struct cast_data *key;
  	if (info->lo_encrypt_key_size != 16)
		return -EINVAL;

	key=(struct cast_data *)kmalloc(sizeof(struct cast_data),GFP_KERNEL);
	if (key == NULL)
		return -ENOMEM;
	
	lo->key_data=key;
	memcpy(key->public.key, info->lo_encrypt_key, 16);
	cast_keySchedule(key->public.key, key->private.skeys);
	MOD_INC_USE_COUNT;
	return 0;
}

static int cast_release(struct loop_device *lo)
{ 
	if (lo->key_data != NULL)
	{ 
		kfree(lo->key_data);
		lo->key_data=NULL;
		MOD_DEC_USE_COUNT;
	}
	return (0);
}

static int cast_transfer(struct loop_device *lo, int cmd, char *raw_buf,
			 char *loop_buf, int size, int real_block)
{ 
	register U32 *K, *M, *C, result;
	K = ((struct cast_data *) (lo->key_data))->private.skeys;
	M = (U32 *)loop_buf;
	C = (U32 *)raw_buf;
	if (cmd == READ) {
		/* Decrypt */
		result = cast_decrypt(K, M, C, size);
	
		/* Pad if necessary */
		if (result) {
			C = C + ((size - result) >> 2);
			M = M + ((size - result) >> 2);
			memcpy(M, C,  result);
			memset(((U8 *)M) + result, 0, 8 - result);
			cast_decrypt(K, M, M, size);
		}
	} else {
		/* Encrypt */
		result = cast_encrypt(K, M, C, size);
		
		/* Pad if necessary */
		if (result) {
			C = C + ((size - result) >> 2);
			M = M + ((size - result) >> 2);
			memcpy(C, M,  result);
			memset(((U8 *)C) + result, 0, 8 - result);
			cast_decrypt(K, C, C, size);
		}
	}
	return 0;
}


/*
 * cast.c
 *
 * PURPOSE
 *	The CAST-128 data encryption/decryption algorithm.
 *
 * DESCRIPTION
 *	The following code should be completely portable.
 *	Refer to RFC 2144 for complete details on the algorithm.
 *
 *	This code _only_ implements keys of 128 bits in length, though the
 *	CAST algorithm supports keys of lengths 40 bits to 128 bits in steps
 *	of 8 bits.
 *
 *	Make sure that the type U32 is a 32 bit unsigned integer, and that
 *	U8 is an 8 bit unsigned integer.
 *
 *	For best performance, define as many of the following as possible:
 *		BIG_ENDIAN	- Native byte order is 1234.
 *	These only apply when using GNU C:
 *		__GNUC__	- GNU C defines this.
 *		i386		- Reduce cast_rol() to "roll".
 *		i486		- Reduce htonl() and ntohl() to "bswapl".
 *
 *	To run a full maintenance test (Appendix B, section 2), and a
 *	processing speed test, uncomment the line "#define DO_TESTS".
 *
 * EXPORT INFO
 *	This software is 100% of Canadian origin (both the algorithm and
 *	the code), and is NOT subject to U.S. ITAR export restrictions.
 *	Only Canadian export restictions apply.
 *
 *	Both the algorithm and this code are both "in the public domain",
 *	with the restriction that the algorithm and code authors maintain
 *	the copyrights to their work.
 *
 * HISTORY
 *	May 23, 1997 - Andrew E. Mileski <aem@netcom.ca>
 *	Implemented RFC 2144 in C, tested, and released.
 */

/* #define BIG_ENDIAN */
/* #define DO_TESTS */
/* #define i386 */
/* #define i486 */

/* These should be adjusted to the machine used */
/* typedef unsigned U32; */
/* typedef unsigned char U8; */

#if !defined(__GNUC__)
#if defined(__STDC__)
#define __inline__ inline
#else /* !defined(__STDC__) */
#error For speed, you should rewrite all the inline functions as macros.
#endif /* !defined(__STDC__) */
#endif /* !defined(__GNUC__) */

#define EXTERNAL_LINKAGE	static
#define INTERNAL_LINKAGE	static

/* Prototypes */
EXTERNAL_LINKAGE void cast_keySchedule(U32 *, U32 *);
EXTERNAL_LINKAGE U32 cast_encrypt(U32 *, U32 *, U32 *, U32);
EXTERNAL_LINKAGE U32 cast_decrypt(U32 *, U32 *, U32 *, U32);

/* First substitution box (S box) */
INTERNAL_LINKAGE U32 S1[256] = {
	0x30fb40d4U, 0x9fa0ff0bU, 0x6beccd2fU, 0x3f258c7aU,
	0x1e213f2fU, 0x9c004dd3U, 0x6003e540U, 0xcf9fc949U,
	0xbfd4af27U, 0x88bbbdb5U, 0xe2034090U, 0x98d09675U,
	0x6e63a0e0U, 0x15c361d2U, 0xc2e7661dU, 0x22d4ff8eU,
	0x28683b6fU, 0xc07fd059U, 0xff2379c8U, 0x775f50e2U,
	0x43c340d3U, 0xdf2f8656U, 0x887ca41aU, 0xa2d2bd2dU,
	0xa1c9e0d6U, 0x346c4819U, 0x61b76d87U, 0x22540f2fU,
	0x2abe32e1U, 0xaa54166bU, 0x22568e3aU, 0xa2d341d0U,
	0x66db40c8U, 0xa784392fU, 0x004dff2fU, 0x2db9d2deU,
	0x97943facU, 0x4a97c1d8U, 0x527644b7U, 0xb5f437a7U,
	0xb82cbaefU, 0xd751d159U, 0x6ff7f0edU, 0x5a097a1fU,
	0x827b68d0U, 0x90ecf52eU, 0x22b0c054U, 0xbc8e5935U,
	0x4b6d2f7fU, 0x50bb64a2U, 0xd2664910U, 0xbee5812dU,
	0xb7332290U, 0xe93b159fU, 0xb48ee411U, 0x4bff345dU,
	0xfd45c240U, 0xad31973fU, 0xc4f6d02eU, 0x55fc8165U,
	0xd5b1caadU, 0xa1ac2daeU, 0xa2d4b76dU, 0xc19b0c50U,
	0x882240f2U, 0x0c6e4f38U, 0xa4e4bfd7U, 0x4f5ba272U,
	0x564c1d2fU, 0xc59c5319U, 0xb949e354U, 0xb04669feU,
	0xb1b6ab8aU, 0xc71358ddU, 0x6385c545U, 0x110f935dU,
	0x57538ad5U, 0x6a390493U, 0xe63d37e0U, 0x2a54f6b3U,
	0x3a787d5fU, 0x6276a0b5U, 0x19a6fcdfU, 0x7a42206aU,
	0x29f9d4d5U, 0xf61b1891U, 0xbb72275eU, 0xaa508167U,
	0x38901091U, 0xc6b505ebU, 0x84c7cb8cU, 0x2ad75a0fU,
	0x874a1427U, 0xa2d1936bU, 0x2ad286afU, 0xaa56d291U,
	0xd7894360U, 0x425c750dU, 0x93b39e26U, 0x187184c9U,
	0x6c00b32dU, 0x73e2bb14U, 0xa0bebc3cU, 0x54623779U,
	0x64459eabU, 0x3f328b82U, 0x7718cf82U, 0x59a2cea6U,
	0x04ee002eU, 0x89fe78e6U, 0x3fab0950U, 0x325ff6c2U,
	0x81383f05U, 0x6963c5c8U, 0x76cb5ad6U, 0xd49974c9U,
	0xca180dcfU, 0x380782d5U, 0xc7fa5cf6U, 0x8ac31511U,
	0x35e79e13U, 0x47da91d0U, 0xf40f9086U, 0xa7e2419eU,
	0x31366241U, 0x051ef495U, 0xaa573b04U, 0x4a805d8dU,
	0x548300d0U, 0x00322a3cU, 0xbf64cddfU, 0xba57a68eU,
	0x75c6372bU, 0x50afd341U, 0xa7c13275U, 0x915a0bf5U,
	0x6b54bfabU, 0x2b0b1426U, 0xab4cc9d7U, 0x449ccd82U,
	0xf7fbf265U, 0xab85c5f3U, 0x1b55db94U, 0xaad4e324U,
	0xcfa4bd3fU, 0x2deaa3e2U, 0x9e204d02U, 0xc8bd25acU,
	0xeadf55b3U, 0xd5bd9e98U, 0xe31231b2U, 0x2ad5ad6cU,
	0x954329deU, 0xadbe4528U, 0xd8710f69U, 0xaa51c90fU,
	0xaa786bf6U, 0x22513f1eU, 0xaa51a79bU, 0x2ad344ccU,
	0x7b5a41f0U, 0xd37cfbadU, 0x1b069505U, 0x41ece491U,
	0xb4c332e6U, 0x032268d4U, 0xc9600accU, 0xce387e6dU,
	0xbf6bb16cU, 0x6a70fb78U, 0x0d03d9c9U, 0xd4df39deU,
	0xe01063daU, 0x4736f464U, 0x5ad328d8U, 0xb347cc96U,
	0x75bb0fc3U, 0x98511bfbU, 0x4ffbcc35U, 0xb58bcf6aU,
	0xe11f0abcU, 0xbfc5fe4aU, 0xa70aec10U, 0xac39570aU,
	0x3f04442fU, 0x6188b153U, 0xe0397a2eU, 0x5727cb79U,
	0x9ceb418fU, 0x1cacd68dU, 0x2ad37c96U, 0x0175cb9dU,
	0xc69dff09U, 0xc75b65f0U, 0xd9db40d8U, 0xec0e7779U,
	0x4744ead4U, 0xb11c3274U, 0xdd24cb9eU, 0x7e1c54bdU,
	0xf01144f9U, 0xd2240eb1U, 0x9675b3fdU, 0xa3ac3755U,
	0xd47c27afU, 0x51c85f4dU, 0x56907596U, 0xa5bb15e6U,
	0x580304f0U, 0xca042cf1U, 0x011a37eaU, 0x8dbfaadbU,
	0x35ba3e4aU, 0x3526ffa0U, 0xc37b4d09U, 0xbc306ed9U,
	0x98a52666U, 0x5648f725U, 0xff5e569dU, 0x0ced63d0U,
	0x7c63b2cfU, 0x700b45e1U, 0xd5ea50f1U, 0x85a92872U,
	0xaf1fbda7U, 0xd4234870U, 0xa7870bf3U, 0x2d3b4d79U,
	0x42e04198U, 0x0cd0ede7U, 0x26470db8U, 0xf881814cU,
	0x474d6ad7U, 0x7c0c5e5cU, 0xd1231959U, 0x381b7298U,
	0xf5d2f4dbU, 0xab838653U, 0x6e2f1e23U, 0x83719c9eU,
	0xbd91e046U, 0x9a56456eU, 0xdc39200cU, 0x20c8c571U,
	0x962bda1cU, 0xe1e696ffU, 0xb141ab08U, 0x7cca89b9U,
	0x1a69e783U, 0x02cc4843U, 0xa2f7c579U, 0x429ef47dU,
	0x427b169cU, 0x5ac9f049U, 0xdd8f0f00U, 0x5c8165bfU
};

/* Second substitution box (S box) */
INTERNAL_LINKAGE U32 S2[256] = {
	0x1f201094U, 0xef0ba75bU, 0x69e3cf7eU, 0x393f4380U,
	0xfe61cf7aU, 0xeec5207aU, 0x55889c94U, 0x72fc0651U,
	0xada7ef79U, 0x4e1d7235U, 0xd55a63ceU, 0xde0436baU,
	0x99c430efU, 0x5f0c0794U, 0x18dcdb7dU, 0xa1d6eff3U,
	0xa0b52f7bU, 0x59e83605U, 0xee15b094U, 0xe9ffd909U,
	0xdc440086U, 0xef944459U, 0xba83ccb3U, 0xe0c3cdfbU,
	0xd1da4181U, 0x3b092ab1U, 0xf997f1c1U, 0xa5e6cf7bU,
	0x01420ddbU, 0xe4e7ef5bU, 0x25a1ff41U, 0xe180f806U,
	0x1fc41080U, 0x179bee7aU, 0xd37ac6a9U, 0xfe5830a4U,
	0x98de8b7fU, 0x77e83f4eU, 0x79929269U, 0x24fa9f7bU,
	0xe113c85bU, 0xacc40083U, 0xd7503525U, 0xf7ea615fU,
	0x62143154U, 0x0d554b63U, 0x5d681121U, 0xc866c359U,
	0x3d63cf73U, 0xcee234c0U, 0xd4d87e87U, 0x5c672b21U,
	0x071f6181U, 0x39f7627fU, 0x361e3084U, 0xe4eb573bU,
	0x602f64a4U, 0xd63acd9cU, 0x1bbc4635U, 0x9e81032dU,
	0x2701f50cU, 0x99847ab4U, 0xa0e3df79U, 0xba6cf38cU,
	0x10843094U, 0x2537a95eU, 0xf46f6ffeU, 0xa1ff3b1fU,
	0x208cfb6aU, 0x8f458c74U, 0xd9e0a227U, 0x4ec73a34U,
	0xfc884f69U, 0x3e4de8dfU, 0xef0e0088U, 0x3559648dU,
	0x8a45388cU, 0x1d804366U, 0x721d9bfdU, 0xa58684bbU,
	0xe8256333U, 0x844e8212U, 0x128d8098U, 0xfed33fb4U,
	0xce280ae1U, 0x27e19ba5U, 0xd5a6c252U, 0xe49754bdU,
	0xc5d655ddU, 0xeb667064U, 0x77840b4dU, 0xa1b6a801U,
	0x84db26a9U, 0xe0b56714U, 0x21f043b7U, 0xe5d05860U,
	0x54f03084U, 0x066ff472U, 0xa31aa153U, 0xdadc4755U,
	0xb5625dbfU, 0x68561be6U, 0x83ca6b94U, 0x2d6ed23bU,
	0xeccf01dbU, 0xa6d3d0baU, 0xb6803d5cU, 0xaf77a709U,
	0x33b4a34cU, 0x397bc8d6U, 0x5ee22b95U, 0x5f0e5304U,
	0x81ed6f61U, 0x20e74364U, 0xb45e1378U, 0xde18639bU,
	0x881ca122U, 0xb96726d1U, 0x8049a7e8U, 0x22b7da7bU,
	0x5e552d25U, 0x5272d237U, 0x79d2951cU, 0xc60d894cU,
	0x488cb402U, 0x1ba4fe5bU, 0xa4b09f6bU, 0x1ca815cfU,
	0xa20c3005U, 0x8871df63U, 0xb9de2fcbU, 0x0cc6c9e9U,
	0x0beeff53U, 0xe3214517U, 0xb4542835U, 0x9f63293cU,
	0xee41e729U, 0x6e1d2d7cU, 0x50045286U, 0x1e6685f3U,
	0xf33401c6U, 0x30a22c95U, 0x31a70850U, 0x60930f13U,
	0x73f98417U, 0xa1269859U, 0xec645c44U, 0x52c877a9U,
	0xcdff33a6U, 0xa02b1741U, 0x7cbad9a2U, 0x2180036fU,
	0x50d99c08U, 0xcb3f4861U, 0xc26bd765U, 0x64a3f6abU,
	0x80342676U, 0x25a75e7bU, 0xe4e6d1fcU, 0x20c710e6U,
	0xcdf0b680U, 0x17844d3bU, 0x31eef84dU, 0x7e0824e4U,
	0x2ccb49ebU, 0x846a3baeU, 0x8ff77888U, 0xee5d60f6U,
	0x7af75673U, 0x2fdd5cdbU, 0xa11631c1U, 0x30f66f43U,
	0xb3faec54U, 0x157fd7faU, 0xef8579ccU, 0xd152de58U,
	0xdb2ffd5eU, 0x8f32ce19U, 0x306af97aU, 0x02f03ef8U,
	0x99319ad5U, 0xc242fa0fU, 0xa7e3ebb0U, 0xc68e4906U,
	0xb8da230cU, 0x80823028U, 0xdcdef3c8U, 0xd35fb171U,
	0x088a1bc8U, 0xbec0c560U, 0x61a3c9e8U, 0xbca8f54dU,
	0xc72feffaU, 0x22822e99U, 0x82c570b4U, 0xd8d94e89U,
	0x8b1c34bcU, 0x301e16e6U, 0x273be979U, 0xb0ffeaa6U,
	0x61d9b8c6U, 0x00b24869U, 0xb7ffce3fU, 0x08dc283bU,
	0x43daf65aU, 0xf7e19798U, 0x7619b72fU, 0x8f1c9ba4U,
	0xdc8637a0U, 0x16a7d3b1U, 0x9fc393b7U, 0xa7136eebU,
	0xc6bcc63eU, 0x1a513742U, 0xef6828bcU, 0x520365d6U,
	0x2d6a77abU, 0x3527ed4bU, 0x821fd216U, 0x095c6e2eU,
	0xdb92f2fbU, 0x5eea29cbU, 0x145892f5U, 0x91584f7fU,
	0x5483697bU, 0x2667a8ccU, 0x85196048U, 0x8c4baceaU,
	0x833860d4U, 0x0d23e0f9U, 0x6c387e8aU, 0x0ae6d249U,
	0xb284600cU, 0xd835731dU, 0xdcb1c647U, 0xac4c56eaU,
	0x3ebd81b3U, 0x230eabb0U, 0x6438bc87U, 0xf0b5b1faU,
	0x8f5ea2b3U, 0xfc184642U, 0x0a036b7aU, 0x4fb089bdU,
	0x649da589U, 0xa345415eU, 0x5c038323U, 0x3e5d3bb9U,
	0x43d79572U, 0x7e6dd07cU, 0x06dfdf1eU, 0x6c6cc4efU,
	0x7160a539U, 0x73bfbe70U, 0x83877605U, 0x4523ecf1U
};

/* Third substitution box (S box) */
INTERNAL_LINKAGE U32 S3[256] = {
	0x8defc240U, 0x25fa5d9fU, 0xeb903dbfU, 0xe810c907U,
	0x47607fffU, 0x369fe44bU, 0x8c1fc644U, 0xaececa90U,
	0xbeb1f9bfU, 0xeefbcaeaU, 0xe8cf1950U, 0x51df07aeU,
	0x920e8806U, 0xf0ad0548U, 0xe13c8d83U, 0x927010d5U,
	0x11107d9fU, 0x07647db9U, 0xb2e3e4d4U, 0x3d4f285eU,
	0xb9afa820U, 0xfade82e0U, 0xa067268bU, 0x8272792eU,
	0x553fb2c0U, 0x489ae22bU, 0xd4ef9794U, 0x125e3fbcU,
	0x21fffceeU, 0x825b1bfdU, 0x9255c5edU, 0x1257a240U,
	0x4e1a8302U, 0xbae07fffU, 0x528246e7U, 0x8e57140eU,
	0x3373f7bfU, 0x8c9f8188U, 0xa6fc4ee8U, 0xc982b5a5U,
	0xa8c01db7U, 0x579fc264U, 0x67094f31U, 0xf2bd3f5fU,
	0x40fff7c1U, 0x1fb78dfcU, 0x8e6bd2c1U, 0x437be59bU,
	0x99b03dbfU, 0xb5dbc64bU, 0x638dc0e6U, 0x55819d99U,
	0xa197c81cU, 0x4a012d6eU, 0xc5884a28U, 0xccc36f71U,
	0xb843c213U, 0x6c0743f1U, 0x8309893cU, 0x0feddd5fU,
	0x2f7fe850U, 0xd7c07f7eU, 0x02507fbfU, 0x5afb9a04U,
	0xa747d2d0U, 0x1651192eU, 0xaf70bf3eU, 0x58c31380U,
	0x5f98302eU, 0x727cc3c4U, 0x0a0fb402U, 0x0f7fef82U,
	0x8c96fdadU, 0x5d2c2aaeU, 0x8ee99a49U, 0x50da88b8U,
	0x8427f4a0U, 0x1eac5790U, 0x796fb449U, 0x8252dc15U,
	0xefbd7d9bU, 0xa672597dU, 0xada840d8U, 0x45f54504U,
	0xfa5d7403U, 0xe83ec305U, 0x4f91751aU, 0x925669c2U,
	0x23efe941U, 0xa903f12eU, 0x60270df2U, 0x0276e4b6U,
	0x94fd6574U, 0x927985b2U, 0x8276dbcbU, 0x02778176U,
	0xf8af918dU, 0x4e48f79eU, 0x8f616ddfU, 0xe29d840eU,
	0x842f7d83U, 0x340ce5c8U, 0x96bbb682U, 0x93b4b148U,
	0xef303cabU, 0x984faf28U, 0x779faf9bU, 0x92dc560dU,
	0x224d1e20U, 0x8437aa88U, 0x7d29dc96U, 0x2756d3dcU,
	0x8b907ceeU, 0xb51fd240U, 0xe7c07ce3U, 0xe566b4a1U,
	0xc3e9615eU, 0x3cf8209dU, 0x6094d1e3U, 0xcd9ca341U,
	0x5c76460eU, 0x00ea983bU, 0xd4d67881U, 0xfd47572cU,
	0xf76cedd9U, 0xbda8229cU, 0x127dadaaU, 0x438a074eU,
	0x1f97c090U, 0x081bdb8aU, 0x93a07ebeU, 0xb938ca15U,
	0x97b03cffU, 0x3dc2c0f8U, 0x8d1ab2ecU, 0x64380e51U,
	0x68cc7bfbU, 0xd90f2788U, 0x12490181U, 0x5de5ffd4U,
	0xdd7ef86aU, 0x76a2e214U, 0xb9a40368U, 0x925d958fU,
	0x4b39fffaU, 0xba39aee9U, 0xa4ffd30bU, 0xfaf7933bU,
	0x6d498623U, 0x193cbcfaU, 0x27627545U, 0x825cf47aU,
	0x61bd8ba0U, 0xd11e42d1U, 0xcead04f4U, 0x127ea392U,
	0x10428db7U, 0x8272a972U, 0x9270c4a8U, 0x127de50bU,
	0x285ba1c8U, 0x3c62f44fU, 0x35c0eaa5U, 0xe805d231U,
	0x428929fbU, 0xb4fcdf82U, 0x4fb66a53U, 0x0e7dc15bU,
	0x1f081fabU, 0x108618aeU, 0xfcfd086dU, 0xf9ff2889U,
	0x694bcc11U, 0x236a5caeU, 0x12deca4dU, 0x2c3f8cc5U,
	0xd2d02dfeU, 0xf8ef5896U, 0xe4cf52daU, 0x95155b67U,
	0x494a488cU, 0xb9b6a80cU, 0x5c8f82bcU, 0x89d36b45U,
	0x3a609437U, 0xec00c9a9U, 0x44715253U, 0x0a874b49U,
	0xd773bc40U, 0x7c34671cU, 0x02717ef6U, 0x4feb5536U,
	0xa2d02fffU, 0xd2bf60c4U, 0xd43f03c0U, 0x50b4ef6dU,
	0x07478cd1U, 0x006e1888U, 0xa2e53f55U, 0xb9e6d4bcU,
	0xa2048016U, 0x97573833U, 0xd7207d67U, 0xde0f8f3dU,
	0x72f87b33U, 0xabcc4f33U, 0x7688c55dU, 0x7b00a6b0U,
	0x947b0001U, 0x570075d2U, 0xf9bb88f8U, 0x8942019eU,
	0x4264a5ffU, 0x856302e0U, 0x72dbd92bU, 0xee971b69U,
	0x6ea22fdeU, 0x5f08ae2bU, 0xaf7a616dU, 0xe5c98767U,
	0xcf1febd2U, 0x61efc8c2U, 0xf1ac2571U, 0xcc8239c2U,
	0x67214cb8U, 0xb1e583d1U, 0xb7dc3e62U, 0x7f10bdceU,
	0xf90a5c38U, 0x0ff0443dU, 0x606e6dc6U, 0x60543a49U,
	0x5727c148U, 0x2be98a1dU, 0x8ab41738U, 0x20e1be24U,
	0xaf96da0fU, 0x68458425U, 0x99833be5U, 0x600d457dU,
	0x282f9350U, 0x8334b362U, 0xd91d1120U, 0x2b6d8da0U,
	0x642b1e31U, 0x9c305a00U, 0x52bce688U, 0x1b03588aU,
	0xf7baefd5U, 0x4142ed9cU, 0xa4315c11U, 0x83323ec5U,
	0xdfef4636U, 0xa133c501U, 0xe9d3531cU, 0xee353783U
};

/* Fourth substitution box (S box) */
INTERNAL_LINKAGE U32 S4[256] = {
	0x9db30420U, 0x1fb6e9deU, 0xa7be7befU, 0xd273a298U,
	0x4a4f7bdbU, 0x64ad8c57U, 0x85510443U, 0xfa020ed1U,
	0x7e287affU, 0xe60fb663U, 0x095f35a1U, 0x79ebf120U,
	0xfd059d43U, 0x6497b7b1U, 0xf3641f63U, 0x241e4adfU,
	0x28147f5fU, 0x4fa2b8cdU, 0xc9430040U, 0x0cc32220U,
	0xfdd30b30U, 0xc0a5374fU, 0x1d2d00d9U, 0x24147b15U,
	0xee4d111aU, 0x0fca5167U, 0x71ff904cU, 0x2d195ffeU,
	0x1a05645fU, 0x0c13fefeU, 0x081b08caU, 0x05170121U,
	0x80530100U, 0xe83e5efeU, 0xac9af4f8U, 0x7fe72701U,
	0xd2b8ee5fU, 0x06df4261U, 0xbb9e9b8aU, 0x7293ea25U,
	0xce84ffdfU, 0xf5718801U, 0x3dd64b04U, 0xa26f263bU,
	0x7ed48400U, 0x547eebe6U, 0x446d4ca0U, 0x6cf3d6f5U,
	0x2649abdfU, 0xaea0c7f5U, 0x36338cc1U, 0x503f7e93U,
	0xd3772061U, 0x11b638e1U, 0x72500e03U, 0xf80eb2bbU,
	0xabe0502eU, 0xec8d77deU, 0x57971e81U, 0xe14f6746U,
	0xc9335400U, 0x6920318fU, 0x081dbb99U, 0xffc304a5U,
	0x4d351805U, 0x7f3d5ce3U, 0xa6c866c6U, 0x5d5bcca9U,
	0xdaec6feaU, 0x9f926f91U, 0x9f46222fU, 0x3991467dU,
	0xa5bf6d8eU, 0x1143c44fU, 0x43958302U, 0xd0214eebU,
	0x022083b8U, 0x3fb6180cU, 0x18f8931eU, 0x281658e6U,
	0x26486e3eU, 0x8bd78a70U, 0x7477e4c1U, 0xb506e07cU,
	0xf32d0a25U, 0x79098b02U, 0xe4eabb81U, 0x28123b23U,
	0x69dead38U, 0x1574ca16U, 0xdf871b62U, 0x211c40b7U,
	0xa51a9ef9U, 0x0014377bU, 0x041e8ac8U, 0x09114003U,
	0xbd59e4d2U, 0xe3d156d5U, 0x4fe876d5U, 0x2f91a340U,
	0x557be8deU, 0x00eae4a7U, 0x0ce5c2ecU, 0x4db4bba6U,
	0xe756bdffU, 0xdd3369acU, 0xec17b035U, 0x06572327U,
	0x99afc8b0U, 0x56c8c391U, 0x6b65811cU, 0x5e146119U,
	0x6e85cb75U, 0xbe07c002U, 0xc2325577U, 0x893ff4ecU,
	0x5bbfc92dU, 0xd0ec3b25U, 0xb7801ab7U, 0x8d6d3b24U,
	0x20c763efU, 0xc366a5fcU, 0x9c382880U, 0x0ace3205U,
	0xaac9548aU, 0xeca1d7c7U, 0x041afa32U, 0x1d16625aU,
	0x6701902cU, 0x9b757a54U, 0x31d477f7U, 0x9126b031U,
	0x36cc6fdbU, 0xc70b8b46U, 0xd9e66a48U, 0x56e55a79U,
	0x026a4cebU, 0x52437effU, 0x2f8f76b4U, 0x0df980a5U,
	0x8674cde3U, 0xedda04ebU, 0x17a9be04U, 0x2c18f4dfU,
	0xb7747f9dU, 0xab2af7b4U, 0xefc34d20U, 0x2e096b7cU,
	0x1741a254U, 0xe5b6a035U, 0x213d42f6U, 0x2c1c7c26U,
	0x61c2f50fU, 0x6552daf9U, 0xd2c231f8U, 0x25130f69U,
	0xd8167fa2U, 0x0418f2c8U, 0x001a96a6U, 0x0d1526abU,
	0x63315c21U, 0x5e0a72ecU, 0x49bafefdU, 0x187908d9U,
	0x8d0dbd86U, 0x311170a7U, 0x3e9b640cU, 0xcc3e10d7U,
	0xd5cad3b6U, 0x0caec388U, 0xf73001e1U, 0x6c728affU,
	0x71eae2a1U, 0x1f9af36eU, 0xcfcbd12fU, 0xc1de8417U,
	0xac07be6bU, 0xcb44a1d8U, 0x8b9b0f56U, 0x013988c3U,
	0xb1c52fcaU, 0xb4be31cdU, 0xd8782806U, 0x12a3a4e2U,
	0x6f7de532U, 0x58fd7eb6U, 0xd01ee900U, 0x24adffc2U,
	0xf4990fc5U, 0x9711aac5U, 0x001d7b95U, 0x82e5e7d2U,
	0x109873f6U, 0x00613096U, 0xc32d9521U, 0xada121ffU,
	0x29908415U, 0x7fbb977fU, 0xaf9eb3dbU, 0x29c9ed2aU,
	0x5ce2a465U, 0xa730f32cU, 0xd0aa3fe8U, 0x8a5cc091U,
	0xd49e2ce7U, 0x0ce454a9U, 0xd60acd86U, 0x015f1919U,
	0x77079103U, 0xdea03af6U, 0x78a8565eU, 0xdee356dfU,
	0x21f05cbeU, 0x8b75e387U, 0xb3c50651U, 0xb8a5c3efU,
	0xd8eeb6d2U, 0xe523be77U, 0xc2154529U, 0x2f69efdfU,
	0xafe67afbU, 0xf470c4b2U, 0xf3e0eb5bU, 0xd6cc9876U,
	0x39e4460cU, 0x1fda8538U, 0x1987832fU, 0xca007367U,
	0xa99144f8U, 0x296b299eU, 0x492fc295U, 0x9266beabU,
	0xb5676e69U, 0x9bd3dddaU, 0xdf7e052fU, 0xdb25701cU,
	0x1b5e51eeU, 0xf65324e6U, 0x6afce36cU, 0x0316cc04U,
	0x8644213eU, 0xb7dc59d0U, 0x7965291fU, 0xccd6fd43U,
	0x41823979U, 0x932bcdf6U, 0xb657c34dU, 0x4edfd282U,
	0x7ae5290cU, 0x3cb9536bU, 0x851e20feU, 0x9833557eU,
	0x13ecf0b0U, 0xd3ffb372U, 0x3f85c5c1U, 0x0aef7ed2U
};

/* Fifth substitution box (S box) */
INTERNAL_LINKAGE U32 S5[256] = {
	0x7ec90c04U, 0x2c6e74b9U, 0x9b0e66dfU, 0xa6337911U,
	0xb86a7fffU, 0x1dd358f5U, 0x44dd9d44U, 0x1731167fU,
	0x08fbf1faU, 0xe7f511ccU, 0xd2051b00U, 0x735aba00U,
	0x2ab722d8U, 0x386381cbU, 0xacf6243aU, 0x69befd7aU,
	0xe6a2e77fU, 0xf0c720cdU, 0xc4494816U, 0xccf5c180U,
	0x38851640U, 0x15b0a848U, 0xe68b18cbU, 0x4caadeffU,
	0x5f480a01U, 0x0412b2aaU, 0x259814fcU, 0x41d0efe2U,
	0x4e40b48dU, 0x248eb6fbU, 0x8dba1cfeU, 0x41a99b02U,
	0x1a550a04U, 0xba8f65cbU, 0x7251f4e7U, 0x95a51725U,
	0xc106ecd7U, 0x97a5980aU, 0xc539b9aaU, 0x4d79fe6aU,
	0xf2f3f763U, 0x68af8040U, 0xed0c9e56U, 0x11b4958bU,
	0xe1eb5a88U, 0x8709e6b0U, 0xd7e07156U, 0x4e29fea7U,
	0x6366e52dU, 0x02d1c000U, 0xc4ac8e05U, 0x9377f571U,
	0x0c05372aU, 0x578535f2U, 0x2261be02U, 0xd642a0c9U,
	0xdf13a280U, 0x74b55bd2U, 0x682199c0U, 0xd421e5ecU,
	0x53fb3ce8U, 0xc8adedb3U, 0x28a87fc9U, 0x3d959981U,
	0x5c1ff900U, 0xfe38d399U, 0x0c4eff0bU, 0x062407eaU,
	0xaa2f4fb1U, 0x4fb96976U, 0x90c79505U, 0xb0a8a774U,
	0xef55a1ffU, 0xe59ca2c2U, 0xa6b62d27U, 0xe66a4263U,
	0xdf65001fU, 0x0ec50966U, 0xdfdd55bcU, 0x29de0655U,
	0x911e739aU, 0x17af8975U, 0x32c7911cU, 0x89f89468U,
	0x0d01e980U, 0x524755f4U, 0x03b63cc9U, 0x0cc844b2U,
	0xbcf3f0aaU, 0x87ac36e9U, 0xe53a7426U, 0x01b3d82bU,
	0x1a9e7449U, 0x64ee2d7eU, 0xcddbb1daU, 0x01c94910U,
	0xb868bf80U, 0x0d26f3fdU, 0x9342ede7U, 0x04a5c284U,
	0x636737b6U, 0x50f5b616U, 0xf24766e3U, 0x8eca36c1U,
	0x136e05dbU, 0xfef18391U, 0xfb887a37U, 0xd6e7f7d4U,
	0xc7fb7dc9U, 0x3063fcdfU, 0xb6f589deU, 0xec2941daU,
	0x26e46695U, 0xb7566419U, 0xf654efc5U, 0xd08d58b7U,
	0x48925401U, 0xc1bacb7fU, 0xe5ff550fU, 0xb6083049U,
	0x5bb5d0e8U, 0x87d72e5aU, 0xab6a6ee1U, 0x223a66ceU,
	0xc62bf3cdU, 0x9e0885f9U, 0x68cb3e47U, 0x086c010fU,
	0xa21de820U, 0xd18b69deU, 0xf3f65777U, 0xfa02c3f6U,
	0x407edac3U, 0xcbb3d550U, 0x1793084dU, 0xb0d70ebaU,
	0x0ab378d5U, 0xd951fb0cU, 0xded7da56U, 0x4124bbe4U,
	0x94ca0b56U, 0x0f5755d1U, 0xe0e1e56eU, 0x6184b5beU,
	0x580a249fU, 0x94f74bc0U, 0xe327888eU, 0x9f7b5561U,
	0xc3dc0280U, 0x05687715U, 0x646c6bd7U, 0x44904db3U,
	0x66b4f0a3U, 0xc0f1648aU, 0x697ed5afU, 0x49e92ff6U,
	0x309e374fU, 0x2cb6356aU, 0x85808573U, 0x4991f840U,
	0x76f0ae02U, 0x083be84dU, 0x28421c9aU, 0x44489406U,
	0x736e4cb8U, 0xc1092910U, 0x8bc95fc6U, 0x7d869cf4U,
	0x134f616fU, 0x2e77118dU, 0xb31b2be1U, 0xaa90b472U,
	0x3ca5d717U, 0x7d161bbaU, 0x9cad9010U, 0xaf462ba2U,
	0x9fe459d2U, 0x45d34559U, 0xd9f2da13U, 0xdbc65487U,
	0xf3e4f94eU, 0x176d486fU, 0x097c13eaU, 0x631da5c7U,
	0x445f7382U, 0x175683f4U, 0xcdc66a97U, 0x70be0288U,
	0xb3cdcf72U, 0x6e5dd2f3U, 0x20936079U, 0x459b80a5U,
	0xbe60e2dbU, 0xa9c23101U, 0xeba5315cU, 0x224e42f2U,
	0x1c5c1572U, 0xf6721b2cU, 0x1ad2fff3U, 0x8c25404eU,
	0x324ed72fU, 0x4067b7fdU, 0x0523138eU, 0x5ca3bc78U,
	0xdc0fd66eU, 0x75922283U, 0x784d6b17U, 0x58ebb16eU,
	0x44094f85U, 0x3f481d87U, 0xfcfeae7bU, 0x77b5ff76U,
	0x8c2302bfU, 0xaaf47556U, 0x5f46b02aU, 0x2b092801U,
	0x3d38f5f7U, 0x0ca81f36U, 0x52af4a8aU, 0x66d5e7c0U,
	0xdf3b0874U, 0x95055110U, 0x1b5ad7a8U, 0xf61ed5adU,
	0x6cf6e479U, 0x20758184U, 0xd0cefa65U, 0x88f7be58U,
	0x4a046826U, 0x0ff6f8f3U, 0xa09c7f70U, 0x5346aba0U,
	0x5ce96c28U, 0xe176eda3U, 0x6bac307fU, 0x376829d2U,
	0x85360fa9U, 0x17e3fe2aU, 0x24b79767U, 0xf5a96b20U,
	0xd6cd2595U, 0x68ff1ebfU, 0x7555442cU, 0xf19f06beU,
	0xf9e0659aU, 0xeeb9491dU, 0x34010718U, 0xbb30cab8U,
	0xe822fe15U, 0x88570983U, 0x750e6249U, 0xda627e55U,
	0x5e76ffa8U, 0xb1534546U, 0x6d47de08U, 0xefe9e7d4U
};
/* Sixth substitution box (S box) */
INTERNAL_LINKAGE U32 S6[256] = {
	0xf6fa8f9dU, 0x2cac6ce1U, 0x4ca34867U, 0xe2337f7cU,
	0x95db08e7U, 0x016843b4U, 0xeced5cbcU, 0x325553acU,
	0xbf9f0960U, 0xdfa1e2edU, 0x83f0579dU, 0x63ed86b9U,
	0x1ab6a6b8U, 0xde5ebe39U, 0xf38ff732U, 0x8989b138U,
	0x33f14961U, 0xc01937bdU, 0xf506c6daU, 0xe4625e7eU,
	0xa308ea99U, 0x4e23e33cU, 0x79cbd7ccU, 0x48a14367U,
	0xa3149619U, 0xfec94bd5U, 0xa114174aU, 0xeaa01866U,
	0xa084db2dU, 0x09a8486fU, 0xa888614aU, 0x2900af98U,
	0x01665991U, 0xe1992863U, 0xc8f30c60U, 0x2e78ef3cU,
	0xd0d51932U, 0xcf0fec14U, 0xf7ca07d2U, 0xd0a82072U,
	0xfd41197eU, 0x9305a6b0U, 0xe86be3daU, 0x74bed3cdU,
	0x372da53cU, 0x4c7f4448U, 0xdab5d440U, 0x6dba0ec3U,
	0x083919a7U, 0x9fbaeed9U, 0x49dbcfb0U, 0x4e670c53U,
	0x5c3d9c01U, 0x64bdb941U, 0x2c0e636aU, 0xba7dd9cdU,
	0xea6f7388U, 0xe70bc762U, 0x35f29adbU, 0x5c4cdd8dU,
	0xf0d48d8cU, 0xb88153e2U, 0x08a19866U, 0x1ae2eac8U,
	0x284caf89U, 0xaa928223U, 0x9334be53U, 0x3b3a21bfU,
	0x16434be3U, 0x9aea3906U, 0xefe8c36eU, 0xf890cdd9U,
	0x80226daeU, 0xc340a4a3U, 0xdf7e9c09U, 0xa694a807U,
	0x5b7c5eccU, 0x221db3a6U, 0x9a69a02fU, 0x68818a54U,
	0xceb2296fU, 0x53c0843aU, 0xfe893655U, 0x25bfe68aU,
	0xb4628abcU, 0xcf222ebfU, 0x25ac6f48U, 0xa9a99387U,
	0x53bddb65U, 0xe76ffbe7U, 0xe967fd78U, 0x0ba93563U,
	0x8e342bc1U, 0xe8a11be9U, 0x4980740dU, 0xc8087dfcU,
	0x8de4bf99U, 0xa11101a0U, 0x7fd37975U, 0xda5a26c0U,
	0xe81f994fU, 0x9528cd89U, 0xfd339fedU, 0xb87834bfU,
	0x5f04456dU, 0x22258698U, 0xc9c4c83bU, 0x2dc156beU,
	0x4f628daaU, 0x57f55ec5U, 0xe2220abeU, 0xd2916ebfU,
	0x4ec75b95U, 0x24f2c3c0U, 0x42d15d99U, 0xcd0d7fa0U,
	0x7b6e27ffU, 0xa8dc8af0U, 0x7345c106U, 0xf41e232fU,
	0x35162386U, 0xe6ea8926U, 0x3333b094U, 0x157ec6f2U,
	0x372b74afU, 0x692573e4U, 0xe9a9d848U, 0xf3160289U,
	0x3a62ef1dU, 0xa787e238U, 0xf3a5f676U, 0x74364853U,
	0x20951063U, 0x4576698dU, 0xb6fad407U, 0x592af950U,
	0x36f73523U, 0x4cfb6e87U, 0x7da4cec0U, 0x6c152daaU,
	0xcb0396a8U, 0xc50dfe5dU, 0xfcd707abU, 0x0921c42fU,
	0x89dff0bbU, 0x5fe2be78U, 0x448f4f33U, 0x754613c9U,
	0x2b05d08dU, 0x48b9d585U, 0xdc049441U, 0xc8098f9bU,
	0x7dede786U, 0xc39a3373U, 0x42410005U, 0x6a091751U,
	0x0ef3c8a6U, 0x890072d6U, 0x28207682U, 0xa9a9f7beU,
	0xbf32679dU, 0xd45b5b75U, 0xb353fd00U, 0xcbb0e358U,
	0x830f220aU, 0x1f8fb214U, 0xd372cf08U, 0xcc3c4a13U,
	0x8cf63166U, 0x061c87beU, 0x88c98f88U, 0x6062e397U,
	0x47cf8e7aU, 0xb6c85283U, 0x3cc2acfbU, 0x3fc06976U,
	0x4e8f0252U, 0x64d8314dU, 0xda3870e3U, 0x1e665459U,
	0xc10908f0U, 0x513021a5U, 0x6c5b68b7U, 0x822f8aa0U,
	0x3007cd3eU, 0x74719eefU, 0xdc872681U, 0x073340d4U,
	0x7e432fd9U, 0x0c5ec241U, 0x8809286cU, 0xf592d891U,
	0x08a930f6U, 0x957ef305U, 0xb7fbffbdU, 0xc266e96fU,
	0x6fe4ac98U, 0xb173ecc0U, 0xbc60b42aU, 0x953498daU,
	0xfba1ae12U, 0x2d4bd736U, 0x0f25faabU, 0xa4f3fcebU,
	0xe2969123U, 0x257f0c3dU, 0x9348af49U, 0x361400bcU,
	0xe8816f4aU, 0x3814f200U, 0xa3f94043U, 0x9c7a54c2U,
	0xbc704f57U, 0xda41e7f9U, 0xc25ad33aU, 0x54f4a084U,
	0xb17f5505U, 0x59357cbeU, 0xedbd15c8U, 0x7f97c5abU,
	0xba5ac7b5U, 0xb6f6deafU, 0x3a479c3aU, 0x5302da25U,
	0x653d7e6aU, 0x54268d49U, 0x51a477eaU, 0x5017d55bU,
	0xd7d25d88U, 0x44136c76U, 0x0404a8c8U, 0xb8e5a121U,
	0xb81a928aU, 0x60ed5869U, 0x97c55b96U, 0xeaec991bU,
	0x29935913U, 0x01fdb7f1U, 0x088e8dfaU, 0x9ab6f6f5U,
	0x3b4cbf9fU, 0x4a5de3abU, 0xe6051d35U, 0xa0e1d855U,
	0xd36b4cf1U, 0xf544edebU, 0xb0e93524U, 0xbebb8fbdU,
	0xa2d762cfU, 0x49c92f54U, 0x38b5f331U, 0x7128a454U,
	0x48392905U, 0xa65b1db8U, 0x851c97bdU, 0xd675cf2fU
};

/* Seventh substitution box (S box) */
INTERNAL_LINKAGE U32 S7[256] = {
	0x85e04019U, 0x332bf567U, 0x662dbfffU, 0xcfc65693U,
	0x2a8d7f6fU, 0xab9bc912U, 0xde6008a1U, 0x2028da1fU,
	0x0227bce7U, 0x4d642916U, 0x18fac300U, 0x50f18b82U,
	0x2cb2cb11U, 0xb232e75cU, 0x4b3695f2U, 0xb28707deU,
	0xa05fbcf6U, 0xcd4181e9U, 0xe150210cU, 0xe24ef1bdU,
	0xb168c381U, 0xfde4e789U, 0x5c79b0d8U, 0x1e8bfd43U,
	0x4d495001U, 0x38be4341U, 0x913cee1dU, 0x92a79c3fU,
	0x089766beU, 0xbaeeadf4U, 0x1286becfU, 0xb6eacb19U,
	0x2660c200U, 0x7565bde4U, 0x64241f7aU, 0x8248dca9U,
	0xc3b3ad66U, 0x28136086U, 0x0bd8dfa8U, 0x356d1cf2U,
	0x107789beU, 0xb3b2e9ceU, 0x0502aa8fU, 0x0bc0351eU,
	0x166bf52aU, 0xeb12ff82U, 0xe3486911U, 0xd34d7516U,
	0x4e7b3affU, 0x5f43671bU, 0x9cf6e037U, 0x4981ac83U,
	0x334266ceU, 0x8c9341b7U, 0xd0d854c0U, 0xcb3a6c88U,
	0x47bc2829U, 0x4725ba37U, 0xa66ad22bU, 0x7ad61f1eU,
	0x0c5cbafaU, 0x4437f107U, 0xb6e79962U, 0x42d2d816U,
	0x0a961288U, 0xe1a5c06eU, 0x13749e67U, 0x72fc081aU,
	0xb1d139f7U, 0xf9583745U, 0xcf19df58U, 0xbec3f756U,
	0xc06eba30U, 0x07211b24U, 0x45c28829U, 0xc95e317fU,
	0xbc8ec511U, 0x38bc46e9U, 0xc6e6fa14U, 0xbae8584aU,
	0xad4ebc46U, 0x468f508bU, 0x7829435fU, 0xf124183bU,
	0x821dba9fU, 0xaff60ff4U, 0xea2c4e6dU, 0x16e39264U,
	0x92544a8bU, 0x009b4fc3U, 0xaba68cedU, 0x9ac96f78U,
	0x06a5b79aU, 0xb2856e6eU, 0x1aec3ca9U, 0xbe838688U,
	0x0e0804e9U, 0x55f1be56U, 0xe7e5363bU, 0xb3a1f25dU,
	0xf7debb85U, 0x61fe033cU, 0x16746233U, 0x3c034c28U,
	0xda6d0c74U, 0x79aac56cU, 0x3ce4e1adU, 0x51f0c802U,
	0x98f8f35aU, 0x1626a49fU, 0xeed82b29U, 0x1d382fe3U,
	0x0c4fb99aU, 0xbb325778U, 0x3ec6d97bU, 0x6e77a6a9U,
	0xcb658b5cU, 0xd45230c7U, 0x2bd1408bU, 0x60c03eb7U,
	0xb9068d78U, 0xa33754f4U, 0xf430c87dU, 0xc8a71302U,
	0xb96d8c32U, 0xebd4e7beU, 0xbe8b9d2dU, 0x7979fb06U,
	0xe7225308U, 0x8b75cf77U, 0x11ef8da4U, 0xe083c858U,
	0x8d6b786fU, 0x5a6317a6U, 0xfa5cf7a0U, 0x5dda0033U,
	0xf28ebfb0U, 0xf5b9c310U, 0xa0eac280U, 0x08b9767aU,
	0xa3d9d2b0U, 0x79d34217U, 0x021a718dU, 0x9ac6336aU,
	0x2711fd60U, 0x438050e3U, 0x069908a8U, 0x3d7fedc4U,
	0x826d2befU, 0x4eeb8476U, 0x488dcf25U, 0x36c9d566U,
	0x28e74e41U, 0xc2610acaU, 0x3d49a9cfU, 0xbae3b9dfU,
	0xb65f8de6U, 0x92aeaf64U, 0x3ac7d5e6U, 0x9ea80509U,
	0xf22b017dU, 0xa4173f70U, 0xdd1e16c3U, 0x15e0d7f9U,
	0x50b1b887U, 0x2b9f4fd5U, 0x625aba82U, 0x6a017962U,
	0x2ec01b9cU, 0x15488aa9U, 0xd716e740U, 0x40055a2cU,
	0x93d29a22U, 0xe32dbf9aU, 0x058745b9U, 0x3453dc1eU,
	0xd699296eU, 0x496cff6fU, 0x1c9f4986U, 0xdfe2ed07U,
	0xb87242d1U, 0x19de7eaeU, 0x053e561aU, 0x15ad6f8cU,
	0x66626c1cU, 0x7154c24cU, 0xea082b2aU, 0x93eb2939U,
	0x17dcb0f0U, 0x58d4f2aeU, 0x9ea294fbU, 0x52cf564cU,
	0x9883fe66U, 0x2ec40581U, 0x763953c3U, 0x01d6692eU,
	0xd3a0c108U, 0xa1e7160eU, 0xe4f2dfa6U, 0x693ed285U,
	0x74904698U, 0x4c2b0eddU, 0x4f757656U, 0x5d393378U,
	0xa132234fU, 0x3d321c5dU, 0xc3f5e194U, 0x4b269301U,
	0xc79f022fU, 0x3c997e7eU, 0x5e4f9504U, 0x3ffafbbdU,
	0x76f7ad0eU, 0x296693f4U, 0x3d1fce6fU, 0xc61e45beU,
	0xd3b5ab34U, 0xf72bf9b7U, 0x1b0434c0U, 0x4e72b567U,
	0x5592a33dU, 0xb5229301U, 0xcfd2a87fU, 0x60aeb767U,
	0x1814386bU, 0x30bcc33dU, 0x38a0c07dU, 0xfd1606f2U,
	0xc363519bU, 0x589dd390U, 0x5479f8e6U, 0x1cb8d647U,
	0x97fd61a9U, 0xea7759f4U, 0x2d57539dU, 0x569a58cfU,
	0xe84e63adU, 0x462e1b78U, 0x6580f87eU, 0xf3817914U,
	0x91da55f4U, 0x40a230f3U, 0xd1988f35U, 0xb6e318d2U,
	0x3ffa50bcU, 0x3d40f021U, 0xc3c0bdaeU, 0x4958c24cU,
	0x518f36b2U, 0x84b1d370U, 0x0fedce83U, 0x878ddadaU,
	0xf2a279c7U, 0x94e01be8U, 0x90716f4bU, 0x954b8aa3U
};

/* Eighth substitution box (S box) */
INTERNAL_LINKAGE U32 S8[256] = {
	0xe216300dU, 0xbbddfffcU, 0xa7ebdabdU, 0x35648095U,
	0x7789f8b7U, 0xe6c1121bU, 0x0e241600U, 0x052ce8b5U,
	0x11a9cfb0U, 0xe5952f11U, 0xece7990aU, 0x9386d174U,
	0x2a42931cU, 0x76e38111U, 0xb12def3aU, 0x37ddddfcU,
	0xde9adeb1U, 0x0a0cc32cU, 0xbe197029U, 0x84a00940U,
	0xbb243a0fU, 0xb4d137cfU, 0xb44e79f0U, 0x049eedfdU,
	0x0b15a15dU, 0x480d3168U, 0x8bbbde5aU, 0x669ded42U,
	0xc7ece831U, 0x3f8f95e7U, 0x72df191bU, 0x7580330dU,
	0x94074251U, 0x5c7dcdfaU, 0xabbe6d63U, 0xaa402164U,
	0xb301d40aU, 0x02e7d1caU, 0x53571daeU, 0x7a3182a2U,
	0x12a8ddecU, 0xfdaa335dU, 0x176f43e8U, 0x71fb46d4U,
	0x38129022U, 0xce949ad4U, 0xb84769adU, 0x965bd862U,
	0x82f3d055U, 0x66fb9767U, 0x15b80b4eU, 0x1d5b47a0U,
	0x4cfde06fU, 0xc28ec4b8U, 0x57e8726eU, 0x647a78fcU,
	0x99865d44U, 0x608bd593U, 0x6c200e03U, 0x39dc5ff6U,
	0x5d0b00a3U, 0xae63aff2U, 0x7e8bd632U, 0x70108c0cU,
	0xbbd35049U, 0x2998df04U, 0x980cf42aU, 0x9b6df491U,
	0x9e7edd53U, 0x06918548U, 0x58cb7e07U, 0x3b74ef2eU,
	0x522fffb1U, 0xd24708ccU, 0x1c7e27cdU, 0xa4eb215bU,
	0x3cf1d2e2U, 0x19b47a38U, 0x424f7618U, 0x35856039U,
	0x9d17dee7U, 0x27eb35e6U, 0xc9aff67bU, 0x36baf5b8U,
	0x09c467cdU, 0xc18910b1U, 0xe11dbf7bU, 0x06cd1af8U,
	0x7170c608U, 0x2d5e3354U, 0xd4de495aU, 0x64c6d006U,
	0xbcc0c62cU, 0x3dd00db3U, 0x708f8f34U, 0x77d51b42U,
	0x264f620fU, 0x24b8d2bfU, 0x15c1b79eU, 0x46a52564U,
	0xf8d7e54eU, 0x3e378160U, 0x7895cda5U, 0x859c15a5U,
	0xe6459788U, 0xc37bc75fU, 0xdb07ba0cU, 0x0676a3abU,
	0x7f229b1eU, 0x31842e7bU, 0x24259fd7U, 0xf8bef472U,
	0x835ffcb8U, 0x6df4c1f2U, 0x96f5b195U, 0xfd0af0fcU,
	0xb0fe134cU, 0xe2506d3dU, 0x4f9b12eaU, 0xf215f225U,
	0xa223736fU, 0x9fb4c428U, 0x25d04979U, 0x34c713f8U,
	0xc4618187U, 0xea7a6e98U, 0x7cd16efcU, 0x1436876cU,
	0xf1544107U, 0xbedeee14U, 0x56e9af27U, 0xa04aa441U,
	0x3cf7c899U, 0x92ecbae6U, 0xdd67016dU, 0x151682ebU,
	0xa842eedfU, 0xfdba60b4U, 0xf1907b75U, 0x20e3030fU,
	0x24d8c29eU, 0xe139673bU, 0xefa63fb8U, 0x71873054U,
	0xb6f2cf3bU, 0x9f326442U, 0xcb15a4ccU, 0xb01a4504U,
	0xf1e47d8dU, 0x844a1be5U, 0xbae7dfdcU, 0x42cbda70U,
	0xcd7dae0aU, 0x57e85b7aU, 0xd53f5af6U, 0x20cf4d8cU,
	0xcea4d428U, 0x79d130a4U, 0x3486ebfbU, 0x33d3cddcU,
	0x77853b53U, 0x37effcb5U, 0xc5068778U, 0xe580b3e6U,
	0x4e68b8f4U, 0xc5c8b37eU, 0x0d809ea2U, 0x398feb7cU,
	0x132a4f94U, 0x43b7950eU, 0x2fee7d1cU, 0x223613bdU,
	0xdd06caa2U, 0x37df932bU, 0xc4248289U, 0xacf3ebc3U,
	0x5715f6b7U, 0xef3478ddU, 0xf267616fU, 0xc148cbe4U,
	0x9052815eU, 0x5e410fabU, 0xb48a2465U, 0x2eda7fa4U,
	0xe87b40e4U, 0xe98ea084U, 0x5889e9e1U, 0xefd390fcU,
	0xdd07d35bU, 0xdb485694U, 0x38d7e5b2U, 0x57720101U,
	0x730edebcU, 0x5b643113U, 0x94917e4fU, 0x503c2fbaU,
	0x646f1282U, 0x7523d24aU, 0xe0779695U, 0xf9c17a8fU,
	0x7a5b2121U, 0xd187b896U, 0x29263a4dU, 0xba510cdfU,
	0x81f47c9fU, 0xad1163edU, 0xea7b5965U, 0x1a00726eU,
	0x11403092U, 0x00da6d77U, 0x4a0cdd61U, 0xad1f4603U,
	0x605bdfb0U, 0x9eedc364U, 0x22ebe6a8U, 0xcee7d28aU,
	0xa0e736a0U, 0x5564a6b9U, 0x10853209U, 0xc7eb8f37U,
	0x2de705caU, 0x8951570fU, 0xdf09822bU, 0xbd691a6cU,
	0xaa12e4f2U, 0x87451c0fU, 0xe0f6a27aU, 0x3ada4819U,
	0x4cf1764fU, 0x0d771c2bU, 0x67cdb156U, 0x350d8384U,
	0x5938fa0fU, 0x42399ef3U, 0x36997b07U, 0x0e84093dU,
	0x4aa93e61U, 0x8360d87bU, 0x1fa98b0cU, 0x1149382cU,
	0xe97625a5U, 0x0614d1b7U, 0x0e25244bU, 0x0c768347U,
	0x589e8d82U, 0x0d2059d1U, 0xa466bb1eU, 0xf8da0a82U,
	0x04f19130U, 0xba6e4ec0U, 0x99265164U, 0x1ee7230dU,
	0x50b2ad80U, 0xeaee6801U, 0x8db2a283U, 0xea8bf59eU
};

#if defined(BIG_ENDIAN)
#define cast_htonl(H)	(H)
#define cast_ntohl(N)	(N)
#else /* !defined(BIG_ENDIAN) */

/*
 * htonl
 *
 * PURPOSE
 *	Convert from the host machine byte order to the byte order used
 *	by by the CAST algorithm (big-endian = network order = 1234).
 *
 * HISTORY
 *	May 23, 1997 - Andrew E. Mileski
 *	Written, tested, and released.
 */
INTERNAL_LINKAGE __inline__
U32 cast_htonl(U32 data)
{
#if defined(__GNUC__) && defined(i486)
	__asm__ (
		"bswapl %0"
		: "=r" (data)
		: "0" (data)
	);
	return data;
#else /* !(defined(__GNUC__) && defined(i486)) */
	union { U8 b[4]; U32 l; } u;
	u.b[3] = (U8) data;
	u.b[2] = (U8)(data >> 8);
	u.b[1] = (U8)(data >> 16);
	u.b[0] = (U8)(data >> 24);
	return u.l;
#endif /* !(defined(__GNUC__) && defined(i486)) */
}

/*
 * ntohl
 *
 * PURPOSE
 *	Convert to the host machine byte order from the byte order used
 *	by by the CAST algorithm (big-endian = network order = 1234).
 *
 * HISTORY
 *	May 23, 1997 - Andrew E. Mileski
 *	Written, tested, and released.
 */
INTERNAL_LINKAGE __inline__
U32 cast_ntohl(U32 data)
{
#if defined(__GNUC__) && defined(i486)
	__asm__ (
		"bswapl %0"
		: "=r" (data)
		: "0" (data)
	);
	return data;
#else /* !(defined(__GNUC__) && defined(i486)) */
	union { U8 b[4]; U32 l; } u;
	u.l = data;
	return (u.b[0] << 24) | (u.b[1] << 16) | (u.b[2] << 8) | u.b[3];
#endif /* !(defined(__GNUC__) && defined(i486)) */
}
#endif /* !defined(BIG_ENDIAN) */

/*
 * cast_rol
 *
 * PURPOSE
 *	Perform a left rotation.
 *
 * DESCRIPTION
 *	The portable code is SLOW. The ix86 code is a single instruction!
 *
 *	Note: This is a macro.
 *
 * HISTORY
 *	May 21, 1997 - Andrew E. Mileski
 *	Written, tested, and released.
 */
INTERNAL_LINKAGE __inline__
U32 cast_rol(U32 data, U32 count)
{
#if defined(__GNUC__) && (defined(i386) || defined(i486))
        U32 _d;
	__asm__ (
		"roll %%cl,%0"
		: "=r" (data), "=c" (_d)
		: "0" (data), "1" (count)
	);
	return data;
#else /* !(defined(__GNUC__) && defined(i386)) */
	return (count &= 31) ? (data << count) | ((data >> (32 - count))
			& (0xffffffff >> (32 - count))): data;
#endif /* !(defined(__GNUC__) && defined(i386)) */
}

/*
 * cast_encrypt
 *
 * PURPOSE
 *	Apply the CAST-128 algorithm to encrypt the specified plaintext.
 *
 * DESCRIPTION
 *	Refer to RFC 2144 for complete details on CAST.
 *
 *	Note that CAST is a block cipher of size 64 bits (8 bytes).
 *
 * PRE-CONDITIONS
 *	K		Pointer to subkeys - see cast_keySchedule().
 *	M		Pointer to plaintext.
 *	size		Bytes in the plaintext.
 *
 * POST-CONDITIONS
 *	C		Pointer to ciphertext.
 *	<return>	Number of bytes plaintext not processed.
 *
 * HISTORY
 *	May 23, 1997 - Andrew E. Mileski
 *	Implemented RFC 2144 in C, tested, and released.
 */
#define Ia	((U8)(I >> 24))
#define Ib	((U8)(I >> 16))
#define Ic	((U8)(I >> 8))
#define Id	((U8) I)

#define ROUND1(X) \
	T = L; \
	L = R; \
	I = cast_rol(K[X] + R, K[(X) + 16]); \
	R = T ^ (((S1[Ia] ^ S2[Ib]) - S3[Ic]) + S4[Id]);

#define ROUND2(X) \
	T = L; \
	L = R; \
	I = cast_rol(K[X] ^ R, K[(X) + 16]); \
	R = T ^ (((S1[Ia] - S2[Ib]) + S3[Ic]) ^ S4[Id]);

#define ROUND3(X) \
	T = L; \
	L = R; \
	I = cast_rol(K[X] - R, K[(X) + 16]); \
	R = T ^ (((S1[Ia] + S2[Ib]) ^ S3[Ic]) - S4[Id]);

EXTERNAL_LINKAGE
U32 cast_encrypt(U32 * K, U32 * M, U32 * C, U32 size)
{
	register U32 I, L, R, T;

	for (; size >= 8; size -= 8) {

		/* Split the plain-text */
		L = cast_htonl(*(M++));
		R = cast_htonl(*(M++));

		/* Do rounds */
		ROUND1(0);
		ROUND2(1);
		ROUND3(2);
		ROUND1(3);
		ROUND2(4);
		ROUND3(5);
		ROUND1(6);
		ROUND2(7);
		ROUND3(8);
		ROUND1(9);
		ROUND2(10);
		ROUND3(11);
		ROUND1(12);
		ROUND2(13);
		ROUND3(14);
		ROUND1(15);

		/* Swap halves */
		*(C++) = cast_ntohl(R);
		*(C++) = cast_ntohl(L);
	}
	return size;
}

/*
 * cast_decrypt
 *
 * PURPOSE
 *	Apply the CAST-128 algorithm to decrypt the specified ciphertext.
 *
 * DESCRIPTION
 *	Refer to RFC 2144 for complete details on CAST.
 *
 *	Note that CAST is a block cipher of size 64 bits (8 bytes).
 *
 * PRE-CONDITIONS
 *	K		Pointer to subkeys - see cast_keySchedule().
 *	C		Pointer to ciphertext.
 *	size		Bytes in the ciphertext.
 *
 * POST-CONDITIONS
 *	M		Pointer to plaintext.
 *	<return>	Number of bytes ciphertext not processed.
 *
 * HISTORY
 *	May 23, 1997 - Andrew E. Mileski
 *	Implemented RFC 2144 in C, tested, and released.
 */
EXTERNAL_LINKAGE
U32 cast_decrypt(U32 * K, U32 * M, U32 * C, U32 size)
{
	register U32 I, L, R, T;

	for (; size >= 8; size -= 8) {

		/* Split the ciphertext */
		L = cast_htonl(*(C++));
		R = cast_htonl(*(C++));

		/* Do rounds */
		ROUND1(15);
		ROUND3(14);
		ROUND2(13);
		ROUND1(12);
		ROUND3(11);
		ROUND2(10);
		ROUND1(9);
		ROUND3(8);
		ROUND2(7);
		ROUND1(6);
		ROUND3(5);
		ROUND2(4);
		ROUND1(3);
		ROUND3(2);
		ROUND2(1);
		ROUND1(0);

		/* Swap halves */
		*(M++) = cast_ntohl(R);
		*(M++) = cast_ntohl(L);
	}
	return size;
}

/* Cleanup the defines */
#undef Ia
#undef Ib
#undef Ic
#undef Id
#undef ROUND1
#undef ROUND2
#undef ROUND3

/*
 * cast_keySchedule
 *
 * PURPOSE
 *	Create the subkeys which are used in each round.
 *
 * DESCRIPTION
 *	Refer to RFC 2144, section 2.4.
 *
 *	IMPORTANT - the macros are required for portability to machines
 *	with different byte ordering (endian-ness), but they badly pollute
 *	the namespace.
 *
 * PRE-CONDITIONS
 *	X		Pointer to a key.
 *
 * POST-CONDITIONS
 *	K		Pointer to sub-keys.
 *
 * HISTORY
 *	May 23, 1997 - Andrew E. Mileski
 *	Implemented RFC 2144 in C, tested, and released.
 */
#define x0	((U8)(X0 >> 24))
#define x1	((U8)(X0 >> 16))
#define x2	((U8)(X0 >>  8))
#define x3	((U8) X0)
#define x4	((U8)(X1 >> 24))
#define x5	((U8)(X1 >> 16))
#define x6	((U8)(X1 >>  8))
#define x7	((U8) X1)
#define x8	((U8)(X2 >> 24))
#define x9	((U8)(X2 >> 16))
#define xa	((U8)(X2 >>  8))
#define xb	((U8) X2)
#define xc	((U8)(X3 >> 24))
#define xd	((U8)(X3 >> 16))
#define xe	((U8)(X3 >>  8))
#define xf	((U8) X3)
#define z0	((U8)(Z0 >> 24))
#define z1	((U8)(Z0 >> 16))
#define z2	((U8)(Z0 >>  8))
#define z3	((U8) Z0)
#define z4	((U8)(Z1 >> 24))
#define z5	((U8)(Z1 >> 16))
#define z6	((U8)(Z1 >>  8))
#define z7	((U8) Z1)
#define z8	((U8)(Z2 >> 24))
#define z9	((U8)(Z2 >> 16))
#define za	((U8)(Z2 >>  8))
#define zb	((U8) Z2)
#define zc	((U8)(Z3 >> 24))
#define zd	((U8)(Z3 >> 16))
#define ze	((U8)(Z3 >>  8))
#define zf	((U8) Z3)

EXTERNAL_LINKAGE
void cast_keySchedule(U32 * x, U32 * K)
{
	U32 X0, X1, X2, X3;
	U32 Z0, Z1, Z2, Z3;

	/* Preserve the passed key */
	X0 = cast_htonl(*(x++));
	X1 = cast_htonl(*(x++));
	X2 = cast_htonl(*(x++));
	X3 = cast_htonl(*(x++));

	Z0 = X0 ^ S5[xd] ^ S6[xf] ^ S7[xc] ^ S8[xe] ^ S7[x8];
	Z1 = X2 ^ S5[z0] ^ S6[z2] ^ S7[z1] ^ S8[z3] ^ S8[xa];
	Z2 = X3 ^ S5[z7] ^ S6[z6] ^ S7[z5] ^ S8[z4] ^ S5[x9];
	Z3 = X1 ^ S5[za] ^ S6[z9] ^ S7[zb] ^ S8[z8] ^ S6[xb];

	K[0] = S5[z8] ^ S6[z9] ^ S7[z7] ^ S8[z6] ^ S5[z2];
	K[1] = S5[za] ^ S6[zb] ^ S7[z5] ^ S8[z4] ^ S6[z6];
	K[2] = S5[zc] ^ S6[zd] ^ S7[z3] ^ S8[z2] ^ S7[z9];
	K[3] = S5[ze] ^ S6[zf] ^ S7[z1] ^ S8[z0] ^ S8[zc];

	X0 = Z2 ^ S5[z5] ^ S6[z7] ^ S7[z4] ^ S8[z6] ^ S7[z0];
	X1 = Z0 ^ S5[x0] ^ S6[x2] ^ S7[x1] ^ S8[x3] ^ S8[z2];
	X2 = Z1 ^ S5[x7] ^ S6[x6] ^ S7[x5] ^ S8[x4] ^ S5[z1];
	X3 = Z3 ^ S5[xa] ^ S6[x9] ^ S7[xb] ^ S8[x8] ^ S6[z3];

	K[4] = S5[x3] ^ S6[x2] ^ S7[xc] ^ S8[xd] ^ S5[x8];
	K[5] = S5[x1] ^ S6[x0] ^ S7[xe] ^ S8[xf] ^ S6[xd];
	K[6] = S5[x7] ^ S6[x6] ^ S7[x8] ^ S8[x9] ^ S7[x3];
	K[7] = S5[x5] ^ S6[x4] ^ S7[xa] ^ S8[xb] ^ S8[x7];

	Z0 = X0 ^ S5[xd] ^ S6[xf] ^ S7[xc] ^ S8[xe] ^ S7[x8];
	Z1 = X2 ^ S5[z0] ^ S6[z2] ^ S7[z1] ^ S8[z3] ^ S8[xa];
	Z2 = X3 ^ S5[z7] ^ S6[z6] ^ S7[z5] ^ S8[z4] ^ S5[x9];
	Z3 = X1 ^ S5[za] ^ S6[z9] ^ S7[zb] ^ S8[z8] ^ S6[xb];

	K[ 8] = S5[z3] ^ S6[z2] ^ S7[zc] ^ S8[zd] ^ S5[z9];
	K[ 9] = S5[z1] ^ S6[z0] ^ S7[ze] ^ S8[zf] ^ S6[zc];
	K[10] = S5[z7] ^ S6[z6] ^ S7[z8] ^ S8[z9] ^ S7[z2];
	K[11] = S5[z5] ^ S6[z4] ^ S7[za] ^ S8[zb] ^ S8[z6];

	X0 = Z2 ^ S5[z5] ^ S6[z7] ^ S7[z4] ^ S8[z6] ^ S7[z0];
	X1 = Z0 ^ S5[x0] ^ S6[x2] ^ S7[x1] ^ S8[x3] ^ S8[z2];
	X2 = Z1 ^ S5[x7] ^ S6[x6] ^ S7[x5] ^ S8[x4] ^ S5[z1];
	X3 = Z3 ^ S5[xa] ^ S6[x9] ^ S7[xb] ^ S8[x8] ^ S6[z3];

	K[12] = S5[x8] ^ S6[x9] ^ S7[x7] ^ S8[x6] ^ S5[x3];
	K[13] = S5[xa] ^ S6[xb] ^ S7[x5] ^ S8[x4] ^ S6[x7];
	K[14] = S5[xc] ^ S6[xd] ^ S7[x3] ^ S8[x2] ^ S7[x8];
	K[15] = S5[xe] ^ S6[xf] ^ S7[x1] ^ S8[x0] ^ S8[xd];

	Z0 = X0 ^ S5[xd] ^ S6[xf] ^ S7[xc] ^ S8[xe] ^ S7[x8];
	Z1 = X2 ^ S5[z0] ^ S6[z2] ^ S7[z1] ^ S8[z3] ^ S8[xa];
	Z2 = X3 ^ S5[z7] ^ S6[z6] ^ S7[z5] ^ S8[z4] ^ S5[x9];
	Z3 = X1 ^ S5[za] ^ S6[z9] ^ S7[zb] ^ S8[z8] ^ S6[xb];

	K[16] = S5[z8] ^ S6[z9] ^ S7[z7] ^ S8[z6] ^ S5[z2];
	K[17] = S5[za] ^ S6[zb] ^ S7[z5] ^ S8[z4] ^ S6[z6];
	K[18] = S5[zc] ^ S6[zd] ^ S7[z3] ^ S8[z2] ^ S7[z9];
	K[19] = S5[ze] ^ S6[zf] ^ S7[z1] ^ S8[z0] ^ S8[zc];

	X0 = Z2 ^ S5[z5] ^ S6[z7] ^ S7[z4] ^ S8[z6] ^ S7[z0];
	X1 = Z0 ^ S5[x0] ^ S6[x2] ^ S7[x1] ^ S8[x3] ^ S8[z2];
	X2 = Z1 ^ S5[x7] ^ S6[x6] ^ S7[x5] ^ S8[x4] ^ S5[z1];
	X3 = Z3 ^ S5[xa] ^ S6[x9] ^ S7[xb] ^ S8[x8] ^ S6[z3];

	K[20] = S5[x3] ^ S6[x2] ^ S7[xc] ^ S8[xd] ^ S5[x8];
	K[21] = S5[x1] ^ S6[x0] ^ S7[xe] ^ S8[xf] ^ S6[xd];
	K[22] = S5[x7] ^ S6[x6] ^ S7[x8] ^ S8[x9] ^ S7[x3];
	K[23] = S5[x5] ^ S6[x4] ^ S7[xa] ^ S8[xb] ^ S8[x7];

	Z0 = X0 ^ S5[xd] ^ S6[xf] ^ S7[xc] ^ S8[xe] ^ S7[x8];
	Z1 = X2 ^ S5[z0] ^ S6[z2] ^ S7[z1] ^ S8[z3] ^ S8[xa];
	Z2 = X3 ^ S5[z7] ^ S6[z6] ^ S7[z5] ^ S8[z4] ^ S5[x9];
	Z3 = X1 ^ S5[za] ^ S6[z9] ^ S7[zb] ^ S8[z8] ^ S6[xb];

	K[24] = S5[z3] ^ S6[z2] ^ S7[zc] ^ S8[zd] ^ S5[z9];
	K[25] = S5[z1] ^ S6[z0] ^ S7[ze] ^ S8[zf] ^ S6[zc];
	K[26] = S5[z7] ^ S6[z6] ^ S7[z8] ^ S8[z9] ^ S7[z2];
	K[27] = S5[z5] ^ S6[z4] ^ S7[za] ^ S8[zb] ^ S8[z6];

	X0 = Z2 ^ S5[z5] ^ S6[z7] ^ S7[z4] ^ S8[z6] ^ S7[z0];
	X1 = Z0 ^ S5[x0] ^ S6[x2] ^ S7[x1] ^ S8[x3] ^ S8[z2];
	X2 = Z1 ^ S5[x7] ^ S6[x6] ^ S7[x5] ^ S8[x4] ^ S5[z1];
	X3 = Z3 ^ S5[xa] ^ S6[x9] ^ S7[xb] ^ S8[x8] ^ S6[z3];

	K[28] = S5[x8] ^ S6[x9] ^ S7[x7] ^ S8[x6] ^ S5[x3];
	K[29] = S5[xa] ^ S6[xb] ^ S7[x5] ^ S8[x4] ^ S6[x7];
	K[30] = S5[xc] ^ S6[xd] ^ S7[x3] ^ S8[x2] ^ S7[x8];
	K[31] = S5[xe] ^ S6[xf] ^ S7[x1] ^ S8[x0] ^ S8[xd];
}

/* Cleanup the defines */
#undef x0
#undef x1
#undef x2
#undef x3
#undef x4
#undef x5
#undef x6
#undef x7
#undef x8
#undef x9
#undef xa
#undef xb
#undef xc
#undef xd
#undef xe
#undef xf
#undef z0
#undef z1
#undef z2
#undef z3
#undef z4
#undef z5
#undef z6
#undef z7
#undef z8
#undef z9
#undef za
#undef zb
#undef zc
#undef zd
#undef ze
#undef zf

/*****************************************************************************/

#if defined(DO_TESTS)

#include <stdio.h>
#include <time.h>

void speed_test(void)
{
	U8 a[] = {
		0x01, 0x23, 0x45, 0x67, 0x12, 0x34, 0x56, 0x78,
		0x23, 0x45, 0x67, 0x89, 0x34, 0x56, 0x78, 0x9a
	};
	time_t start;
	double elapsed;
	U32 K[32];
	register int i;

	cast_keySchedule((U32 * ) a, K);

	printf("Performing Processing speed test...");
	fflush(stdout);
	start = clock();
	for (i = 1; i <= 1000000; i++)
		cast_encrypt(K, (U32 * ) a, (U32 * ) a, 16);
	elapsed = clock() - start;
	printf("Done.\nProcessed 16,000,000 bytes.\n");

	printf("Elapsed time = %f\n", elapsed / CLOCKS_PER_SEC);
	printf("Processing speed (million bytes/sec) = %.6f\n",
		(CLOCKS_PER_SEC * 16) / elapsed);
}

INTERNAL_LINKAGE __inline__
int cmp(U8 *a, U8 *b)
{
	register int i;

	for (i = 0; i < 16 && *(a++) == *(b++); i++) {
		;
	}

	return (i == 16);
}

INTERNAL_LINKAGE __inline__
void maintenance_test(void)
{
	U8 a[] = {
		0x01, 0x23, 0x45, 0x67, 0x12, 0x34, 0x56, 0x78,
		0x23, 0x45, 0x67, 0x89, 0x34, 0x56, 0x78, 0x9a
	};
	U8 b[] = {
		0x01, 0x23, 0x45, 0x67, 0x12, 0x34, 0x56, 0x78,
		0x23, 0x45, 0x67, 0x89, 0x34, 0x56, 0x78, 0x9a
	};
	U8 r1_a[] = {
		0xee, 0xa9, 0xd0, 0xa2, 0x49, 0xfd, 0x3b, 0xa6,
		0xb3, 0x43, 0x6f, 0xb8, 0x9d, 0x6d, 0xca, 0x92
	};
	U8 r1_b[] = {
		0xb2, 0xc9, 0x5e, 0xb0, 0x0c, 0x31, 0xad, 0x71,
		0x80, 0xac, 0x05, 0xb8, 0xe8, 0x3d, 0x69, 0x6e
	};
	U8 r2_ab[] = {
		0x01, 0x23, 0x45, 0x67, 0x12, 0x34, 0x56, 0x78,
		0x23, 0x45, 0x67, 0x89, 0x34, 0x56, 0x78, 0x9a
	};
	U32 K[32];
	register int i;

	printf("Testing encryption...");
	fflush(stdout);
	for (i = 1; i <= 1000000; i++) {
		cast_keySchedule((U32 *) b, K);
		cast_encrypt(K, (U32 *) a, (U32 *) a, 16);

		cast_keySchedule((U32 *) a, K);
		cast_encrypt(K, (U32 *) b, (U32 *) b, 16);
	}
	if (cmp(a, r1_a) && cmp(b, r1_b))
		printf("Passed!\n");
	else
		printf("Failed!\n");

	printf("Testing decryption...");
	fflush(stdout);
	for (i = 1; i <= 1000000; i++) {
		cast_keySchedule((U32 *) a, K);
		cast_decrypt(K, (U32 *) b, (U32 *) b, 16);

		cast_keySchedule((U32 *) b, K);
		cast_decrypt(K, (U32 *) a, (U32 *) a, 16);
	}
	if (cmp(a, r2_ab) && cmp(b, r2_ab))
		printf("Passed!\n");
	else
		printf("Failed!\n");
}

int main(void)
{
	maintenance_test();
	speed_test();
	return 0;
}
#endif /* defined(DO_TESTS) */
