#include <linux/module.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/string.h> 
#include <linux/mm.h>
#include <linux/slab.h>
#include <asm/byteorder.h>
#include <linux/loop.h>
#include <linux/crypto.h>


static char loop_gen_xfers[MAX_CIPHER] = {0};

static int loop_gen_transfer(struct loop_device *lo, int cmd, char *raw_buf,
			     char *loop_buf, int size, int real_block)
{ 
	struct cipher_context * cx = (struct cipher_context *) lo->key_data;
	struct cipher_implementation *ci = cx->ci;
	int (*encdecfunc)(struct cipher_context *cx, const u8 *in, u8 *out, int size);
	char *in, *out;

	if (cmd == READ) {
		encdecfunc = ci->decrypt;
		in = raw_buf;
		out = loop_buf;
	} else {
		encdecfunc = ci->encrypt;
		in = loop_buf;
		out = raw_buf;
	}
	if (!encdecfunc) {
		printk("ouch! encdecfunc == NULL!\n");
		return -EINVAL;
	}
	((u32 *)cx->iv)[0] = cpu_to_le32(real_block);
	encdecfunc(cx, in, out, size);
	return 0;
}


static int loop_gen_init2(struct loop_device *lo, struct loop_info *info)
{
	int cipher,err = -EINVAL;
	int mode = CIPHER_CBC;
	struct cipher_implementation *ci;
	struct cipher_context *cx;

	switch (info->lo_encrypt_type) {
	case LO_CRYPT_NONE:     cipher = CIPHER_NONE; break;  
	case LO_CRYPT_XOR:      cipher = CIPHER_XOR; break;
	case LO_CRYPT_DES:      cipher = CIPHER_DES | CIPHER_CBC; break;
	case LO_CRYPT_FISH2:    cipher = CIPHER_FISH2 | CIPHER_CBC; break;
	default:                cipher = info->lo_encrypt_type | mode; break;
	}

	ci = find_cipher_by_id(cipher);
	if (!ci)
		goto out;

	if (ci->trans.t_id != cipher) {
		printk("find_cipher_by_id gave me the wrong cipher!\n");
		goto out;
	}

	ci->lock();
	err = -ENOMEM;
	cx = (struct cipher_context *) kmalloc(sizeof(struct cipher_context), 
					       GFP_KERNEL);
	if (!cx)
		goto out_ci;
	cx->ci = ci;
	memset(&cx->iv, 0, cx->ci->ivsize);
	cx->keyinfo = (u32 *) kmalloc(ci->key_schedule_size, GFP_KERNEL);
	if (!cx->keyinfo)
		goto out_cx;
	if (!ci->set_key) {
		printk("Invalid cipher - set_key is NULL\n");
		err = -EINVAL;
		goto out_key;
	}
	if (ci->set_key(cx, info->lo_encrypt_key, 
			info->lo_encrypt_key_size)) {
		err = -EINVAL;
		goto out_key;
	}
       	lo->key_data = cx;
	return 0;

 out_key:
	kfree(cx->keyinfo);
 out_cx:
	kfree(cx);
 out_ci:
	ci->unlock();
 out:
	return err;
}


static int loop_gen_release(struct loop_device *lo)
{ 
	struct cipher_context *cx = (struct cipher_context *) lo->key_data;
	if (cx != NULL)	{ 
		memset(cx->keyinfo, 0, cx->ci->key_schedule_size);
		kfree(cx->keyinfo);
		cx->ci->unlock();
		memset(cx, 0, sizeof(*cx));
		kfree(cx);
		lo->key_data=NULL;
	}
	return 0;
}


static void loop_gen_lock(struct loop_device *lo)
{
	MOD_INC_USE_COUNT;
}

static void loop_gen_unlock(struct loop_device *lo)
{
	MOD_DEC_USE_COUNT;
}   

#ifdef MODULE
int __init init_module(void)
#else
int __init loop_gen_init(void)
#endif
{ 
       int i, n=0, err;
       struct loop_func_table *funcs;

        for (i=0; i<MAX_CIPHER; i++) {
               funcs = kmalloc(sizeof (struct loop_func_table), GFP_KERNEL);
               memset(funcs, 0, sizeof(*funcs));
               if (funcs) {
                       funcs->number = i;
                       funcs->transfer = loop_gen_transfer;
                       funcs->init = loop_gen_init2;
                       funcs->release = loop_gen_release;
                       funcs->lock = loop_gen_lock;
                       funcs->unlock = loop_gen_unlock;

                       err = loop_register_transfer(funcs);
                       if (!err) {
                               loop_gen_xfers[i] = 1;
                               n++;
                       }
		}
	}
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
       int i;

       for (i=0; i<MAX_CIPHER; i++) {
               if (loop_gen_xfers[i]) {
                       loop_unregister_transfer(i);
                       loop_gen_xfers[i] = 0;
               }
       }
}
#endif

