/*
 * see http://gtf.org/garzik/drivers/kcompat24/
 * this is version 0.9.3
 *				-- kraxel
 */

/*
 *	$Id: kcompat24.c,v 1.2 2000/06/01 22:18:36 jgarzik Exp $
 *
 *	PCI Bus Services, see include/linux/pci.h for further explanation.
 *
 *	Copyright 1993 -- 1997 Drew Eckhardt, Frederic Potter,
 *	David Mosberger-Tang
 *
 *	Copyright 1997 -- 2000 Martin Mares <mj@suse.cz>
 */

#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)

#ifndef KCOMPAT_INCLUDED
#include <linux/module.h>
#endif

#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/pci.h>
#include <linux/ioport.h>
#include "kcompat24.h"

/*
 *  Registration of PCI drivers and handling of hot-pluggable devices.
 */

static LIST_HEAD(pci_drivers);

struct pci_driver_mapping {
	struct pci_dev *dev;
	struct pci_driver *drv;
	void *driver_data;
};

#define PCI_MAX_MAPPINGS 16
static struct pci_driver_mapping drvmap [PCI_MAX_MAPPINGS] = { { NULL, } , };

struct net_device_mapping {
	struct net_device *dev;
	struct timer_list dev_watchdog;
	int watchdog_timeo;
	void (*tx_timeout) (struct net_device *dev);
	int timer_started;
};

#define NETDEV_MAX_MAPPINGS 16
static struct net_device_mapping netdevmap [NETDEV_MAX_MAPPINGS] = { { NULL, } , };


static inline struct net_device_mapping *netif_map_find (struct net_device *dev)
{
	int i;

	for (i = 0; i < NETDEV_MAX_MAPPINGS; i++)
		if (netdevmap[i].dev == dev)
			return &netdevmap[i];

	panic ("mapping not found");
}


void netif_watchdog (unsigned long data)
{
	struct net_device_mapping *cdev = (struct net_device_mapping *) data;
	struct net_device *dev = cdev->dev;

	if (netif_running(dev) && netif_queue_stopped(dev) &&
	    ((jiffies - dev->trans_start) > cdev->watchdog_timeo)) {
		printk(KERN_INFO "NETDEV WATCHDOG: %s: transmit timed out\n", dev->name);
		cdev->tx_timeout(dev);
	}

	mod_timer(&cdev->dev_watchdog, jiffies + cdev->watchdog_timeo);
}


void netif_init_watchdog(struct net_device *dev, int timeout,
                         void (*tx_timeout) (struct net_device *dev))
{
	struct net_device_mapping *cdev;

	cdev = netif_map_find (NULL);
	cdev->dev = dev;
	cdev->watchdog_timeo = timeout;
	cdev->tx_timeout = tx_timeout;
	cdev->timer_started = 0;

	init_timer(&cdev->dev_watchdog);
	cdev->dev_watchdog.data = (unsigned long) cdev;
	cdev->dev_watchdog.function = netif_watchdog;
}


void netif_start_watchdog(struct net_device *dev)
{
	struct net_device_mapping *cdev = netif_map_find (dev);

	if (!cdev->timer_started) {
		cdev->timer_started = 1;
		mod_timer(&cdev->dev_watchdog, jiffies + cdev->watchdog_timeo);
	}
}


void netif_stop_watchdog(struct net_device *dev)
{
	struct net_device_mapping *cdev = netif_map_find (dev);

	if (!cdev->timer_started) {
		cdev->timer_started = 0;
		del_timer(&cdev->dev_watchdog);
	}
}


KCOMPINC void * pci_compat_get_driver_data (struct pci_dev *dev)
{
	int i;
	
	for (i = 0; i < PCI_MAX_MAPPINGS; i++)
		if (drvmap[i].dev == dev)
			return drvmap[i].driver_data;
	
	return NULL;
}


KCOMPINC void pci_compat_set_driver_data (struct pci_dev *dev, void *driver_data)
{
	int i;
	
	for (i = 0; i < PCI_MAX_MAPPINGS; i++)
		if (drvmap[i].dev == dev) {
			drvmap[i].driver_data = driver_data;
			return;
		} 
}


KCOMPINC const struct pci_device_id *
pci_compat_match_device(const struct pci_device_id *ids, struct pci_dev *dev)
{
	u16 subsystem_vendor, subsystem_device;

	pci_read_config_word(dev, PCI_SUBSYSTEM_VENDOR_ID, &subsystem_vendor);
	pci_read_config_word(dev, PCI_SUBSYSTEM_ID, &subsystem_device);

	while (ids->vendor || ids->subvendor || ids->class_mask) {
		if ((ids->vendor == PCI_ANY_ID || ids->vendor == dev->vendor) &&
		    (ids->device == PCI_ANY_ID || ids->device == dev->device) &&
		    (ids->subvendor == PCI_ANY_ID || ids->subvendor == subsystem_vendor) &&
		    (ids->subdevice == PCI_ANY_ID || ids->subdevice == subsystem_device) &&
		    !((ids->class ^ dev->class) & ids->class_mask))
			return ids;
		ids++;
	}
	return NULL;
}

static int
pci_announce_device(struct pci_driver *drv, struct pci_dev *dev)
{
	const struct pci_device_id *id;
	int found, i;

	if (drv->id_table) {
		id = pci_compat_match_device(drv->id_table, dev);
		if (!id)
			return 0;
	} else
		id = NULL;

	found = 0;
	for (i = 0; i < PCI_MAX_MAPPINGS && !found; i++)
		if (!drvmap[i].dev) {
			drvmap[i].dev = dev;
			drvmap[i].drv = drv;
			found = 1;
		}

	if (drv->probe(dev, id) >= 0) { 
		if(found)
			return 1;
	} else
               drvmap[i - 1].dev = NULL; 

	return 0;
}

KCOMPINC int
pci_compat_register_driver(struct pci_driver *drv)
{
	struct pci_dev *dev;
	int count = 0, found, i;

	list_add_tail(&drv->node, &pci_drivers);
	pci_for_each_dev(dev) {
		found = 0;
		for (i = 0; i < PCI_MAX_MAPPINGS && !found; i++)
			if (drvmap[i].dev == dev)
				found = 1;
		if (!found)
			count += pci_announce_device(drv, dev);
	}
	return count;
}

KCOMPINC void
pci_compat_unregister_driver(struct pci_driver *drv)
{
	struct pci_dev *dev;
	int i;

	list_del(&drv->node);
	pci_for_each_dev(dev) {
		for (i = 0; i < PCI_MAX_MAPPINGS; i++)
			if (drvmap[i].dev == dev &&
			    drvmap[i].drv == drv)
				break;
		if (PCI_MAX_MAPPINGS == i)
			continue;
		if (drv->remove)
			drv->remove(dev);
               	drvmap[i].dev = NULL;
	}
}

KCOMPINC unsigned long pci_compat_get_size (struct pci_dev *dev, int n_base)
{
	u32 l, sz;
	int reg = PCI_BASE_ADDRESS_0 + (n_base << 2);

	/* XXX temporarily disable I/O and memory decoding for this device? */

	pci_read_config_dword (dev, reg, &l);
	if (l == 0xffffffff)
		return 0;

	pci_write_config_dword (dev, reg, ~0);
	pci_read_config_dword (dev, reg, &sz);
	pci_write_config_dword (dev, reg, l);

	if (!sz || sz == 0xffffffff)
		return 0;
	if ((l & PCI_BASE_ADDRESS_SPACE) == PCI_BASE_ADDRESS_SPACE_MEMORY) {
		sz = ~(sz & PCI_BASE_ADDRESS_MEM_MASK);
	} else {
		sz = ~(sz & PCI_BASE_ADDRESS_IO_MASK) & 0xffff;
	}
	
	return sz + 1;
}

KCOMPINC int pci_compat_get_flags (struct pci_dev *dev, int n_base)
{
	unsigned long foo = dev->base_address[n_base] & PCI_BASE_ADDRESS_SPACE;
	int flags = 0;
	
	if (foo == 0)
		flags |= IORESOURCE_MEM;
	if (foo == 1)
		flags |= IORESOURCE_IO;
	
	return flags;
}

/*
 *  Set power management state of a device.  For transitions from state D3
 *  it isn't as straightforward as one could assume since many devices forget
 *  their configuration space during wakeup.  Returns old power state.
 */
KCOMPINC int
pci_compat_set_power_state(struct pci_dev *dev, int new_state)
{
	u32 base[5], romaddr;
	u16 pci_command, pwr_command;
	u8  pci_latency, pci_cacheline;
	int i, old_state;
	int pm = pci_compat_find_capability(dev, PCI_CAP_ID_PM);

	if (!pm)
		return 0;
	pci_read_config_word(dev, pm + PCI_PM_CTRL, &pwr_command);
	old_state = pwr_command & PCI_PM_CTRL_STATE_MASK;
	if (old_state == new_state)
		return old_state;
	if (old_state == 3) {
		pci_read_config_word(dev, PCI_COMMAND, &pci_command);
		pci_write_config_word(dev, PCI_COMMAND, pci_command & ~(PCI_COMMAND_IO | PCI_COMMAND_MEMORY));
		for (i = 0; i < 5; i++)
			pci_read_config_dword(dev, PCI_BASE_ADDRESS_0 + i*4, &base[i]);
		pci_read_config_dword(dev, PCI_ROM_ADDRESS, &romaddr);
		pci_read_config_byte(dev, PCI_LATENCY_TIMER, &pci_latency);
		pci_read_config_byte(dev, PCI_CACHE_LINE_SIZE, &pci_cacheline);
		pci_write_config_word(dev, pm + PCI_PM_CTRL, new_state);
		for (i = 0; i < 5; i++)
			pci_write_config_dword(dev, PCI_BASE_ADDRESS_0 + i*4, base[i]);
		pci_write_config_dword(dev, PCI_ROM_ADDRESS, romaddr);
		pci_write_config_byte(dev, PCI_INTERRUPT_LINE, dev->irq);
		pci_write_config_byte(dev, PCI_CACHE_LINE_SIZE, pci_cacheline);
		pci_write_config_byte(dev, PCI_LATENCY_TIMER, pci_latency);
		pci_write_config_word(dev, PCI_COMMAND, pci_command);
	} else
		pci_write_config_word(dev, pm + PCI_PM_CTRL, (pwr_command & ~PCI_PM_CTRL_STATE_MASK) | new_state);
	return old_state;
}

/*
 *  Initialize device before it's used by a driver. Ask low-level code
 *  to enable I/O and memory. Wake up the device if it was suspended.
 *  Beware, this function can fail.
 */
KCOMPINC int
pci_compat_enable_device(struct pci_dev *dev)
{
#if 0
	pci_compat_set_power_state(dev, 0);
#endif
	return 0;
}

KCOMPINC int
pci_compat_find_capability(struct pci_dev *dev, int cap)
{
	u16 status;
	u8 pos, id;
	int ttl = 48;

	pci_read_config_word(dev, PCI_STATUS, &status);
	if (!(status & PCI_STATUS_CAP_LIST))
		return 0;
	pci_read_config_byte(dev, PCI_CAPABILITY_LIST, &pos);
	while (ttl-- && pos >= 0x40) {
		pos &= ~3;
		pci_read_config_byte(dev, pos + PCI_CAP_LIST_ID, &id);
		if (id == 0xff)
			break;
		if (id == cap)
			return pos;
		pci_read_config_byte(dev, pos + PCI_CAP_LIST_NEXT, &pos);
	}
	return 0;
}


void daemonize(void)
{
        struct fs_struct *fs;
	
	exit_mm(current);
	current->session = 1;
	current->pgrp = 1;
        exit_fs(current);
	fs = init_task.fs;
	current->fs = fs;
	atomic_inc(&fs->count);
}

#ifndef KCOMPAT_INCLUDED
EXPORT_SYMBOL(pci_compat_match_device);
EXPORT_SYMBOL(pci_compat_register_driver);
EXPORT_SYMBOL(pci_compat_unregister_driver);
EXPORT_SYMBOL(pci_compat_get_size);
EXPORT_SYMBOL(pci_compat_get_flags);
EXPORT_SYMBOL(pci_compat_set_power_state);
EXPORT_SYMBOL(pci_compat_enable_device);
EXPORT_SYMBOL(pci_compat_find_capability);
EXPORT_SYMBOL(pci_compat_get_driver_data);
EXPORT_SYMBOL(pci_compat_set_driver_data);
EXPORT_SYMBOL(compat_request_region);
EXPORT_SYMBOL(daemonize);
#endif

#endif /* kernel version < 2.3.0 */

