/*           -*- linux-c -*-
 * INET		An implementation of the TCP/IP protocol suite for the LINUX
 *		operating system.  INET is implemented using the  BSD Socket
 *		interface as the means of communication with the user level.
 *
 *		Ethernet-type device handling.
 *
 * Version:	@(#)vlan.c	started	12/21/98
 *                              
 * Authors:	Ben Greear <greearb@candelatech.com>, <greearb@agcs.com>
 * 
 * Fixes:
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 */

#include <asm/uaccess.h> /* for copy_from_user */
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <net/datalink.h>
#include <linux/mm.h>
#include <linux/in.h>
#include <linux/init.h>
#include <net/p8022.h>
#include <net/arp.h>

#include <linux/if_vlan.h>
#include "vlan.h"
#include "vlanproc.h"

extern int register_netdevice(struct device *dev); /* found in dev.c */
extern int unregister_netdevice(struct device *dev); /* found in dev.c */
extern int dev_new_index(void); /* dev.c */

extern int eth_header_parse(struct sk_buff *skb, unsigned char *haddr); /* eth.c */

extern struct Qdisc noqueue_qdisc;

/* Global VLAN variables */

/* Our listing of VLAN group(s) */
struct vlan_group *p802_1Q_vlan_list = NULL;

static char vlan_fullname[] = "802.1Q VLAN Support";
static unsigned int vlan_version = 1;
static unsigned int vlan_release = 0;
static char vlan_copyright[] = "(c) 2000 Ben Greear (GPL)";

/** These may be changed at run-time through IOCTLs */
unsigned short vlan_name_type = 0; /* determines interface naming scheme */
unsigned long vlan_bad_proto_recvd = 0; /* Counter for how many NON-VLAN protos we've received on a VLAN. */


static struct packet_type vlan_packet_type = 
{
        0,      /* MUTTER ntohs(ETH_P_802_1Q),*/
        NULL,
        vlan_dev_type_trans, /* VLAN receive method */
        NULL,
        NULL,
};

/* End of global variables definitions. */

#ifdef MODULE

/*
 *	Kernel Loadable Module Entry Points
 *
 *	Module 'insert' entry point.
 *	o print announcement
 *	o initialize static data
 *	o create /proc/net/vlan directory and static entries
 *
 *	Return:	0	Ok
 *		< 0	error.
 *	Context:	process
 */
int init_module	(void) {
        printk(VLAN_INF __FUNCTION__);

        vlan_proto_init(NULL);
        return 0;
}

/*
 *	Module 'remove' entry point.
 *	o delete /proc/net/router directory and static entries.
 */ 
void cleanup_module (void) {
        vlan_proto_cleanup(); // TODO: Define this so modules work.
}

#else


/** Non-module init entry point. */
__initfunc(void vlan_system_init(void)) {
        printk(VLAN_INF __FUNCTION__);

        /* protocol initialization */
        vlan_proto_init(NULL);

}
#endif

/*
 * Function vlan_proto_init (pro)
 *
 *    Initialize VLAN protocol layer, 
 *
 */
void vlan_proto_init(struct net_proto *pro) {

        int err;
        printk(VLAN_INF "%s v%u.%u %s\n",
               vlan_fullname, vlan_version, vlan_release, vlan_copyright);

        /* proc file system initialization */
        err = vlan_proc_init();
        if (err < 0) {
                printk(KERN_ERR __FUNCTION__
                       "%s: can't create entry in proc filesystem!\n", VLAN_NAME);
        }

        /* network byte order!! */
        vlan_packet_type.type = htons(ETH_P_802_1Q);
        dev_add_pack(&vlan_packet_type);
        printk(VLAN_INF "%s Initialization complete.\n", VLAN_NAME);
}



/**  Will search linearly for now, based on device index.  Could
 * hash, or directly link, this some day. --Ben
 */
struct vlan_group* vlan_find_group(int real_dev_ifindex) {
        struct vlan_group* grp = NULL;

        for (grp = p802_1Q_vlan_list;
             ((grp != NULL) && (grp->real_dev_ifindex != real_dev_ifindex));
             grp = grp->next) {
#ifdef VLAN_DEBUG
                printk(VLAN_DBG __FUNCTION__ ": grp_idx: %i  real_dev_idx: %i\n",
                       grp->real_dev_ifindex, real_dev_ifindex);
#endif
           ;
        } /* for */
         
        return grp;
}

/*  Find the protocol handler.  Assumes VID < 0xFFF.
 */
struct device *find_802_1Q_vlan_dev(struct device* real_dev, unsigned short VID) {

        struct vlan_group* grp = vlan_find_group(real_dev->ifindex);

#ifdef VLAN_DEBUG
        printk(VLAN_DBG __FUNCTION__ ": idx: %i  grp: %p\n", real_dev->ifindex, grp);
#endif

        /*  When here, we have found the correct group, if it exists. */

        if (grp) { /* then we found one */
                return grp->vlan_devices[VID]; /* return the vlan device */
        }//if
   
        return NULL;
}/* find_802_1Q_vlan_dev */



int unregister_802_1Q_vlan_dev(int real_dev_ifindex, unsigned short vlan_id) {
        struct vlan_group* grp;
        struct device* dev = NULL;

#ifdef VLAN_DEBUG
        printk(VLAN_DBG __FUNCTION__ ": VID: %i\n", vlan_id);
#endif
   
        /* sanity check */
        if ((vlan_id >= 0xFFF) || (vlan_id <= 0)) {
                return -EINVAL;
        }
   
        grp = vlan_find_group(real_dev_ifindex);
        /*  When here, we have found the correct group, if it exists. */

        if (grp) {
                dev = grp->vlan_devices[vlan_id];
                if (dev) {

                        /* Remove proc entry */
                        vlan_proc_rem_dev(dev);
         
                        /* take it out of our own structures */
                        grp->vlan_devices[vlan_id] = NULL;

                        /* Take it out of the global list of devices.
                         *  NOTE:  This deletes dev, don't access it again!!
                         */
                        unregister_netdevice(dev);

                }/* if */
        }/* if */
        return 0;
}/* unregister vlan device */



int unregister_802_1Q_vlan_device(const char* vlan_IF_name) {
        struct device* dev = NULL;

#ifdef VLAN_DEBUG
        printk(VLAN_DBG __FUNCTION__ ": unregister VLAN by name, name -:%s:-\n",
               vlan_IF_name);
#endif
   
        dev = dev_get(vlan_IF_name);

        if (dev && dev->vlan_dev) {
                return unregister_802_1Q_vlan_dev(dev->vlan_dev->real_dev->ifindex,
                                                  (unsigned short)(dev->vlan_dev->vlan_id));
        }
        else {
#ifdef VLAN_DEBUG
                printk(VLAN_DBG __FUNCTION__ ": WARNING: Could not find dev\n");
#endif
                return -EINVAL;
        }
}/* unregister vlan device */


/*
  TODO:  This for modules or something?? --BLG
  
  EXPORT_SYMBOL(register_802_1Q_vlan_device);
  EXPORT_SYMBOL(unregister_802_1Q_vlan_device);

*/

/*  Attach a VLAN device to a mac address (ie Ethernet Card).
 *  Returns the device that was created, or NULL if there was
 *  an error of some kind.
 */
struct device *register_802_1Q_vlan_device(const char* eth_IF_name,
                                           unsigned short VLAN_ID) {
        struct vlan_group* grp;
        struct device *new_dev;
        struct device *real_dev; /* the ethernet device */
        int malloc_size = 0;
   
#ifdef VLAN_DEBUG
        printk(VLAN_DBG __FUNCTION__ ": if_name -:%s:-  vid: %i\n",
               eth_IF_name, VLAN_ID);
#endif
   
        /* find the device relating to eth_IF_name.
    * TODO:  Make sure it's an ethernet device. */
        real_dev = dev_get(eth_IF_name);

        if (real_dev != NULL) {
                /* printk(KERN_ALERT "Found real_dev"); */
            
                if ((VLAN_ID > 0) && (VLAN_ID < 0xFFF)) {

                        /* printk(KERN_ALERT "VID is in range"); */

                        if (find_802_1Q_vlan_dev(real_dev, VLAN_ID)) {
                                /* was already registered. */
                                printk(VLAN_DBG __FUNCTION__ ": ALREADY had VLAN registered\n");
                                return NULL;
                        }

                        malloc_size = (sizeof(struct device));

                        new_dev = (struct device*) kmalloc(malloc_size, GFP_KERNEL);
                        VLAN_MEM_DBG("device malloc, addr: %p  size: %i\n", new_dev, malloc_size);
         
                        if (new_dev != NULL) {
                                /* printk(KERN_ALERT "Got a new device.."); */
            
                                memset(new_dev, 0, malloc_size); /* zero everything out */

                                /* set us up to not use a Qdisc, as the underlying Hardware device
                                 * can do all the queueing we could want.
                                 */
                                new_dev->qdisc_sleeping = &noqueue_qdisc;
            
                                /* Gotta set up the fields for the device. */
                                new_dev->name = (char*)(kmalloc(IFNAMSIZ + 1, GFP_KERNEL));
                                VLAN_MEM_DBG("new_dev->name malloc, addr: %p  size: %i\n", new_dev->name, IFNAMSIZ + 1);
            
                                if (new_dev->name) {
                                        memset(new_dev->name, 0, IFNAMSIZ + 1); /* zero everything out */
                                }
                                else {
                                        kfree(new_dev);
                                        VLAN_FMEM_DBG("new_dev free, addr: %p\n", new_dev);
                                        return NULL;
                                }

                                if (vlan_name_type == VLAN_NAME_TYPE_RAW_PLUS_VID) {
                                        /* name will look like:  eth1.0005 */
                                        sprintf(new_dev->name, "%s.%.4i", real_dev->name, VLAN_ID);
                                }
                                else if (vlan_name_type == VLAN_NAME_TYPE_PLUS_VID_NO_PAD) {
                                        /* Put our vlan.VID in the name.  Name will look like:  vlan5 */
                                        sprintf(new_dev->name, "vlan%i", VLAN_ID);
                                }
                                else if (vlan_name_type == VLAN_NAME_TYPE_RAW_PLUS_VID_NO_PAD) {
                                        /* Put our vlan.VID in the name.  Name will look like:  eth0.5 */
                                        sprintf(new_dev->name, "%s.%i", real_dev->name, VLAN_ID);
                                }
                                else { /* (vlan_name_type == VLAN_NAME_TYPE_PLUS_VID) { */
                                        /* Put our vlan.VID in the name.  Name will look like:  vlan0005 */
                                        /* default case */
                                        sprintf(new_dev->name, "vlan%.4i", VLAN_ID);
                                }
            
            
#ifdef VLAN_DEBUG
                                printk(VLAN_DBG "Allocated new name -:%s:-\n", new_dev->name);
#endif
                                /* set up method calls */
                                new_dev->init = vlan_dev_init;
                                new_dev->destructor = vlan_dev_destruct;
            
                                /* new_dev->ifindex = 0;  it will be set when added to
                                 * the global list.
                                 * iflink is set as well. */
            
                                new_dev->get_stats = vlan_dev_get_stats;
            
                                /* IFF_BROADCAST|IFF_MULTICAST; ??? */
                                new_dev->flags = real_dev->flags;
                                new_dev->flags &= ~IFF_UP;
            
                                /* need 4 bytes for extra VLAN header info, hope
                                 * underlying device can handle it. */
                                new_dev->mtu = real_dev->mtu;
            
                                new_dev->type = real_dev->type; /* TODO: is this true? */

                                /* Regular ethernet + 4 bytes (18 total). */
                                new_dev->hard_header_len = VLAN_ETH_HLEN;

                                new_dev->priv = kmalloc(sizeof(struct net_device_stats),
                                                        GFP_KERNEL);
                                VLAN_MEM_DBG("new_dev->priv malloc, addr: %p  size: %i\n", new_dev->priv,
                                             sizeof(struct net_device_stats));
            
                                if (new_dev->priv) {
                                        memset(new_dev->priv, 0, sizeof(struct net_device_stats));
                                }//if

                                memcpy(new_dev->broadcast, real_dev->broadcast, real_dev->addr_len);
                                memcpy(new_dev->dev_addr, real_dev->dev_addr, real_dev->addr_len);
                                new_dev->addr_len = real_dev->addr_len;

                                new_dev->open = vlan_dev_open;
                                new_dev->stop = vlan_dev_stop;
                                new_dev->hard_header = vlan_dev_hard_header;
                                /*new_dev->hard_header_cache = vlan_header_cache;*/
                                /*new_dev->header_cache_update = vlan_header_cache_update;*/
                                new_dev->hard_start_xmit = vlan_dev_hard_start_xmit;
                                new_dev->rebuild_header = vlan_dev_rebuild_header;
                                new_dev->hard_header_parse = eth_header_parse; /* trivial. */
                                new_dev->set_mac_address = vlan_dev_set_mac_address;
                                new_dev->set_multicast_list = vlan_dev_set_multicast_list;

                                new_dev->vlan_dev = (struct vlan_dev_info*) kmalloc(sizeof(struct vlan_dev_info),
                                                                                    GFP_KERNEL);
                                VLAN_MEM_DBG("new_dev->vlan_dev malloc, addr: %p  size: %i\n", new_dev->vlan_dev,
                                             sizeof(struct vlan_dev_info));
                                if (new_dev->vlan_dev == NULL) {
                                        kfree(new_dev->priv);
                                        VLAN_FMEM_DBG("new_dev->priv free, addr: %p\n", new_dev->priv);
                                        kfree(new_dev->name);
                                        VLAN_FMEM_DBG("new_dev->name free, addr: %p\n", new_dev->name);
                                        kfree(new_dev);
                                        VLAN_FMEM_DBG("new_dev free, addr: %p\n", new_dev);
                                        return NULL;
                                }
                                else {
                                        /* Initialize it. */
                                        memset(new_dev->vlan_dev, 0, sizeof(struct vlan_dev_info));
               
                                        new_dev->vlan_dev->vlan_id = VLAN_ID; /* 1 through 0xFFF */
                                        /* TODO: have to be careful deleting real devices now. */
                                        new_dev->vlan_dev->real_dev = real_dev;

                                        memset(&(new_dev->vlan_dev->dent), 0, sizeof(struct proc_dir_entry));
                                }
            
                                /* So, got the sucker initialized, now lets place it into our local
                                 * structure.
                                 */

                                grp = vlan_find_group(real_dev->ifindex);

                                /*  When here, we have found the correct group, if it exists. */

                                if (!grp) { /* need to add a new group */
                                        /* printk(VLAN_DBG "VLAN REGISTER: "
                                           "Need to add new vlan group.\n");*/

                                        grp = kmalloc(sizeof(struct vlan_group), GFP_KERNEL);
                                        VLAN_MEM_DBG("grp malloc, addr: %p  size: %i\n", grp, sizeof(struct vlan_group));

                                        if (grp) {
                                                printk(KERN_ALERT "VLAN REGISTER:  Allocated new group, idx: %i\n",
                                                       real_dev->ifindex);
                                                memset(grp, 0, sizeof(struct vlan_group));
                                                grp->real_dev_ifindex = real_dev->ifindex;
                                                grp->next = p802_1Q_vlan_list;
                                                p802_1Q_vlan_list = grp;
                                        }
                                        else {
                                                kfree(new_dev->name);
                                                VLAN_FMEM_DBG("new_dev->name free, addr: %p\n", new_dev->name);
                                                kfree(new_dev->priv);
                                                VLAN_FMEM_DBG("new_dev->priv free, addr: %p\n", new_dev->priv);
                                                kfree(new_dev);
                                                VLAN_FMEM_DBG("new_dev free, addr: %p\n", new_dev);
                                                return NULL;
                                        }
                                }/* if */
            
                                grp->vlan_devices[VLAN_ID] = new_dev;

                                /*  Now, add it to the global list of devices. */
                                /* printk(KERN_ALERT "Registering new device."); */
                                register_netdevice(new_dev);
                                vlan_proc_add_dev(new_dev); /* create it's proc entry */
                                return new_dev;
                        }
                }//if
        }//if

        return NULL;
}/* register (create) VLAN device */
