/* -*- linux-c -*-
 * INET		An implementation of the TCP/IP protocol suite for the LINUX
 *		operating system.  INET is implemented using the  BSD Socket
 *		interface as the means of communication with the user level.
 *
 *		Ethernet-type device handling.
 *
 * Version:	@(#)vlan_dev.c	Started	3/29/99
 *
 * Authors:	Ben Greear <greearb@candelatech.com>, <greearb@agcs.com>
 * 
 * Fixes:
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 */

#include <asm/uaccess.h> /* for copy_from_user */
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <net/datalink.h>
#include <linux/mm.h>
#include <linux/in.h>
#include <linux/init.h>
#include <net/p8022.h>
#include <net/arp.h>
#include "vlan.h"
#include "vlanproc.h"
#include <linux/if_vlan.h>
#include <net/ip.h>
#include <asm/checksum.h>


struct net_device_stats* vlan_dev_get_stats(struct device* dev) {
        return (struct net_device_stats*)(dev->priv);
}


/*
 *	Rebuild the Ethernet MAC header. This is called after an ARP
 *	(or in future other address resolution) has completed on this
 *	sk_buff. We now let ARP fill in the other fields.
 *
 *	This routine CANNOT use cached dst->neigh!
 *	Really, it is used only when dst->neigh is wrong.
 *
 * TODO:  This needs a checkup, I'm ignorant here. --BLG
 */
int vlan_dev_rebuild_header(struct sk_buff *skb) {

        struct device *dev = skb->dev;
        struct vlan_ethhdr *veth = (struct vlan_ethhdr*)(skb->data);

        switch (veth->h_vlan_encapsulated_proto)
        {
#ifdef CONFIG_INET
        case __constant_htons(ETH_P_IP):

                /* TODO:  Confirm this will work with VLAN headers... */
                return arp_find(veth->h_dest, skb);
#endif	
        default:
                printk(VLAN_DBG
                       "%s: unable to resolve type %X addresses.\n", 
                       dev->name, (int)veth->h_vlan_encapsulated_proto);
         
                memcpy(veth->h_source, dev->dev_addr, ETH_ALEN);
                break;
        }/* switch */

        return 0;
}/* vlan_dev_rebuild_header */



/*
 *	Determine the packet's protocol ID. The rule here is that we 
 *	assume 802.3 if the type field is short enough to be a length.
 *	This is normal practice and works for any 'now in use' protocol.
 *
 *  Also, at this point we assume that we ARE dealing exclusively with
 *  VLAN packets, or packets that should be made into VLAN packets based
 *  on a default VLAN ID.
 *
 *  NOTE:  Should be similar to ethernet/eth.c.
 *
 *  SANITY NOTE:  This method is called when a packet is moving up the stack
 *                towards userland.  To get here, it would have already passed
 *                through the ethernet/eth.c eth_type_trans() method.
 */
int vlan_dev_type_trans(struct sk_buff *skb, struct device *dev,
                        struct packet_type* ptype) {
        unsigned char* rawp = NULL;
        struct vlan_ethhdr *veth = (struct vlan_ethhdr*)(skb->mac.ethernet);
        unsigned short vid = 0;
        struct net_device_stats* stats;
   
        /* Do we have a VLAN packet?  If not, then throw it away, after printing an error.
         *
         */
        if (veth->h_vlan_proto != __constant_htons(ETH_P_802_1Q)) {
                printk(VLAN_INF __FUNCTION__ ": VLAN device received NON-VLAN protocol: %hx\n", htons(veth->h_vlan_proto));
                vlan_bad_proto_recvd++;
                kfree_skb(skb);
                return -EINVAL;
        }
        else {
                vid = ((unsigned short)(ntohs(veth->h_vlan_TCI)) & 0xFFF);
        }

#ifdef VLAN_DEBUG
        printk(VLAN_DBG __FUNCTION__ ": skb: %p vlan_id: %hx dev: %s, encap_proto: %hx\n",
               skb, vid, dev->name, veth->h_vlan_encapsulated_proto);
#endif
   
        /*  Ok, we will find the correct VLAN device, strip the header,
            and then go on as usual.
        */
        
        /* we have 12 bits of vlan ID. */
        /* If it's NULL, we will tag the skb to be junked below */
        skb->dev = find_802_1Q_vlan_dev(dev, vid);

        if (!skb->dev) {
#ifdef VLAN_DEBUG
                printk(VLAN_DBG __FUNCTION__ ": ERROR:  No device for VID: %i on dev: %s [%i]\n",
                       (unsigned int)(vid), dev->name, dev->ifindex);
#endif
                kfree_skb(skb);
                return -1;
        }

        stats = (struct net_device_stats*)(skb->dev->priv);

        /*
         * Deal with ingress priority mapping.
         */
        skb->priority = skb->dev->vlan_dev->ingress_priority_map[(ntohs(veth->h_vlan_TCI) >> 13) & 0x7];
   
#ifdef VLAN_DEBUG
        printk(VLAN_DBG __FUNCTION__ ": priority: %lu  for TCI: %hu (hbo) on vlan_dev: %s\n",
               (unsigned long)(skb->priority), ntohs(veth->h_vlan_TCI), skb->dev->name);
#endif

        /* Bump the rx counters for the VLAN device. */
        stats->rx_packets++;
        stats->rx_bytes += skb->len;
      
        /* NOTE: The underlying device SHOULD NOT PULL THE MAC BYTES OFF.
           (it doesn't seem to.)
        */
        skb_pull(skb, VLAN_ETH_HLEN); /* take off the VLAN header */
   
   
        /*  VLAN and regular Ethernet headers have the addresses in the same place.
         *  TODO:  Add code to deal with VLAN control packets?? --BLG
         *         Is there such a thing??
         */
        if (*(veth->h_dest) & 1) {
                stats->multicast++;
                if (memcmp(veth->h_dest, dev->broadcast, ETH_ALEN) == 0)
                        skb->pkt_type = PACKET_BROADCAST;
                else
                        skb->pkt_type = PACKET_MULTICAST;
        }
	
        /*
         *	This ALLMULTI check should be redundant by 1.4
         *	so don't forget to remove it.
         *
         *	Seems, you forgot to remove it. All silly devices
         *	seems to set IFF_PROMISC.
         */
   
        else if (dev->flags & (IFF_PROMISC/*|IFF_ALLMULTI*/)) {
                if (memcmp(veth->h_dest, dev->dev_addr, ETH_ALEN) != 0)
                        skb->pkt_type = PACKET_OTHERHOST;
        }

        /*  Was a VLAN packet, grab the encapsulated protocol, which the layer
         * three protocols care about.
         */
        if (ntohs(veth->h_vlan_encapsulated_proto) >= 1536) {
           
           skb->protocol = veth->h_vlan_encapsulated_proto;
           /* place it back on the queue to be handled by true layer 3 protocols.
            */
           
           /* See if we are configured to re-write the VLAN header to make it look like
            * ethernet...
            */
           if (skb->dev->vlan_dev->flags & 1) {
                   /* Lifted from Gleb's VLAN code... */
                   memmove(skb->data - (VLAN_ETH_HLEN - 4), skb->data - VLAN_ETH_HLEN, 12);
                   skb->mac.raw += 4;
           }
           netif_rx(skb);
           return 0;
        }

        rawp = skb->data;
	
        /*
         *	This is a magic hack to spot IPX packets. Older Novell breaks
         *	the protocol design and runs IPX over 802.3 without an 802.2 LLC
         *	layer. We look for FFFF which isn't a used 802.2 SSAP/DSAP. This
         *	won't work for fault tolerant netware but does for the rest.
         */
        if (*(unsigned short *)rawp == 0xFFFF) {
                skb->protocol = __constant_htons(ETH_P_802_3);
                /* place it back on the queue to be handled by true layer 3 protocols.
                 */
      
                /* See if we are configured to re-write the VLAN header to make it look like
                 * ethernet...
                 */
                if (skb->dev->vlan_dev->flags & 1) {
                        /* Lifted from Gleb's VLAN code... */
                        memmove(skb->data - (VLAN_ETH_HLEN - 4), skb->data - VLAN_ETH_HLEN, 12);
                        skb->mac.raw += 4;
                }
                netif_rx(skb);
                return 0;
        }
		
        /*
         *	Real 802.2 LLC
         */
        skb->protocol = __constant_htons(ETH_P_802_2);
        /* place it back on the queue to be handled by upper layer protocols.
         */

        /* See if we are configured to re-write the VLAN header to make it look like
         * ethernet...
         */
        if (skb->dev->vlan_dev->flags & 1) {
                /* Lifted from Gleb's VLAN code... */
                memmove(skb->data - (VLAN_ETH_HLEN - 4), skb->data - VLAN_ETH_HLEN, 12);
                skb->mac.raw += 4;
        }
        netif_rx(skb);
        return 0;
}


/*
 *	 Create the Ethernet VLAN MAC header for an arbitrary protocol layer 
 *
 *	saddr=NULL	means use device source address
 *	daddr=NULL	means leave destination address (eg unresolved arp)
 *
 *  This is called when the SKB is moving down the stack towards the
 *  physical devices.
 */
int vlan_dev_hard_header(struct sk_buff *skb, struct device *dev,
                         unsigned short type, void *daddr, void *saddr,
                         unsigned len) {
        struct vlan_ethhdr *veth;
        unsigned short veth_TCI = 0;
        
#ifdef VLAN_DEBUG
        printk(VLAN_DBG __FUNCTION__ ": skb: %p type: %hx len: %x vlan_id: %hx, daddr: %p\n",
               skb, type, len, dev->vlan_dev->vlan_id, daddr);
#endif
   
        veth = (struct vlan_ethhdr*)skb_push(skb, VLAN_ETH_HLEN);

        /* build the four bytes that make this a VLAN header. */

        /* first, the ethernet type */
        veth->h_vlan_proto = __constant_htons(ETH_P_802_1Q);

        /* Now, construct the second two bytes. This field looks something
         * like:
         * usr_priority: 3 bits  (high bits)
         * CFI           1 bit
         * VLAN ID       12 bits (low bits)
         *
         */
        veth_TCI = dev->vlan_dev->vlan_id;
        veth_TCI |= vlan_dev_get_egress_qos_mask(dev, skb);
   
        veth->h_vlan_TCI = htons(veth_TCI);
           
        /* Rest should be the same as a normal header. */
        /* 
         *  Set the protocol type. For a packet of type ETH_P_802_3 we put the length
         *  in here instead. It is up to the 802.2 layer to carry protocol information.
         *
         */

        if (type != ETH_P_802_3) 
                veth->h_vlan_encapsulated_proto = htons(type);
        else
                veth->h_vlan_encapsulated_proto = htons(len);

        /*
         *	Set the source hardware address. 
         */
           
        if (saddr)
                memcpy(veth->h_source, saddr, ETH_ALEN);
        else
                memcpy(veth->h_source, dev->dev_addr, ETH_ALEN);
   
        /*
         *	Anyway, the loopback-device should never use this function... 
         *   This is especially true with VLAN's. --BLG
         */
   
        if (dev->flags & (IFF_LOOPBACK | IFF_NOARP)) {
                memset(veth->h_dest, 0, ETH_ALEN);
                return (VLAN_ETH_HLEN); /* was: dev->hard_header_len */
        }
   
        if (daddr) {
                memcpy(veth->h_dest, daddr, ETH_ALEN);
                return (VLAN_ETH_HLEN); /* was: dev->hard_header_len */
        }
   
        return -(VLAN_ETH_HLEN); /* was: dev->hard_header_len */
   
} /* vlan_hard_header, put on the VLAN hardware header */


int vlan_dev_hard_start_xmit(struct sk_buff *skb, struct device *dev) {
        struct net_device_stats* stats = (struct net_device_stats*)(dev->priv);
        struct vlan_ethhdr *veth = (struct vlan_ethhdr*)(skb->data);
        
                /* Handle non-VLAN frames if they are sent to us, for example by DHCP. */
        if (veth->h_vlan_proto != __constant_htons(ETH_P_802_1Q)) {
                /* This is not a VLAN frame...but we can fix that! */
                unsigned short veth_TCI = 0;
                dev->vlan_dev->cnt_encap_on_xmit++;
                
                if (skb_headroom(skb) < 4) {
                        struct sk_buff* sk_tmp = skb;
                        skb = skb_realloc_headroom(sk_tmp, 4);
                        kfree_skb(sk_tmp);
                        if (skb == NULL) {
                                stats->tx_dropped++;
                                kfree_skb(sk_tmp);
                                return -ENOMEM;
                        }
                        dev->vlan_dev->cnt_inc_headroom_on_tx++;
                }
                else {
                        if( !(skb = skb_unshare(skb, GFP_ATOMIC)) ) {
                                printk(KERN_ERR "vlan: failed to unshare skbuff\n");
                                stats->tx_dropped++;
                                return -ENOMEM;
                        }
                }
                veth = (struct vlan_ethhdr*)skb_push(skb, 4);

                /* Move the mac addresses to the beginning of the new header. */
                memmove(skb->data, skb->data + 4, 12);

                /* first, the ethernet type */
                veth->h_vlan_proto = __constant_htons(ETH_P_802_1Q);

                /* Now, construct the second two bytes. This field looks something
                 * like:
                 * usr_priority: 3 bits  (high bits)
                 * CFI           1 bit
                 * VLAN ID       12 bits (low bits)
                 *
                 */
                veth_TCI = dev->vlan_dev->vlan_id;
                veth_TCI |= vlan_dev_get_egress_qos_mask(dev, skb);
   
                veth->h_vlan_TCI = htons(veth_TCI);
        }/* If we needed to encapsulate the frame */

        skb->dev = dev->vlan_dev->real_dev;


#ifdef VLAN_DEBUG
        printk(VLAN_DBG __FUNCTION__ ": about to send skb: %p  to dev: %s\n", skb, skb->dev->name);
#endif

        dev_queue_xmit(skb);
        stats->tx_packets++; /* for statics only */
        stats->tx_bytes += skb->len;
        return 0;
}/* vlan_dev_hard_start_xmit */


int vlan_dev_change_mtu(struct device *dev, int new_mtu) {
        /* TODO: gotta make sure the underlying layer can handle it,
         * maybe an IFF_VLAN_CAPABLE flag for devices?
    */

        dev->mtu = new_mtu;
        return new_mtu;
}

int vlan_dev_open(struct device* dev) {
        dev->flags |= IFF_UP;
        return 0;
}

int vlan_dev_stop(struct device* dev) {
        dev->flags &= ~IFF_UP;
        return 0;
}

int vlan_dev_init(struct device* dev) {
        /* TODO:  figure this out, maybe do nothing?? */
        return 0;
}

void vlan_dev_destruct(struct device* dev) {
        kfree(dev->name);
        VLAN_FMEM_DBG("dev->name free, addr: %p\n", dev->name);
        dev->name = NULL; /* better safe than hosed */
   
        kfree(dev->priv);
        VLAN_FMEM_DBG("dev->priv free, addr: %p\n", dev->priv);
        dev->priv = NULL;
   
        kfree(dev->vlan_dev);
        VLAN_FMEM_DBG("dev->vlan_dev free, addr: %p\n", dev->vlan_dev);
        dev->vlan_dev = NULL;

        kfree(dev);
        VLAN_FMEM_DBG("device free, addr: %p\n", dev);
        dev = NULL;

        return;
}


/* TODO:  Not to sure if the VLAN stuff works here.  Need to understand
 * this better. --BLG
 */
/*
int vlan_dev_header_cache(struct neighbour *neigh, struct hh_cache *hh) {
   unsigned short type = hh->hh_type;
   struct vlan_ethhdr *veth = (struct vlan_ethhdr*)(((u8*)hh->hh_data) + 2);
   struct device *dev = neigh->dev;

   if (type == __constant_htons(ETH_P_802_3)) {
      return -1;
   }

   veth->h_vlan_proto = __constant_htons(ETH_P_802_1Q);
   memcpy(veth->h_source, dev->dev_addr, ETH_ALEN);
   memcpy(veth->h_dest, neigh->ha, ETH_ALEN);
      
   * VLAN specific attributes. *
   veth->h_vlan_TCI = htons(dev->VLAN_id); * TODO:  Add priority control (high 3 bits.) * 
   veth->h_vlan_encapsulated_proto = type; * should already be in network order *
        
   return 0;
}
*/

/*
 * Called by Address Resolution module to notify changes in address.
 */
/*
void vlan_dev_header_cache_update(struct hh_cache *hh, struct device *dev,
                                  unsigned char * haddr) {
   memcpy(((u8*)hh->hh_data) + 2, haddr, VLAN_ETH_HLEN);
}
*/

#ifndef CONFIG_IP_ROUTER

/*
 *	Copy from an ethernet device memory space to an sk_buff while
 *      checksumming if IP
 *
 *  TODO:  Find out who calls this:  This was lifted from eth.c, and
 *         was called eth_copy_and_sum. --BLG
 */

void vlan_dev_copy_and_sum(struct sk_buff *dest, unsigned char *src,
                           int length, int base) {
        struct vlan_ethhdr* veth;
        struct iphdr *iph;
        int ip_length;
   
        veth = (struct vlan_ethhdr*)(src);
   
        /*  This grabs the VLAN part of the header too. */
        if (veth->h_vlan_encapsulated_proto != __constant_htons(ETH_P_IP)) {
                memcpy(dest->data, src, length);
                return;
        }
      
        /*
    * We have to watch for padded packets. The csum doesn't include the
    * padding, and there is no point in copying the padding anyway.
    * We have to use the smaller of length and ip_length because it
    * can happen that ip_length > length.
    */
      
   /* ethernet is always >= 34 */
        memcpy(dest->data, src, sizeof(struct iphdr) + VLAN_ETH_HLEN);
      
        length -= sizeof(struct iphdr) + VLAN_ETH_HLEN;
        iph = (struct iphdr*)(src + VLAN_ETH_HLEN);
        ip_length = ntohs(iph->tot_len) - sizeof(struct iphdr);
   
   /* Also watch out for bogons - min IP size is 8 (rfc-1042) */
        if ((ip_length <= length) && (ip_length > 7))
                length=ip_length;
   
        dest->csum = csum_partial_copy(src + sizeof(struct iphdr) + VLAN_ETH_HLEN,
                                       dest->data + sizeof(struct iphdr) + VLAN_ETH_HLEN,
                                       length, base);
        dest->ip_summed=1;

} /* vlan_copy_and_sum */

#endif //! CONFIG_IP_ROUTER


int vlan_dev_set_ingress_priority(char* dev_name, __u32 skb_prio, short vlan_prio) {
        struct device* dev = dev_get(dev_name);
   
        if (dev) {
                if (dev->vlan_dev) { /* can't put a dflt ID on a vlan device */
                        /* see if a priority mapping exists.. */
                        dev->vlan_dev->ingress_priority_map[vlan_prio & 0x7] = skb_prio;
                        return 0;
                }
        }
        return -EINVAL;
}

int vlan_dev_set_egress_priority(char* dev_name, __u32 skb_prio, short vlan_prio) {
        struct device* dev = dev_get(dev_name);
        struct vlan_priority_tci_mapping* mp = NULL;
        struct vlan_priority_tci_mapping* np;
   
        if (dev) {
                if (dev->vlan_dev) { /* can't put a dflt ID on a vlan device */
                        /* see if a priority mapping exists.. */
                        mp = dev->vlan_dev->egress_priority_map[skb_prio & 0xF];
                        while (mp) {
                                if (mp->priority == skb_prio) {
                                        mp->vlan_qos = ((vlan_prio << 13) & 0xE000);
                                        return 0;
                                }
                        }
                        /* create a new mapping then. */
                        mp = dev->vlan_dev->egress_priority_map[skb_prio & 0xF];
                        np = kmalloc(sizeof(struct vlan_priority_tci_mapping), GFP_KERNEL);
                        if (np) {
                                np->next = mp;
                                np->priority = skb_prio;
                                np->vlan_qos = ((vlan_prio << 13) & 0xE000);
                                dev->vlan_dev->egress_priority_map[skb_prio & 0xF] = np;
                                return 0;
                        }
                        else {
                                return -ENOBUFS;
                        }
                }
        }
        return -EINVAL;
}

/* Flags are defined in the vlan_dev_info class in include/linux/if_vlan.h file. */
int vlan_dev_set_vlan_flag(char* dev_name, __u32 flag, short flag_val) {
        struct device* dev = dev_get(dev_name);
   
        if (dev) {
                if (dev->vlan_dev) {
                        /* verify flag is supported */
                        if (flag == 1) {
                                if (flag_val) {
                                        dev->vlan_dev->flags |= 1;
                                }
                                else {
                                        dev->vlan_dev->flags &= ~1;
                                }
                                return 0;
                        }
                        else {
                                return -EINVAL;
                        }
                }/* if it's a vlan device */
        }/* if we found the device */
        return -EINVAL;
}


int vlan_dev_set_mac_address(struct device *dev, void* addr_struct_p) {
        int i;
        struct sockaddr *addr = (struct sockaddr*)(addr_struct_p);

        if (dev->start) {
                return -EBUSY;
        }

	memcpy(dev->dev_addr, addr->sa_data, dev->addr_len);
        
        printk("%s: Setting MAC address to ", dev->name);
        for (i = 0; i < 6; i++) {
                printk(" %2.2x", dev->dev_addr[i]);
        }
        printk(".\n");

        if (memcmp(dev->vlan_dev->real_dev->dev_addr, dev->dev_addr, dev->addr_len) != 0) {
                if (dev->vlan_dev->real_dev->flags & IFF_PROMISC) {
                        /* Already promiscious...leave it alone. */
                        printk("VLAN (%s):  Good, underlying device (%s) is already promiscious.\n",
                               dev->name, dev->vlan_dev->real_dev->name);
                }
                else {
                        int flgs = dev->vlan_dev->real_dev->flags;
                        printk("VLAN (%s):  Setting underlying device (%s) to promiscious mode.\n",
                               dev->name, dev->vlan_dev->real_dev->name);
                        flgs |= IFF_PROMISC;
                        dev_change_flags(dev->vlan_dev->real_dev, flgs);
                        /*  This should work, but doesn't:
                            dev_set_promiscuity(dev->vlan_dev->real_dev, 1);
                        */
                }
        }
        else {
                printk("VLAN (%s):  Underlying device (%s) has same MAC, not checking promiscious mode.\n",
                       dev->name, dev->vlan_dev->real_dev->name);
        }
        return 0;
}


/** Taken from Gleb + Lennert's VLAN code, and modified... */
void vlan_dev_set_multicast_list(struct device *vlan_dev) {
        struct dev_mc_list *dmi;
        struct device *real_dev;
        int inc;

        if (vlan_dev && vlan_dev->vlan_dev) {
                /* Then it's a real vlan device, as far as we can tell.. */
                real_dev = vlan_dev->vlan_dev->real_dev;

                /* compare the current promiscuity to the last promisc we had.. */
                inc = vlan_dev->promiscuity - vlan_dev->vlan_dev->old_promiscuity;
   
                if (inc) {
                        printk(KERN_INFO "vlan: dev_set_promiscuity(master, %d)\n", inc);
                        dev_set_promiscuity(real_dev, inc); /* found in dev.c */
                        vlan_dev->vlan_dev->old_promiscuity = vlan_dev->promiscuity;
                }
   
                inc = vlan_dev->allmulti - vlan_dev->vlan_dev->old_allmulti;
   
                if (inc) {
                        printk(KERN_INFO "vlan: dev_set_allmulti(master, %d)\n", inc);
                        dev_set_allmulti(real_dev, inc); /* dev.c */
                        vlan_dev->vlan_dev->old_allmulti = vlan_dev->allmulti;
                }
   
                /* looking for addresses to add to master's list */
                for (dmi = vlan_dev->mc_list; dmi!=NULL; dmi=dmi->next) {
                        if (vlan_should_add_mc(dmi, vlan_dev->vlan_dev->old_mc_list)) {
                                dev_mc_add(real_dev, dmi->dmi_addr, dmi->dmi_addrlen, 0);
                                printk(KERN_INFO "vlan: add %.2x:%.2x:%.2x:%.2x:%.2x:%.2x mcast address to master interface\n",
                                       dmi->dmi_addr[0],
                                       dmi->dmi_addr[1],
                                       dmi->dmi_addr[2],
                                       dmi->dmi_addr[3],
                                       dmi->dmi_addr[4],
                                       dmi->dmi_addr[5]);
                        }
                }
   
                /* looking for addresses to delete from master's list */
                for (dmi = vlan_dev->mc_list; dmi!=NULL; dmi=dmi->next) {
                        if (vlan_should_add_mc(dmi, vlan_dev->mc_list)) {
                                /* if we think we should add it to the new list, then we should really
                                 * delete it from the real list on the underlying device.
                                 */
                                dev_mc_delete(real_dev, dmi->dmi_addr, dmi->dmi_addrlen, 0);
                                printk(KERN_INFO "vlan: del %.2x:%.2x:%.2x:%.2x:%.2x:%.2x mcast address from master interface\n",
                                       dmi->dmi_addr[0],
                                       dmi->dmi_addr[1],
                                       dmi->dmi_addr[2],
                                       dmi->dmi_addr[3],
                                       dmi->dmi_addr[4],
                                       dmi->dmi_addr[5]);
                        }
                }
   
                /* save multicast list */
                vlan_copy_mc_list(vlan_dev->mc_list, vlan_dev->vlan_dev);
        }/* if we were sent a valid device */
}/* vlan_dev_set_multicast */


/** dmi is a single entry into a dev_mc_list, a single node.  mc_list is
 *  an entire list, and we'll iterate through it.
 */
int vlan_should_add_mc(struct dev_mc_list *dmi, struct dev_mc_list *mc_list) {
        struct dev_mc_list *idmi; /* iterator */

        for (idmi=mc_list; idmi!=NULL;) {
                if (vlan_dmi_equals(dmi, idmi)) {
                        if (dmi->dmi_users > idmi->dmi_users)
                                return 1;
                        else
                                return 0;
                }
                else {
                        idmi = idmi->next;
                }
        }
   
        return 1;
}


void vlan_copy_mc_list(struct dev_mc_list *mc_list, struct vlan_dev_info *vlan_info) {
        struct dev_mc_list *dmi, *new_dmi;
   
        vlan_destroy_mc_list(vlan_info->old_mc_list);
        vlan_info->old_mc_list = NULL;

        for (dmi=mc_list; dmi!=NULL; dmi=dmi->next) {
                new_dmi = kmalloc(sizeof(*new_dmi), GFP_KERNEL);
                if (new_dmi == NULL) {
                        printk(KERN_ERR "vlan: cannot allocate memory. Multicast may not work properly from now.\n");
                        return;
                }
      
                new_dmi->next = vlan_info->old_mc_list;
                vlan_info->old_mc_list = new_dmi;
      
                new_dmi->dmi_addrlen = dmi->dmi_addrlen;
                memcpy(new_dmi->dmi_addr, dmi->dmi_addr, dmi->dmi_addrlen);
                new_dmi->dmi_users = dmi->dmi_users;
                new_dmi->dmi_gusers = dmi->dmi_gusers;
        }
}

void vlan_flush_mc_list(struct device *dev) {
        struct dev_mc_list *dmi = dev->mc_list;

        while (dmi) {
                dev_mc_delete(dev, dmi->dmi_addr, dmi->dmi_addrlen, 0);
                printk(KERN_INFO "vlan: del %.2x:%.2x:%.2x:%.2x:%.2x:%.2x mcast address from master interface\n",
                       dmi->dmi_addr[0],
                       dmi->dmi_addr[1],
                       dmi->dmi_addr[2],
                       dmi->dmi_addr[3],
                       dmi->dmi_addr[4],
                       dmi->dmi_addr[5]);
                dmi = dev->mc_list;
        }
   
        vlan_destroy_mc_list(dev->mc_list);
        if (dev->vlan_dev) {
                vlan_destroy_mc_list(dev->vlan_dev->old_mc_list);
                dev->vlan_dev->old_mc_list = NULL;
        }
        dev->mc_list = NULL;
}/* vlan_flush_mc_list */
