/* * -*- linux-c -*- */
/*****************************************************************************
 * vlanproc.c	VLAN Module. /proc filesystem interface.
*
* Author:	Ben Greear, <greearb@candelatech.com> coppied from wanproc.c
*               by: Gene Kozin	<genek@compuserve.com>
*
* Copyright:	(c) 1998-2000 Ben Greear
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
* Jan 20, 1998        Ben Greear     Initial Version
*****************************************************************************/

#include <linux/config.h>
#include <linux/stddef.h>	/* offsetof(), etc. */
#include <linux/errno.h>	/* return codes */
#include <linux/kernel.h>
#include <linux/malloc.h>	/* kmalloc(), kfree() */
#include <linux/mm.h>		/* verify_area(), etc. */
#include <linux/string.h>	/* inline mem*, str* functions */
#include <linux/init.h>		/* __initfunc et al. */
#include <asm/segment.h>	/* kernel <-> user copy */
#include <asm/byteorder.h>	/* htons(), etc. */
#include <asm/uaccess.h>	/* copy_to_user */
#include <asm/io.h>
#include <linux/proc_fs.h>
#include <linux/fs.h>
#include <linux/netdevice.h>
#include <linux/if_vlan.h>
#include "vlanproc.h"
#include "vlan.h"

/****** Defines and Macros **************************************************/

#ifndef	min
#define min(a,b) (((a)<(b))?(a):(b))
#endif
#ifndef	max
#define max(a,b) (((a)>(b))?(a):(b))
#endif


/****** Function Prototypes *************************************************/

#ifdef CONFIG_PROC_FS

/* Proc filesystem interface */
static int vlan_proc_perms(struct inode *, int);
static ssize_t vlan_proc_read(struct file* file, char* buf, size_t count,
                              loff_t *ppos);

/* Methods for preparing data for reading proc entries */

static int vlan_config_get_info(char* buf, char** start, off_t offs, int len,
                                int dummy);
static int vlandev_get_info(char* buf, char** start, off_t offs, int len,
                            int dummy);


/* Miscellaneous */

/*
 *	Global Data
 */

/*
 *	Names of the proc directory entries 
 */

static char name_root[]	 = "vlan";
static char name_conf[]	 = "config";
static char term_msg[]   = "***KERNEL:  Out of buffer space!***\n";


/*
 *	VLAN device IOCTL.
 *	o execute requested action or pass command to the device driver
 */

int vlan_ioctl(struct inode* inode, struct file* file,
               unsigned int cmd, unsigned long arg) {
        int err = 0;
        /*
          struct proc_dir_entry* dent;
          struct device* dev;
        */
        struct vlan_ioctl_args args;
   
        printk(VLAN_DBG __FUNCTION__ ": cmd: %x\n", cmd);

   /* everything here needs root permissions, except aguably the
    * hack ioctls for sending packets.  However, I know _I_ don't
    * want users running that on my network! --BLG
    */
        if (!capable(CAP_NET_ADMIN)){
                return -EPERM;
        }
		
        if ((cmd >> 8) != VLAN_IOCTL) {
                printk(VLAN_DBG __FUNCTION__ ": Not a VLAN IOCTL: %x \n", cmd);
                return -EINVAL;
        }
   
        if (copy_from_user(&args, (void*)arg, sizeof(struct vlan_ioctl_args)))
                return -EFAULT;

   /* Null terminate this sucker, just in case. */
        args.dev1[23] = 0;
        args.u.dev2[23] = 0;

   /*
   dent = inode->u.generic_ip;
   if ((dent == NULL) || (dent->data == NULL))
     return -EINVAL;

   dev = dent->data;
   */
   
        switch (cmd)
        {
        case SET_INGRESS_PRIORITY_IOCTL:
                err = vlan_dev_set_ingress_priority(args.dev1, args.u.skb_priority, args.vlan_qos);
                break;

        case SET_EGRESS_PRIORITY_IOCTL:
                err = vlan_dev_set_egress_priority(args.dev1, args.u.skb_priority, args.vlan_qos);
                break;

        case SET_VLAN_FLAG_IOCTL:
                err = vlan_dev_set_vlan_flag(args.dev1, args.u.flag, args.vlan_qos);
                break;

        case SET_NAME_TYPE_IOCTL:
                if ((args.u.name_type >= 0) && (args.u.name_type < VLAN_NAME_TYPE_HIGHEST)) {
                        vlan_name_type = args.u.name_type;
                        err = 0;
                }
                else {
                        err = -EINVAL;
                }
                break;
         
                /* TODO:  Figure out how to pass info back...
                   case GET_INGRESS_PRIORITY_IOCTL:
                   err = vlan_dev_get_ingress_priority(args);
                   break;

                   case GET_EGRESS_PRIORITY_IOCTL:
                   err = vlan_dev_get_egress_priority(args);
                   break;
                */
         
        case ADD_VLAN_IOCTL:
                /* we have been given the name of the Ethernet Device we want to
                 * talk to:  args.dev1   We also have the
                 * VLAN ID:  args.u.VID
                 */
                if (register_802_1Q_vlan_device(args.dev1, args.u.VID)) {
                        err = 0;
                }
                else {
                        err = -EINVAL;
                }
                break;

        case DEL_VLAN_IOCTL:
                /* Here, the args.dev1 is the actual VLAN we want to get rid of. */

                err = unregister_802_1Q_vlan_device(args.dev1);
                break;

        default:
                /* pass on to underlying device instead?? */
                printk(VLAN_DBG __FUNCTION__ ": Unknown VLAN IOCTL: %x \n", cmd);
                return -EINVAL;
        }/* switch */
        return err;
}

/*
 *	Structures for interfacing with the /proc filesystem.
 *	VLAN creates its own directory /proc/net/vlan with the folowing
 *	entries:
 *	config		device status/configuration
 *	<device>	entry for each  device
 */

/*
 *	Generic /proc/net/vlan/<file> file and inode operations 
 */

static struct file_operations vlan_fops = {
        NULL,			/* lseek   */
        vlan_proc_read,	        /* read	   */
        NULL,        		/* write   */
        NULL,			/* readdir */
        NULL,			/* select  */
        vlan_ioctl,			/* ioctl   */
        NULL,			/* mmap	   */
        NULL,			/* no special open code	   */
        NULL,			/* flush */
        NULL,			/* no special release code */
        NULL  			/* can't fsync */
};

static struct inode_operations vlan_inode = {
        &vlan_fops,                  
        NULL,			/* create */
        NULL,			/* lookup */
        NULL,			/* link */
        NULL,			/* unlink */
        NULL,			/* symlink */
        NULL,			/* mkdir */
        NULL,			/* rmdir */
        NULL,			/* mknod */
        NULL,			/* rename */
        NULL,			/* follow link */
        NULL,			/* readlink */
        NULL,			/* readpage */
        NULL,			/* writepage */
        NULL,			/* bmap */
        NULL,			/* truncate */
        vlan_proc_perms
};

/*
 *	/proc/net/vlan/<device> file and inode operations
 */

static struct file_operations vlandev_fops = {
        NULL,			/* lseek   */
        vlan_proc_read,	        /* read	   */
        NULL,        		/* write   */
        NULL,			/* readdir */
        NULL,			/* select  */
        vlan_ioctl,	                /* ioctl   */
        NULL,			/* mmap	   */
        NULL,			/* no special open code	   */
        NULL,			/* flush */
        NULL,			/* no special release code */
        NULL			        /* can't fsync */
};

static struct inode_operations vlandev_inode = {
        &vlandev_fops,
        NULL,			/* create */
        NULL,			/* lookup */
        NULL,			/* link */
        NULL,			/* unlink */
        NULL,			/* symlink */
        NULL,			/* mkdir */
        NULL,			/* rmdir */
        NULL,			/* mknod */
        NULL,			/* rename */
        NULL,			/* readlink */
        NULL,			/* follow_link */
        NULL,			/* readpage */
        NULL,			/* writepage */
        NULL,			/* bmap */
        NULL,			/* truncate */
        vlan_proc_perms
};


/*
 * Proc filesystem derectory entries.
 */

/*
 *	/proc/net/vlan 
 */

static struct proc_dir_entry proc_vlan = {
        0,			/* .low_ino */
        sizeof(name_root) - 1,	/* .namelen */
        name_root,		/* .name */
        0555 | S_IFDIR,		/* .mode */
        2,			/* .nlink */
        0,			/* .uid */
        0,			/* .gid */
        0,			/* .size */
        &proc_dir_inode_operations, /* .ops */
        NULL,			/* .get_info */
        NULL,			/* .fill_node */
        NULL,			/* .next */
        NULL,			/* .parent */
        NULL,			/* .subdir */
        NULL,			/* .data */
};

/*
 *	/proc/net/vlan/config 
 */
 
static struct proc_dir_entry proc_vlan_conf = {
        0,    			/* .low_ino */
        sizeof(name_conf) - 1,	/* .namelen */
        name_conf,   		/* .name */
        0444 | S_IFREG,		/* .mode */
        1,   			/* .nlink */
        0,	        		/* .uid */
        0,		        	/* .gid */
        0,   			/* .size */
        &vlan_inode,	        	/* .ops */
        &vlan_config_get_info,	/* .get_info */
        NULL,			/* .fill_node */
        NULL,			/* .next */
        NULL,			/* .parent */
        NULL,			/* .subdir */
        NULL,			/* .data */
};


/* Strings */
static char conf_hdr[] = "VLAN Dev name  | VLAN ID\n";


/*
 *	Interface functions
 */

/*
 *	Initialize vlan proc interface.
 */

__initfunc(int vlan_proc_init (void)) {
        int err = proc_register(proc_net, &proc_vlan);

        if (!err) {
                proc_register(&proc_vlan, &proc_vlan_conf);
        }
        return err;
}

/*
 *	Clean up router proc interface.
 */

void vlan_proc_cleanup (void) {
        proc_unregister(&proc_vlan, proc_vlan_conf.low_ino);
        proc_unregister(proc_net, proc_vlan.low_ino);
}


/*
 *	Add directory entry for VLAN device.
 */

int vlan_proc_add_dev (struct device* vlandev) {
        if (!vlandev->vlan_dev) {
                printk(KERN_ERR "ERROR:  vlan_proc_add, device -:%s:- is NOT a VLAN\n",
                       vlandev->name);
                return -EINVAL;
        }
   
        memset(&(vlandev->vlan_dev->dent), 0, sizeof(vlandev->vlan_dev->dent));
        vlandev->vlan_dev->dent.namelen = strlen(vlandev->name);
        vlandev->vlan_dev->dent.name = vlandev->name;
        vlandev->vlan_dev->dent.mode = 0444 | S_IFREG;
        vlandev->vlan_dev->dent.nlink = 1;
        vlandev->vlan_dev->dent.ops = &vlandev_inode;
        vlandev->vlan_dev->dent.get_info = &vlandev_get_info;
        vlandev->vlan_dev->dent.data = vlandev;

#ifdef VLAN_DEBUG
        printk(KERN_ERR "vlan_proc_add, device -:%s:- being added.\n",
               vlandev->name);
#endif
   
        return proc_register(&proc_vlan, &vlandev->vlan_dev->dent);
}



/*
 *	Delete directory entry for VLAN device.
 */
int vlan_proc_rem_dev(struct device* vlandev) {
        if (!vlandev || !vlandev->vlan_dev) {
#ifdef VLAN_DEBUG
                printk(VLAN_DBG __FUNCTION__ ": invalid argument: %p\n", vlandev);
#endif
                return -EINVAL;
        }
   
#ifdef VLAN_DEBUG
        printk(VLAN_DBG __FUNCTION__ ": calling proc_unregister for dev: %p\n",
               vlandev);
#endif
        return proc_unregister(&proc_vlan, vlandev->vlan_dev->dent.low_ino);
}


/****** Proc filesystem entry points ****************************************/

/*
 *	Verify access rights.
 */

static int vlan_proc_perms (struct inode* inode, int op) {
        return 0;
}

/*
 *	Read VLAN proc directory entry.
 *	This is universal routine for reading all entries in /proc/net/vlan
 *	directory.  Each directory entry contains a pointer to the 'method' for
 *	preparing data for that entry.
 *	o verify arguments
 *	o allocate kernel buffer
 *	o call get_info() to prepare data
 *	o copy data to user space
 *	o release kernel buffer
 *
 *	Return:	number of bytes copied to user space (0, if no data)
 *		<0	error
 */
static ssize_t vlan_proc_read(struct file* file, char* buf, size_t count,
                              loff_t *ppos) {
        struct inode *inode = file->f_dentry->d_inode;
        struct proc_dir_entry* dent;
        char* page;
        int pos, offs, len;

        if (count <= 0)
                return 0;
		
        dent = inode->u.generic_ip;
        if ((dent == NULL) || (dent->get_info == NULL))
                return 0;
		
        page = kmalloc(VLAN_PROC_BUFSZ, GFP_KERNEL);
        VLAN_MEM_DBG("page malloc, addr: %p  size: %i\n", page, VLAN_PROC_BUFSZ);
   
        if (page == NULL)
                return -ENOBUFS;
		
        pos = dent->get_info(page, dent->data, 0, 0, 0);
        offs = file->f_pos;
        if (offs < pos) {
                len = min(pos - offs, count);
                if (copy_to_user(buf, (page + offs), len)) {
                        return -EFAULT;
                }
                file->f_pos += len;
        }
        else {
                len = 0;
        }

        kfree(page);
        VLAN_FMEM_DBG("page free, addr: %p\n", page);
        return len;
}/* vlan_proc_read */


static int vlan_proc_get_vlan_info(char* buf, unsigned int cnt) {
        struct device* vlandev = NULL;
        struct vlan_group* grp = NULL;
        int i = 0;
        char* nm_type = NULL;
   
        printk(VLAN_DBG __FUNCTION__ ": cnt == %i\n", cnt);

        if (vlan_name_type == VLAN_NAME_TYPE_RAW_PLUS_VID) {
                nm_type = "VLAN_NAME_TYPE_RAW_PLUS_VID";
        }
        else if (vlan_name_type == VLAN_NAME_TYPE_PLUS_VID_NO_PAD) {
                nm_type = "VLAN_NAME_TYPE_PLUS_VID_NO_PAD";
        }
        else if (vlan_name_type == VLAN_NAME_TYPE_RAW_PLUS_VID_NO_PAD) {
                nm_type = "VLAN_NAME_TYPE_RAW_PLUS_VID_NO_PAD";
        }
        else if (vlan_name_type == VLAN_NAME_TYPE_PLUS_VID) {
                nm_type = "VLAN_NAME_TYPE_PLUS_VID";
        }
        else {
                nm_type = "UNKNOWN";
        }
   
        cnt += sprintf(buf + cnt, "Name-Type: %s  bad_proto_recvd: %lu\n",
                       nm_type, vlan_bad_proto_recvd);
   
        for (grp = p802_1Q_vlan_list; grp != NULL; grp = grp->next) {
                /* loop through all devices for this device */
                printk(VLAN_DBG __FUNCTION__ ": found a group, addr: %p\n", grp);
                for (i = 0; i < VLAN_GROUP_ARRAY_LEN; i++) {
                        /* printk(VLAN_DBG __FUNCTION__ ": checking index[%i]\n", i); */
                        if ((vlandev = grp->vlan_devices[i])) {
                                printk(VLAN_DBG __FUNCTION__ ": found a vlan_dev, addr: %p\n", vlandev);
                                if ((cnt + 100) > VLAN_PROC_BUFSZ) {
                                        if ((cnt + strlen(term_msg)) >= VLAN_PROC_BUFSZ) {
                                                /* should never get here */
                                                return cnt;
                                        }
                                        else {
                                                cnt += sprintf(buf + cnt, "%s", term_msg);
                                                return cnt;
                                        }
                                }/* if running out of buffer space */
                                else {
                                        if (!vlandev->vlan_dev) {
                                                printk(KERN_ERR __FUNCTION__ ": ERROR: vlandev->vlan_dev is NULL\n");
                                        }
                                        else {
                                                printk(VLAN_DBG __FUNCTION__ ": got a good vlandev, addr: %p\n", vlandev->vlan_dev);
                                                cnt += sprintf(buf + cnt, "%-15s| %d | %s\n",
                                                               vlandev->name, vlandev->vlan_dev->vlan_id, vlandev->vlan_dev->real_dev->name);
                                        }/* else */
                                }/* else */
                        }/* if we have a vlan of this number */
                }/* for all VLAN's */
        }/* for each vlan group, default is only one.*/
   
        return cnt;
}/* vlan_proc_get_vlan_info */

/*
 *	Prepare data for reading 'Config' entry.
 *	Return length of data.
 */

static int vlan_config_get_info(char* buf, char** start, off_t offs, int len, 
                                int dummy) {
        strcpy(buf, conf_hdr);
        return vlan_proc_get_vlan_info(buf, (unsigned int)(strlen(conf_hdr)));
}


/*
 *	Prepare data for reading <device> entry.
 *	Return length of data.
 *
 *	On entry, the 'start' argument will contain a pointer to VLAN device
 *	data space.
 */

static int vlandev_get_info(char* buf, char** start, off_t offs, int len, 
                            int dummy) {
        struct device* vlandev = (void*)start;
        struct net_device_stats* stats;
        int cnt = 0;
        struct vlan_priority_tci_mapping* mp;
        int i;
   
#ifdef VLAN_DEBUG
        printk(VLAN_DBG __FUNCTION__ ": vlandev: %p\n", vlandev);
#endif
   
        if ((vlandev == NULL) || (!vlandev->vlan_dev)) {
                return 0;
        }

        cnt += sprintf(buf + cnt, "%s  VID: %d  REORDER_HDR: %i\n",
                       vlandev->name, vlandev->vlan_dev->vlan_id, (int)(vlandev->vlan_dev->flags & 1));
        stats = (struct net_device_stats*)(vlandev->priv);
   
        cnt += sprintf(buf + cnt, "%30s: %12lu\n",
                       "total frames received", stats->rx_packets);

        cnt += sprintf(buf + cnt, "%30s: %12lu\n",
                       "total bytes received", stats->rx_bytes);
   
        cnt += sprintf(buf + cnt, "%30s: %12lu\n",
                       "Broadcast/Multicast Rcvd", stats->multicast);
   
        cnt += sprintf(buf + cnt, "\n%30s: %12lu\n",
                       "total frames transmitted", stats->tx_packets);
   
        cnt += sprintf(buf + cnt, "%30s: %12lu\n",
                       "total bytes transmitted", stats->tx_bytes);

        cnt += sprintf(buf + cnt, "%30s: %12lu\n",
                       "total headroom inc", vlandev->vlan_dev->cnt_inc_headroom_on_tx);
        
        cnt += sprintf(buf + cnt, "%30s: %12lu\n",
                       "total encap on xmit", vlandev->vlan_dev->cnt_encap_on_xmit);

        cnt += sprintf(buf + cnt, "Device: %s", vlandev->vlan_dev->real_dev->name);

   /* now show all PRIORITY mappings relating to this VLAN */
        cnt += sprintf(buf + cnt, "\nINGRESS priority mappings: 0:%lu  1:%lu  2:%lu  3:%lu  4:%lu  5:%lu  6:%lu 7:%lu\n",
                       vlandev->vlan_dev->ingress_priority_map[0],
                       vlandev->vlan_dev->ingress_priority_map[1],
                       vlandev->vlan_dev->ingress_priority_map[2],
                       vlandev->vlan_dev->ingress_priority_map[3],
                       vlandev->vlan_dev->ingress_priority_map[4],
                       vlandev->vlan_dev->ingress_priority_map[5],
                       vlandev->vlan_dev->ingress_priority_map[6],
                       vlandev->vlan_dev->ingress_priority_map[7]);

        cnt += sprintf(buf + cnt, "EGRESSS priority Mappings: ");
   
        for (i = 0; i<16; i++) {
                mp = vlandev->vlan_dev->egress_priority_map[i];
                while (mp) {
                        cnt += sprintf(buf + cnt, "%lu:%hu ", mp->priority, ((mp->vlan_qos >> 13) & 0x7));

                        if ((cnt + 100) > VLAN_PROC_BUFSZ) {
                                if ((cnt + strlen(term_msg)) >= VLAN_PROC_BUFSZ) {
                                        /* should never get here */
                                        return cnt;
                                }
                                else {
                                        cnt += sprintf(buf + cnt, "%s", term_msg);
                                        return cnt;
                                }
                        }/* if running out of buffer space */
                        mp = mp->next;
                }
        }/* for */

        cnt += sprintf(buf + cnt, "\n");

        return cnt;
}


/*
 *	End
 */
 
#else

/*
 *	No /proc - output stubs
 */
 
__initfunc(int vlan_proc_init(void))
{
        return 0;
}

void vlan_proc_cleanup(void)
{
        return;
}


int vlan_proc_add_dev(struct device *vlandev)
{
        return 0;
}

int vlan_proc_rem_dev(struct device *vlandev)
{
        return 0;
}

#endif
