/*
 * ip_masq_generic.c
 *
 * Attempt to masquerade any IP protocol.
 * Since we have no protocol details, we can only support one session
 * with any particular foreign host.
 *
 * Copyright (c) 1999 John D. Hardin
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/skbuff.h>
#include <asm/system.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/inet.h>
#include <net/checksum.h>
#include <net/icmp.h>
#include <net/ip_masq.h>
#include <net/ip_masq_mod.h>


/*
 *	Ensure we're not configured to masquerade a protocol that
 *	we already know the details about...
 *
 *	Drat. enums can't be used in preprocessor stuff; this'd be
 *      much nicer as "CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL == IPPROTO_ICMP", etc.
 */

#if CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL == 0
#error The generic IP masquerade protocol number has not been specified.
#endif
#if CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL == 1
#error Cannot specify a known protocol (ICMP = 1).
#endif
#if CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL == 6
#error Cannot specify a known protocol (TCP = 6).
#endif
#if CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL == 17
#error Cannot specify a known protocol (UDP = 17).
#endif
#if CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL == 47
#error Cannot specify a known protocol (GRE = 47).
#endif
#if CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL == 50
#error Cannot specify a known protocol (ESP = 50).
#endif

/*
 * If we handle the packet, return 1.  If we do not, return 0.  If there
 * is some sort of error, return -1
 */
int ip_masq_generic(struct sk_buff **skb_p, __u32 maddr)
{
        struct sk_buff  *skb   = *skb_p;
        struct iphdr    *iph   = skb->nh.iph;
        struct ip_masq  *ms;

#ifdef CONFIG_IP_MASQ_DEBUG_GENERIC
        printk(KERN_DEBUG "ip_masq_generic(): proto %d ", CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL);
        printk("%d.%d.%d.%d", NIPQUAD(iph->saddr));
        printk(" -> %d.%d.%d.%d\n", NIPQUAD(iph->daddr));
#endif /* CONFIG_IP_MASQ_DEBUG_GENERIC */

	/*
	 *	look for existing outbound entry in masq table
	 */

        ms = ip_masq_out_get(iph->protocol,
                iph->saddr, 0,
                iph->daddr, 0);

        if (ms!=NULL) {
                /*
                 *      Make sure that the masq IP address is correct
                 *      for dynamic IP...
                 */
		if ( (ms->maddr != maddr) && ip_masq_get_dynaddr()) {
			if (ip_masq_get_dynaddr() > 1) {
				printk(KERN_INFO "ip_masq_generic(): proto %d ", CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL);
				printk("change maddr from %d.%d.%d.%d", NIPQUAD(ms->maddr));
				printk(" to %d.%d.%d.%d\n", NIPQUAD(maddr));
			}

			write_lock(&__ip_masq_lock);

                        ip_masq_unhash(ms);
                        ms->maddr = maddr;
                        ip_masq_hash(ms);

			write_unlock(&__ip_masq_lock);
                }
        } else {
                /*
                 *      Nope, not found, create a new entry for it, maybe
		 *
		 *	Look for existing outbound entry from different masq'd host,
		 *	we can only support one masq'd host per remote host
		 */

		read_lock(&__ip_masq_lock);
		ms = __ip_masq_in_get(iph->protocol,
			iph->daddr, 0,
			iph->saddr, 0);
		read_unlock(&__ip_masq_lock);

		if (ms!=NULL) {
			atomic_dec(&ms->refcnt);	/* just looking... */
			if (ms->saddr != iph->saddr) {
				/* somebody else is talking to them, sorry. */
				printk(KERN_INFO "ip_masq_generic(): proto %d ", CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL);
				printk("remote is busy, discarding ");
				printk("%d.%d.%d.%d", NIPQUAD(iph->saddr));
				printk(" -> %d.%d.%d.%d\n", NIPQUAD(iph->daddr));
				icmp_send(skb, ICMP_DEST_UNREACH, ICMP_PKT_FILTERED, 0);
				return -1;
			}
                } else  /* nothing else, make new entry */
                        ms = ip_masq_new(iph->protocol,
				maddr, 0xFFFF,
				iph->saddr, 0,
				iph->daddr, 0,
				IP_MASQ_F_MPORT | IP_MASQ_F_MPORT_NOT_UNIQUE | IP_MASQ_F_SIMPLE_HASH );

                if (ms == NULL) {
                        printk(KERN_NOTICE "ip_masq_generic(): proto %d ", CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL);
			printk("Couldn't create masq table entry.\n");
                        return -1;
                }
        }

        /*
         *      Set iph source addr from ip_masq obj.
         */
        iph->saddr = maddr;

        /*
         *      set timeout and check IP header
         */
        ms->timeout = CONFIG_IP_MASQUERADE_GENERIC_EXPIRE * 60 * HZ;
        ip_send_check(iph);

	/* housekeeping */
	ip_masq_put(ms);

#ifdef CONFIG_IP_MASQ_DEBUG_GENERIC_VERBOSE
        printk(KERN_DEBUG "ip_masq_generic(): proto %d ", CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL);
        printk("O-routed from %d.%d.%d.%d\n", NIPQUAD(iph->saddr));
#endif /* CONFIG_IP_MASQ_DEBUG_GENERIC_VERBOSE */

        return 1;
}

int ip_demasq_generic(struct sk_buff **skb_p)
{
        struct sk_buff		*skb   = *skb_p;
        struct iphdr		*iph   = skb->nh.iph;
        struct ip_masq		*ms;

#ifdef CONFIG_IP_MASQ_DEBUG_GENERIC
        printk(KERN_DEBUG "ip_demasq_generic(): proto %d ", CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL);
        printk("%d.%d.%d.%d", NIPQUAD(iph->saddr));
        printk(" -> %d.%d.%d.%d\n", NIPQUAD(iph->daddr));
#endif /* CONFIG_IP_MASQ_DEBUG_GENERIC */

        /*
         *      Look for a masq table entry and reroute if found
         */

	read_lock(&__ip_masq_lock);
        ms = ip_masq_in_get(iph->protocol,
                iph->saddr, 0,
                iph->daddr, 0xFFFF);
	read_unlock(&__ip_masq_lock);

        if (ms != NULL)
        {
                iph->daddr = ms->saddr;

                /*
                 * resum checksums and set timeout
                 */
                ms->timeout = CONFIG_IP_MASQUERADE_GENERIC_EXPIRE * 60 * HZ;
                ip_send_check(iph);

		/* housekeeping */
		ip_masq_put(ms);

#ifdef CONFIG_IP_MASQ_DEBUG_GENERIC_VERBOSE
		printk(KERN_DEBUG "ip_demasq_generic(): proto %d ", CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL);
                printk("I-routed to %d.%d.%d.%d\n", NIPQUAD(iph->daddr));
#endif /* CONFIG_IP_MASQ_DEBUG_GENERIC_VERBOSE */
                return 1;
        }

        /* sorry, all this trouble for a no-hit :) */
        printk(KERN_INFO "ip_demasq_generic(): proto %d ", CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL);
        printk("Inbound from %d.%d.%d.%d has no masq table entry.\n", NIPQUAD(iph->saddr));
        return 0;
}


static struct ip_masq_proto_mod ip_masq_generic_proto = {
	NULL,				/* next */
	CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL,		/* proto */
	ip_masq_generic,		/* masquerade */
	ip_demasq_generic,		/* de-masquerade */
	ATOMIC_INIT(0)			/* refcount */
};

/*
 *    generic masq module initialization
 */

__initfunc(int ip_masq_generic_init(void))
{
        int     retcode;

        if ((retcode = register_ip_masq_proto_mod(&ip_masq_generic_proto))) {
                printk(KERN_ERR "Register generic masq (proto %d) failed: %d\n", CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL, retcode);
        }
	return retcode;
}

/*
 *    generic masq module exit
 */

int ip_masq_generic_done(void)
{
   return unregister_ip_masq_proto_mod(&ip_masq_generic_proto);
}


#ifdef MODULE

int init_module(void)
{
#ifdef CONFIG_IP_MASQ_DEBUG_GENERIC_VERBOSE
        printk(KERN_INFO "ip_masq_generic init_module(): proto %d ", CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL);
        printk("entry\n");
#endif /* CONFIG_IP_MASQ_DEBUG_GENERIC_VERBOSE */

	printk(KERN_INFO "ip_masq_generic: proto %d ", CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL );
	printk("loading\n" );
        if (ip_masq_generic_init())
                return -EIO;
        return 0;
}

void cleanup_module(void)
{
#ifdef CONFIG_IP_MASQ_DEBUG_GENERIC_VERBOSE
        printk(KERN_INFO "ip_masq_generic cleanup_module(): proto %d ", CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL);
        printk("entry\n");
#endif /* CONFIG_IP_MASQ_DEBUG_GENERIC_VERBOSE */

        if (ip_masq_generic_done()) {
		printk(KERN_ERR "ip_masq_generic: proto %d ", CONFIG_IP_MASQUERADE_GENERIC_PROTOCOL );
                printk("can't remove module\n");
	}
}

#else /* MODULE */

EXPORT_SYMBOL(ip_masq_generic);
EXPORT_SYMBOL(ip_demasq_generic);
EXPORT_SYMBOL(ip_masq_generic_init);
EXPORT_SYMBOL(ip_masq_generic_done);

#endif /* MODULE */

/* eof */
