/*
 * ip_masq_ipsec.c
 *
 * Masquerading for IPSec (IP-level Security),
 * as defined by RFC2406 (ESP protocol) and RFC2408 (ISAKMP protocol)
 *
 * Copyright (c) 1997-1999 John D. Hardin
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/skbuff.h>
#include <asm/system.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/inet.h>
#include <net/checksum.h>
#include <net/icmp.h>
#include <net/ip_masq.h>
#include <net/ip_masq_mod.h>

static struct ip_masq_app ip_masq_esp_app;

/*
 * We can't know the inbound SPI until it comes in (the ISAKMP exchange
 * is encrypted so we can't sniff it out of that), so we associate inbound
 * and outbound traffic by inspection. If somebody sends a new packet to a
 * remote server, then block all other new traffic to that server until we
 * get a response from that server with a SPI we haven't seen yet. It is
 * assumed that this is the correct response - we have no way to verify it,
 * as everything else is encrypted.
 *
 * If there is a collision, the block will last for up to two minutes (or
 * whatever MASQUERADE_EXPIRE_IPSEC_INIT is set to), and if the client
 * retries during that time the timer will be reset. This could easily lead
 * to a Denial of Service, so we limit the number of retries that will
 * reset the timer. This means the maximum time the server could be blocked
 * is ((IPSEC_INIT_RETRIES + 1) * MASQUERADE_EXPIRE_IPSEC_INIT).
 *
 * Note: blocking will not affect already-established traffic (i.e. where
 * the inbound SPI has been associated with an outbound SPI).
 */
#if CONFIG_IP_MASQUERADE_IPSEC_EXPIRE > 15
#define MASQUERADE_EXPIRE_IPSEC CONFIG_IP_MASQUERADE_IPSEC_EXPIRE*60*HZ
#else
#define MASQUERADE_EXPIRE_IPSEC 15*60*HZ
#endif

#define MASQUERADE_EXPIRE_IPSEC_INIT 2*60*HZ
#define IPSEC_INIT_RETRIES 5

/*
 * ...connections that don't get an answer are squelched
 * (recognized but ignored) for a short time to prevent DoS.
 */
#define MASQUERADE_EXPIRE_IPSEC_SQUELCHED 30*HZ

/*
 * Once we see inbound traffic with a new SPI, we can safely assume that
 * there won't be any more traffic with old SPIs. To reduce masq table clutter,
 * chop the lifetime on the old SPI entries so that they'll expire a lot
 * sooner. If there *is* traffic with the old SPIs, the table entry will either
 * be refreshed or recreated.
 */
#define MASQUERADE_EXPIRE_IPSEC_OLDSPI 5*60*HZ


/************************************************************
 *                     ISAKMP MASQ                          *
 *  This has to work with the existing UDP masq, which is   *
 *  why we go through all of the timeout table contortions  *
 *  and so forth...                                         *
 ************************************************************/

/*
 * ISAKMP uses 500/udp, and the traffic must come from
 * 500/udp (i.e. 500/udp <-> 500/udp), so we need to
 * check for ISAKMP UDP traffic and avoid changing the
 * source port number. In order to associate the data streams
 * we need to sniff the ISAKMP cookies as well.
 */
#define UDP_PORT_ISAKMP 500     /* ISAKMP default UDP port */

struct isakmp_data {
	__u64	icookie;	/* initiator cookie */
	__u64	rcookie;	/* responder cookie */
};

/*
 * Split the 64-bit cookie into two 32-bit chunks for display
 */
#ifdef __LITTLE_ENDIAN
#define splitcookie(x)	(long unsigned int) ntohl(((x) & 0xFFFFFFFF)), (long unsigned int) ntohl(((x) >> 32))
#else
#define splitcookie(x)	(long unsigned int) ntohl(((x) >> 32)), (long unsigned int) ntohl(((x) & 0xFFFFFFFF))
#endif

/*
 * A table holding these values...
 * Pick states that vaguely map to what we're doing.
 * This is used for ISAKMP table entries. ESP masq
 * doesn't fiddle around with a timeout table and doesn't need to fake one
 * since the timeout table code in ip_masq.c will never see any ESP packets.
 */

static struct ip_masq_timeout_table ipsec_timeout_table = {
	ATOMIC_INIT(0), /* refcnt */
	0,              /* scale  */
	{
		MASQUERADE_EXPIRE_IPSEC_INIT,	/*      IP_MASQ_S_NONE, */
		MASQUERADE_EXPIRE_IPSEC,	/*      IP_MASQ_S_ESTABLISHED,  */
		MASQUERADE_EXPIRE_IPSEC_INIT,	/*      IP_MASQ_S_SYN_SENT,     */
		MASQUERADE_EXPIRE_IPSEC_INIT,	/*      IP_MASQ_S_SYN_RECV,     */
		MASQUERADE_EXPIRE_IPSEC_OLDSPI,	/*      IP_MASQ_S_FIN_WAIT,     */
		MASQUERADE_EXPIRE_IPSEC_OLDSPI,	/*      IP_MASQ_S_TIME_WAIT,    */
		MASQUERADE_EXPIRE_IPSEC_OLDSPI,	/*      IP_MASQ_S_CLOSE,        */
		MASQUERADE_EXPIRE_IPSEC_OLDSPI,	/*      IP_MASQ_S_CLOSE_WAIT,   */
		MASQUERADE_EXPIRE_IPSEC_OLDSPI,	/*      IP_MASQ_S_LAST_ACK,     */
		MASQUERADE_EXPIRE_IPSEC_INIT,	/*      IP_MASQ_S_LISTEN,       */
		MASQUERADE_EXPIRE_IPSEC,	/*      IP_MASQ_S_UDP,  */
		MASQUERADE_EXPIRE_IPSEC_INIT,	/*      IP_MASQ_S_ICMP, */
		2*HZ,	/* IP_MASQ_S_LAST  */
	},	/* timeout */
};


struct ip_masq *
ip_masq_out_get_isakmp(const struct sk_buff *skb, const struct iphdr *iph)
{
	int			protocol;
	union ip_masq_tphdr	h;
	__u16			s_port, d_port;
	__u32			s_addr, d_addr;
	__u64			*dptr, icookie, rcookie;
	unsigned		hash, dhash;
	struct isakmp_data	*priv = NULL;
	struct list_head	*l, *e;
	struct ip_masq		*ms = NULL;


	/* ensure we've been given an ISAKMP packet to work with */

        if ((protocol = iph->protocol) != IPPROTO_UDP) {
		return NULL;
	}

	h.raw = (char*) iph + iph->ihl * 4;
        s_port = h.uh->source;
        d_port = h.uh->dest;
        if (ntohs(s_port) != UDP_PORT_ISAKMP || ntohs(d_port) != UDP_PORT_ISAKMP) {
		return NULL;
	}

	/* Okay, it's ISAKMP */

        s_addr = iph->saddr;
        d_addr = iph->daddr;

	/* If it's not long enough to contain cookies, give up */

	if ((ntohs(iph->tot_len) - (iph->ihl * 4) - sizeof(struct udphdr)) < 16) {
		return NULL;
	}

	dptr = (__u64 *) (h.raw + sizeof(struct udphdr));
	icookie = dptr[0];
	rcookie = dptr[1];

#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC
        printk(KERN_DEBUG "ip_masq_out_get_isakmp(): ");
        printk("%d.%d.%d.%d -> ", NIPQUAD(s_addr));
        printk("%d.%d.%d.%d ", NIPQUAD(d_addr));
        printk("i%08lX%08lX r%08lX%08lX\n",
		splitcookie(icookie),
		splitcookie(rcookie));
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC */

        if (icookie == 0) {
                printk(KERN_INFO "ip_masq_out_get_isakmp(): ");
                printk("zero icookie");
		printk(", pkt fm %d.%d.%d.%d discarded\n",
			NIPQUAD(s_addr));
		return &ms_discard_pkt;
        }

        hash = ip_masq_hash_key(protocol, s_addr, s_port);

	/* look for fully-qualified entries (icookie + rcookie) first */
	l = &ip_masq_s_table[hash];
	for (e=l->next; e!=l; e=e->next) {
		ms = list_entry(e, struct ip_masq, s_list);
                if (protocol == ms->protocol &&
                    s_addr == ms->saddr && s_port == ms->sport &&
                    d_addr == ms->daddr && d_port == ms->dport) {
			if (ms->app_data) {
				priv = ms->app_data;
				if (icookie == priv->icookie &&
				    rcookie == priv->rcookie) {
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
					printk(KERN_DEBUG "ip_masq_out_get_isakmp(): ");
					printk("lookup OK %08X:%04hX->%08X:%04hX ",
					       s_addr, s_port, d_addr, d_port);
					printk("i%08lX%08lX r%08lX%08lX\n",
						splitcookie(icookie),
						splitcookie(rcookie));
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE */
					atomic_inc(&ms->refcnt);
					return ms;
				}
			} else {
				printk(KERN_NOTICE "ip_masq_out_get_isakmp(): ");
				printk("lookup %08X:%04hX->%08X:%04hX %04hX has no app_data\n",
				       s_addr, s_port, d_addr, d_port, ms->mport);
			}
                }
        }

	/* no match, try entries w/o rcookie */
	l = &ip_masq_s_table[hash];
	for (e=l->next; e!=l; e=e->next) {
		ms = list_entry(e, struct ip_masq, s_list);
                if (protocol == ms->protocol &&
                    s_addr == ms->saddr && s_port == ms->sport &&
                    d_addr == ms->daddr && d_port == ms->dport) {
			if (ms->app_data) {
				priv = ms->app_data;
				if (icookie == priv->icookie &&
				    priv->rcookie == 0) {
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
					printk(KERN_DEBUG "ip_masq_out_get_isakmp(): ");
					printk("lookup OK %08X:%04hX->%08X:%04hX ",
					       s_addr, s_port, d_addr, d_port);
					printk("i%08lX%08lX r%08X%08X\n",
						splitcookie(icookie),
						0, 0);
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE */
					atomic_inc(&ms->refcnt);
					return ms;
				}
			}
                }
        }

        dhash = ip_masq_hash_key(protocol, d_addr, d_port);

#ifndef CONFIG_IP_MASQUERADE_IPSEC_PAROK
	/* no match, prevent parallel connections */
	l = &ip_masq_d_table[dhash];
	for (e=l->next; e!=l; e=e->next) {
		ms = list_entry(e, struct ip_masq, d_list);
		if (protocol == ms->protocol &&
		    s_addr != ms->saddr && s_port == ms->sport &&
		    d_addr == ms->daddr && d_port == ms->dport) {
			printk(KERN_INFO "ip_masq_out_get_isakmp(): ");
			printk("remote SG %d.%d.%d.%d busy",
				NIPQUAD(d_addr));
			printk(", pkt fm %d.%d.%d.%d discarded\n",
				NIPQUAD(s_addr));
			/* the cast below just quiets a compiler warning */
			icmp_send((struct sk_buff *) skb, ICMP_DEST_UNREACH, ICMP_PKT_FILTERED, 0);
			return &ms_discard_pkt;
		    }
	}
#endif /* CONFIG_IP_MASQUERADE_IPSEC_PAROK */

	/* no match, prevent icookie collisions */
	l = &ip_masq_d_table[dhash];
	for (e=l->next; e!=l; e=e->next) {
		ms = list_entry(e, struct ip_masq, d_list);
                if (protocol == ms->protocol &&
                    s_addr != ms->saddr && s_port == ms->sport &&
                    d_addr == ms->daddr && d_port == ms->dport) {
			if (ms->app_data) {
				priv = ms->app_data;
				if (icookie == priv->icookie &&
				    priv->rcookie == 0) {
					printk(KERN_INFO "ip_masq_out_get_isakmp(): ");
					printk("avoiding icookie collision");
					printk(", pkt fm %d.%d.%d.%d discarded\n",
						NIPQUAD(s_addr));
					return &ms_discard_pkt;
				}
			}
                }
        }

#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
	printk(KERN_DEBUG "ip_masq_out_get_isakmp(): ");
	printk("lookup FAIL %08X:%04hX->%08X:%04hX ",
	       s_addr, s_port, d_addr, d_port);
	printk("i%08lX%08lX r%08lX%08lX\n",
		splitcookie(icookie),
		splitcookie(rcookie));
#endif

        return &ms_not_found;

}

struct ip_masq *
ip_masq_out_create_isakmp(const struct sk_buff *skb, const struct iphdr *iph, __u32 maddr)
{
	int			protocol;
	union ip_masq_tphdr	h;
	__u16			s_port, d_port;
	__u32			s_addr, d_addr;
	__u64			*dptr, icookie, rcookie;
	struct isakmp_data	*priv = NULL;
	struct ip_masq		*ms = NULL;


	/* ensure we've been given an ISAKMP packet to work with */

        if ((protocol = iph->protocol) != IPPROTO_UDP) {
		return NULL;
	}

	h.raw = (char*) iph + iph->ihl * 4;
        s_port = h.uh->source;
        d_port = h.uh->dest;
        if (ntohs(s_port) != UDP_PORT_ISAKMP || ntohs(d_port) != UDP_PORT_ISAKMP) {
		return NULL;
	}

	/* Okay, it's ISAKMP */

        s_addr = iph->saddr;
        d_addr = iph->daddr;

	/* If it's not long enough to contain cookies, give up */

	if ((ntohs(iph->tot_len) - (iph->ihl * 4) - sizeof(struct udphdr)) < 16) {
		return NULL;
	}

	dptr = (__u64 *) (h.raw + sizeof(struct udphdr));
	icookie = dptr[0];
	rcookie = dptr[1];

#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC
        printk(KERN_DEBUG "ip_masq_out_create_isakmp(): ");
        printk("%d.%d.%d.%d -> ", NIPQUAD(s_addr));
        printk("%d.%d.%d.%d ", NIPQUAD(d_addr));
        printk("i%08lX%08lX r%08lX%08lX\n",
		splitcookie(icookie),
		splitcookie(rcookie));
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC */

	ms = ip_masq_new(iph->protocol,
			maddr, s_port,
			s_addr, s_port,
			d_addr, d_port,
			IP_MASQ_F_MPORT | IP_MASQ_F_MPORT_NOT_UNIQUE | IP_MASQ_F_SIMPLE_HASH);

	if (ms != NULL && (priv = ms->app_data) != NULL) {
		priv->icookie = icookie;
		priv->rcookie = rcookie;
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
		printk(KERN_DEBUG "ip_masq_out_create_isakmp(): ");
		printk("created ");
		printk("%d.%d.%d.%d -> ", NIPQUAD(ms->saddr));
		printk("%d.%d.%d.%d ", NIPQUAD(ms->daddr));
		printk("i%08lX%08lX r%08lX%08lX\n",
			splitcookie(priv->icookie),
			splitcookie(priv->rcookie));
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE */
		if (ms->mport != ms->sport) {
			printk(KERN_NOTICE "ip_masq_out_create_isakmp(): ");
			printk("ip_masq_new() assigned m_port!\n");
		}
	}

	return ms;
}

struct ip_masq *
ip_masq_in_get_isakmp(const struct sk_buff *skb, const struct iphdr *iph)
{
	int			protocol;
	union ip_masq_tphdr	h;
	__u16			s_port, d_port;
	__u32			s_addr, d_addr;
	__u64			*dptr, icookie, rcookie;
	struct isakmp_data	*priv = NULL;
	struct list_head	*l, *e, *ol, *oe;
	struct ip_masq		*ms = NULL;
        struct ip_masq		*oms;
        unsigned long		cutoff;
        unsigned		hash;


	/* ensure we've been given an ISAKMP packet to work with */

        if ((protocol = iph->protocol) != IPPROTO_UDP) {
		return NULL;
	}

	h.raw = (char*) iph + iph->ihl * 4;
        s_port = h.uh->source;
        d_port = h.uh->dest;
        if (ntohs(s_port) != UDP_PORT_ISAKMP || ntohs(d_port) != UDP_PORT_ISAKMP) {
		return NULL;
	}

	/* Okay, it's ISAKMP */

        s_addr = iph->saddr;
        d_addr = iph->daddr;

	/* If it's not long enough to contain cookies, give up */

	if ((ntohs(iph->tot_len) - (iph->ihl * 4) - sizeof(struct udphdr)) < 16) {
		return NULL;
	}

	dptr = (__u64 *) (h.raw + sizeof(struct udphdr));
	icookie = dptr[0];
	rcookie = dptr[1];

#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC
        printk(KERN_DEBUG "ip_masq_in_get_isakmp(): ");
        printk("%d.%d.%d.%d -> ", NIPQUAD(s_addr));
        printk("%d.%d.%d.%d ", NIPQUAD(d_addr));
        printk("i%08lX%08lX r%08lX%08lX\n",
		splitcookie(icookie),
		splitcookie(rcookie));
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC */

        if (icookie == 0) {
                printk(KERN_INFO "ip_masq_in_get_isakmp(): ");
                printk("zero icookie");
		printk(", pkt fm %d.%d.%d.%d discarded\n",
			NIPQUAD(s_addr));
		return &ms_discard_pkt;
        }

	/* search by remote IP, since all ISAKMP entries will have the same mport */
        hash = ip_masq_hash_key(protocol, s_addr, s_port);

	/* look for fully-qualified entries (icookie + rcookie) first */
	l = &ip_masq_d_table[hash];
	for (e=l->next; e!=l; e=e->next) {
		ms = list_entry(e, struct ip_masq, d_list);
                if (protocol == ms->protocol &&
                    s_addr == ms->daddr && s_port == ms->dport &&
                    d_addr == ms->maddr && d_port == ms->mport) {
			if (ms->app_data) {
				priv = ms->app_data;
				if (icookie == priv->icookie &&
				    rcookie == priv->rcookie) {
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
					printk(KERN_DEBUG "ip_masq_in_get_isakmp(): ");
					printk("lookup OK %08X:%04hX->%08X:%04hX ",
					       s_addr, s_port, d_addr, d_port);
					printk("i%08lX%08lX r%08lX%08lX\n",
						splitcookie(icookie),
						splitcookie(rcookie));
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE */
					atomic_inc(&ms->refcnt);
					return ms;
				}
			} else {
				printk(KERN_NOTICE "ip_masq_in_get_isakmp(): ");
				printk("lookup %08X:%04hX->%08X:%04hX %04hX has no app_data\n",
				       s_addr, s_port, d_addr, d_port, ms->mport);
			}
                }
        }

	/* no match, try entries w/o rcookie */
	l = &ip_masq_d_table[hash];
	for (e=l->next; e!=l; e=e->next) {
		ms = list_entry(e, struct ip_masq, d_list);
                if (protocol == ms->protocol &&
                    s_addr == ms->daddr && s_port == ms->dport &&
                    d_addr == ms->maddr && d_port == ms->mport) {
			if (ms->app_data) {
				priv = ms->app_data;
				if (icookie == priv->icookie &&
				    priv->rcookie == 0) {
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
					printk(KERN_DEBUG "ip_masq_in_get_isakmp(): ");
					printk("lookup OK %08X:%04hX->%08X:%04hX ",
					       s_addr, s_port, d_addr, d_port);
					printk("i%08lX%08lX r%08X%08X\n",
						splitcookie(icookie),
						0, 0);
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE */
					/* save the rcookie */
					priv->rcookie = rcookie;
					/* remain until next rekey */
					ms->timeout_table = &ipsec_timeout_table;
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
					printk(KERN_DEBUG "ip_masq_in_get_isakmp(): ");
					printk("rcookie now %08lX%08lX\n", splitcookie(priv->rcookie));
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE */

					/* clean up old ISAKMP masq table entries */
					cutoff = jiffies + MASQUERADE_EXPIRE_IPSEC_OLDSPI;
					hash = ip_masq_hash_key(IPPROTO_UDP, ms->saddr, ms->sport);
					ol = &ip_masq_s_table[hash];
					for (oe=ol->next; oe!=ol; oe=oe->next) {
						oms = list_entry(oe, struct ip_masq, s_list);
						if (oms->protocol == IPPROTO_UDP &&
						    oms->saddr == ms->saddr &&
						    oms->sport == ms->sport &&
						    oms->daddr == ms->daddr &&
						    oms->dport == ms->dport &&
						    oms->timer.expires > cutoff) {
							if (oms->app_data) {
								priv = oms->app_data;
								if (priv->icookie != icookie ||
								    priv->rcookie != rcookie) {
									/* an old ISAKMP SA with a long life:
									   reduce its lifetime... */
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
									printk(KERN_DEBUG "ip_masq_in_get_isakmp(): ");
									printk("accelerating i%08lX%08lX\n", splitcookie(priv->icookie));
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE */
									__ip_masq_set_expire(oms, 0);
									oms->timeout = MASQUERADE_EXPIRE_IPSEC_OLDSPI;
									__ip_masq_set_expire(oms, oms->timeout);
								}
							}
						}
					}

					atomic_inc(&ms->refcnt);
					return ms;
				}
			}
                }
        }

#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
	printk(KERN_DEBUG "ip_masq_in_get_isakmp(): ");
	printk("lookup FAIL %08X:%04hX->%08X:%04hX ",
	       s_addr, s_port, d_addr, d_port);
	printk("i%08lX%08lX r%08lX%08lX\n",
		splitcookie(icookie),
		splitcookie(rcookie));
#endif
        return &ms_not_found;

}

static int masq_isakmp_init_1(struct ip_masq_app *mapp, struct ip_masq *ms)
{
	struct isakmp_data *priv;

	MOD_INC_USE_COUNT;

	if ((ms->app_data = kmalloc(sizeof(struct isakmp_data), GFP_ATOMIC)) == NULL) {
		printk(KERN_INFO "masq_isakmp_init_1(): no memory\n");
	} else {
		priv = ms->app_data;
		priv->icookie = priv->rcookie = 0;
	}

	return 0;
}

static int masq_isakmp_done_1(struct ip_masq_app *mapp, struct ip_masq *ms)
{
	MOD_DEC_USE_COUNT;

	if (ms->app_data)
		kfree_s(ms->app_data, sizeof(struct isakmp_data));

	return 0;
}


/********************************************************
 *                     ESP MASQ                         *
 ********************************************************/

struct esp_data {
	__u32	ospi;
	__u32	ispi;
	int	blocking;
	int	ocnt;
	int	squelched;
};

static struct ip_masq *
ip_masq_out_get_esp(__u32 s_addr, __u16 s_port, __u32 d_addr, __u16 d_port, __u32 o_spi)
{
        unsigned		hash;
	struct list_head	*l, *e;
        struct ip_masq		*ms;
	struct esp_data		*priv;

        hash = ip_masq_hash_key(IPPROTO_ESP, s_addr, s_port);
	l = &ip_masq_s_table[hash];
	for (e=l->next; e!=l; e=e->next) {
		ms = list_entry(e, struct ip_masq, s_list);
                if (ms->protocol == IPPROTO_ESP &&
                    ms->saddr == s_addr &&
                    ms->daddr == d_addr) {
			if (ms->app_data) {
				priv = ms->app_data;
				if (o_spi == priv->ospi) {
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
					printk(KERN_DEBUG "ip_masq_out_get_esp(): ");
					printk("lookup %08X:%08X->%08X OK\n",
					       s_addr,
					       o_spi,
					       d_addr);
#endif
					atomic_inc(&ms->refcnt);
					return ms;
				}
			} else {
				printk(KERN_NOTICE "ip_masq_out_get_esp(): ");
				printk("lookup %08X:%04hX->%08X:%04hX has no app_data\n",
				       s_addr, s_port, d_addr, d_port);
			}
                }
        }

#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
	printk(KERN_DEBUG "ip_masq_out_get_esp(): ");
        printk("lookup %08X:%08X->%08X fail\n",
               s_addr,
               o_spi,
               d_addr);
#endif
        return NULL;
}

static struct ip_masq *
ip_masq_in_get_esp(__u32 s_addr, __u16 s_port, __u32 d_addr, __u16 d_port, __u32 i_spi)
{
        unsigned		hash;
	struct list_head	*l, *e;
        struct ip_masq		*ms;
	struct esp_data		*priv;

        /* 
	 * find an entry for a packet coming in from outside,
         * or find whether there's a setup pending
         */

        if (i_spi != 0) {
                /* there's a SPI - look for a completed entry */
                hash = ip_masq_hash_key(IPPROTO_ESP, s_addr, s_port);
		l = &ip_masq_d_table[hash];
		for (e=l->next; e!=l; e=e->next) {
			ms = list_entry(e, struct ip_masq, d_list);
                        if (ms->protocol == IPPROTO_ESP &&
                            ms->daddr == s_addr &&
                            ms->maddr == d_addr) {
				if (ms->app_data) {
					priv = ms->app_data;
					if (priv->ispi != 0 && priv->ispi == i_spi) {
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
						printk(KERN_DEBUG "ip_masq_in_get_esp(): ");
						printk("lookup %08X->%08X:%08X OK\n",
						       s_addr,
						       d_addr,
						       i_spi);
#endif
						atomic_inc(&ms->refcnt);
						return ms;
					}
				}
                        }
                }
        }

        /* no joy. look for a pending connection - maybe somebody else's
         * if we're checking for a pending setup, the d_addr will be zero
         * to avoid having to know the masq IP.
         */
        hash = ip_masq_hash_key(IPPROTO_ESP, s_addr, 0);
	l = &ip_masq_d_table[hash];
	for (e=l->next; e!=l; e=e->next) {
		ms = list_entry(e, struct ip_masq, d_list);
		if (ms->protocol == IPPROTO_ESP &&
                    ms->daddr == s_addr &&
                    (d_addr == 0 || d_addr == ms->maddr)) {
			if (ms->app_data) {
				priv = ms->app_data;
				if (priv->ispi == 0) {
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
					printk(KERN_DEBUG "ip_masq_in_get_esp(): ");
					printk("lookup %08X->%08X:0 OK\n",
					       s_addr,
					       d_addr
					       );
#endif
					atomic_inc(&ms->refcnt);
					return ms;
				}
			}
                }
        }

#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
	printk(KERN_DEBUG "ip_masq_in_get_esp(): ");
	printk("lookup %08X->%08X:%08X fail\n",
		s_addr,
		d_addr,
		i_spi);
#endif

	return NULL;
}

static int masq_esp_done_1(struct ip_masq_app *mapp, struct ip_masq *ms)
{

	MOD_DEC_USE_COUNT;

	if (ms->app_data)
		kfree_s(ms->app_data, sizeof(struct esp_data));

	return 0;
}

/*
 * If we handle the packet, return 1.  If we do not, return 0.  If there
 * is some sort of error, return -1
 */
int ip_masq_esp(struct sk_buff **skb_p, __u32 maddr)
{
        struct sk_buff  *skb   = *skb_p;
        struct iphdr    *iph   = skb->nh.iph;
        __u32		 o_spi;
        __u16		 fake_port;
        struct ip_masq  *ms;
	struct esp_data	*priv;
        unsigned long    timeout = MASQUERADE_EXPIRE_IPSEC;

        o_spi = *((__u32 *)&(((char *)iph)[iph->ihl*4]));
        fake_port = (__u16) ntohl(o_spi) & 0xFFFF;

#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC
        printk(KERN_DEBUG "ip_masq_esp(): ");
        printk("%d.%d.%d.%d", NIPQUAD(iph->saddr));
        printk(" -> %d.%d.%d.%d SPI %lX (fakeport %X)\n", NIPQUAD(iph->daddr), ntohl(o_spi), fake_port);
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC */

        if (o_spi == 0) {
                /* illegal SPI - discard */
                printk(KERN_INFO "ip_masq_esp(): ");
                printk("zero SPI from %d.%d.%d.%d discarded\n", NIPQUAD(iph->saddr));
                return -1;
        }

#ifdef CONFIG_IP_MASQUERADE_VS
	/*
	 * hack around VS' playing with maddr
	 * maddr should always be nonzero otherwise
	 */
 	if (!maddr && (ip_masq_select_addr(skb, &maddr) < 0)) {
		printk(KERN_NOTICE "ip_masq_esp(): ");
		printk("determine masq addr failed\n");
 		return -1;
 	}
#endif

        /*
         *      Look for masq table entry
         */

	read_lock(&__ip_masq_lock);
        ms = ip_masq_out_get_esp(
                iph->saddr, fake_port,
                iph->daddr, 0,
                o_spi);
	read_unlock(&__ip_masq_lock);

        if (ms!=NULL) {
		priv = ms->app_data;

                if (priv->squelched) {
                        /* squelched: toss the packet without changing the timer */
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC
                        printk(KERN_INFO "ip_masq_esp(): ");
                        printk("init %d.%d.%d.%d ", NIPQUAD(iph->saddr));
                        printk("-> %d.%d.%d.%d SPI %lX ", NIPQUAD(iph->daddr), ntohl(o_spi));
                        printk("squelched\n");
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC */
			atomic_dec(&ms->refcnt);
                        return -1;
                }

                /* delete the expiration timer */
		__ip_masq_set_expire(ms, 0);
		
                /*
                 *      Make sure that the masq IP address is correct
                 *      for dynamic IP...
                 */
		if ( (ms->maddr != maddr) && ip_masq_get_dynaddr()) {
			if (ip_masq_get_dynaddr() > 1) {
				printk(KERN_INFO "ip_masq_esp(): ");
				printk("change maddr from %d.%d.%d.%d", NIPQUAD(ms->maddr));
				printk(" to %d.%d.%d.%d\n", NIPQUAD(maddr));
			}

			write_lock(&__ip_masq_lock);

                        ip_masq_unhash(ms);
                        ms->maddr = maddr;
                        ip_masq_hash(ms);

			write_unlock(&__ip_masq_lock);
                }

                if (priv->ispi == 0) {
                        /* no response yet, keep timeout short */
                        timeout = MASQUERADE_EXPIRE_IPSEC_INIT;
                        if (priv->blocking) {
                                /* prevent DoS: limit init packet timer resets */
                                priv->ocnt++;
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC
                                printk(KERN_INFO "ip_masq_esp(): ");
                                printk("init %d.%d.%d.%d ", NIPQUAD(iph->saddr));
                                printk("-> %d.%d.%d.%d SPI %lX ", NIPQUAD(iph->daddr), ntohl(o_spi));
                                printk("retry %d\n", priv->ocnt);
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC */
                                if (priv->ocnt > IPSEC_INIT_RETRIES) {
                                        /* more than IPSEC_INIT_RETRIES tries, give up */
                                        printk(KERN_INFO "ip_masq_esp(): ");
                                        printk("init %d.%d.%d.%d ", NIPQUAD(iph->saddr));
                                        printk("-> %d.%d.%d.%d SPI %lX ", NIPQUAD(iph->daddr), ntohl(o_spi));
                                        printk("no response after %d tries, unblocking & squelching\n", priv->ocnt);
                                        /* squelch that source+SPI for a bit */
                                        priv->squelched = 1;
                                        ms->timeout = MASQUERADE_EXPIRE_IPSEC_SQUELCHED;
					/* set dport to nonzero to keep it from blocking other inits */
					write_lock(&__ip_masq_lock);
                                        ip_masq_unhash(ms);
                                        ms->dport = 1;
                                        ip_masq_hash(ms);
					write_unlock(&__ip_masq_lock);
					/* housekeeping */
					ip_masq_put(ms);
                                        /* toss the packet */
                                        return -1;
                                }
                        }
                }
        } else {
                /*
                 *      Nope, not found, create a new entry for it, maybe
                 */

                /* see if there are any pending inits with the same destination... */
		read_lock(&__ip_masq_lock);
                ms = ip_masq_in_get_esp(
                        iph->daddr, 0,
                        0, 0,
                        0);
		read_unlock(&__ip_masq_lock);

                if (ms != NULL) {
			priv = ms->app_data;

                        /* found one with ispi == 0 */
                        if (ms->saddr != iph->saddr) {
                                /* it's not ours, don't step on their toes */
                                printk(KERN_INFO "ip_masq_esp(): ");
                                printk("init %d.%d.%d.%d ", NIPQUAD(iph->saddr));
                                printk("-> %d.%d.%d.%d ", NIPQUAD(iph->daddr));
                                printk("temporarily blocked by pending ");
                                printk("%d.%d.%d.%d init\n", NIPQUAD(ms->saddr));
                                /* let it know it has competition */
                                priv->blocking = 1;
                                /* toss the packet */
				atomic_dec(&ms->refcnt);
                                return -1;
                        }
                        if (priv->ospi != o_spi) {
                                /* SPIs differ, still waiting for a previous attempt to expire */
                                printk(KERN_INFO "ip_masq_esp(): ");
                                printk("init %d.%d.%d.%d ", NIPQUAD(iph->saddr));
                                printk("-> %d.%d.%d.%d SPI %lX ", NIPQUAD(iph->daddr), ntohl(o_spi));
                                printk("temporarily blocked by pending ");
                                printk("init w/ SPI %lX\n", ntohl(priv->ospi));
                                /* let it know it has competition */
                                priv->blocking = 1;
                                /* toss the packet */
				atomic_dec(&ms->refcnt);
                                return -1;
                        }
			/* shouldn't get here... */
			atomic_dec(&ms->refcnt);
			printk(KERN_INFO "ip_masq_esp(): ");
			printk(KERN_INFO "internal logic error...\n");
                } else  /* nothing pending, make new entry, pending response */
                        ms = ip_masq_new(iph->protocol,
				maddr, 0xFFFF,
				iph->saddr, fake_port,
				iph->daddr, 0,
				IP_MASQ_F_MPORT | IP_MASQ_F_MPORT_NOT_UNIQUE | IP_MASQ_F_SIMPLE_HASH );

                if (ms == NULL) {
                        printk(KERN_NOTICE "ip_masq_esp(): ");
			printk("Couldn't create masq table entry.\n");
                        return -1;
                }

		/* this isn't TCP or UDP so we have to do the mapp stuff ourselves... */
		if ((ms->app_data = kmalloc(sizeof(struct esp_data), GFP_ATOMIC)) == NULL) {
                        printk(KERN_NOTICE "ip_masq_esp(): ");
			printk("no memory\n");
		} else {
			MOD_INC_USE_COUNT;
			ms->app = &ip_masq_esp_app;	/* to find done_1() when it expires */
			ip_masq_esp_app.n_attach++;
			priv = ms->app_data;
			priv->ispi = priv->blocking = priv->ocnt = priv->squelched = 0;
			priv->ospi = o_spi;
		}
                timeout = MASQUERADE_EXPIRE_IPSEC_INIT;      /* fairly brief timeout while waiting for a response */
        }

        /*
         *      Set iph source addr from ip_masq obj.
         */
        iph->saddr = ms->maddr;

        /*
         *      set timeout and check IP header
         */
        ms->timeout = timeout;
        ip_send_check(iph);

	/* housekeeping */
	ip_masq_put(ms);

#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
        printk(KERN_DEBUG "ip_masq_esp(): ");
        printk("O-routed from %d.%d.%d.%d\n", NIPQUAD(iph->saddr));
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE */

        return 1;
}

int ip_demasq_esp(struct sk_buff **skb_p)
{
        struct sk_buff		*skb   = *skb_p;
        struct iphdr		*iph   = skb->nh.iph;
        struct ip_masq		*ms;
        __u32			i_spi, o_spi;
        __u16			fake_port;
	struct esp_data		*priv;
	struct list_head	*l, *e;
        struct ip_masq		*oms;
        unsigned long		cutoff;
        unsigned		hash;
#define ESP_GUESS_SZ 5                  /* minimum 3, please */
#define ESP_CAND_MIN_TM 5*60*HZ         /* max 10*60*HZ? */
        int			i, ii,
				ncand = 0, nguess = 0;
        __u16			isakmp;
        __u32			cand_ip,
				guess_ip[ESP_GUESS_SZ];
        unsigned long		cand_tm,
				guess_tm[ESP_GUESS_SZ];
	struct isakmp_data	*priv_isakmp;
        struct sk_buff		*skb_cl;
        struct iphdr		*iph_cl;

        i_spi = *((__u32 *)&(((char *)iph)[iph->ihl*4]));
        fake_port = (__u16) ntohl(i_spi) & 0xFFFF;

#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC
        printk(KERN_DEBUG "ip_demasq_esp(): ");
        printk("%d.%d.%d.%d", NIPQUAD(iph->saddr));
        printk(" -> %d.%d.%d.%d SPI %lX (fakeport %X)\n", NIPQUAD(iph->daddr), ntohl(i_spi), fake_port);
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC */

        if (i_spi == 0) {
                /* illegal SPI - discard */
                printk(KERN_INFO "ip_demasq_esp(): ");
                printk("zero SPI from %d.%d.%d.%d discarded\n", NIPQUAD(iph->saddr));
                return -1;
        }

        /*
         *      Look for a masq table entry and reroute if found
         */

	read_lock(&__ip_masq_lock);
        ms = ip_masq_in_get_esp(
                iph->saddr, fake_port,
                iph->daddr, 0,
                i_spi);
	read_unlock(&__ip_masq_lock);

        if (ms != NULL)
        {
		priv = ms->app_data;

                /* delete the expiration timer */
                __ip_masq_set_expire(ms,0);

                iph->daddr = ms->saddr;

                if (priv->ispi == 0) {
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC
                        printk(KERN_INFO "ip_demasq_esp(): ");
                        printk("resp from %d.%d.%d.%d SPI %lX", NIPQUAD(iph->saddr), ntohl(i_spi));
                        printk(" routed to %d.%d.%d.%d (SPI %lX)\n", NIPQUAD(ms->saddr), ntohl(priv->ospi));
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC */
			write_lock(&__ip_masq_lock);
                        ip_masq_unhash(ms);
                        priv->ispi = i_spi;
                        ms->dport = fake_port;
                        ip_masq_hash(ms);
			write_unlock(&__ip_masq_lock);

                        /*
                         * To reduce mask table clutter, reduce the expire timer
                         * on all other ESP masq table entries for this saddr/daddr combination
                         */
			o_spi = priv->ospi;
                        cutoff = jiffies + MASQUERADE_EXPIRE_IPSEC_OLDSPI;
                        hash = ip_masq_hash_key(IPPROTO_ESP, ms->maddr, 0xFFFF);
			l = &ip_masq_m_table[hash];
			for (e=l->next; e!=l; e=e->next) {
				oms = list_entry(e, struct ip_masq, m_list);
                                if (oms->protocol == IPPROTO_ESP &&
                                    oms->saddr == ms->saddr &&
                                    oms->daddr == ms->daddr &&
                                    oms->timer.expires > cutoff) {
					if (oms->app_data) {
						priv = oms->app_data;
						if ((priv->ospi != o_spi || priv->ispi != i_spi) &&
						    priv->ispi != 0) {
							/* an old IPSEC SA with a long life:
							   reduce its lifetime... */
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
							printk(KERN_DEBUG "ip_demasq_esp(): ");
							printk("accelerating SPI %08X\n", priv->ospi);
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE */
							__ip_masq_set_expire(oms, 0);
							oms->timeout = MASQUERADE_EXPIRE_IPSEC_OLDSPI;
							__ip_masq_set_expire(oms, oms->timeout);
						}
					}
                                }
                        }
                }

                /*
                 * resum checksums and set timeout
                 */
                ms->timeout = MASQUERADE_EXPIRE_IPSEC;
                ip_send_check(iph);

		/* housekeeping */
		ip_masq_put(ms);

#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
		printk(KERN_DEBUG "ip_demasq_esp(): ");
                printk("I-routed to %d.%d.%d.%d\n", NIPQUAD(iph->daddr));
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE */
                return 1;
        }

        /* 
         * Guess who this packet is likely intended for:
         * Scan the UDP masq table for local hosts that have communicated via
         * ISAKMP with the host who sent this packet.
         * Using an insertion sort with duplicate IP suppression, build a list
         * of the ESP_GUESS_SZ most recent ISAKMP sessions (determined by
         * sorting in decreasing order of timeout timer).
         * Clone the original packet and send it to those hosts, but DON'T make
         * a masq table entry, as we're only guessing. It is assumed that the correct
         * host will respond to the traffic and that will create a masq table entry.

         * host will respond to the traffic and that will create a masq table entry.
         * To limit the list a bit, don't consider any ISAKMP masq entries with
         * less than ESP_CAND_MIN_TM time to live. This should be some value less
         * than the IPSEC table timeout or *all* entries will be ignored...
         */

#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
        printk(KERN_DEBUG "ip_demasq_esp(): ");
        printk("guessing from %d.%d.%d.%d SPI %lX\n", NIPQUAD(iph->saddr), ntohl(i_spi));
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE */

        /* zero out the guess table */
        for (i = 0;i < ESP_GUESS_SZ; i++) {
                guess_ip[i] = 0;
                guess_tm[i] = 0;
        }

        /* scan for ISAKMP sessions with the remote host */
        isakmp = htons(UDP_PORT_ISAKMP);
	read_lock(&__ip_masq_lock);
        hash = ip_masq_hash_key(IPPROTO_UDP, iph->saddr, isakmp);
	l = &ip_masq_d_table[hash];
	for (e=l->next; e!=l; e=e->next) {
		ms = list_entry(e, struct ip_masq, d_list);
                if (ms->protocol == IPPROTO_UDP &&
                    ms->daddr == iph->saddr &&
                    ms->sport == isakmp &&
                    ms->dport == isakmp &&
                    ms->mport == isakmp) {
			if (ms->app_data) {
				priv_isakmp = ms->app_data;
				if (priv_isakmp->icookie != 0 &&
				    priv_isakmp->rcookie != 0) {
					/* a candidate... */
					ncand++;
					cand_ip = ms->saddr;
					cand_tm = ms->timer.expires - jiffies;
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
					printk(KERN_DEBUG "ip_demasq_esp(): ");
					printk("cand %d: IP %d.%d.%d.%d TM %ld\n", ncand, NIPQUAD(cand_ip), cand_tm);
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE */
					if (cand_tm > ESP_CAND_MIN_TM) {
						/* traffic is recent enough, add to list (maybe) */
						for (i = 0; i < ESP_GUESS_SZ; i++) {
							if (cand_tm > guess_tm[i]) {
								/* newer */
								if (guess_ip[i] != 0 && cand_ip != guess_ip[i]) {
									/* newer and IP different - insert */
									if (i < (ESP_GUESS_SZ - 1)) {
										/* move entries down the list,
										 * find first entry after this slot
										 * where the IP is 0 (unused) or
										 * IP == candidate (older traffic, same host)
										 * rather than simply going to the end of the list,
										 * for efficiency (don't shift zeros) and
										 * duplicate IP suppression (don't keep older entries
										 * having the same IP)
										 */
										for (ii = i + 1; ii < (ESP_GUESS_SZ - 1); ii++) {
											if (guess_ip[ii] == 0 || guess_ip[ii] == cand_ip)
												break;
										}
										for (ii-- ; ii >= i; ii--) {
											guess_ip[ii+1] = guess_ip[ii];
											guess_tm[ii+1] = guess_tm[ii];
										}

									}
								}
								guess_ip[i] = cand_ip;
								guess_tm[i] = cand_tm;
								break;
							}
							if (cand_ip == guess_ip[i]) {
								/* fresher entry already there */
								break;
							}
						}
					}
				}
                        }
                }
        }
	read_unlock(&__ip_masq_lock);

        if (guess_ip[0]) {
                /* had guesses - send */
                if (guess_ip[1]) {
                        /* multiple guesses, send a copy to all */
                        for (i = 0; i < ESP_GUESS_SZ && guess_ip[i] != 0; i++) {
                                nguess++;
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
                                printk(KERN_DEBUG "ip_demasq_esp(): ");
                                printk("guess %d: IP %d.%d.%d.%d TM %ld\n", nguess, NIPQUAD(guess_ip[i]), guess_tm[i]);
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE */
                                /* duplicate and send the skb */
                                if ((skb_cl = skb_copy(skb, GFP_ATOMIC)) == NULL) {
                                        printk(KERN_INFO "ip_demasq_esp(): ");
                                        printk("guessing: cannot copy skb\n");
                                } else {
                                        iph_cl = skb_cl->nh.iph;
                                        iph_cl->daddr = guess_ip[i];
                                        ip_send_check(iph_cl);
					IPCB(skb_cl)->flags |= IPSKB_MASQUERADED;
					dst_release(skb_cl->dst);
					skb_cl->dst = NULL;
					if (ip_route_input(skb_cl, iph_cl->daddr, iph_cl->saddr, iph_cl->tos, skb_cl->dev)) {
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC
						printk(KERN_INFO "ip_demasq_esp(): ");
						printk("guessing: cannot route to %d.%d.%d.%d\n", NIPQUAD(guess_ip[i]));
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC */
					} else {
						skb_cl->dst->input(skb_cl);
					}
					kfree_skb(skb_cl);
                                }
                        }
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC
                        printk(KERN_INFO "ip_demasq_esp(): ");
                        printk("guessing from %d.%d.%d.%d SPI %lX sent to", NIPQUAD(iph->saddr), ntohl(i_spi));
                        printk(" %d hosts (%d cand)\n", nguess, ncand);
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC */
                        return -1;      /* discard original packet */
                } else {
                        /* only one guess, send original packet to that host */
                        iph->daddr = guess_ip[0];
                        ip_send_check(iph);

#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC
                        printk(KERN_INFO "ip_demasq_esp(): ");
                        printk("guessing from %d.%d.%d.%d SPI %lX sent to", NIPQUAD(iph->saddr), ntohl(i_spi));
                        printk(" %d.%d.%d.%d (%d cand)\n", NIPQUAD(guess_ip[0]), ncand);
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC */
                        return 1;
                }
        }

        /* sorry, all this trouble for a no-hit :) */
        printk(KERN_INFO "ip_demasq_esp(): ");
        printk("Inbound from %d.%d.%d.%d SPI %lX has no masq table entry.\n", NIPQUAD(iph->saddr), ntohl(i_spi));
        return 0;
}


static struct ip_masq_mod ip_masq_isakmp_mod = {
	NULL,				/* next */
	NULL,				/* next_reg */
	"isakmp",			/* name */
	ATOMIC_INIT(0),			/* nent */
	ATOMIC_INIT(0),			/* refcnt */
	NULL,				/* /proc entries */
	NULL,				/* control */
	NULL,				/* masq_mod_init */
	NULL,				/* masq_mod_done */
	NULL,				/* in rule */
	NULL,				/* in update */
	NULL,				/* in create */
	NULL,				/* out rule */
	NULL,				/* out update */
	ip_masq_out_create_isakmp,	/* out create */
	ip_masq_in_get_isakmp,		/* in get */
	ip_masq_out_get_isakmp,		/* out get */
};

static struct ip_masq_app ip_masq_isakmp_app = {
	NULL,				/* next */
	"isakmp",			/* name */
	0,				/* type */
	0,				/* n_attach */
	masq_isakmp_init_1,		/* ip_masq_init_1 */
	masq_isakmp_done_1,		/* ip_masq_done_1 */
	NULL,				/* pkt_out */
	NULL				/* pkt_in */
};

static struct ip_masq_app ip_masq_esp_app = {
	NULL,				/* next */
	"esp",				/* name */
	0,				/* type */
	0,				/* n_attach */
	NULL,				/* ip_masq_init_1 */
	masq_esp_done_1,		/* ip_masq_done_1 */
	NULL,				/* pkt_out */
	NULL				/* pkt_in */
};

static struct ip_masq_proto_mod ip_masq_esp_proto = {
	NULL,				/* next */
	IPPROTO_ESP,			/* proto */
	ip_masq_esp,			/* masquerade */
	ip_demasq_esp,			/* de-masquerade */
	ATOMIC_INIT(0)			/* refcount */
};

/*
 *    ipsec masq module initialization
 */

__initfunc(int ip_masq_ipsec_init(void))
{
        int     retcode;

        if ((retcode = register_ip_masq_app(&ip_masq_isakmp_app, IPPROTO_UDP, UDP_PORT_ISAKMP))) {
                printk(KERN_ERR "Register %s masq failed: %d\n", "ISAKMP app", retcode);
                return retcode;
        }

        if ((retcode = register_ip_masq_mod(&ip_masq_isakmp_mod))) {
                printk(KERN_ERR "Register %s masq failed: %d\n", "ISAKMP mod", retcode);
                return retcode;
        }
	ip_masq_mod_inc_nent(&ip_masq_isakmp_mod);

        if ((retcode = register_ip_masq_proto_mod(&ip_masq_esp_proto))) {
                printk(KERN_ERR "Register %s masq failed: %d\n", "ESP", retcode);
        }
	return retcode;
}

/*
 *    ipsec masq module exit
 */

int ip_masq_ipsec_done(void)
{

   ip_masq_mod_dec_nent(&ip_masq_isakmp_mod);
   unregister_ip_masq_mod(&ip_masq_isakmp_mod);

   unregister_ip_masq_app(&ip_masq_isakmp_app);

   return unregister_ip_masq_proto_mod(&ip_masq_esp_proto);
}


#ifdef MODULE

int init_module(void)
{
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
        printk(KERN_INFO "ip_masq_ipsec init_module(): ");
        printk("entry\n");
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE */

	printk(KERN_INFO "ip_masq_ipsec: " );
	printk("loading\n" );
        if (ip_masq_ipsec_init())
                return -EIO;
        return 0;
}

void cleanup_module(void)
{
#ifdef CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE
        printk(KERN_INFO "ip_masq_ipsec cleanup_module(): ");
        printk("entry\n");
#endif /* CONFIG_IP_MASQ_DEBUG_IPSEC_VERBOSE */

        if (ip_masq_ipsec_done()) {
		printk(KERN_ERR "ip_masq_ipsec: " );
                printk("can't remove module\n");
	}
}

#else /* MODULE */

EXPORT_SYMBOL(ip_masq_out_get_isakmp);
EXPORT_SYMBOL(ip_masq_out_create_isakmp);
EXPORT_SYMBOL(ip_masq_in_get_isakmp);
EXPORT_SYMBOL(ip_masq_out_get_esp);
EXPORT_SYMBOL(ip_masq_in_get_esp);
EXPORT_SYMBOL(ip_masq_esp);
EXPORT_SYMBOL(ip_demasq_esp);
EXPORT_SYMBOL(ip_masq_ipsec_init);
EXPORT_SYMBOL(ip_masq_ipsec_done);

#endif /* MODULE */

/* eof */
