/*
 * ip_masq_pptp.c
 *
 * Masquerading for PPTP (Point to Point Tunneling Protocol).
 * PPTP is a a protocol for creating virtual private networks.
 * It is a specification defined by Microsoft and some vendors
 * working with Microsoft.  PPTP is built on top of a modified
 * version of the Internet Generic Routing Encapsulation Protocol.
 * GRE is defined in RFC 1701 and RFC 1702.  Documentation of
 * PPTP can be found on the Microsoft web site.
 *
 * Copyright (c) 1997-1998 Gordon Chaffee
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/skbuff.h>
#include <asm/system.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/inet.h>
#include <net/checksum.h>
#include <net/ip_masq.h>
#include <net/ip_masq_mod.h>

#define PPTP_GRE_VERSION	0x1
#define PPTP_GRE_PROTOCOL	0x880B

#define PPTP_GRE_FLAG_C		0x80
#define PPTP_GRE_FLAG_R		0x40
#define PPTP_GRE_FLAG_K		0x20
#define PPTP_GRE_FLAG_S		0x10
#define PPTP_GRE_FLAG_A		0x80

#define PPTP_GRE_IS_C(f)	((f)&PPTP_GRE_FLAG_C)
#define PPTP_GRE_IS_R(f)	((f)&PPTP_GRE_FLAG_R)
#define PPTP_GRE_IS_K(f)	((f)&PPTP_GRE_FLAG_K)
#define PPTP_GRE_IS_S(f)	((f)&PPTP_GRE_FLAG_S)
#define PPTP_GRE_IS_A(f)	((f)&PPTP_GRE_FLAG_A)

struct pptp_gre_hdr {
	__u8  flags;		/* bitfield */
	__u8  version;		/* should be PPTP_GRE_VER (enhanced GRE) */
	__u16 protocol;		/* should be PPTP_GRE_PROTO (ppp-encaps) */
	__u16 payload_len;	/* size of ppp payload, not inc. gre header */
	__u16 call_id;		/* peer's call_id for this session */
	__u32 seq;		/* sequence number.  Present if S==1 */
	__u32 ack;		/* seq number of highest packet recieved by */
				/*  sender in this session */
};

#define PPTP_CONTROL_PACKET	1
#define PPTP_MGMT_PACKET	2
#define PPTP_MAGIC_COOKIE	0x1a2b3c4d

struct pptp_pkt_hdr {
	__u16	packetLength;
	__u16	packetType;
	__u32	magicCookie;
};

/* PptpControlMessageType values */
#define PPTP_START_SESSION_REQUEST	1
#define PPTP_START_SESSION_REPLY	2
#define PPTP_STOP_SESSION_REQUEST	3
#define PPTP_STOP_SESSION_REPLY		4
#define PPTP_ECHO_REQUEST		5
#define PPTP_ECHO_REPLY			6
#define PPTP_OUT_CALL_REQUEST		7
#define PPTP_OUT_CALL_REPLY		8
#define PPTP_IN_CALL_REQUEST		9
#define PPTP_IN_CALL_REPLY		10
#define PPTP_CALL_CLEAR_REQUEST		11
#define PPTP_CALL_DISCONNECT_NOTIFY	12
#define PPTP_CALL_ERROR_NOTIFY		13
#define PPTP_WAN_ERROR_NOTIFY		14
#define PPTP_SET_LINK_INFO		15

#define PPTP_MSG_MAX			15

/* PptpControlMessageType names */
static const char *strMName[] = {
	"UNKNOWN_MESSAGE",
	"START_SESSION_REQUEST",
	"START_SESSION_REPLY",
	"STOP_SESSION_REQUEST",
	"STOP_SESSION_REPLY",
	"ECHO_REQUEST",
	"ECHO_REPLY",
	"OUT_CALL_REQUEST",
	"OUT_CALL_REPLY",
	"IN_CALL_REQUEST",
	"IN_CALL_REPLY",
	"CALL_CLEAR_REQUEST",
	"CALL_DISCONNECT_NOTIFY",
	"CALL_ERROR_NOTIFY",
	"WAN_ERROR_NOTIFY",
	"SET_LINK_INFO"
};

/* PptpGeneralError values */
#define PPTP_ERROR_CODE_NONE		0
#define PPTP_NOT_CONNECTED		1
#define PPTP_BAD_FORMAT			2
#define PPTP_BAD_VALUE			3
#define PPTP_NO_RESOURCE		4
#define PPTP_BAD_CALLID			5
#define PPTP_REMOVE_DEVICE_ERROR	6

struct PptpControlHeader {
	__u16	messageType;
	__u16	reserved;
};

/* FramingCapability Bitmap Values */
#define PPTP_FRAME_CAP_ASYNC		0x1
#define PPTP_FRAME_CAP_SYNC		0x2

/* BearerCapability Bitmap Values */
#define PPTP_BEARER_CAP_ANALOG		0x1
#define PPTP_BEARER_CAP_DIGITAL		0x2

struct PptpStartSessionRequest {
	__u16	protocolVersion;
	__u8	reserved1;
	__u8	reserved2;
	__u32	framingCapability;
	__u32	bearerCapability;
	__u16	maxChannels;
	__u16	firmwareRevision;
	__u8	hostName[64];
	__u8	vendorString[64];
};

/* PptpStartSessionResultCode Values */
#define PPTP_START_OK			1
#define PPTP_START_GENERAL_ERROR	2
#define PPTP_START_ALREADY_CONNECTED	3
#define PPTP_START_NOT_AUTHORIZED	4
#define PPTP_START_UNKNOWN_PROTOCOL	5

struct PptpStartSessionReply {
	__u16	protocolVersion;
	__u8	resultCode;
	__u8	generalErrorCode;
	__u32	framingCapability;
	__u32	bearerCapability;
	__u16	maxChannels;
	__u16	firmwareRevision;
	__u8	hostName[64];
	__u8	vendorString[64];
};

/* PptpStopReasons */
#define PPTP_STOP_NONE			1
#define PPTP_STOP_PROTOCOL		2
#define PPTP_STOP_LOCAL_SHUTDOWN	3

struct PptpStopSessionRequest {
	__u8	reason;
};

/* PptpStopSessionResultCode */
#define PPTP_STOP_OK			1
#define PPTP_STOP_GENERAL_ERROR		2

struct PptpStopSessionReply {
	__u8	resultCode;
	__u8	generalErrorCode;
};

struct PptpEchoRequest {
	__u32 identNumber;
};

/* PptpEchoReplyResultCode */
#define PPTP_ECHO_OK			1
#define PPTP_ECHO_GENERAL_ERROR		2

struct PptpEchoReply {
	__u32	identNumber;
	__u8	resultCode;
	__u8	generalErrorCode;
	__u16	reserved;
};

/* PptpFramingType */
#define PPTP_ASYNC_FRAMING		1
#define PPTP_SYNC_FRAMING		2
#define PPTP_DONT_CARE_FRAMING		3

/* PptpCallBearerType */
#define PPTP_ANALOG_TYPE		1
#define PPTP_DIGITAL_TYPE		2
#define PPTP_DONT_CARE_BEARER_TYPE	3

struct PptpOutCallRequest {
	__u16	callID;
	__u16	callSerialNumber;
	__u32	minBPS;
	__u32	maxBPS;
	__u32	bearerType;
	__u32	framingType;
	__u16	packetWindow;
	__u16	packetProcDelay;
	__u16	reserved1;
	__u16	phoneNumberLength;
	__u16	reserved2;
	__u8	phoneNumber[64];
	__u8	subAddress[64];
};

/* PptpCallResultCode */
#define PPTP_OUTCALL_CONNECT		1
#define PPTP_OUTCALL_GENERAL_ERROR	2
#define PPTP_OUTCALL_NO_CARRIER		3
#define PPTP_OUTCALL_BUSY		4
#define PPTP_OUTCALL_NO_DIAL_TONE	5
#define PPTP_OUTCALL_TIMEOUT		6
#define PPTP_OUTCALL_DONT_ACCEPT	7

struct PptpOutCallReply {
	__u16	callID;
	__u16	peersCallID;
	__u8	resultCode;
	__u8	generalErrorCode;
	__u16	causeCode;
	__u32	connectSpeed;
	__u16	packetWindow;
	__u16	packetProcDelay;
	__u32	physChannelID;
};

struct PptpInCallRequest {
	__u16	callID;
	__u16	callSerialNumber;
	__u32	callBearerType;
	__u32	physChannelID;
	__u16	dialedNumberLength;
	__u16	dialingNumberLength;
	__u8	dialedNumber[64];
	__u8	dialingNumber[64];
	__u8	subAddress[64];
};

/* PptpInCallResultCode */
#define PPTP_INCALL_ACCEPT		1
#define PPTP_INCALL_GENERAL_ERROR	2
#define PPTP_INCALL_DONT_ACCEPT		3

struct PptpInCallReply {
	__u16	callID;
	__u16	peersCallID;
	__u8	resultCode;
	__u8	generalErrorCode;
	__u16	packetWindow;
	__u16	packetProcDelay;
	__u16	reserved;
};

struct PptpInCallConnected {
	__u16	callID;
	__u16	peersCallID;
	__u32	connectSpeed;
	__u16	packetWindow;
	__u16	packetProcDelay;
	__u32	callFramingType;
};

struct PptpClearCallRequest {
	__u16	callID;
	__u16	reserved;
};

struct PptpCallDisconnectNotify {
	__u16	callID;
	__u8	resultCode;
	__u8	generalErrorCode;
	__u16	causeCode;
	__u16	reserved;
	__u8	callStatistics[128];
};

struct PptpWanErrorNotify {
	__u16	peersCallID;
	__u16	reserved;
	__u32	crcErrors;
	__u32	framingErrors;
	__u32	hardwareOverRuns;
	__u32	bufferOverRuns;
	__u32	timeoutErrors;
	__u32	alignmentErrors;
};

struct PptpSetLinkInfo {
	__u16	peersCallID;
	__u16	reserved;
	__u32	sendAccm;
	__u32	recvAccm;
};


struct pptp_priv_data {
	__u16	call_id;
	__u16	mcall_id;
	__u16	pcall_id;
};


/*
 *	Due to changes in the hashing and lookup semantics,
 *	it easier to write our own lookup routines rather than
 *	try to shoehorn this into the existing TCP/UDP lookup routines.
 */

#define FAKE_MPORT(x) (ntohl((x)) & 0x7FFF)

static struct ip_masq *
ip_masq_out_get_gre(__u32 s_addr, __u32 d_addr, __u16 call_id)
{
        unsigned		hash;
	struct list_head	*l, *e;
        struct ip_masq		*ms;

	/* look up by d{addr,port} (dport will be other guy's callID) */
        hash = ip_masq_hash_key(IPPROTO_GRE, d_addr, call_id);
	l = &ip_masq_d_table[hash];
	for (e=l->next; e!=l; e=e->next) {
		ms = list_entry(e, struct ip_masq, d_list);
                if (ms->protocol == IPPROTO_GRE &&
                    ms->saddr == s_addr &&
                    ms->daddr == d_addr &&
                    ms->dport == call_id) {
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP_VERBOSE_X
			printk(KERN_DEBUG "ip_masq_out_get_gre(): ");
			printk("lookup %d.%d.%d.%d->%d.%d.%d.%d CID=%X OK\n",
			       NIPQUAD(s_addr),
			       NIPQUAD(d_addr),
			       ntohs(call_id));
#endif
			atomic_inc(&ms->refcnt);
			__ip_masq_set_expire(ms, 0);
			return ms;
		}
        }

#ifdef CONFIG_IP_MASQ_DEBUG_PPTP_VERBOSE_X
	printk(KERN_DEBUG "ip_masq_out_get_gre(): ");
	printk("lookup %d.%d.%d.%d->%d.%d.%d.%d CID=%X OK\n",
	       NIPQUAD(s_addr),
	       NIPQUAD(d_addr),
	       ntohs(call_id));
#endif
        return NULL;
}

static struct ip_masq *
ip_masq_in_get_gre(__u32 s_addr, __u32 d_addr, __u16 call_id)
{
	unsigned		hash;
	struct list_head	*l, *e;
	struct ip_masq		*ms;
	__u16			fakemport;

	/* look up by m{addr,port} where mport is a MCID */
	hash = ip_masq_hash_key(IPPROTO_GRE, d_addr, call_id);
	l = &ip_masq_m_table[hash];
	for (e=l->next; e!=l; e=e->next) {
		ms = list_entry(e, struct ip_masq, m_list);
		if (ms->protocol == IPPROTO_GRE &&
		    ms->daddr == s_addr &&
		    ms->maddr == d_addr &&
		    ms->mport == call_id) {
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP_VERBOSE
			printk(KERN_DEBUG "ip_masq_in_get_gre(): ");
			printk("lookup %d.%d.%d.%d->%d.%d.%d.%d CID=%X OK\n",
			       NIPQUAD(s_addr),
			       NIPQUAD(d_addr),
			       ntohs(call_id));
#endif
			atomic_inc(&ms->refcnt);
			__ip_masq_set_expire(ms, 0);
			return ms;
		}
	}

	/* look up by m{addr,port} where MCID not done (masq'd server) */
	fakemport = FAKE_MPORT(s_addr);
	hash = ip_masq_hash_key(IPPROTO_GRE, d_addr, fakemport);
	l = &ip_masq_m_table[hash];
	for (e=l->next; e!=l; e=e->next) {
		ms = list_entry(e, struct ip_masq, m_list);
		if (ms->protocol == IPPROTO_GRE &&
		    ms->daddr == s_addr &&
		    ms->maddr == d_addr &&
		    ms->mport == fakemport &&
		    ms->control == NULL) {
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP_VERBOSE
			printk(KERN_DEBUG "ip_masq_in_get_gre(): ");
			printk("lookup %d.%d.%d.%d->%d.%d.%d.%d CID=%X OK\n",
			       NIPQUAD(s_addr),
			       NIPQUAD(d_addr),
			       ntohs(fakemport));
#endif
			atomic_inc(&ms->refcnt);
			__ip_masq_set_expire(ms, 0);
			return ms;
		}
	}

#ifdef CONFIG_IP_MASQ_DEBUG_PPTP_VERBOSE
	printk(KERN_DEBUG "ip_masq_in_get_gre(): ");
	printk("lookup %d.%d.%d.%d->%d.%d.%d.%d CID=%X FAIL\n",
	       NIPQUAD(s_addr),
	       NIPQUAD(d_addr),
	       ntohs(call_id));
#endif
	return NULL;
}


/*
 * If we handle the packet, return 1.  If we do not, return 0.  If there
 * is some sort of error, return -1
 */
static int
ip_masq_gre(struct sk_buff **skb_ptr, __u32 maddr)
{
	struct sk_buff		*skb = *skb_ptr;
	struct iphdr		*iph = skb->nh.iph;
	struct ip_masq		*ms;
	struct pptp_gre_hdr	*grehdr;
	__u16			fakemport;


	if ((skb->len - iph->ihl*4 - sizeof(struct pptp_gre_hdr)) < 0) {
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
		printk(KERN_DEBUG "ip_masq_gre(): ");
		printk("short pkt\n");
#endif
		return 0;
	}

	grehdr = (struct pptp_gre_hdr *) &(((char *)iph)[iph->ihl*4]);

#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
	printk(KERN_DEBUG "ip_masq_gre(): ");
        printk("%d.%d.%d.%d -> ", NIPQUAD(iph->saddr));
        printk("%d.%d.%d.%d ", NIPQUAD(iph->daddr));
	printk("CID=%X VER=%X PROTO=%X\n", 
		ntohs(grehdr->call_id), grehdr->version, ntohs(grehdr->protocol));
#endif

	if ((grehdr->version & 0x7F) != PPTP_GRE_VERSION ||
	    ntohs(grehdr->protocol) != PPTP_GRE_PROTOCOL) return 0;

#ifdef CONFIG_IP_MASQUERADE_VS
	/*
	 * hack around VS' playing with maddr
	 * maddr should always be nonzero otherwise
	 */
 	if (!maddr && (ip_masq_select_addr(skb, &maddr) < 0)) {
		printk(KERN_NOTICE "ip_masq_gre(): ");
		printk("determine masq addr failed\n");
 		return -1;
 	}
#endif

	/*
	 *	Now hunt the list to see if we have an entry
	 */

	ms = ip_masq_out_get_gre(iph->saddr, iph->daddr, grehdr->call_id);

	if (ms == NULL)
	{
		/*
		 *	Not found.
		 *	This is a problem if we're masquerading a client,
		 *	but is expected if we're masquerading a server.
		 *	Since we didn't sniff the control channel to create
		 *	this entry, fake up a mport value from the remote IP address
		 */
		fakemport = FAKE_MPORT(iph->daddr);
		printk(KERN_INFO "ip_masq_gre(): ");
		printk("creating GRE masq for ");
		printk("%d.%d.%d.%d -> ", NIPQUAD(iph->saddr));
		printk("%d.%d.%d.%d ", NIPQUAD(iph->daddr));
		printk("CID=%X MCID=%X\n", ntohs(grehdr->call_id), fakemport);
      
		ms = ip_masq_new(IPPROTO_GRE,
			maddr, fakemport,
			iph->saddr, 0,	/* CID only known from control channel */
			iph->daddr, grehdr->call_id,
			IP_MASQ_F_MPORT | IP_MASQ_F_MPORT_NOT_UNIQUE | IP_MASQ_F_SIMPLE_HASH );
		if (ms == NULL) {
			printk(KERN_NOTICE "ip_masq_gre(): ");
			printk("create masq failed\n");
			return -1;
		}
	}

	/*
	 *      Make sure that the masq IP address is correct
	 *      for dynamic IP...
	 */
	if ( (ms->maddr != maddr) && ip_masq_get_dynaddr()) {
		if (ip_masq_get_dynaddr() > 1) {
			printk(KERN_INFO "ip_masq_gre(): ");
			printk("change maddr from %d.%d.%d.%d", NIPQUAD(ms->maddr));
			printk(" to %d.%d.%d.%d\n", NIPQUAD(maddr));
		}

		write_lock(&__ip_masq_lock);

		ip_masq_unhash(ms);
		ms->maddr = maddr;
		ip_masq_hash(ms);

		write_unlock(&__ip_masq_lock);
	}

	iph->saddr = maddr;

	iph->check=0;
	iph->check=ip_fast_csum((unsigned char *)iph, iph->ihl);

	ip_send_check(iph);

#ifdef CONFIG_IP_MASQ_DEBUG_PPTP_VERBOSE
	printk(KERN_DEBUG "ip_masq_gre(): ");
	printk("O-routed from %d.%d.%d.%d\n", NIPQUAD(iph->saddr));
#endif

	if (ms->control) {
		/* masq'd client from control channel - long timeout */
		ms->timeout = 30*60*HZ;
	} else {
		/* masq'd server, no call-ID masq, short timeout to avoid clutter */
		ms->timeout = 2*60*HZ;  
	}
	ip_masq_put(ms);

	return 1;
}


/*
 * If we handle the packet, return 1.  If we do not, return 0.  If there
 * is some sort of error, return -1
 */
static int
ip_demasq_gre(struct sk_buff **skb_p)
{
	struct sk_buff		*skb = *skb_p;
	struct iphdr		*iph = skb->nh.iph;
	struct ip_masq		*ms;
	struct pptp_gre_hdr	*grehdr;


	if ((skb->len - iph->ihl*4 - sizeof(struct pptp_gre_hdr)) < 0) {
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
		printk(KERN_DEBUG "ip_demasq_gre(): ");
		printk("short pkt\n");
#endif
		return 0;
	}

	grehdr = (struct pptp_gre_hdr *) &(((char *)iph)[iph->ihl*4]);

#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
	printk(KERN_DEBUG "ip_demasq_gre(): ");
        printk("%d.%d.%d.%d -> ", NIPQUAD(iph->saddr));
        printk("%d.%d.%d.%d ", NIPQUAD(iph->daddr));
	printk("CID=%X VER=%X PROTO=%X\n", 
		ntohs(grehdr->call_id), grehdr->version, ntohs(grehdr->protocol));
#endif

	if ((grehdr->version & 0x7F) != PPTP_GRE_VERSION ||
	    ntohs(grehdr->protocol) != PPTP_GRE_PROTOCOL) return 0;

	/* look for masq'd client traffic first (to unmasq callid) */
	ms = ip_masq_in_get_gre(iph->saddr, iph->daddr, grehdr->call_id);

	if (ms != NULL)
	{
		iph->daddr = ms->saddr;

		/*
		 * Adjust IP checksums.
		 */
		iph->check=0;
		iph->check=ip_fast_csum((unsigned char *)iph, iph->ihl);
		ip_send_check(iph);

#ifdef CONFIG_IP_MASQ_DEBUG_PPTP_VERBOSE
		printk(KERN_DEBUG "ip_demasq_gre(): ");
		printk("I-routed to ");
		printk("%d.%d.%d.%d ", NIPQUAD(iph->daddr));
		printk("CID=%X\n", ntohs(grehdr->call_id));
#endif

		if (ms->control) {
			/* masq'd client from control channel - long timeout */
			ms->timeout = 30*60*HZ;
			/* and set the call ID back to the original */
			grehdr->call_id = ms->sport;
		} else {
			/* masq'd server, no call-ID masq, short timeout to avoid clutter */
			ms->timeout = 5*60*HZ;
		}
		ip_masq_put(ms);

		return 1;
	}

#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
	printk(KERN_INFO "ip_demasq_gre(): ");
	printk("%d.%d.%d.%d -> ", NIPQUAD(iph->saddr));
	printk("%d.%d.%d.%d ", NIPQUAD(iph->daddr));
	printk("CID=%X no masq table, discarding\n", ntohs(grehdr->call_id));
#endif

	/* sorry, all this trouble for a no-hit :) */
	return 0;
}

static int
masq_pptp_init_1( struct ip_masq_app *mapp, struct ip_masq *ms )
{
	struct pptp_priv_data	*priv;

	MOD_INC_USE_COUNT;

	if ((ms->app_data = kmalloc(sizeof(struct pptp_priv_data), GFP_ATOMIC)) == NULL) {
		printk(KERN_NOTICE "masq_pptp_init_1(): no memory\n");
	} else {
		priv = ms->app_data;
		priv->call_id = priv->mcall_id = priv->pcall_id = 0;
	}

	return 0;
}

static int
masq_pptp_done_1( struct ip_masq_app *mapp, struct ip_masq *ms )
{

	MOD_DEC_USE_COUNT;

	if (ms->app_data)
		kfree_s(ms->app_data, sizeof(struct pptp_priv_data));

	return 0;
}


#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
static void
printk_pptp_hdr(char *from_txt, struct iphdr *iph, struct pptp_pkt_hdr *pptph)
{
	struct PptpControlHeader	*ctlh;
	__u16				msg;

	printk("%s", from_txt);
        printk("%d.%d.%d.%d -> ", NIPQUAD(iph->saddr));
        printk("%d.%d.%d.%d ", NIPQUAD(iph->daddr));
	printk("LEN=%d TY=%d MC=%lX", ntohs(pptph->packetLength),
		ntohs(pptph->packetType), ntohl(pptph->magicCookie));

	if (ntohs(pptph->packetType) == PPTP_CONTROL_PACKET) {
		ctlh = (struct PptpControlHeader *) ((char *)pptph + sizeof(struct pptp_pkt_hdr));

		msg = htons(ctlh->messageType);
		printk(" CTL=%s", (msg <= PPTP_MSG_MAX)? strMName[msg]:strMName[0]);
	}

	printk("\n");
}
#endif

/* 
 *	masquerade outbound control packets (masq callID)
 *	from masqed client through masq gateway to server
 *	we'll never see masq'd server traffic since the ports are backwards from what we'd expect
 */

int
ip_masq_pptp_tcp(struct ip_masq_app *mapp, struct ip_masq *ms, struct sk_buff **skb_p, __u32 maddr)
{
	struct sk_buff			*skb = *skb_p;
	struct iphdr			*iph = skb->nh.iph;
	struct pptp_pkt_hdr		*pptph;
	struct pptp_priv_data		*priv;
	struct PptpControlHeader	*ctlh;
        union {
                void				*rawreq;
                struct PptpOutCallRequest       *ocreq;
                struct PptpOutCallReply         *ocack;
                struct PptpInCallRequest        *icreq;
                struct PptpInCallReply          *icack;
		struct PptpClearCallRequest	*clrreq;
                struct PptpCallDisconnectNotify *disc;
                struct PptpWanErrorNotify       *wanerr;
                struct PptpSetLinkInfo          *setlink;
        } pptpReq;
	__u16				msg, *cid;
	struct ip_masq			*ms_gre;


	/* verify we have app_data allocated */
	if (!(priv = ms->app_data)) {
		printk(KERN_NOTICE "ip_masq_pptp_tcp(): ");
		printk("no app_data - what now?\n");
		return 0;
	}

#ifdef CONFIG_IP_MASQ_DEBUG_PPTP_VERBOSE
	printk(KERN_DEBUG "ip_masq_pptp_tcp(): ");
	printk("app_data=%p CID=%X MCID=%X PCID=%X\n",
		ms->app_data, ntohs(priv->call_id), ntohs(priv->mcall_id), ntohs(priv->pcall_id));
#endif

	pptph = (struct pptp_pkt_hdr *) ((char *) iph + sizeof(struct iphdr) + sizeof(struct tcphdr));

	/* verify we have data (i.e. pptph points before end of packet) */
	if ((void *) pptph >= (void *) (skb->h.raw + skb->len)) {
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
		printk(KERN_DEBUG "ip_masq_pptp_tcp(): ");
		printk("no TCP data in pkt\n");
#endif
		return 0;
	}

#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
	printk_pptp_hdr(KERN_DEBUG "ip_masq_pptp_tcp(): ", iph, pptph);
#endif

	/* if it's not a control message we can't do anything with it */
        if (ntohs(pptph->packetType) != PPTP_CONTROL_PACKET ||
	    ntohl(pptph->magicCookie) != PPTP_MAGIC_COOKIE) {
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
		printk(KERN_DEBUG "ip_masq_pptp_tcp(): ");
		printk("not a control pkt\n");
#endif
		return 0;
	}

	ctlh = (struct PptpControlHeader *) ((char *) pptph + sizeof(struct pptp_pkt_hdr));
	pptpReq.rawreq = (void *) ((char*) ctlh + sizeof(struct PptpControlHeader));
	switch (msg = htons(ctlh->messageType)) {
		case PPTP_OUT_CALL_REQUEST:
			/* masq'd client initiating connection to server */
			cid = &pptpReq.ocreq->callID;
			break;		/* create GRE masq table entry and get MCID */

		case PPTP_IN_CALL_REQUEST:
			/* masq'd client initiating connection to server */
			cid = &pptpReq.icreq->callID;
			break;		/* create GRE masq table entry and get MCID */

		case PPTP_CALL_CLEAR_REQUEST:
			/* masq'd client sending to server */
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
			printk(KERN_DEBUG "ip_masq_pptp_tcp(): ");
			printk("%s, CID=%X\n", strMName[msg], ntohs(pptpReq.clrreq->callID));
#endif /* CONFIG_IP_MASQ_DEBUG_PPTP */
			pptpReq.clrreq->callID = priv->mcall_id;
			/* no need to create GRE masq table entry */
			return 0;

		case PPTP_CALL_DISCONNECT_NOTIFY:
			/* masq'd client sending to server */
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
			printk(KERN_DEBUG "ip_masq_pptp_tcp(): ");
			printk("%s, CID=%X\n", strMName[msg], ntohs(pptpReq.disc->callID));
#endif /* CONFIG_IP_MASQ_DEBUG_PPTP */
			ms_gre = ip_masq_in_get_gre(ms->daddr, maddr, priv->mcall_id);
			if (ms_gre != NULL)
			{
				/*
				 * expire the data channel entry quickly now.
				 */
				ms_gre->timeout = 30*HZ;
				if (ip_masq_control_get(ms_gre)) ip_masq_control_del(ms_gre);
				printk(KERN_INFO "ip_masq_pptp_tcp(): ");
				printk("%s ", strMName[msg]);
				printk("%d.%d.%d.%d -> ", NIPQUAD(ms->saddr));
				printk("%d.%d.%d.%d ", NIPQUAD(ms->daddr));
				printk("CID=%X MCID=%X\n", ntohs(ms_gre->sport), ntohs(ms_gre->mport));
				ip_masq_put(ms_gre);
			}
			pptpReq.disc->callID = priv->mcall_id;
			/* no need to create GRE masq table entry */
			return 0;

		default:
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
			printk(KERN_DEBUG "ip_masq_pptp_tcp(): ");
			printk("%s (TY=%d)\n", (msg <= PPTP_MSG_MAX)? strMName[msg]:strMName[0], msg);
#endif /* CONFIG_IP_MASQ_DEBUG_PPTP */
			/* unknown need to create GRE masq table entry */
			return 0;
	}

#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
	printk(KERN_DEBUG "ip_masq_pptp_tcp(): ");
	printk("%s, CID=%X\n", strMName[msg], ntohs(*cid));
#endif /* CONFIG_IP_MASQ_DEBUG_PPTP */

	/*
	 * create GRE masq table entry
	 * hash using CID as sport
	 * MCID is taken from mport
	 */
	ms_gre = ip_masq_new(IPPROTO_GRE,
		maddr, 0,
		ms->saddr, *cid,
		ms->daddr, 0,		/* will fill in dport later */
		IP_MASQ_F_SIMPLE_HASH);
	if (ms_gre != NULL) {
		priv->call_id = *cid; 
		*cid = priv->mcall_id = ms_gre->mport;
		ip_masq_control_add(ms_gre, ms);
		ms_gre->timeout = 2*60*HZ;	/* 2 min for data */
		printk(KERN_INFO "ip_masq_pptp_tcp(): ");
		printk("%s ", strMName[msg]);
		printk("%d.%d.%d.%d -> ", NIPQUAD(ms->saddr));
		printk("%d.%d.%d.%d ", NIPQUAD(ms->daddr));
		printk("CID=%X MCID=%X\n", ntohs(ms_gre->sport), ntohs(ms_gre->mport));
		if (!ip_masq_control_get(ms_gre)) {
			printk(KERN_INFO "ip_masq_pptp_tcp(): ");
			printk("failed to set control\n");
		}
		ip_masq_put(ms_gre);
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP_VERBOSE
		printk(KERN_DEBUG "ip_masq_pptp_tcp(): ");
		printk("app_data=%p CID=%X MCID=%X\n",
			ms->app_data, ntohs(priv->call_id), ntohs(*cid));
#endif
	} else {
		printk(KERN_NOTICE "ip_masq_pptp_tcp(): ");
		printk("create GRE masq failed (%s)\n", strMName[msg]);
	}

	return 0;
}



/* 
 *	demasquerade inbound control packets (demasq peersCallID)
 *	from server through masq gateway to masqed client
 */

int ip_demasq_pptp_tcp(struct ip_masq_app *mapp, struct ip_masq *ms, struct sk_buff **skb_p, __u32 maddr)
{
	struct sk_buff			*skb = *skb_p;
	struct iphdr			*iph = skb->nh.iph;
	struct pptp_pkt_hdr		*pptph;
	struct pptp_priv_data		*priv;
	struct PptpControlHeader	*ctlh;
        union {
                void				*rawreq;
                struct PptpOutCallRequest       *ocreq;
                struct PptpOutCallReply         *ocack;
                struct PptpInCallRequest        *icreq;
                struct PptpInCallReply          *icack;
		struct PptpClearCallRequest	*clrreq;
                struct PptpCallDisconnectNotify *disc;
                struct PptpWanErrorNotify       *wanerr;
                struct PptpSetLinkInfo          *setlink;
        } pptpReq;
	__u16				msg, *cid, *pcid;
	struct ip_masq			*ms_gre;


	/* verify we have app_data allocated */
	if (!(priv = ms->app_data)) {
		printk(KERN_NOTICE "ip_demasq_pptp_tcp(): ");
		printk("no app_data - what now?\n");
		return 0;
	}

#ifdef CONFIG_IP_MASQ_DEBUG_PPTP_VERBOSE
	printk(KERN_DEBUG "ip_demasq_pptp_tcp(): ");
	printk("app_data=%p CID=%X MCID=%X PCID=%X\n",
		ms->app_data, ntohs(priv->call_id), ntohs(priv->mcall_id), ntohs(priv->pcall_id));
#endif

	pptph = (struct pptp_pkt_hdr *) ((char *) iph + sizeof(struct iphdr) + sizeof(struct tcphdr));

	/* verify we have data (i.e. pptph points before end of packet) */
	if ((void *) pptph >= (void *) (skb->h.raw + skb->len)) {
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
		printk(KERN_DEBUG "ip_demasq_pptp_tcp(): ");
		printk("no TCP data in pkt\n");
#endif
		return 0;
	}

#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
	printk_pptp_hdr(KERN_DEBUG "ip_demasq_pptp_tcp(): ", iph, pptph);
#endif

	/* if it's not a control message we can't do anything with it */
        if (ntohs(pptph->packetType) != PPTP_CONTROL_PACKET ||
	    ntohl(pptph->magicCookie) != PPTP_MAGIC_COOKIE) {
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
		printk(KERN_DEBUG "ip_demasq_pptp_tcp(): ");
		printk("not a control pkt\n");
#endif
		return 0;
	}

	ctlh = (struct PptpControlHeader *) ((char *) pptph + sizeof(struct pptp_pkt_hdr));
	pptpReq.rawreq = (void *) ((char*) ctlh + sizeof(struct PptpControlHeader));
	switch (msg = htons(ctlh->messageType)) {
		case PPTP_OUT_CALL_REPLY:
			/* server responding to masq'd client */
			cid = &pptpReq.ocack->callID;
			pcid = &pptpReq.ocack->peersCallID;
			/* fix up GRE masq table entry */
			break;

		case PPTP_IN_CALL_REPLY:
			/* server responding to masq'd client */
			cid = &pptpReq.icack->callID;
			pcid = &pptpReq.icack->peersCallID;
			/* fix up GRE masq table entry */
			break;

		case PPTP_WAN_ERROR_NOTIFY:
			/* server notifying masq'd client */
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
			printk(KERN_DEBUG "ip_demasq_pptp_tcp(): ");
			printk("%s, CID=%X\n", strMName[msg], ntohs(pptpReq.wanerr->peersCallID));
#endif /* CONFIG_IP_MASQ_DEBUG_PPTP */
			pptpReq.wanerr->peersCallID = priv->call_id;
			/* no need to alter GRE masq table entry */
			return 0;

		case PPTP_SET_LINK_INFO:
			/* server notifying masq'd client */
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
			printk(KERN_DEBUG "ip_demasq_pptp_tcp(): ");
			printk("%s, CID=%X\n", strMName[msg], ntohs(pptpReq.setlink->peersCallID));
#endif /* CONFIG_IP_MASQ_DEBUG_PPTP */
			pptpReq.setlink->peersCallID = priv->call_id;
			/* no need to alter GRE masq table entry */
			return 0;

		default:
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
			printk(KERN_DEBUG "ip_demasq_pptp_tcp(): ");
			printk("%s (TY=%d)\n", (msg <= PPTP_MSG_MAX)? strMName[msg]:strMName[0], msg);
#endif /* CONFIG_IP_MASQ_DEBUG_PPTP */
			/* unknown need to alter GRE masq table entry */
			return 0;
	}

#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
	printk(KERN_DEBUG "ip_demasq_pptp_tcp(): ");
	printk("%s, CID=%X, PCID=%X\n", strMName[msg], ntohs(*cid), ntohs(*pcid));
#endif /* CONFIG_IP_MASQ_DEBUG_PPTP */

	/*
	 * save server's callID in dport for hashing and lookup speed
	 */
	if (priv->pcall_id != *cid) {
		priv->pcall_id = *cid;
		ms_gre = ip_masq_in_get_gre(iph->saddr, maddr, priv->mcall_id);
		if (ms_gre != NULL)
		{
			write_lock(&__ip_masq_lock);
			ip_masq_unhash(ms_gre);
			ms_gre->dport = priv->pcall_id;
			ip_masq_hash(ms_gre);
			write_unlock(&__ip_masq_lock);
			ms_gre->timeout = 2*60*HZ;	/* 2 min for data */
			ip_masq_put(ms_gre);
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
			printk(KERN_DEBUG "ip_demasq_pptp_tcp(): ");
			printk("saved CID=%X\n", ntohs(priv->pcall_id));
#endif /* CONFIG_IP_MASQ_DEBUG_PPTP */
		} else {
			printk(KERN_NOTICE "ip_demasq_pptp_tcp(): ");
			printk("couldn't find GRE masq for ");
			printk("%d.%d.%d.%d -> ", NIPQUAD(ms->saddr));
			printk("%d.%d.%d.%d ", NIPQUAD(ms->daddr));
			printk("CID=%X MCID=%X\n", ntohs(priv->call_id), ntohs(*pcid));
		}
	}

	/*
	 * demasquerade peersCallID
	 */
	printk(KERN_INFO "ip_demasq_pptp_tcp(): ");
	printk("%s ", strMName[msg]);
	printk("%d.%d.%d.%d -> ", NIPQUAD(ms->saddr));
	printk("%d.%d.%d.%d ", NIPQUAD(ms->daddr));
	printk("CID=%X MCID=%X\n", ntohs(priv->call_id), ntohs(*pcid));
	*pcid = priv->call_id;

	return 0;
}


struct ip_masq_proto_mod ip_masq_gre_mod =
{
	NULL,				/* next */
	IPPROTO_GRE,			/* proto */
	ip_masq_gre,			/* masquerade */
	ip_demasq_gre,			/* de-masquerade */
	ATOMIC_INIT(0)			/* refcount */
};

struct ip_masq_app ip_masq_pptp_tcp_app =
{
	NULL,				/* next */
	"pptp-tcp",			/* name */
	0,				/* type */
	0,				/* n_attach */
	masq_pptp_init_1,		/* ip_masq_init_1 */
	masq_pptp_done_1,		/* ip_masq_done_1 */
	ip_masq_pptp_tcp,		/* pkt_out */
	ip_demasq_pptp_tcp		/* pkt_in */
};


/*
 *    ip_masq_pptp initialization
 */

__initfunc(int ip_masq_pptp_init(void))
{
	int retcode;

	if ((retcode = register_ip_masq_app(&ip_masq_pptp_tcp_app, IPPROTO_TCP, 1723))) {
                printk(KERN_ERR "Register %s masq failed: %d\n", "PPTP app", retcode);
		return retcode;
	}

	if ((retcode = register_ip_masq_proto_mod(&ip_masq_gre_mod))) {
                printk(KERN_ERR "Register %s masq failed: %d\n", "GRE", retcode);
	}
	return retcode;
}

/*
 *    ip_masq_pptp fin.
 */

int ip_masq_pptp_done(void)
{
	unregister_ip_masq_app(&ip_masq_pptp_tcp_app);

	return unregister_ip_masq_proto_mod(&ip_masq_gre_mod);
}

#ifdef MODULE

int init_module(void)
{
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
	printk(KERN_DEBUG "ip_masq_pptp init_module(): " );
        printk("entry\n");
#endif /* CONFIG_IP_MASQ_DEBUG_PPTP */

        if (ip_masq_pptp_init() != 0)
                return -EIO;
        return 0;
}

void cleanup_module(void)
{
#ifdef CONFIG_IP_MASQ_DEBUG_PPTP
	printk(KERN_DEBUG "ip_masq_pptp cleanup_module(): " );
        printk("entry\n");
#endif /* CONFIG_IP_MASQ_DEBUG_PPTP */

        if (ip_masq_pptp_done() != 0) {
                printk(KERN_ERR "ip_masq_pptp: ");
                printk("can't remove module\n");
	}
}

#else /* MODULE */

EXPORT_SYMBOL(ip_masq_out_get_gre);
EXPORT_SYMBOL(ip_masq_in_get_gre);
EXPORT_SYMBOL(ip_masq_pptp_tcp);
EXPORT_SYMBOL(ip_demasq_pptp_tcp);
EXPORT_SYMBOL(ip_masq_gre);
EXPORT_SYMBOL(ip_demasq_gre);

#endif /* MODULE */
/* eof */
