/*
 * IPVS         An implementation of the IP virtual server support for the
 *              LINUX operating system.  IPVS is now implemented as a part
 *              of IP masquerading code. IPVS can be used to build a
 *              high-performance and highly available server based on a
 *              cluster of servers.
 *
 * Version:     $Id: ip_vs.c,v 1.29 2001/05/14 14:04:50 wensong Exp $
 *
 * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
 *              Peter Kese <peter.kese@ijs.si>
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Changes:
 *     Wensong Zhang           :    fixed the overflow bug in ip_vs_procinfo
 *     Wensong Zhang           :    added editing dest and service functions
 *     Wensong Zhang           :    changed the names of some functions
 *     Wensong Zhang           :    fixed the unlocking bug in ip_vs_del_dest
 *     Wensong Zhang           :    added a separate hash table for IPVS
 *     Wensong Zhang           :    added slow timer for IPVS masq entries
 *     Julian Anastasov        :    fixed the number of active connections
 *     Wensong Zhang           :    added persistent port
 *     Wensong Zhang           :    fixed the incorrect lookup in hash table
 *     Wensong Zhang           :    added server status checking
 *     Wensong Zhang           :    fixed the incorrect slow timer vector layout
 *     Wensong Zhang           :    fixed the sltimer added twice bug of mst
 *     Julian Anastasov        :    fixed the IP_MASQ_F_VS_INACTIVE cleared bug after editing dest
 *     Wensong Zhang           :    added the inactive connection counter
 *     Wensong Zhang           :    changed the body of ip_vs_schedule
 *     Julian Anastasov        :    fixed the unlocking bug in ip_vs_schedule
 *     Julian Anastasov        :    fixed the uncounting bug in creating masqs by template
 *     Wensong Zhang           :    changed some condition orders for a bit performance
 *     Julian Anastasov        :    don't touch counters in ip_vs_unbind_masq for templates
 *     Wensong Zhang           :    added the hash table for virtual services
 *     Wensong Zhang           :    changed destination lists to d-linked lists
 *     Wensong Zhang           :    changed the scheduler list to the d-linked list
 *     Wensong Zhang           :    added new persistent service handling
 *     Julian Anastasov        :    fixed the counting bug in ip_vs_unbind_masq again
 *                                    (don't touch counters for templates)
 *     Wensong Zhang           :    changed some IP_VS_ERR to IP_VS_DBG in the ip_vs_tunnel_xmit
 *     Wensong Zhang           :    added different timeout support for persistent svc
 *     Wensong Zhang           :    fixed the bug that persistent svc cannot be edited
 *     Julian Anastasov        :    removed extra read_unlock in __ip_vs_lookup_service
 *     Julian Anastasov        :    changed not to restart template timers if dest is unavailable
 *     Julian Anastasov        :    added the destination trash
 *     Wensong Zhang           :    added the update_service call in ip_vs_del_dest
 *     Wensong Zhang           :    added the ip_vs_leave function
 *     Lars Marowsky-Bree      :    added persistence granularity support
 *     Julian Anastasov        :    changed some comestics things for debugging
 *     Wensong Zhang           :    use vmalloc to allocate big ipvs hash table
 *     Wensong Zhang           :    changed the tunneling/direct routing methods a little
 *     Julian Anastasov        :    fixed the return bug of ip_vs_leave(-2 instead of -3)
 *     Roberto Nibali          :    fixed the undefined variable bug in the IP_VS_DBG of ip_vs_dr_xmit
 *     Julian Anastasov        :    changed ICMP_PROT_UNREACH to ICMP_PORT_UNREACH in ip_vs_leave
 *     Wensong Zhang           :    added port zero support for persistent services
 *     Wensong Zhang           :    fixed the bug that virtual ftp service blocks other services not listed in ipvs table
 *     Wensong Zhang           :    invalidate a persistent template when its dest is unavailable
 *     Julian Anastasov        :    changed two IP_VS_ERR calls to IP_VS_DBG
 *     Wensong Zhang           :    added random drop of syn entries
 *     Wensong Zhang           :    added random drop of UDP entris
 *     Julian Anastasov        :    added droprate defense against DoS attack
 *     Julian Anastasov        :    added secure_tcp defense against DoS attack
 *     Wensong Zhang           :    revisited dropentry defense against DoS attach
 *     Horms                   :    added the fwmark service feature
 *     Wensong Zhang           :    changed to two service hash tables
 *     Julian Anastasov        :    corrected trash_dest lookup for both
 *                                  normal service and fwmark service
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/vmalloc.h>
#include <linux/swap.h>
#include <net/ip_masq.h>

#include <linux/sysctl.h>
#include <linux/ip_fw.h>
#include <linux/ip_masq.h>
#include <linux/proc_fs.h>

#include <linux/inetdevice.h>
#include <linux/ip.h>
#include <net/icmp.h>
#include <net/ip.h>
#include <net/route.h>
#include <net/ip_vs.h>

#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif

EXPORT_SYMBOL(register_ip_vs_scheduler);
EXPORT_SYMBOL(unregister_ip_vs_scheduler);
EXPORT_SYMBOL(ip_vs_bind_masq);
EXPORT_SYMBOL(ip_vs_unbind_masq);
EXPORT_SYMBOL(ip_vs_lookup_dest);
#ifdef CONFIG_IP_VS_DEBUG
EXPORT_SYMBOL(ip_vs_get_debug_level);
#endif

int sysctl_ip_vs_drop_entry = 0;
int sysctl_ip_vs_drop_packet = 0;
int sysctl_ip_vs_secure_tcp = 0;
int sysctl_ip_vs_amemthresh = 1024;
int sysctl_ip_vs_am_droprate = 10;

#ifdef CONFIG_IP_VS_DEBUG
static int sysctl_ip_vs_debug_level = 0;

int ip_vs_get_debug_level(void)
{
	return sysctl_ip_vs_debug_level;
}
#endif


int ip_vs_dropentry = 0;

static inline void update_defense_level(void)
{
	int ip_vs_amem = nr_free_pages+page_cache_size+(buffermem>>PAGE_SHIFT);
	int nomem = (ip_vs_amem < sysctl_ip_vs_amemthresh);

	/* drop_entry */
	switch (sysctl_ip_vs_drop_entry) {
	case 0:
		ip_vs_dropentry = 0;
		break;
	case 1:
		if (nomem) {
			ip_vs_dropentry = 1;
			sysctl_ip_vs_drop_entry = 2;
		} else {
			ip_vs_dropentry = 0;
		}
		break;
	case 2:
		if (nomem) {
			ip_vs_dropentry = 1;
		} else {
			ip_vs_dropentry = 0;
			sysctl_ip_vs_drop_entry = 1;
		};
		break;
	case 3:
		ip_vs_dropentry = 1;
		break;
	}

	/* drop_packet */
	switch (sysctl_ip_vs_drop_packet) {
	case 0:
		ip_masq_drop_rate = 0;
		break;
	case 1:
		if (nomem) {
			ip_masq_drop_rate = ip_masq_drop_counter
				= sysctl_ip_vs_amemthresh /
				(sysctl_ip_vs_amemthresh-ip_vs_amem);
			sysctl_ip_vs_drop_packet = 2;
		} else {
			ip_masq_drop_rate = 0;
		}
		break;
	case 2:
		if (nomem) {
			ip_masq_drop_rate = ip_masq_drop_counter
				= sysctl_ip_vs_amemthresh /
				(sysctl_ip_vs_amemthresh-ip_vs_amem);
		} else {
			ip_masq_drop_rate = 0;
			sysctl_ip_vs_drop_packet = 1;
		}
		break;
	case 3:
		ip_masq_drop_rate = sysctl_ip_vs_am_droprate;
		break;
	}

	/* secure_tcp */
	switch (sysctl_ip_vs_secure_tcp) {
	case 0:
		ip_masq_secure_tcp_set(0);
		break;
	case 1:
		if (nomem) {
			ip_masq_secure_tcp_set(1);
			sysctl_ip_vs_secure_tcp = 2;
		} else {
			ip_masq_secure_tcp_set(0);
		}
		break;
	case 2:
		if (nomem) {
			ip_masq_secure_tcp_set(1);
		} else {
			ip_masq_secure_tcp_set(0);
			sysctl_ip_vs_secure_tcp = 1;
		}
		break;
	case 3:
		ip_masq_secure_tcp_set(1);
		break;
	}
}


static inline int todrop_entry(struct ip_masq *ms)
{
	/*
	 * The drop rate array needs tuning for real environments.
	 */
	static char todrop_rate[9] = {0, 1, 2, 3, 4, 5, 6, 7, 8};
	static char todrop_counter[9] = {0};
	int i;

	if (ms->timeout+jiffies-ms->timer.expires < 60*HZ)
		return 0;

	i = atomic_read(&ms->in_pkts);
	if (i > 8) return 0;

	if (!todrop_rate[i]) return 0;
	if (--todrop_counter[i] > 0) return 0;

	todrop_counter[i] = todrop_rate[i];
	return 1;
}

static inline void ip_vs_random_dropentry(void)
{
	int i;
	struct ip_masq *ms;
	struct list_head *l,*e;
	struct ip_masq *mst;
	void (*fn)(unsigned long);

	/*
	 * Randomly scan 1/32 of the whole table every second
	 */
	for (i=0; i < (IP_VS_TAB_SIZE>>5); i++) {
		/*
		 *  Lock is actually needed in this loop.
		 */
		write_lock(&__ip_masq_lock);

		l = &ip_vs_table[net_random()&IP_VS_TAB_MASK];
		for (e=l->next; e!=l; e=e->next) {
			ms = list_entry(e, struct ip_masq, m_list);
			if (ms->dport == 0)
				/* masq template */
				continue;
			switch(ms->state) {
			case IP_MASQ_S_SYN_RECV:
			case IP_MASQ_S_SYNACK:
				break;

			case IP_MASQ_S_ESTABLISHED:
			case IP_MASQ_S_UDP:
				if (todrop_entry(ms))
					break;
				continue;

			default:
				continue;
			}

			/*
			 * Drop the entry, and drop its mst if not referenced
			 */
			write_unlock(&__ip_masq_lock);
			IP_VS_DBG(4, "Drop masq\n");
			mst = ms->control;
			fn = (ms->timer).function;
			del_sltimer(&ms->timer);
			fn((unsigned long)ms);
			if (mst && !atomic_read(&mst->n_control)) {
				IP_VS_DBG(4, "Drop masq template\n");
				del_sltimer(&mst->timer);
				fn((unsigned long)mst);
			}
			write_lock(&__ip_masq_lock);
		}
		write_unlock(&__ip_masq_lock);
	}
}


/*
 * The following block implements slow timers for IPVS, most code is stolen
 * from linux/kernel/sched.c
 * Slow timer is used to avoid the overhead of cascading timers, when lots
 * of masq entries (>50,000) are cluttered in the system.
 */
#define SHIFT_BITS 6
#define TVN_BITS 8
#define TVR_BITS 10
#define TVN_SIZE (1 << TVN_BITS)
#define TVR_SIZE (1 << TVR_BITS)
#define TVN_MASK (TVN_SIZE - 1)
#define TVR_MASK (TVR_SIZE - 1)

struct sltimer_vec {
	int index;
	struct timer_list *vec[TVN_SIZE];
};

struct sltimer_vec_root {
	int index;
	struct timer_list *vec[TVR_SIZE];
};

static struct sltimer_vec sltv3 = { 0 };
static struct sltimer_vec sltv2 = { 0 };
static struct sltimer_vec_root sltv1 = { 0 };

static struct sltimer_vec * const sltvecs[] = {
	(struct sltimer_vec *)&sltv1, &sltv2, &sltv3
};

#define NOOF_SLTVECS (sizeof(sltvecs) / sizeof(sltvecs[0]))

static unsigned long sltimer_jiffies = 0;

static inline void insert_sltimer(struct timer_list *timer,
				struct timer_list **vec, int idx)
{
	if ((timer->next = vec[idx]))
		vec[idx]->prev = timer;
	vec[idx] = timer;
	timer->prev = (struct timer_list *)&vec[idx];
}

static inline void internal_add_sltimer(struct timer_list *timer)
{
	/*
	 * must be cli-ed when calling this
	 */
	unsigned long expires = timer->expires;
	unsigned long idx = (expires - sltimer_jiffies) >> SHIFT_BITS;

	if (idx < TVR_SIZE) {
		int i = (expires >> SHIFT_BITS) & TVR_MASK;
		insert_sltimer(timer, sltv1.vec, i);
	} else if (idx < 1 << (TVR_BITS + TVN_BITS)) {
		int i = (expires >> (SHIFT_BITS+TVR_BITS)) & TVN_MASK;
		insert_sltimer(timer, sltv2.vec, i);
	} else if ((signed long) idx < 0) {
		/*
		 * can happen if you add a timer with expires == jiffies,
		 * or you set a timer to go off in the past
		 */
		insert_sltimer(timer, sltv1.vec, sltv1.index);
	} else if (idx <= 0xffffffffUL) {
		int i = (expires >> (SHIFT_BITS+TVR_BITS+TVN_BITS)) & TVN_MASK;
		insert_sltimer(timer, sltv3.vec, i);
	} else {
		/* Can only get here on architectures with 64-bit jiffies */
		timer->next = timer->prev = timer;
	}
}

rwlock_t  sltimerlist_lock = RW_LOCK_UNLOCKED;

void add_sltimer(struct timer_list *timer)
{
	write_lock(&sltimerlist_lock);
	if (timer->prev)
		goto bug;
	internal_add_sltimer(timer);
out:
	write_unlock(&sltimerlist_lock);
	return;

bug:
	printk("bug: kernel sltimer added twice at %p.\n",
	       __builtin_return_address(0));
	goto out;
}

static inline int detach_sltimer(struct timer_list *timer)
{
	struct timer_list *prev = timer->prev;
	if (prev) {
		struct timer_list *next = timer->next;
		prev->next = next;
		if (next)
			next->prev = prev;
		return 1;
	}
	return 0;
}

void mod_sltimer(struct timer_list *timer, unsigned long expires)
{
	write_lock(&sltimerlist_lock);
	timer->expires = expires;
	detach_sltimer(timer);
	internal_add_sltimer(timer);
	write_unlock(&sltimerlist_lock);
}

int del_sltimer(struct timer_list * timer)
{
	int ret;

	write_lock(&sltimerlist_lock);
	ret = detach_sltimer(timer);
	timer->next = timer->prev = 0;
	write_unlock(&sltimerlist_lock);
	return ret;
}


static inline void cascade_sltimers(struct sltimer_vec *tv)
{
	/*
	 * cascade all the timers from tv up one level
	 */
	struct timer_list *timer;
	timer = tv->vec[tv->index];
	/*
	 * We are removing _all_ timers from the list, so we don't  have to
	 * detach them individually, just clear the list afterwards.
	 */
	while (timer) {
		struct timer_list *tmp = timer;
		timer = timer->next;
		internal_add_sltimer(tmp);
	}
	tv->vec[tv->index] = NULL;
	tv->index = (tv->index + 1) & TVN_MASK;
}

static inline void run_sltimer_list(void)
{
	write_lock(&sltimerlist_lock);
	while ((long)(jiffies - sltimer_jiffies) >= 0) {
		struct timer_list *timer;
		if (!sltv1.index) {
			int n = 1;
			do {
				cascade_sltimers(sltvecs[n]);
			} while (sltvecs[n]->index == 1 && ++n < NOOF_SLTVECS);
		}
		while ((timer = sltv1.vec[sltv1.index])) {
			void (*fn)(unsigned long) = timer->function;
			unsigned long data = timer->data;
			detach_sltimer(timer);
			timer->next = timer->prev = NULL;
			write_unlock(&sltimerlist_lock);
			fn(data);
			write_lock(&sltimerlist_lock);
		}
		sltimer_jiffies += 1<<SHIFT_BITS;
		sltv1.index = (sltv1.index + 1) & TVR_MASK;
	}
	write_unlock(&sltimerlist_lock);
}

static void sltimer_handler(unsigned long data);

struct timer_list       slow_timer = {
	NULL, NULL,
	0, 0,
	sltimer_handler,
};

/*
 *  Slow timer handler is activated every second
 */
#define SLTIMER_PERIOD       1*HZ

void sltimer_handler(unsigned long data)
{
	run_sltimer_list();

	update_defense_level();
	if (ip_vs_dropentry)
		ip_vs_random_dropentry();

	mod_timer(&slow_timer, (jiffies + SLTIMER_PERIOD));
}


/*
 *  The port number of FTP service (in network order).
 */
#define FTPPORT  __constant_htons(21)
#define FTPDATA  __constant_htons(20)

/*
 *  Lock for IPVS
 */
rwlock_t __ip_vs_lock = RW_LOCK_UNLOCKED;

/*
 *  Hash table: for input and output packets lookups of IPVS
 */
#define IP_MASQ_NTABLES 3

struct list_head *ip_vs_table;

/*
 *  Hash table: for virtual service lookups
 */
#define IP_VS_SVC_TAB_BITS 8
#define IP_VS_SVC_TAB_SIZE (1 << IP_VS_SVC_TAB_BITS)
#define IP_VS_SVC_TAB_MASK (IP_VS_SVC_TAB_SIZE - 1)

/* the service table hashed by <protocol, addr, port> */
struct list_head ip_vs_svc_table[IP_VS_SVC_TAB_SIZE];
/* the service table hashed by fwmark */
struct list_head ip_vs_svc_fwm_table[IP_VS_SVC_TAB_SIZE];

/*
 *  Hash table: for real service lookups
 */
#define IP_VS_RTAB_BITS 4
#define IP_VS_RTAB_SIZE (1 << IP_VS_RTAB_BITS)
#define IP_VS_RTAB_MASK (IP_VS_RTAB_SIZE - 1)

struct list_head ip_vs_rtable[IP_VS_RTAB_SIZE];

/*
 * IPVS scheduler list
 */
struct list_head ip_vs_schedulers;

/*
 * Trash for destinations
 */
struct list_head ip_vs_dest_trash;

/*
 * FTP & NULL virtual service counters
 */
atomic_t ip_vs_ftpsvc_counter = ATOMIC_INIT(0);
atomic_t ip_vs_nullsvc_counter = ATOMIC_INIT(0);

/*
 *  Register a scheduler in the scheduler list
 */
int register_ip_vs_scheduler(struct ip_vs_scheduler *scheduler)
{
	if (!scheduler) {
		IP_VS_ERR("register_ip_vs_scheduler(): NULL arg\n");
		return -EINVAL;
	}

	if (!scheduler->name) {
		IP_VS_ERR("register_ip_vs_scheduler(): NULL scheduler_name\n");
		return -EINVAL;
	}

	if (scheduler->n_list.next != &scheduler->n_list) {
		IP_VS_ERR("register_ip_vs_scheduler(): scheduler already linked\n");
		return -EINVAL;
	}

	/*
	 *      Add it into the d-linked scheduler list
	 */
	list_add(&scheduler->n_list, &ip_vs_schedulers);

	return 0;
}


/*
 *  Unregister a scheduler in the scheduler list
 */
int unregister_ip_vs_scheduler(struct ip_vs_scheduler *scheduler)
{
	if (!scheduler) {
		IP_VS_ERR( "unregister_ip_vs_scheduler(): NULL arg\n");
		return -EINVAL;
	}

	/*
	 *	Only allow unregistration if it is not referenced
	 */
	if (atomic_read(&scheduler->refcnt))  {
		IP_VS_ERR("unregister_ip_vs_scheduler(): is in use by %d guys. failed\n",
			  atomic_read(&scheduler->refcnt));
		return -EINVAL;
	}

	if (scheduler->n_list.next == &scheduler->n_list) {
		IP_VS_ERR("unregister_ip_vs_scheduler(): scheduler is not in the list. failed\n");
		return -EINVAL;
	}

	/*
	 *	Removed it from the d-linked scheduler list
	 */
	list_del(&scheduler->n_list);

	return 0;
}


/*
 *  Bind a service with a scheduler
 *  Must called with the __ip_vs_lock lock, and return bool.
 */
int ip_vs_bind_scheduler(struct ip_vs_service *svc,
			 struct ip_vs_scheduler *scheduler)
{
	if (svc == NULL) {
		IP_VS_ERR("ip_vs_bind_scheduler(): svc arg NULL\n");
		return -EINVAL;
	}
	if (scheduler == NULL) {
		IP_VS_ERR("ip_vs_bind_scheduler(): scheduler arg NULL\n");
		return -EINVAL;
	}

	svc->scheduler = scheduler;
	atomic_inc(&scheduler->refcnt);

	if(scheduler->init_service)
		if(scheduler->init_service(svc) != 0) {
			IP_VS_ERR("ip_vs_bind_scheduler(): init error\n");
			return -EINVAL;
		}

	return 0;
}


/*
 *  Unbind a service with its scheduler
 *  Must called with the __ip_vs_lock lock, and return bool.
 */
int ip_vs_unbind_scheduler(struct ip_vs_service *svc)
{
	struct ip_vs_scheduler *sched;

	if (svc == NULL) {
		IP_VS_ERR("ip_vs_unbind_scheduler(): svc arg NULL\n");
		return -EINVAL;
	}

	sched = svc->scheduler;
	if (sched == NULL) {
		IP_VS_ERR("ip_vs_unbind_scheduler(): svc isn't bound\n");
		return -EINVAL;
	}

	if(sched->done_service)
		if(sched->done_service(svc) != 0) {
			IP_VS_ERR("ip_vs_unbind_scheduler(): done error\n");
			return -EINVAL;
		}

	atomic_dec(&sched->refcnt);
	svc->scheduler = NULL;

	return 0;
}


/*
 *    Get scheduler in the scheduler list by name
 */
struct ip_vs_scheduler * ip_vs_sched_getbyname(const char *sched_name)
{
	struct ip_vs_scheduler *sched;
	struct list_head *l, *e;

	IP_VS_DBG(6, "ip_vs_sched_getbyname(): sched_name \"%s\"\n",
		  sched_name);

	read_lock_bh(&__ip_vs_lock);

	l = &ip_vs_schedulers;
	for (e=l->next; e!=l; e=e->next) {
		sched = list_entry(e, struct ip_vs_scheduler, n_list);
		if (strcmp(sched_name, sched->name)==0) {
			/* HIT */
			read_unlock_bh(&__ip_vs_lock);
			return sched;
		}
	}

	read_unlock_bh(&__ip_vs_lock);
	return NULL;
}


/*
 *  Lookup scheduler and try to load it if it doesn't exist
 */
struct ip_vs_scheduler * ip_vs_lookup_scheduler(const char *sched_name)
{
	struct ip_vs_scheduler *sched;

	/*
	 *  Search for the scheduler by sched_name
	 */
	sched = ip_vs_sched_getbyname(sched_name);

	/*
	 *  If scheduler not found, load the module and search again
	 */
	if (sched == NULL) {
		char module_name[IP_MASQ_TNAME_MAX+8];
		sprintf(module_name,"ip_vs_%s",sched_name);
#ifdef CONFIG_KMOD
		request_module(module_name);
#endif /* CONFIG_KMOD */
		sched = ip_vs_sched_getbyname(sched_name);
	}

	return sched;
}


/*
 *	Returns hash value for IPVS masq entry
 */

static __inline__ unsigned
ip_vs_hash_key(unsigned proto, __u32 addr, __u16 port)
{
	unsigned addrh = ntohl(addr);

	return (proto^addrh^(addrh>>IP_VS_TAB_BITS)^ntohs(port))
		& IP_VS_TAB_MASK;
}


/*
 *	Hashes ip_masq in ip_vs_table by proto,addr,port.
 *	should be called with locked tables.
 *	returns bool success.
 */
int ip_vs_hash(struct ip_masq *ms)
{
	unsigned hash;

	if (ms->flags & IP_MASQ_F_HASHED) {
		IP_VS_ERR("ip_vs_hash(): request for already hashed, "
			  "called from %p\n", __builtin_return_address(0));
		return 0;
	}

	/*
	 * Note: because ip_masq_put sets masq expire only if its
	 *       refcnt==IP_MASQ_NTABLES, otherwise the masq entry
	 *       will never expire.
	 */
	atomic_add(IP_MASQ_NTABLES, &ms->refcnt);

	/*
	 *	Hash by proto,d{addr,port},
	 *      which are client address and port in IPVS.
	 */
	hash = ip_vs_hash_key(ms->protocol, ms->daddr, ms->dport);
	list_add(&ms->m_list, &ip_vs_table[hash]);

	ms->flags |= IP_MASQ_F_HASHED;
	return 1;
}


/*
 *	Unhashes ip_masq from ip_vs_table.
 *	should be called with locked tables.
 *	returns bool success.
 */
int ip_vs_unhash(struct ip_masq *ms)
{
	if (!(ms->flags & IP_MASQ_F_HASHED)) {
		IP_VS_ERR("ip_vs_unhash(): request for unhash flagged, "
			  "called from %p\n", __builtin_return_address(0));
		return 0;
	}

	/*
	 * Remove it from the list and decrease its reference counter.
	 */
	list_del(&ms->m_list);
	atomic_sub(IP_MASQ_NTABLES, &ms->refcnt);

	ms->flags &= ~IP_MASQ_F_HASHED;
	return 1;
}


/*
 *  Gets ip_masq associated with supplied parameters in the ip_vs_table.
 *  Called for pkts coming from OUTside-to-INside.
 *	s_addr, s_port: pkt source address (foreign host)
 *	d_addr, d_port: pkt dest address (load balancer)
 *  Caller must lock tables
 */
struct ip_masq * __ip_vs_in_get(int protocol, __u32 s_addr, __u16 s_port, __u32 d_addr, __u16 d_port)
{
	unsigned hash;
	struct ip_masq *ms;
	struct list_head *l,*e;

	hash = ip_vs_hash_key(protocol, s_addr, s_port);

	l = &ip_vs_table[hash];
	for (e=l->next; e!=l; e=e->next) {
		ms = list_entry(e, struct ip_masq, m_list);
		if (s_addr==ms->daddr && s_port==ms->dport &&
		    d_port==ms->mport && d_addr==ms->maddr &&
		    protocol==ms->protocol) {
			/* HIT */
			atomic_inc(&ms->refcnt);
			goto out;
		}
	}
	ms = NULL;

  out:
	IP_VS_DBG(7, "look/in %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d %s\n",
		  masq_proto_name(protocol),
		  NIPQUAD(s_addr), ntohs(s_port),
		  NIPQUAD(d_addr), ntohs(d_port),
		  ms?"hit":"not hit");

	return ms;
}


/*
 *  Gets ip_masq associated with supplied parameters in the ip_vs_table.
 *  Called for pkts coming from inside-to-OUTside.
 *	s_addr, s_port: pkt source address (inside host)
 *	d_addr, d_port: pkt dest address (foreign host)
 *  Caller must lock tables
 */
struct ip_masq * __ip_vs_out_get(int protocol, __u32 s_addr, __u16 s_port, __u32 d_addr, __u16 d_port)
{
	unsigned hash;
	struct ip_masq *ms;
	struct list_head *l,*e;

	/*
	 *	Check for "full" addressed entries
	 */
	hash = ip_vs_hash_key(protocol, d_addr, d_port);

	l = &ip_vs_table[hash];
	for (e=l->next; e!=l; e=e->next) {
		ms = list_entry(e, struct ip_masq, m_list);
		if (d_addr == ms->daddr && d_port == ms->dport &&
		    s_port == ms->sport && s_addr == ms->saddr &&
		    protocol == ms->protocol) {
			/* HIT */
			atomic_inc(&ms->refcnt);
			goto out;
		}
	}
	ms = NULL;

  out:
	IP_VS_DBG(7, "look/out %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d %s\n",
		  masq_proto_name(protocol),
		  NIPQUAD(s_addr), ntohs(s_port),
		  NIPQUAD(d_addr), ntohs(d_port),
		  ms?"hit":"not hit");

	return ms;
}


/*
 *	Called by ip_vs_sched_persist to look for masq template.
 */
static __inline__ struct ip_masq *ip_vs_in_get
(int protocol, __u32 s_addr, __u16 s_port, __u32 d_addr, __u16 d_port)
{
	struct ip_masq *ms;

	read_lock(&__ip_masq_lock);
	ms = __ip_vs_in_get(protocol, s_addr, s_port, d_addr, d_port);
	read_unlock(&__ip_masq_lock);

	return ms;
}


/*
 *	Returns hash value for virtual service
 */
static __inline__ unsigned
ip_vs_svc_hashkey(unsigned proto, __u32 addr, __u16 port)
{
	register unsigned porth = ntohs(port);

	return (proto^ntohl(addr)^(porth>>IP_VS_SVC_TAB_BITS)^porth)
		& IP_VS_SVC_TAB_MASK;
}

/*
 *	Returns hash value of fwmark for virtual service lookup
 */
static __inline__ unsigned ip_vs_svc_fwm_hashkey(__u32 fwmark)
{
	return fwmark & IP_VS_SVC_TAB_MASK;
}

/*
 *	Hashes ip_vs_service in the ip_vs_svc_table by <proto,addr,port>
 *      or in the ip_vs_svc_fwm_table by fwmark.
 *      Should be called with locked tables.
 *	Returns bool success.
 */
int ip_vs_svc_hash(struct ip_vs_service *svc)
{
	unsigned hash;

	if (svc->flags & IP_VS_SVC_F_HASHED) {
		IP_VS_ERR("ip_vs_svc_hash(): request for already hashed, "
			  "called from %p\n", __builtin_return_address(0));
		return 0;
	}

	if (svc->fwmark == 0) {
		/*
		 *	Hash by <protocol,addr,port> in ip_vs_svc_table
		 */
		hash = ip_vs_svc_hashkey(svc->protocol, svc->addr, svc->port);
		list_add(&svc->s_list, &ip_vs_svc_table[hash]);
	} else {
		/*
		 *      Hash by fwmark in ip_vs_svc_fwm_table
		 */
		hash = ip_vs_svc_fwm_hashkey(svc->fwmark);
		list_add(&svc->f_list, &ip_vs_svc_fwm_table[hash]);
	}

	svc->flags |= IP_VS_SVC_F_HASHED;
	atomic_inc(&svc->refcnt);
	return 1;
}


/*
 *	Unhashes ip_vs_service from ip_vs_svc_table/ip_vs_svc_fwm_table.
 *	Should be called with locked tables.
 *	Returns bool success.
 */
int ip_vs_svc_unhash(struct ip_vs_service *svc)
{
	if (!(svc->flags & IP_VS_SVC_F_HASHED)) {
		IP_VS_ERR("ip_vs_svc_unhash(): request for unhash flagged, "
			  "called from %p\n", __builtin_return_address(0));
		return 0;
	}

	if (svc->fwmark == 0) {
		/*
		 * Remove it from the ip_vs_svc_table table.
		 */
		list_del(&svc->s_list);
	} else {
		/*
		 * Remove it from the ip_vs_svc_fwm_table table.
		 */
		list_del(&svc->f_list);
	}

	svc->flags &= ~IP_VS_SVC_F_HASHED;
	atomic_dec(&svc->refcnt);
	return 1;
}


/*
 *  Lookup service by {proto,addr,port} in the service table.
 */
static __inline__ struct ip_vs_service *
__ip_vs_lookup_service(__u16 protocol, __u32 vaddr, __u16 vport)
{
	unsigned hash;
	struct ip_vs_service *svc;
	struct list_head *l,*e;

	/*
	 *	Check for "full" addressed entries
	 *      Note: as long as IP_VS_SVC_TAB_BITS is larger than zero,
	 *      <TCP,addr,port> and <UDP,addr,port> have different hash
	 *      keys, there is no need to do protcol checking.
	 */
	hash = ip_vs_svc_hashkey(protocol, vaddr, vport);

	l = &ip_vs_svc_table[hash];
	for (e=l->next; e!=l; e=e->next) {
		svc = list_entry(e, struct ip_vs_service, s_list);
		if ((svc->addr == vaddr)
		    && (svc->port == vport)) {
			/* HIT */
			return svc;
		}
	}

	return NULL;
}


/*
 *  Lookup service by fwmark in the service table.
 */
static __inline__ struct ip_vs_service * __ip_vs_lookup_svc_fwm(__u32 fwmark)
{
	unsigned hash;
	struct ip_vs_service *svc;
	struct list_head *l,*e;

	/*
	 *	Check for fwmark-indexed entries
	 */
	hash = ip_vs_svc_fwm_hashkey(fwmark);

	l = &ip_vs_svc_fwm_table[hash];
	for (e=l->next; e!=l; e=e->next) {
		svc = list_entry(e, struct ip_vs_service, f_list);
		if (svc->fwmark == fwmark) {
			/* HIT */
			return svc;
		}
	}

	return NULL;
}

struct ip_vs_service *
ip_vs_lookup_service(__u32 fwmark, __u16 protocol, __u32 vaddr, __u16 vport)
{
	struct ip_vs_service *svc;

	read_lock(&__ip_vs_lock);

	if (fwmark) {
		/*
		 *	Check the table hashed by fwmark first
		 */
		svc = __ip_vs_lookup_svc_fwm(fwmark);
		if (svc)
			goto out;
	}

	/*
	 *	Check the table hashed by <protocol,addr,port>
	 *      first for "full" addressed entries
	 */
	svc = __ip_vs_lookup_service(protocol, vaddr, vport);

	if (svc == NULL
	    && protocol == IPPROTO_TCP
	    && atomic_read(&ip_vs_ftpsvc_counter)
	    && (vport==FTPDATA || ntohs(vport)>=PROT_SOCK)){
		/*
		 * Check if ftp service entry exists, the packet
		 * might belong to FTP data connections.
		 */
		svc = __ip_vs_lookup_service(protocol, vaddr, FTPPORT);
	}

	if (svc == NULL
	    && atomic_read(&ip_vs_nullsvc_counter)) {
		/*
		 * Check if the catch-all port (port zero) exists
		 */
		svc = __ip_vs_lookup_service(protocol, vaddr, 0);
	}

 out:
	read_unlock(&__ip_vs_lock);

	IP_VS_DBG(5, "lookup_service fwm %d %s %u.%u.%u.%u:%d %s\n",
		  fwmark,
		  masq_proto_name(protocol),
		  NIPQUAD(vaddr), ntohs(vport),
		  svc?"hit":"not hit");

	return svc;
}


/*
 *      Bind a destination with a service
 */
static inline void
__ip_vs_bind_svc(struct ip_vs_dest *dest, struct ip_vs_service *svc)
{
	atomic_inc(&svc->refcnt);
	dest->svc = svc;
}

/*
 *      Unbind a destination with its service
 */
static inline void
__ip_vs_unbind_svc(struct ip_vs_dest *dest)
{
	struct ip_vs_service *svc = dest->svc;

	dest->svc = NULL;
	if (atomic_dec_and_test(&svc->refcnt)) {
		IP_VS_DBG(2, "release svc %s %u.%u.%u.%u:%d\n",
			  masq_proto_name(svc->protocol),
			  NIPQUAD(svc->addr), ntohs(svc->port));
		kfree_s(svc, sizeof(struct ip_vs_service));
	}
}


/*
 *	Returns hash value for real service
 */
static __inline__ unsigned
ip_vs_rs_hashkey(__u32 addr, __u16 port)
{
	register unsigned porth = ntohs(port);

	return (ntohl(addr)^(porth>>IP_VS_RTAB_BITS)^porth) & IP_VS_RTAB_MASK;
}

/*
 *	Hashes ip_vs_dest in ip_vs_rtable by proto,addr,port.
 *	should be called with locked tables.
 *	returns bool success.
 */
int ip_vs_rs_hash(struct ip_vs_dest *dest)
{
	unsigned hash;

	if (!list_empty(&dest->d_list)) {
		return 0;
	}

	/*
	 *	Hash by proto,addr,port,
	 *      which are the parameters of the real service.
	 */
	hash = ip_vs_rs_hashkey(dest->addr, dest->port);
	list_add(&dest->d_list, &ip_vs_rtable[hash]);

	return 1;
}

/*
 *	UNhashes ip_vs_dest from ip_vs_rtable.
 *	should be called with locked tables.
 *	returns bool success.
 */
int ip_vs_rs_unhash(struct ip_vs_dest *dest)
{
	/*
	 * Remove it from the ip_vs_rtable table.
	 */
	if (!list_empty(&dest->d_list)) {
		list_del(&dest->d_list);
		INIT_LIST_HEAD(&dest->d_list);
	}

	return 1;
}

/*
 *  Lookup real service by {proto,addr,port} in the real service table.
 */
struct ip_vs_dest * __ip_vs_lookup_real_service(__u16 protocol,
						__u32 daddr, __u16 dport)
{
	unsigned hash;
	struct ip_vs_dest *dest;
	struct list_head *l,*e;

	/*
	 *	Check for "full" addressed entries
	 *	Return the first found entry
	 */
	hash = ip_vs_rs_hashkey(daddr, dport);

	l = &ip_vs_rtable[hash];
	for (e=l->next; e!=l; e=e->next) {
		dest = list_entry(e, struct ip_vs_dest, d_list);
		if ((dest->addr == daddr)
		    && (dest->port == dport)
		    && ((dest->protocol == protocol) || dest->vfwmark)) {
			/* HIT */
			return dest;
		}
	}

	return NULL;
}

/*
 *  Lookup destination by {addr,port} in the given service
 */
struct ip_vs_dest * ip_vs_lookup_dest(struct ip_vs_service *svc,
				      __u32 daddr, __u16 dport)
{
	struct ip_vs_dest *dest;
	struct list_head *l, *e;

	read_lock_bh(&__ip_vs_lock);

	/*
	 * Find the destination for the given service
	 */
	l = &svc->destinations;
	for (e=l->next; e!=l; e=e->next) {
		dest = list_entry(e, struct ip_vs_dest, n_list);
		if ((dest->addr == daddr) && (dest->port == dport)) {
			/* HIT */
			read_unlock_bh(&__ip_vs_lock);
			return dest;
		}
	}

	read_unlock_bh(&__ip_vs_lock);
	return NULL;
}


/*
 *  Lookup dest by {svc,addr,port} in the destination trash.
 *  Called by ip_vs_add_dest with the __ip_vs_lock.
 *  The destination trash is used to hold the destinations that are removed
 *  from the service table but are still referenced by some masq entries.
 *  The reason to add the destination trash is when the dest is temporary
 *  down (either by administrator or by monitor program), the dest can be
 *  picked back from the trash, the remaining connections to the dest can
 *  continue, and the counting information of the dest is also useful for
 *  scheduling.
 */
struct ip_vs_dest * __ip_vs_get_trash_dest(struct ip_vs_service *svc,
					   __u32 daddr, __u16 dport)
{
	struct ip_vs_dest *dest;
	struct list_head *l, *e;

	/*
	 * Find the destination in trash
	 */
	l = &ip_vs_dest_trash;
	for (e=l->next; e!=l; e=e->next) {
		dest = list_entry(e, struct ip_vs_dest, n_list);
		IP_VS_DBG(3, "Destination %u/%u.%u.%u.%u:%d still in trash, "
			  "refcnt=%d\n",
			  dest->vfwmark,
			  NIPQUAD(dest->addr), ntohs(dest->port),
			  atomic_read(&dest->refcnt));
		if (dest->addr == daddr &&
		    dest->port == dport &&
		    dest->vfwmark == svc->fwmark &&
		    (svc->fwmark ||
		     (dest->protocol == svc->protocol &&
		      dest->vaddr == svc->addr &&
		      dest->vport == svc->port))) {
			/* HIT */
			return dest;
		}

		/*
		 * Try to purge the destination from trash if not referenced
		 */
		if (atomic_read(&dest->refcnt) == 1) {
			IP_VS_DBG(3, "Remove destination %u/%u.%u.%u.%u:%d "
				  "from trash\n",
				  dest->vfwmark,
				  NIPQUAD(dest->addr), ntohs(dest->port));
			e = e->prev;
			list_del(&dest->n_list);
			__ip_vs_unbind_svc(dest);
			kfree_s(dest, sizeof(*dest));
		}
	}
	return NULL;
}


/*
 *  Update a destination in the given service
 */
void __ip_vs_update_dest(struct ip_vs_service *svc, struct ip_vs_dest *dest,
	struct ip_masq_ctl *mctl)
{
	struct ip_vs_user *mm =  &mctl->u.vs_user;

	/*
	 *    Set the weight and the flags
	 */
	dest->weight = mm->weight;
	dest->masq_flags = mm->masq_flags;

	dest->masq_flags |= IP_MASQ_F_VS;
	dest->masq_flags |= IP_MASQ_F_VS_INACTIVE;

	/*
	 *    Check if local node and update the flags
	 */
	if (inet_addr_type(mm->daddr) == RTN_LOCAL) {
		dest->masq_flags = (dest->masq_flags & ~IP_MASQ_F_VS_FWD_MASK)
			| IP_MASQ_F_VS_LOCALNODE;
	}

	/*
	 *    Set the IP_MASQ_F_VS_NO_OUTPUT flag if not masquerading
	 */
	if ((dest->masq_flags & IP_MASQ_F_VS_FWD_MASK) != 0) {
		dest->masq_flags |= IP_MASQ_F_VS_NO_OUTPUT;
	} else {
		/*
		 *    Put the real service in ip_vs_rtable if not present.
		 *    For now only for NAT!
		 */
		ip_vs_rs_hash(dest);
	}


	/* bind the service */
	if (!dest->svc) {
		__ip_vs_bind_svc(dest, svc);
	} else {
		if (dest->svc != svc) {
			__ip_vs_unbind_svc(dest);
			__ip_vs_bind_svc(dest, svc);
		}
	}

	/*
	 *    Set the dest status flags
	 */
	dest->flags |= IP_VS_DEST_F_AVAILABLE;
}


/*
 *  Create a destination for the given service
 */
struct ip_vs_dest *ip_vs_new_dest(struct ip_vs_service *svc,
				  struct ip_masq_ctl *mctl)
{
	struct ip_vs_dest *dest;
	struct ip_vs_user *mm =  &mctl->u.vs_user;

	EnterFunction(2);

	dest = (struct ip_vs_dest*) kmalloc(sizeof(struct ip_vs_dest),
					    GFP_ATOMIC);
	if (dest == NULL) {
		IP_VS_ERR("ip_vs_new_dest: kmalloc failed.\n");
		return NULL;
	}
	memset(dest, 0, sizeof(struct ip_vs_dest));

	dest->protocol = svc->protocol;
	dest->vaddr = svc->addr;
	dest->vport = svc->port;
	dest->vfwmark = svc->fwmark;
	dest->addr = mm->daddr;
	dest->port = mm->dport;

	atomic_set(&dest->activeconns, 0);
	atomic_set(&dest->inactconns, 0);
	atomic_set(&dest->refcnt, 0);

	INIT_LIST_HEAD(&dest->d_list);
	dest->stats.lock = SPIN_LOCK_UNLOCKED;
	__ip_vs_update_dest(svc, dest, mctl);

	LeaveFunction(2);

	return dest;
}


/*
 *  Add a destination into an existing service
 */
int ip_vs_add_dest(struct ip_vs_service *svc, struct ip_masq_ctl *mctl)
{
	struct ip_vs_dest *dest;
	struct ip_vs_user *mm =  &mctl->u.vs_user;
	__u32 daddr = mm->daddr;
	__u16 dport = mm->dport;

	EnterFunction(2);

	if (mm->weight < 0) {
		IP_VS_ERR("ip_vs_add_dest(): server weight less than zero\n");
		return -ERANGE;
	}

	/*
	 * Check if the dest already exists in the list
	 */
	dest = ip_vs_lookup_dest(svc, daddr, dport);
	if (dest != NULL) {
		IP_VS_DBG(1, "ip_vs_add_dest(): dest already exists\n");
		return -EEXIST;
	}

	write_lock_bh(&__ip_vs_lock);

	/*
	 * Check if the dest already exists in the trash and
	 * is from the same service
	 */
	dest = __ip_vs_get_trash_dest(svc, daddr, dport);
	if (dest != NULL) {
		IP_VS_DBG(3, "Get destination %u.%u.%u.%u:%d from trash, "
			  "refcnt=%d, service %u.%u.%u.%u:%d\n",
			  NIPQUAD(daddr), ntohs(dport),
			  atomic_read(&dest->refcnt),
			  NIPQUAD(dest->vaddr),
			  ntohs(dest->vport));

		/*
		 * Get the destination from the trash
		 */
		list_del(&dest->n_list);
		list_add(&dest->n_list, &svc->destinations);

		__ip_vs_update_dest(svc, dest, mctl);

		write_unlock_bh(&__ip_vs_lock);
		return 0;
	}

	/*
	 * Allocate and initialize the dest structure
	 */
	dest = ip_vs_new_dest(svc, mctl);
	if (dest == NULL) {
		write_unlock_bh(&__ip_vs_lock);
		IP_VS_ERR("ip_vs_add_dest(): out of memory\n");
		return -ENOMEM;
	}

	/*
	 * Add the dest entry into the list
	 */
	list_add(&dest->n_list, &svc->destinations);
	atomic_inc(&dest->refcnt);

	write_unlock_bh(&__ip_vs_lock);

	LeaveFunction(2);
	return 0;
}


/*
 *  Edit a destination in the given service
 */
int ip_vs_edit_dest(struct ip_vs_service *svc, struct ip_masq_ctl *mctl)
{
	struct ip_vs_dest *dest;
	struct ip_vs_user *mm =  &mctl->u.vs_user;
	__u32 daddr = mm->daddr;
	__u16 dport = mm->dport;

	EnterFunction(2);

	if (mm->weight < 0) {
		IP_VS_ERR("ip_vs_add_dest(): server weight less than zero\n");
		return -ERANGE;
	}

	/*
	 *    Lookup the destination list
	 */
	dest = ip_vs_lookup_dest(svc, daddr, dport);
	if (dest == NULL) {
		IP_VS_DBG(1, "ip_vs_edit_dest(): dest doesn't exist\n");
		return -ENOENT;
	}

	write_lock_bh(&__ip_vs_lock);

	__ip_vs_update_dest(svc, dest, mctl);

	write_unlock_bh(&__ip_vs_lock);

	LeaveFunction(2);
	return 0;
}


/*
 *  Delete a destination from the given service
 */
void __ip_vs_del_dest(struct ip_vs_dest *dest)
{
	dest->flags &= ~IP_VS_DEST_F_AVAILABLE;

	/*
	 *  Remove it from the d-linked destination list.
	 */
	list_del(&dest->n_list);

	/*
	 *  Remove it from the d-linked list with the real services.
	 */
	ip_vs_rs_unhash(dest);

	/*
	 *  Decrease the refcnt of the dest, and free the dest
	 *  if nobody refers to it (refcnt=0). Otherwise, throw
	 *  the destination into the trash.
	 */
	if (atomic_dec_and_test(&dest->refcnt)) {
		/* simply decrease svc->refcnt here, let the caller check
		   and release the service if nobody refers to it.
		   Only user context can release destination and service,
		   and only user context can update virtual service at a
		   time, so the operation here is OK */
		atomic_dec(&dest->svc->refcnt);
		kfree_s(dest, sizeof(*dest));
	} else {
		IP_VS_DBG(3, "Move dest %u.%u.%u.%u:%d into trash, "
			  "refcnt=%d\n",
			  NIPQUAD(dest->addr), ntohs(dest->port),
			  atomic_read(&dest->refcnt));
		list_add(&dest->n_list, &ip_vs_dest_trash);
		atomic_inc(&dest->refcnt);
	}
}

int ip_vs_del_dest(struct ip_vs_service *svc, struct ip_masq_ctl *mctl)
{
	struct ip_vs_dest *dest;
	struct ip_vs_user *mm =  &mctl->u.vs_user;
	__u32 daddr = mm->daddr;
	__u16 dport = mm->dport;

	EnterFunction(2);

	/*
	 *    Lookup the destination list
	 */
	dest = ip_vs_lookup_dest(svc, daddr, dport);
	if (dest == NULL) {
		IP_VS_DBG(1, "ip_vs_del_dest(): destination not found!\n");
		return -ENOENT;
	}

	write_lock_bh(&__ip_vs_lock);

	/*
	 *  Remove dest from the destination list
	 */
	__ip_vs_del_dest(dest);

	/*
	 *  Called the update_service function of its scheduler
	 */
	svc->scheduler->update_service(svc);

	write_unlock_bh(&__ip_vs_lock);

	LeaveFunction(2);

	return 0;
}


/*
 *  Add a service into the service hash table
 */
int ip_vs_add_service(struct ip_masq_ctl *mctl)
{
	struct ip_vs_user *mm =  &mctl->u.vs_user;
	__u16  protocol = mm->protocol;
	__u32 vaddr = mm->vaddr;
	__u16 vport = mm->vport;
	__u32 vfwmark = mm->vfwmark;

	int ret = 0;
	struct ip_vs_scheduler *sched;
	struct ip_vs_service *svc;

	EnterFunction(2);

	/*
	 * Lookup the scheduler, by 'mctl->m_tname'
	 */
	sched = ip_vs_lookup_scheduler(mctl->m_tname);
	if (sched == NULL) {
		IP_VS_INFO("Scheduler module ip_vs_%s.o not found\n",
			   mctl->m_tname);
		return -ENOENT;
	}

	write_lock_bh(&__ip_vs_lock);

	/*
	 * Check if the service already exists
	 */
	if (vfwmark == 0)
		svc = __ip_vs_lookup_service(protocol, vaddr, vport);
	else
		svc = __ip_vs_lookup_svc_fwm(vfwmark);

	if (svc != NULL) {
		IP_VS_DBG(1, "ip_vs_add_service: service already exists.\n");
		ret = -EEXIST;
		goto out;
	}

	svc = (struct ip_vs_service*)
		kmalloc(sizeof(struct ip_vs_service), GFP_ATOMIC);
	if (svc == NULL) {
		IP_VS_DBG(1, "ip_vs_add_service: kmalloc failed.\n");
		ret = -ENOMEM;
		goto out;
	}
	memset(svc, 0, sizeof(struct ip_vs_service));

	svc->protocol = protocol;
	svc->addr = vaddr;
	svc->port = vport;
	svc->fwmark = vfwmark;
	svc->flags = mm->vs_flags;
	svc->timeout = mm->timeout;
	svc->netmask = mm->netmask;

	INIT_LIST_HEAD(&svc->destinations);
	atomic_set(&svc->refcnt, 0);
	svc->stats.lock = SPIN_LOCK_UNLOCKED;

	/*
	 *    Bind the scheduler
	 */
	ip_vs_bind_scheduler(svc, sched);

	/*
	 *    Hash the service into the service table
	 */
	ip_vs_svc_hash(svc);

	/*
	 *    Update the virtual service counters
	 */
	if (vport == FTPPORT)
		atomic_inc(&ip_vs_ftpsvc_counter);
	else if (vport == 0)
		atomic_inc(&ip_vs_nullsvc_counter);

  out:
	write_unlock_bh(&__ip_vs_lock);
	LeaveFunction(2);
	return ret;
}


/*
 *  Edit a service and bind it with a new scheduler
 */
int ip_vs_edit_service(struct ip_vs_service *svc, struct ip_masq_ctl *mctl)
{
	struct ip_vs_user *mm =  &mctl->u.vs_user;
	struct ip_vs_scheduler *sched;

	EnterFunction(2);

	/*
	 * Lookup the scheduler, by 'mctl->m_tname'
	 */
	sched = ip_vs_lookup_scheduler(mctl->m_tname);
	if (sched == NULL) {
		IP_VS_INFO("Scheduler module ip_vs_%s.o not found\n",
			   mctl->m_tname);
		return -ENOENT;
	}

	write_lock_bh(&__ip_vs_lock);

	/*
	 *    Set the flags and timeout value
	 */
	svc->flags = mm->vs_flags | IP_VS_SVC_F_HASHED;
	svc->timeout = mm->timeout;
	svc->netmask = mm->netmask;

	/*
	 *    Unbind the old scheduler
	 */
	ip_vs_unbind_scheduler(svc);

	/*
	 *    Bind the new scheduler
	 */
	ip_vs_bind_scheduler(svc, sched);

	write_unlock_bh(&__ip_vs_lock);

	LeaveFunction(2);
	return 0;
}


/*
 *  Delete a service from the service list
 */
int __ip_vs_del_service(struct ip_vs_service *svc)
{
	struct list_head *l;
	struct ip_vs_dest *dest;

	/*
	 *    Unbind scheduler
	 */
	ip_vs_unbind_scheduler(svc);

	/*
	 *    Unlink the whole destination list
	 */
	l = &svc->destinations;
	while (l->next != l) {
		dest = list_entry(l->next, struct ip_vs_dest, n_list);
		__ip_vs_del_dest(dest);
	}

	/*
	 *    Unhash it from the service table
	 */
	if (ip_vs_svc_unhash(svc)) {
		/*
		 *    Update the virtual service counters
		 */
		if (svc->port == FTPPORT)
			atomic_dec(&ip_vs_ftpsvc_counter);
		else if (svc->port == 0)
			atomic_dec(&ip_vs_nullsvc_counter);

		/*
		 *    Free the service if nobody refers to it
		 */
		if (atomic_read(&svc->refcnt) == 0) {
			IP_VS_DBG(2, "release svc %s %u.%u.%u.%u:%d\n",
				  masq_proto_name(svc->protocol),
				  NIPQUAD(svc->addr), ntohs(svc->port));
			kfree_s(svc, sizeof(struct ip_vs_service));
		}
	} else {
		/*
		 *  Called the update_service function of its scheduler
		 */
		svc->scheduler->update_service(svc);
		return -EPERM;
	}

	return 0;
}

int ip_vs_del_service(struct ip_vs_service *svc)
{
	EnterFunction(2);

	if (svc == NULL)
		return -EEXIST;

	write_lock_bh(&__ip_vs_lock);

	__ip_vs_del_service(svc);

	write_unlock_bh(&__ip_vs_lock);
	LeaveFunction(2);
	return 0;
}


/*
 *  Flush all the virtual services
 */
int ip_vs_flush(void)
{
	int idx;
	struct ip_vs_service *svc;
	struct list_head *l;

	write_lock_bh(&__ip_vs_lock);

	/*
	 *    Flush the service table hashed by <protocol,addr,port>
	 */
	for(idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
		l = &ip_vs_svc_table[idx];
		while (l->next != l) {
			svc = list_entry(l->next,struct ip_vs_service,s_list);

			if (__ip_vs_del_service(svc))
				goto out;
		}
	}

	/*
	 *    Flush the service table hashed by fwmark
	 */
	for(idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
		l = &ip_vs_svc_fwm_table[idx];
		while (l->next != l) {
			svc = list_entry(l->next,struct ip_vs_service,f_list);

			if (__ip_vs_del_service(svc))
				goto out;
		}
	}

  out:
	write_unlock_bh(&__ip_vs_lock);
	return 0;
}


/*
 *  Change the connection counter and the flags if the masq state changes
 *  Called by the masq_tcp_state function.
 */
void ip_vs_set_state(struct ip_masq *ms, int new_state)
{
	struct ip_vs_dest *dest = ms->dest;

	if (dest &&
	    (ms->flags & IP_MASQ_F_VS) && (new_state != ms->state)) {
		if (!(ms->flags & IP_MASQ_F_VS_INACTIVE) &&
		    (new_state != IP_MASQ_S_ESTABLISHED)) {
			atomic_dec(&dest->activeconns);
			atomic_inc(&dest->inactconns);
			ms->flags |= IP_MASQ_F_VS_INACTIVE;
		} else if ((ms->flags & IP_MASQ_F_VS_INACTIVE) &&
			   (new_state == IP_MASQ_S_ESTABLISHED)) {
			atomic_inc(&dest->activeconns);
			atomic_dec(&dest->inactconns);
			ms->flags &= ~IP_MASQ_F_VS_INACTIVE;
		}

		IP_VS_DBG(8, "Set-state masq fwd:%c s:%s c:%u.%u.%u.%u:%d "
			  "v:%u.%u.%u.%u:%d d:%u.%u.%u.%u:%d flg:%X cnt:%d\n",
			  ip_vs_fwd_tag(ms), ip_masq_state_name(ms->state),
			  NIPQUAD(ms->daddr), ntohs(ms->dport),
			  NIPQUAD(ms->maddr), ntohs(ms->mport),
			  NIPQUAD(ms->saddr), ntohs(ms->sport),
			  ms->flags, atomic_read(&ms->refcnt));
	}
}


/*
 *  Bind a masq entry with a virtual service destination
 *  Called when a new masq entry is created for VS.
 */
void ip_vs_bind_masq(struct ip_masq *ms, struct ip_vs_dest *dest)
{
	ms->flags |= dest->masq_flags;
	ms->dest = dest;

	/*
	 *    Increase the refcnt counter of the dest.
	 */
	atomic_inc(&dest->refcnt);

	IP_VS_DBG(9, "Bind-masq fwd:%c s:%s c:%u.%u.%u.%u:%d v:%u.%u.%u.%u:%d "
		  "d:%u.%u.%u.%u:%d flg:%X cnt:%d destcnt:%d\n",
		  ip_vs_fwd_tag(ms), ip_masq_state_name(ms->state),
		  NIPQUAD(ms->daddr), ntohs(ms->dport),
		  NIPQUAD(ms->maddr), ntohs(ms->mport),
		  NIPQUAD(ms->saddr), ntohs(ms->sport),
		  ms->flags, atomic_read(&ms->refcnt),
		  atomic_read(&dest->refcnt));
}


/*
 *  Unbind a masq entry with its VS destination
 *  Called by the masq_expire function.
 */
void ip_vs_unbind_masq(struct ip_masq *ms)
{
	struct ip_vs_dest *dest = ms->dest;

	IP_VS_DBG(9, "Unbind-masq fwd:%c s:%s c:%u.%u.%u.%u:%d v:%u.%u.%u.%u:%d "
		  "d:%u.%u.%u.%u:%d flg:%X cnt:%d destcnt:%d\n",
		  ip_vs_fwd_tag(ms), ip_masq_state_name(ms->state),
		  NIPQUAD(ms->daddr),ntohs(ms->dport),
		  NIPQUAD(ms->maddr),ntohs(ms->mport),
		  NIPQUAD(ms->saddr),ntohs(ms->sport),
		  ms->flags, atomic_read(&ms->refcnt),
		  atomic_read(&dest->refcnt));

	if (dest) {
		/*
		 * Decrease the inactconns or activeconns counter
		 * if it is not a masq template (ms->dport!=0).
		 */
		if (ms->dport) {
			if (ms->flags & IP_MASQ_F_VS_INACTIVE) {
				atomic_dec(&dest->inactconns);
			} else {
				atomic_dec(&dest->activeconns);
			}
		}

		/*
		 * Simply decrease the refcnt of the dest, because the
		 * dest will be either in service's destination list
		 * or in the trash.
		 */
		atomic_dec(&dest->refcnt);
	}
}


/*
 *  Checking if the destination of a masq template is available.
 *  If available, return 1, otherwise return 0 and invalidate this
 *  masq template.
 */
int ip_vs_check_template(struct ip_masq *mst)
{
	struct ip_vs_dest *dest = mst->dest;

	/*
	 * Checking the dest server status.
	 */
	if ((dest == NULL) ||
	    !(dest->flags & IP_VS_DEST_F_AVAILABLE)) {
		IP_VS_DBG(9, "check_template: dest not available for prot %s "
			  "src %u.%u.%u.%u:%d dest %u.%u.%u.%u:%d -> %X:%X\n",
			  masq_proto_name(mst->protocol),
			  NIPQUAD(mst->daddr), ntohs(mst->dport),
			  NIPQUAD(mst->maddr), ntohs(mst->mport),
			  (dest!=NULL)? ntohl(dest->addr):0,
			  (dest!=NULL)? ntohs(dest->port):0);

		/*
		 * Invalidate the masq template
		 */
		ip_vs_unhash(mst);
		mst->sport = 65535;
		mst->mport = 65535;
		mst->dport = 0;
		ip_vs_hash(mst);

		/*
		 * Simply decrease the refcnt of the template,
		 * don't restart its timer.
		 */
		atomic_dec(&mst->refcnt);
		return 0;
	}
	return 1;
}


/*
 *  IPVS persistent scheduling function
 *  It creates a masq entry according to its template if exists, or selects
 *  a server and creates a masq entry plus a template.
 */
struct ip_masq *
ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
{
	struct ip_masq *ms = NULL;
	struct ip_vs_dest *dest;
	const __u16 *portp;
	struct ip_masq *mst;
	__u16  dport;    /* destination port to forward */
	__u32  snet;	 /* source network of the client, after masking */

	portp = (__u16 *)&(((char *)iph)[iph->ihl*4]);

	/* Mask saddr with the netmask to adjust template granularity */
	snet = iph->saddr & svc->netmask;

	IP_VS_DBG(6, "P-schedule: src %u.%u.%u.%u:%d dest %u.%u.%u.%u:%d "
		  "snet %u.%u.%u.%u/%u.%u.%u.%u\n",
		  NIPQUAD(iph->saddr), ntohs(portp[0]),
		  NIPQUAD(iph->daddr), ntohs(portp[1]),
		  NIPQUAD(snet), NIPQUAD(svc->netmask));

	/*
	 * As far as we know, FTP is a very complicated network protocol, and
	 * it uses control connection and data connections. For active FTP,
	 * FTP server initilize data connection to the client, its source port
	 * is often 20. For passive FTP, FTP server tells the clients the port
	 * that it passively listens to,  and the client issues the data
	 * connection. In the tunneling or direct routing mode, the load
	 * balancer is on the client-to-server half of connection, the port
	 * number is unknown to the load balancer. So, a template masq like
	 * <daddr, 0, maddr, 0, saddr, 0> is created for persistent FTP
	 * service, and a template like <daddr, 0, maddr, mport, saddr, sport>
	 * is created for other persistent services.
	 */
	if (portp[1] == svc->port) {
		/* Check if a template already exists */
		if (svc->port != FTPPORT)
			mst = ip_vs_in_get(iph->protocol, snet, 0,
					   iph->daddr, portp[1]);
		else
			mst = ip_vs_in_get(iph->protocol, snet, 0,
					   iph->daddr, 0);

		if (!mst || !ip_vs_check_template(mst)) {
			/*
			 * No template found or the dest of the masq
			 * template is not available.
			 */
			read_lock(&__ip_vs_lock);

			dest = svc->scheduler->schedule(svc, iph);
			if (dest == NULL) {
				IP_VS_DBG(1, "P-schedule: no dest found.\n");
				read_unlock(&__ip_vs_lock);
				return NULL;
			}

			/*
			 * Create a template like <protocol,daddr,0,
			 * maddr,mport,saddr,sport> for non-ftp service,
			 * and <protocol,daddr,0,maddr,0,saddr,0>
			 * for ftp service.
			 */
			if (svc->port != FTPPORT)
				mst = ip_masq_new_vs(iph->protocol,
						     iph->daddr, portp[1],
						     dest->addr, dest->port,
						     snet, 0,
						     0);
			else
				mst = ip_masq_new_vs(iph->protocol,
						     iph->daddr, 0,
						     dest->addr, 0,
						     snet, 0,
						     0);
			if (mst == NULL) {
				IP_VS_ERR("ip_masq_new_vs template failed\n");
				read_unlock(&__ip_vs_lock);
				return NULL;
			}

			/*
			 * Bind the template with dest and set timeout.
			 */
			ip_vs_bind_masq(mst, dest);
			mst->timeout = svc->timeout;

			read_unlock(&__ip_vs_lock);
		} else {
			/*
			 * Template found and its destination is available.
			 */
			dest = mst->dest;

			/*
			 * Delete its timer so that it can be put back.
			 */
			del_sltimer(&mst->timer);
		}
		dport = dest->port;
	} else {
		/*
		 * Note: persistent fwmark-based services and persistent
		 * port zero service are handled here.
		 * fwmark template: <IPPROTO_IP,daddr,0,fwmark,0,saddr,0>
		 * port zero template: <protocol,daddr,0,maddr,0,saddr,0>
		 */
		if (svc->fwmark)
			mst = ip_vs_in_get(IPPROTO_IP, snet, 0,
					   htonl(svc->fwmark), 0);
		else
			mst = ip_vs_in_get(iph->protocol,
					   snet, 0, iph->daddr, 0);

		if (!mst || !ip_vs_check_template(mst)) {
			/*
			 * If it is not persistent port zero, return NULL.
			 */
			if (svc->port)
				return NULL;

			read_lock(&__ip_vs_lock);

			dest = svc->scheduler->schedule(svc, iph);
			if (dest == NULL) {
				IP_VS_DBG(1, "P-schedule: no dest found.\n");
				read_unlock(&__ip_vs_lock);
				return NULL;
			}

			/*
			 * Create a template according to the service
			 */
			if (svc->fwmark)
				mst = ip_masq_new_vs(IPPROTO_IP,
						     htonl(svc->fwmark), 0,
						     dest->addr, 0,
						     snet, 0,
						     0);
			else
				mst = ip_masq_new_vs(iph->protocol,
						     iph->daddr, 0,
						     dest->addr, 0,
						     snet, 0,
						     0);
			if (mst == NULL) {
				IP_VS_ERR("ip_masq_new_vs template failed\n");
				read_unlock(&__ip_vs_lock);
				return NULL;
			}

			/*
			 * Bind the template with dest and set timeout.
			 */
			ip_vs_bind_masq(mst, dest);
			mst->timeout = svc->timeout;
			read_unlock(&__ip_vs_lock);
		} else {
			dest = mst->dest;

			/*
			 * Delete its timer so that it can be put back.
			 */
			del_sltimer(&mst->timer);
		}
		dport = portp[1];
	}

	/*
	 *    Create a new masq according to the template
	 */
	ms = ip_masq_new_vs(iph->protocol,
			    iph->daddr, portp[1],
			    dest->addr, dport,
			    iph->saddr, portp[0],
			    0);
	if (ms == NULL) {
		IP_VS_ERR("ip_masq_new_vs failed\n");
		ip_masq_put(mst);
		return NULL;
	}

	/*
	 *    Bind the masq entry with the vs dest.
	 */
	ip_vs_bind_masq(ms, dest);

	/*
	 *    Increase the inactive connection counter
	 *    because it is in Syn-Received
	 *    state (inactive) when the masq is created.
	 */
	atomic_inc(&dest->inactconns);

	/*
	 *    Add its control
	 */
	ip_masq_control_add(ms, mst);

	ip_masq_put(mst);
	return ms;
}


/*
 *  IPVS main scheduling function
 *  It selects a server according to the virtual service, and
 *  creates a masq entry.
 */
struct ip_masq *ip_vs_schedule(struct ip_vs_service *svc, struct iphdr *iph)
{
	struct ip_masq *ms = NULL;
	struct ip_vs_dest *dest;
	const __u16 *portp;

	/*
	 *    Persistent service
	 */
	if (svc->flags & IP_VS_SVC_F_PERSISTENT)
		return ip_vs_sched_persist(svc, iph);

	/*
	 *    Non-persistent service
	 */
	portp = (__u16 *)&(((char *)iph)[iph->ihl*4]);
	if (!svc->fwmark && portp[1] != svc->port) {
		if (!svc->port)
			IP_VS_ERR("Schedule: port zero only supported in persistent services, check your ipvs configuration\n");
		return NULL;
	}

	read_lock(&__ip_vs_lock);

	dest = svc->scheduler->schedule(svc, iph);
	if (dest == NULL) {
		IP_VS_DBG(1, "Schedule: no dest found.\n");
		read_unlock(&__ip_vs_lock);
		return NULL;
	}

	/*
	 *    Create a masquerading entry.
	 */
	ms = ip_masq_new_vs(iph->protocol,
			    iph->daddr, portp[1],
			    dest->addr, dest->port?dest->port:portp[1],
			    iph->saddr, portp[0],
			    0);
	if (ms == NULL) {
		IP_VS_ERR("Schedule: ip_masq_new_vs failed\n");
		read_unlock(&__ip_vs_lock);
		return NULL;
	}

	/*
	 *    Bind the masq entry with the vs dest.
	 */
	ip_vs_bind_masq(ms, dest);

	/*
	 *    Increase the inactive connection counter because it is in
	 *    Syn-Received state (inactive) when the masq is created.
	 */
	atomic_inc(&dest->inactconns);

	IP_VS_DBG(9, "Schedule masq fwd:%c s:%s c:%u.%u.%u.%u:%d "
		  "v:%u.%u.%u.%u:%d d:%u.%u.%u.%u:%d flg:%X cnt:%d\n",
		  ip_vs_fwd_tag(ms), ip_masq_state_name(ms->state),
		  NIPQUAD(ms->daddr),ntohs(ms->dport),
		  NIPQUAD(ms->maddr),ntohs(ms->mport),
		  NIPQUAD(ms->saddr),ntohs(ms->sport),
		  ms->flags, atomic_read(&ms->refcnt));

	read_unlock(&__ip_vs_lock);

	return ms;
}


/*
 *  Pass or drop the packet.
 *  Called by ip_fw_demasquerade, when the virtual service is available but
 *  no destination is available for a new connection.
 */
int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb)
{
	struct iphdr *iph = skb->nh.iph;
	__u16 *portp = (__u16 *)&(((char *)iph)[iph->ihl*4]);

	/*
	 * When the virtual ftp service is presented, packets destined
	 * for other services on the VIP may get here (except services
	 * listed in the ipvs table), pass the packets, because it is
	 * not ipvs job to decide to drop the packets.
	 */
	if ((svc->port == FTPPORT) && (portp[1] != FTPPORT))
		return 0;

	/*
	 * Notify the client that the destination is unreachable, and
	 * release the socket buffer.
	 * Since it is in IP layer, the TCP socket is not actually
	 * created, the TCP RST packet cannot be sent, instead that
	 * ICMP_PORT_UNREACH is sent here no matter it is TCP/UDP. --WZ
	 */
	icmp_send(skb, ICMP_DEST_UNREACH, ICMP_PORT_UNREACH, 0);
	kfree_skb(skb);
	return -2;
}


/*
 *	IPVS user control entry
 */
int ip_vs_ctl(int optname, struct ip_masq_ctl *mctl, int optlen)
{
	struct ip_vs_service *svc = NULL;
	struct ip_vs_user *mm =  &mctl->u.vs_user;
	__u32 vaddr = mm->vaddr;
	__u16 vport = mm->vport;
	int proto_num = masq_proto_num(mm->protocol);

	/*
	 * Check the size of mctl, no overflow...
	 */
	if (optlen != sizeof(*mctl))
		return -EINVAL;

	/*
	 * Flush all the virtual service...
	 */
	if (mctl->m_cmd == IP_MASQ_CMD_FLUSH)
		return ip_vs_flush();

	/*
	 * Check for valid protocol: TCP or UDP
	 */
	if (mm->vfwmark == 0 && (proto_num < 0 || proto_num > 1)) {
		IP_VS_INFO("vs_ctl: invalid protocol: %d %d.%d.%d.%d:%d %s",
			   ntohs(mm->protocol),
			   NIPQUAD(vaddr), ntohs(vport), mctl->m_tname);
		return -EFAULT;
	}

	/*
	 * Lookup the exact service by (protocol, vaddr, vport)
	 */
	read_lock(&__ip_vs_lock);

	if (mm->vfwmark == 0)
		svc = __ip_vs_lookup_service(mm->protocol, vaddr, vport);
	else
		svc = __ip_vs_lookup_svc_fwm(mm->vfwmark);

	read_unlock(&__ip_vs_lock);

	switch (mctl->m_cmd) {
	case IP_MASQ_CMD_ADD:
		if (svc != NULL)
			return -EEXIST;

		return ip_vs_add_service(mctl);

	case IP_MASQ_CMD_SET:
		if (svc == NULL)
			return -ESRCH;

		return ip_vs_edit_service(svc, mctl);

	case IP_MASQ_CMD_DEL:
		if (svc == NULL)
			return  -ESRCH;
		else
			return ip_vs_del_service(svc);

	case IP_MASQ_CMD_ADD_DEST:
		if (svc == NULL)
			return  -ESRCH;
		else
			return ip_vs_add_dest(svc, mctl);

	case IP_MASQ_CMD_SET_DEST:
		if (svc == NULL)
			return  -ESRCH;
		else
			return ip_vs_edit_dest(svc, mctl);

	case IP_MASQ_CMD_DEL_DEST:
		if (svc == NULL)
			return  -ESRCH;
		else
			return ip_vs_del_dest(svc, mctl);
	}
	return -EINVAL;
}


#ifdef CONFIG_SYSCTL

static int ip_vs_sysctl_defense_mode(ctl_table *ctl, int write,
	struct file * filp,void *buffer, size_t *lenp)
{
	int *valp = ctl->data;
	int val = *valp;
	int ret;

	ret = proc_dointvec(ctl, write, filp, buffer, lenp);
	if (write && (*valp != val)) {
		if ((*valp < 0) || (*valp > 3)) {
			/* Restore the correct value */
			*valp = val;
		} else {
			update_defense_level();
		}
	}
	return ret;
}

ctl_table ipv4_vs_table[] = {
#ifdef CONFIG_IP_VS_DEBUG
	{NET_IPV4_VS_DEBUG_LEVEL, "debug_level",
	 &sysctl_ip_vs_debug_level, sizeof(int), 0644, NULL,
	 &proc_dointvec},
#endif
	{NET_IPV4_VS_AMEMTHRESH, "amemthresh",
	 &sysctl_ip_vs_amemthresh, sizeof(int), 0644, NULL,
	 &proc_dointvec},
	{NET_IPV4_VS_AMDROPRATE, "am_droprate",
	 &sysctl_ip_vs_am_droprate, sizeof(int), 0644, NULL,
	 &proc_dointvec},
	{NET_IPV4_VS_DROP_ENTRY, "drop_entry",
	 &sysctl_ip_vs_drop_entry, sizeof(int), 0644, NULL,
	 &ip_vs_sysctl_defense_mode},
	{NET_IPV4_VS_DROP_PACKET, "drop_packet",
	 &sysctl_ip_vs_drop_packet, sizeof(int), 0644, NULL,
	 &ip_vs_sysctl_defense_mode},
	{NET_IPV4_VS_SECURE_TCP, "secure_tcp",
	 &sysctl_ip_vs_secure_tcp, sizeof(int), 0644, NULL,
	 &ip_vs_sysctl_defense_mode},
	{NET_IPV4_VS_TO_ES, "timeout_established",
	 &masq_timeout_table_dos.timeout[IP_MASQ_S_ESTABLISHED],
	 sizeof(int), 0644, NULL, &proc_dointvec_jiffies},
	{NET_IPV4_VS_TO_SS, "timeout_synsent",
	 &masq_timeout_table_dos.timeout[IP_MASQ_S_SYN_SENT],
	 sizeof(int), 0644, NULL, &proc_dointvec_jiffies},
	{NET_IPV4_VS_TO_SR, "timeout_synrecv",
	 &masq_timeout_table_dos.timeout[IP_MASQ_S_SYN_RECV],
	 sizeof(int), 0644, NULL, &proc_dointvec_jiffies},
	{NET_IPV4_VS_TO_FW, "timeout_finwait",
	 &masq_timeout_table_dos.timeout[IP_MASQ_S_FIN_WAIT],
	 sizeof(int), 0644, NULL, &proc_dointvec_jiffies},
	{NET_IPV4_VS_TO_TW, "timeout_timewait",
	 &masq_timeout_table_dos.timeout[IP_MASQ_S_TIME_WAIT],
	 sizeof(int), 0644, NULL, &proc_dointvec_jiffies},
	{NET_IPV4_VS_TO_CL, "timeout_close",
	 &masq_timeout_table_dos.timeout[IP_MASQ_S_CLOSE],
	 sizeof(int), 0644, NULL, &proc_dointvec_jiffies},
	{NET_IPV4_VS_TO_CW, "timeout_closewait",
	 &masq_timeout_table_dos.timeout[IP_MASQ_S_CLOSE_WAIT],
	 sizeof(int), 0644, NULL, &proc_dointvec_jiffies},
	{NET_IPV4_VS_TO_LA, "timeout_lastack",
	 &masq_timeout_table_dos.timeout[IP_MASQ_S_LAST_ACK],
	 sizeof(int), 0644, NULL, &proc_dointvec_jiffies},
	{NET_IPV4_VS_TO_LI, "timeout_listen",
	 &masq_timeout_table_dos.timeout[IP_MASQ_S_LISTEN],
	 sizeof(int), 0644, NULL, &proc_dointvec_jiffies},
	{NET_IPV4_VS_TO_SA, "timeout_synack",
	 &masq_timeout_table_dos.timeout[IP_MASQ_S_SYNACK],
	 sizeof(int), 0644, NULL, &proc_dointvec_jiffies},
	{NET_IPV4_VS_TO_UDP, "timeout_udp",
	 &masq_timeout_table_dos.timeout[IP_MASQ_S_UDP],
	 sizeof(int), 0644, NULL, &proc_dointvec_jiffies},
	{NET_IPV4_VS_TO_ICMP, "timeout_icmp",
	 &masq_timeout_table_dos.timeout[IP_MASQ_S_ICMP],
	 sizeof(int), 0644, NULL, &proc_dointvec_jiffies},
	{0}
};
#endif

#ifdef CONFIG_PROC_FS
/*
 *	Write the contents of the VS rule table to a PROCfs file.
 */
static int ip_vs_procinfo(char *buf, char **start, off_t offset,
			  int length, int *eof, void *data)
{
	int len=0;
	off_t pos=0;
	char temp[64], temp2[32];
	int idx;
	struct ip_vs_service *svc;
	struct ip_vs_dest *dest;
	struct list_head *l, *e, *p, *q;

	/*
	 * Note: since the length of the buffer is usually the multiple
	 * of 512, it is good to use fixed record of the divisor of 512,
	 * so that records won't be truncated at buffer boundary.
	 */
	pos = 192;
	if (pos > offset) {
		sprintf(temp,
			"IP Virtual Server version %d.%d.%d (size=%d)",
			NVERSION(IP_VS_VERSION_CODE), IP_VS_TAB_SIZE);
		len += sprintf(buf+len, "%-63s\n", temp);
		len += sprintf(buf+len, "%-63s\n",
			       "Prot LocalAddress:Port Scheduler Flags");
		len += sprintf(buf+len, "%-63s\n",
			       "  -> RemoteAddress:Port Forward Weight ActiveConn InActConn");
	}

	read_lock_bh(&__ip_vs_lock);

	/* print the service table hashed by <protocol,addr,port> */
	for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
		l = &ip_vs_svc_table[idx];
		for (e=l->next; e!=l; e=e->next) {
			svc = list_entry(e, struct ip_vs_service, s_list);
			pos += 64;
			if (pos > offset) {
				if (svc->flags & IP_VS_SVC_F_PERSISTENT)
					sprintf(temp2, "persistent %d %08X",
						svc->timeout,
						ntohl(svc->netmask));
				else
					temp2[0] = '\0';

				sprintf(temp, "%s  %08X:%04X %s %s",
					masq_proto_name(svc->protocol),
					ntohl(svc->addr),
					ntohs(svc->port),
					svc->scheduler->name, temp2);
				len += sprintf(buf+len, "%-63s\n", temp);
				if (len >= length)
					goto done;
			}

			p = &svc->destinations;
			for (q=p->next; q!=p; q=q->next) {
				dest = list_entry(q, struct ip_vs_dest, n_list);
				pos += 64;
				if (pos <= offset)
					continue;
				sprintf(temp,
					"  -> %08X:%04X      %-7s %-6d %-10d %-10d",
					ntohl(dest->addr),
					ntohs(dest->port),
					ip_vs_fwd_name(dest->masq_flags),
					dest->weight,
					atomic_read(&dest->activeconns),
					atomic_read(&dest->inactconns));
				len += sprintf(buf+len, "%-63s\n", temp);
				if (len >= length)
					goto done;
			}
		}
	}

	/* print the service table hashed by fwmark */
	for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
		l = &ip_vs_svc_fwm_table[idx];
		for (e=l->next; e!=l; e=e->next) {
			svc = list_entry(e, struct ip_vs_service, f_list);
			pos += 64;
			if (pos > offset) {
				if (svc->flags & IP_VS_SVC_F_PERSISTENT)
					sprintf(temp2, "persistent %d %08X",
						svc->timeout,
						ntohl(svc->netmask));
				else
					temp2[0] = '\0';

				sprintf(temp, "FWM  %08X %s %s",
					svc->fwmark,
					svc->scheduler->name, temp2);
				len += sprintf(buf+len, "%-63s\n", temp);
				if (len >= length)
					goto done;
			}

			p = &svc->destinations;
			for (q=p->next; q!=p; q=q->next) {
				dest = list_entry(q, struct ip_vs_dest, n_list);
				pos += 64;
				if (pos <= offset)
					continue;
				sprintf(temp,
					"  -> %08X:%04X      %-7s %-6d %-10d %-10d",
					ntohl(dest->addr),
					ntohs(dest->port),
					ip_vs_fwd_name(dest->masq_flags),
					dest->weight,
					atomic_read(&dest->activeconns),
					atomic_read(&dest->inactconns));
				len += sprintf(buf+len, "%-63s\n", temp);
				if (len >= length)
					goto done;
			}
		}
	}

  done:
	read_unlock_bh(&__ip_vs_lock);

	*start = buf+len-(pos-offset);          /* Start of wanted data */
	len = pos-offset;
	if (len > length)
		len = length;
	if (len < 0)
		len = 0;
	return len;
}

struct proc_dir_entry ip_vs_proc_entry = {
	0,			/* dynamic inode */
	2, "vs",		/* namelen and name */
	S_IFREG | S_IRUGO,	/* mode */
	1, 0, 0, 0,		/* nlinks, owner, group, size */
	&proc_net_inode_operations, /* operations */
	NULL,			/* get_info */
	NULL,			/* fill_inode */
	NULL, NULL, NULL,	/* next, parent, subdir */
	NULL,			/* data */
	&ip_vs_procinfo,	/* function to generate proc data */
};


/*
 *	Write the IPVS statistic information to a PROCfs file.
 */
struct ip_vs_stats ip_vs_stats = {SPIN_LOCK_UNLOCKED, 0, 0};

static int
ip_vs_stats_get_info(char *buf, char **start, off_t offset,
		     int length, int *eof, void *data)
{
	int idx;
	int len=0;
	off_t pos=0;
	char temp[128];
	struct ip_vs_service *svc;
	struct ip_vs_dest *dest;
	struct list_head *l, *e, *p, *q;

	pos += 128;
	if (pos > offset) {
		len += sprintf(buf+len, "%-63s\n",
/*                                01234567 01234567 01234567 0123456701234567 0123456701234567 */
			       "TotalConns   InPkts  OutPkts          InBytes         OutBytes");
		spin_lock(&ip_vs_stats.lock);
		sprintf(temp, "  %8X %8X %8X %8X%08X %8X%08X",
			ip_vs_stats.conns,
			ip_vs_stats.inpkts,
			ip_vs_stats.outpkts,
			(__u32)(ip_vs_stats.inbytes >> 32),
			(__u32)ip_vs_stats.inbytes,
			(__u32)(ip_vs_stats.outbytes >> 32),
			(__u32)ip_vs_stats.outbytes);
		spin_unlock(&ip_vs_stats.lock);
		len += sprintf(buf+len, "%-63s\n", temp);
	}

	read_lock_bh(&__ip_vs_lock);

	/* print the service statistics */
	pos += 128;
	if (pos > offset) {
		len += sprintf(buf+len, "%-127s\n",
			       "\nVirtual Service\n"
			       "Pro VirtService      Conns   InPkts  OutPkts          InBytes         OutBytes");
	}

	for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
		l = &ip_vs_svc_table[idx];
		for (e=l->next; e!=l; e=e->next) {
			svc = list_entry(e, struct ip_vs_service, s_list);
			pos += 128;
			if (pos <= offset)
				continue;
			spin_lock(&svc->stats.lock);
			sprintf(temp, "%3s %08X:%04X %8X %8X %8X %8X%08X %8X%08X",
				masq_proto_name(svc->protocol),
				ntohl(svc->addr),
				ntohs(svc->port),
				svc->stats.conns,
				svc->stats.inpkts,
				svc->stats.outpkts,
				(__u32)(svc->stats.inbytes >> 32),
				(__u32)svc->stats.inbytes,
				(__u32)(svc->stats.outbytes >> 32),
				(__u32)svc->stats.outbytes);
			spin_unlock(&svc->stats.lock);
			len += sprintf(buf+len, "%-127s\n", temp);
			if (pos >= offset+length)
				goto done;
		}
	}

	for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
		l = &ip_vs_svc_fwm_table[idx];
		for (e=l->next; e!=l; e=e->next) {
			svc = list_entry(e, struct ip_vs_service, f_list);
			pos += 128;
			if (pos <= offset)
				continue;
			spin_lock(&svc->stats.lock);
			sprintf(temp, "FWM %08X      %8X %8X %8X %8X%08X %8X%08X",
				svc->fwmark,
				svc->stats.conns,
				svc->stats.inpkts,
				svc->stats.outpkts,
				(__u32)(svc->stats.inbytes >> 32),
				(__u32)svc->stats.inbytes,
				(__u32)(svc->stats.outbytes >> 32),
				(__u32)svc->stats.outbytes);
			spin_unlock(&svc->stats.lock);
			len += sprintf(buf+len, "%-127s\n", temp);
			if (pos >= offset+length)
				goto done;
		}
	}

	/* print the real server statistics */
	pos += 128;
	if (pos > offset) {
		len += sprintf(buf+len, "%-127s\n",
			       "\nReal Service\n"
			       "Pro VirtService   RealService      Conns   InPkts  OutPkts          InBytes         OutBytes");
	}

	for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
		l = &ip_vs_svc_table[idx];
		for (e=l->next; e!=l; e=e->next) {
			svc = list_entry(e, struct ip_vs_service, s_list);
			p = &svc->destinations;
			for (q=p->next; q!=p; q=q->next) {
				dest = list_entry(q, struct ip_vs_dest, n_list);
				pos += 128;
				if (pos <= offset)
					continue;
				spin_lock(&dest->stats.lock);
				sprintf(temp,
					"%3s %08X:%04X %08X:%04X %8X %8X %8X %8X%08X %8X%08X",
					masq_proto_name(svc->protocol),
					ntohl(svc->addr),
					ntohs(svc->port),
					ntohl(dest->addr),
					ntohs(dest->port),
					dest->stats.conns,
					dest->stats.inpkts,
					dest->stats.outpkts,
					(__u32)(dest->stats.inbytes >> 32),
					(__u32)dest->stats.inbytes,
					(__u32)(dest->stats.outbytes >> 32),
					(__u32)dest->stats.outbytes);
				spin_unlock(&dest->stats.lock);
				len += sprintf(buf+len, "%-127s\n", temp);
				if (pos >= offset+length)
					goto done;
			}
		}
	}

	for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
		l = &ip_vs_svc_fwm_table[idx];
		for (e=l->next; e!=l; e=e->next) {
			svc = list_entry(e, struct ip_vs_service, f_list);
			p = &svc->destinations;
			for (q=p->next; q!=p; q=q->next) {
				dest = list_entry(q,struct ip_vs_dest,n_list);
				pos += 128;
				if (pos <= offset)
					continue;
				spin_lock(&dest->stats.lock);
				sprintf(temp,
					"FWM %08X      %08X:%04X %8X %8X %8X %8X%08X %8X%08X",
					svc->fwmark,
					ntohl(dest->addr),
					ntohs(dest->port),
					dest->stats.conns,
					dest->stats.inpkts,
					dest->stats.outpkts,
					(__u32)(dest->stats.inbytes >> 32),
					(__u32)dest->stats.inbytes,
					(__u32)(dest->stats.outbytes >> 32),
					(__u32)dest->stats.outbytes);
				spin_unlock(&dest->stats.lock);
				len += sprintf(buf+len, "%-127s\n", temp);
				if (pos >= offset+length)
					goto done;
			}
		}
	}
 done:
	read_unlock_bh(&__ip_vs_lock);

	*start = buf+len-(pos-offset);          /* Start of wanted data */
	len = pos-offset;
	if (len > length)
		len = length;
	if (len < 0)
		len = 0;
	return len;
}

struct proc_dir_entry ip_vs_stat_proc_entry = {
	0,			/* dynamic inode */
	8, "vs_stats",		/* namelen and name */
	S_IFREG | S_IRUGO,	/* mode */
	1, 0, 0, 0,		/* nlinks, owner, group, size */
	&proc_net_inode_operations, /* operations */
	NULL,			/* get_info */
	NULL,			/* fill_inode */
	NULL, NULL, NULL,	/* next, parent, subdir */
	NULL,			/* data */
	&ip_vs_stats_get_info,	/* function to generate proc data */
};

#endif


/*
 *   This function encapsulates the packet in a new IP header, its destination
 *   will be set to the daddr. Most code of this function is from ipip.c.
 *   Usage:
 *     It is called in the ip_vs_forward() function. The load balancer
 *     selects a real server from a cluster based on a scheduling algorithm,
 *     encapsulates the packet and forwards it to the selected server. All real
 *     servers are configured with "ifconfig tunl0 <Virtual IP Address> up".
 *     When the server receives the encapsulated packet, it decapsulates the
 *     packet, processes the request and return the reply packets directly to
 *     the client without passing the load balancer. This can greatly
 *     increase the scalability of virtual server.
 *   Returns:
 *     if succeeded, return 1; otherwise, return 0.
 */

int ip_vs_tunnel_xmit(struct sk_buff *skb, __u32 daddr)
{
	struct rtable *rt;			/* Route to the other host */
	struct device *tdev;			/* Device to other host */
	struct iphdr  *old_iph = skb->nh.iph;
	u8     tos = old_iph->tos;
	u16    df = old_iph->frag_off;
	struct iphdr  *iph;			/* Our new IP header */
	int    max_headroom;			/* The extra header space needed */
	u32    dst = daddr;
	u32    src = 0;
	int    mtu;

	if (skb->protocol != __constant_htons(ETH_P_IP)) {
		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): protocol error, ETH_P_IP: %d, skb protocol: %d\n",
			  __constant_htons(ETH_P_IP),skb->protocol);
		goto tx_error;
	}

	if (ip_route_output(&rt, dst, src, RT_TOS(tos), 0)) {
		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): route error, dest: "
			  "%u.%u.%u.%u\n", NIPQUAD(dst));
		goto tx_error_icmp;
	}
	tdev = rt->u.dst.dev;

	mtu = rt->u.dst.pmtu - sizeof(struct iphdr);
	if (mtu < 68) {
		ip_rt_put(rt);
		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): mtu less than 68\n");
		goto tx_error;
	}
	if (skb->dst && mtu < skb->dst->pmtu)
		skb->dst->pmtu = mtu;

	df |= (old_iph->frag_off&__constant_htons(IP_DF));

	if ((old_iph->frag_off&__constant_htons(IP_DF)) && mtu < ntohs(old_iph->tot_len)) {
		icmp_send(skb, ICMP_DEST_UNREACH, ICMP_FRAG_NEEDED, htonl(mtu));
		ip_rt_put(rt);
		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): frag needed\n");
		goto tx_error;
	}

	skb->h.raw = skb->nh.raw;

	/*
	 * Okay, now see if we can stuff it in the buffer as-is.
	 */
	max_headroom = (((tdev->hard_header_len+15)&~15)+sizeof(struct iphdr));

	if (skb_headroom(skb) < max_headroom || skb_cloned(skb) || skb_shared(skb)) {
		struct sk_buff *new_skb = skb_realloc_headroom(skb, max_headroom);
		if (!new_skb) {
			ip_rt_put(rt);
			kfree_skb(skb);
			IP_VS_ERR("ip_vs_tunnel_xmit(): no memory for new_skb\n");
			return 0;
		}
		kfree_skb(skb);
		skb = new_skb;
	}

	skb->nh.raw = skb_push(skb, sizeof(struct iphdr));
	memset(&(IPCB(skb)->opt), 0, sizeof(IPCB(skb)->opt));
	dst_release(skb->dst);
	skb->dst = &rt->u.dst;

	/*
	 *	Push down and install the IPIP header.
	 */

	iph			=	skb->nh.iph;
	iph->version		=	4;
	iph->ihl		=	sizeof(struct iphdr)>>2;
	iph->frag_off		=	df;
	iph->protocol		=	IPPROTO_IPIP;
	iph->tos		=	tos;
	iph->daddr		=	rt->rt_dst;
	iph->saddr		=	rt->rt_src;
	iph->ttl		=	old_iph->ttl;
	iph->tot_len		=	htons(skb->len);
	iph->id			=	htons(ip_id_count++);
	ip_send_check(iph);

	IPCB(skb)->flags |= IPSKB_REDIRECTED;
	IPCB(skb)->flags |= IPSKB_MASQUERADED;

	ip_send(skb);
	return 1;

  tx_error_icmp:
	dst_link_failure(skb);
  tx_error:
	kfree_skb(skb);
	return 0;
}


/*
 *      Direct Routing
 */
int ip_vs_dr_xmit(struct sk_buff *skb, __u32 daddr)
{
	struct rtable *rt;			/* Route to the other host */
	struct iphdr  *iph = skb->nh.iph;
	u8     tos = iph->tos;
	int    mtu;

	if (ip_route_output(&rt, daddr, 0, RT_TOS(tos), 0)) {
		IP_VS_DBG(0, "ip_vs_dr_xmit(): route error, dest: %u.%u.%u.%u\n",
			  NIPQUAD(daddr));
		goto tx_error_icmp;
	}

	/* MTU checking */
	mtu = rt->u.dst.pmtu;
	if ((skb->len > mtu) && (iph->frag_off&__constant_htons(IP_DF))) {
		icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
		ip_rt_put(rt);
		IP_VS_DBG(0, "ip_vs_dr_xmit(): frag needed\n");
		goto tx_error;
	}

	dst_release(skb->dst);
	skb->dst = &rt->u.dst;

	IPCB(skb)->flags |= IPSKB_REDIRECTED;
	IPCB(skb)->flags |= IPSKB_MASQUERADED;

	ip_send(skb);
	return 1;

  tx_error_icmp:
	dst_link_failure(skb);
  tx_error:
	kfree_skb(skb);
	return 0;
}


/*
 *	Initialize IP virtual server
 */
__initfunc(int ip_vs_init(void))
{
	int idx;

	/*
	 * Allocate the ip_vs_table and initialize its list head.
	 * Initilize ip_vs_svc_table, ip_vs_svc_fwm_table, ip_vs_rtable,
	 * ip_vs_schedulers and ip_vs_dest_trash.
	 */
	if (!(ip_vs_table =
	      vmalloc(IP_VS_TAB_SIZE*sizeof(struct list_head)))) {
		return -ENOMEM;
	}
	for(idx = 0; idx < IP_VS_TAB_SIZE; idx++)  {
		INIT_LIST_HEAD(&ip_vs_table[idx]);
	}
	IP_VS_INFO("Connection hash table configured "
		   "(size=%d, memory=%ldKbytes)\n",
		   IP_VS_TAB_SIZE,
		   (long) (IP_VS_TAB_SIZE*sizeof(struct list_head))/1024);

	for(idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++)  {
		INIT_LIST_HEAD(&ip_vs_svc_table[idx]);
		INIT_LIST_HEAD(&ip_vs_svc_fwm_table[idx]);
	}
	for(idx = 0; idx < IP_VS_RTAB_SIZE; idx++)  {
		INIT_LIST_HEAD(&ip_vs_rtable[idx]);
	}
	INIT_LIST_HEAD(&ip_vs_schedulers);
	INIT_LIST_HEAD(&ip_vs_dest_trash);

	/*
	 * Hook the slow_timer handler in the system timer.
	 */
	slow_timer.expires = jiffies+SLTIMER_PERIOD;
	add_timer(&slow_timer);

#ifdef CONFIG_PROC_FS
	ip_masq_proc_register(&ip_vs_proc_entry);
	ip_masq_proc_register(&ip_vs_stat_proc_entry);
#endif

#ifdef CONFIG_IP_MASQUERADE_VS_RR
	ip_vs_rr_init();
#endif
#ifdef CONFIG_IP_MASQUERADE_VS_WRR
	ip_vs_wrr_init();
#endif
#ifdef CONFIG_IP_MASQUERADE_VS_LC
	ip_vs_lc_init();
#endif
#ifdef CONFIG_IP_MASQUERADE_VS_WLC
	ip_vs_wlc_init();
#endif
#ifdef CONFIG_IP_MASQUERADE_VS_LBLC
	ip_vs_lblc_init();
#endif
#ifdef CONFIG_IP_MASQUERADE_VS_LBLCR
	ip_vs_lblcr_init();
#endif
	return 0;
}
