/*
 * IPVS:        Locality-Based Least-Connection with Replication scheduler
 *
 * Version:     $Id: ip_vs_lblcr.c,v 1.4 2001/03/27 09:37:00 wensong Exp $
 *
 * Authors:     Wensong Zhang <wensong@gnuchina.org>
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Changes:
 *     Julian Anastasov        :    Added the missing (dest->weight>0)
 *                                  condition in the ip_vs_dest_set_max.
 *
 */

/*
 * The lblc/r algorithm is as follows (pseudo code):
 *
 *       if serverSet[dest_ip] is null then
 *               n, serverSet[dest_ip] <- {weighted least-conn node};
 *       else
 *               n <- {least-conn (alive) node in serverSet[dest_ip]};
 *               if (n is null) OR
 *                  (n.conns>n.weight AND 
 *                   there is a node m with m.conns<m.weight/2) then
 *                   n <- {weighted least-conn node};
 *                   add n to serverSet[dest_ip];
 *               if |serverSet[dest_ip]| > 1 AND
 *                   now - serverSet[dest_ip].lastMod > T then
 *                   m <- {most conn node in serverSet[dest_ip]};
 *                   remove m from serverSet[dest_ip];
 *       if serverSet[dest_ip] changed then
 *               serverSet[dest_ip].lastMod <- now;
 *
 *       return n;
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/vmalloc.h>
#include <net/ip_masq.h>
#ifdef CONFIG_IP_MASQUERADE_MOD
#include <net/ip_masq_mod.h>
#endif
#include <linux/sysctl.h>
#include <linux/proc_fs.h>
#include <linux/ip_fw.h>
#include <net/ip_vs.h>


/*
 *    It is for garbage collection of stale IPVS lblcr entries,
 *    when the table is full.
 */
#define CHECK_EXPIRE_INTERVAL   (60*HZ)
#define ENTRY_TIMEOUT           (6*60*HZ)

/*
 *    It is for full expiration check.
 *    When there is no partial expiration check (garbage collection)
 *    in a half hour, do a full expiration check to collect stale
 *    entries that haven't been touched for a day.
 */
#define COUNT_FOR_FULL_EXPIRATION   30
int sysctl_ip_vs_lblcr_expiration = 24*60*60*HZ;


/*
 *     for IPVS lblcr entry hash table
 */
#ifndef CONFIG_IP_VS_LBLCR_TAB_BITS
#define CONFIG_IP_VS_LBLCR_TAB_BITS      10
#endif
#define IP_VS_LBLCR_TAB_BITS     CONFIG_IP_VS_LBLCR_TAB_BITS
#define IP_VS_LBLCR_TAB_SIZE     (1 << IP_VS_LBLCR_TAB_BITS)
#define IP_VS_LBLCR_TAB_MASK     (IP_VS_LBLCR_TAB_SIZE - 1)


/*
 *      IPVS destination set structure and operations
 */
struct ip_vs_dest_list {
        struct ip_vs_dest_list  *next;          /* list link */
        struct ip_vs_dest       *dest;          /* destination server */
};

struct ip_vs_dest_set {
        atomic_t                size;           /* set size */
        unsigned long           lastmod;        /* last modified time */
        struct ip_vs_dest_list  *list;          /* destination list */
        rwlock_t	        lock;           /* lock for this list */
};


static struct ip_vs_dest_list *
ip_vs_dest_set_insert(struct ip_vs_dest_set *set, struct ip_vs_dest *dest)
{
        struct ip_vs_dest_list *e;

        for (e=set->list; e!=NULL; e=e->next) {
                if (e->dest == dest)
                        /* already existed */
                        return NULL;
        }

        e = kmalloc(sizeof(struct ip_vs_dest_list), GFP_ATOMIC);
        if (e == NULL) {
                IP_VS_ERR("ip_vs_dest_set_insert(): no memory\n");
		return NULL;
        }

        atomic_inc(&dest->refcnt);
        e->dest = dest;

        /* link it to the list */
        write_lock(&set->lock);
        if (set->list != NULL) {
                e->next = set->list->next;
                set->list = e;
        } else {
                e->next = NULL;
                set->list = e;
        }
        write_unlock(&set->lock);

        atomic_inc(&set->size);
        set->lastmod = jiffies;
        return e;
}

static void
ip_vs_dest_set_erase(struct ip_vs_dest_set *set, struct ip_vs_dest *dest)
{
        struct ip_vs_dest_list *e, **ep;

        write_lock(&set->lock);
        for (ep=&set->list, e=*ep; e!=NULL; e=*ep) {
                if (e->dest == dest) {
                        /* HIT */
                        *ep = e->next;
                        atomic_dec(&set->size);
                        set->lastmod = jiffies;
                        atomic_dec(&e->dest->refcnt);
                        kfree(e);
                        break;
                }
                ep = &e->next;
        }
        write_unlock(&set->lock);
}

static void ip_vs_dest_set_eraseall(struct ip_vs_dest_set *set)
{
        struct ip_vs_dest_list *e, **ep;

        write_lock(&set->lock);
        for (ep=&set->list, e=*ep; e!=NULL; e=*ep) {
                *ep = e->next;
                /*
                 * We don't kfree dest because it is refered either
                 * by its service or by the trash dest list.
                 */
                atomic_dec(&e->dest->refcnt);
                kfree(e);
        }
        write_unlock(&set->lock);
}

/* get weighted least-connection node in the destination set */
static inline struct ip_vs_dest *ip_vs_dest_set_min(struct ip_vs_dest_set *set)
{
        register struct ip_vs_dest_list *e;
        struct ip_vs_dest *dest, *least;
        int loh, doh;

        if (set == NULL)
                return NULL;

        read_lock(&set->lock);
        /* select the first destination server, whose weight > 0 */
        for (e=set->list; e!=NULL; e=e->next) {
                least = e->dest;
                if ((least->weight > 0)
                    && (least->flags & IP_VS_DEST_F_AVAILABLE)) {
                        loh = atomic_read(&least->activeconns) * 50
                                + atomic_read(&least->inactconns);
                        goto nextstage;
                }
        }
        read_unlock(&set->lock);
        return NULL;
        
        /* find the destination with the weighted least load */
  nextstage:
        for (e=e->next; e!=NULL; e=e->next) {
                dest = e->dest;
                doh = atomic_read(&dest->activeconns) * 50
                        + atomic_read(&dest->inactconns);
                if ((loh*dest->weight > doh*least->weight)
                    && (dest->flags & IP_VS_DEST_F_AVAILABLE)) {
                        least = dest;
                        loh = doh;
                }
        }
        read_unlock(&set->lock);

        IP_VS_DBG(6, "ip_vs_dest_set_min: server %d.%d.%d.%d:%d "
                  "activeconns %d refcnt %d weight %d overhead %d\n",
                  NIPQUAD(least->addr), ntohs(least->port),
                  atomic_read(&least->activeconns),
                  atomic_read(&least->refcnt), least->weight, loh);
        return least;
}


/* get weighted most-connection node in the destination set */
static inline struct ip_vs_dest *ip_vs_dest_set_max(struct ip_vs_dest_set *set)
{
        register struct ip_vs_dest_list *e;
        struct ip_vs_dest *dest, *most;
        int moh, doh;

        if (set == NULL)
                return NULL;

        read_lock(&set->lock);
        /* select the first destination server, whose weight > 0 */
        for (e=set->list; e!=NULL; e=e->next) {
                most = e->dest;
                if (most->weight > 0) {
                        moh = atomic_read(&most->activeconns) * 50
                                + atomic_read(&most->inactconns);
                        goto nextstage;
                }
        }
        read_unlock(&set->lock);
        return NULL;
        
        /* find the destination with the weighted most load */
  nextstage:
        for (e=e->next; e!=NULL; e=e->next) {
                dest = e->dest;
                doh = atomic_read(&dest->activeconns) * 50
                        + atomic_read(&dest->inactconns);
                /* moh/mw < doh/dw ==> moh*dw < doh*mw, where mw,dw>0 */ 
                if (moh*dest->weight < doh*most->weight
                    && dest->weight > 0) {
                        most = dest;
                        moh = doh;
                }
        }
        read_unlock(&set->lock);

        IP_VS_DBG(6, "ip_vs_dest_set_max: server %d.%d.%d.%d:%d "
                  "activeconns %d refcnt %d weight %d overhead %d\n",
                  NIPQUAD(most->addr), ntohs(most->port),
                  atomic_read(&most->activeconns),
                  atomic_read(&most->refcnt), most->weight, moh);
        return most;
}


/*
 *      IPVS lblcr entry represents an association between destination
 *      IP address and its destination server set
 */
struct ip_vs_lblcr_entry {
        struct list_head        list;
        __u32                   addr;           /* destination IP address */
        struct ip_vs_dest_set   set;            /* destination server set */
        unsigned long           lastuse;        /* last used time */
};


/*
 *      IPVS lblcr hash table
 */
struct ip_vs_lblcr_table {
        rwlock_t	        lock;           /* lock for this table */
        struct list_head        bucket[IP_VS_LBLCR_TAB_SIZE];  /* hash bucket */
        atomic_t                entries;        /* number of entries */
        int                     max_size;       /* maximum size of entries */
        struct timer_list       periodic_timer; /* collect stale entries */
        int                     rover;          /* rover for expire check */
        int                     counter;        /* counter for no expire */
};


/*
 *      IPVS LBLCR sysctl table
 */
struct ip_vs_lblcr_sysctl_table {
	struct ctl_table_header *sysctl_header;
	ctl_table vs_vars[2];
	ctl_table vs_dir[2];
	ctl_table ipv4_dir[2];
	ctl_table root_dir[2];
};


static struct ip_vs_lblcr_sysctl_table lblcr_sysctl_table = {
        NULL,
	{{NET_IPV4_VS_LBLCR_EXPIRE, "lblcr_expiration",
          &sysctl_ip_vs_lblcr_expiration,
          sizeof(int), 0644, NULL, &proc_dointvec_jiffies},
         {0}},
        {{NET_IPV4_VS, "vs", NULL, 0, 0555, lblcr_sysctl_table.vs_vars},
         {0}},
        {{NET_IPV4, "ipv4", NULL, 0, 0555, lblcr_sysctl_table.vs_dir},
         {0}},
        {{CTL_NET, "net", NULL, 0, 0555, lblcr_sysctl_table.ipv4_dir},
         {0}}
};


/*
 *      new/free a ip_vs_lblcr_entry, which is a mapping of a destination
 *      IP address to a server.
 */
static inline struct ip_vs_lblcr_entry *ip_vs_lblcr_new(__u32 daddr)
{
        struct ip_vs_lblcr_entry *en;

        en = kmalloc(sizeof(struct ip_vs_lblcr_entry), GFP_ATOMIC);
        if (en == NULL) {
                IP_VS_ERR("ip_vs_lblcr_new(): no memory\n");
		return NULL;
        }
        
        INIT_LIST_HEAD(&en->list);
        en->addr = daddr;

        /* initilize its dest set */
        atomic_set(&(en->set.size), 0);
        en->set.list = NULL;
        en->set.lock = RW_LOCK_UNLOCKED;

        return en;
}


static inline void ip_vs_lblcr_free(struct ip_vs_lblcr_entry *en)
{
        list_del(&en->list);
        ip_vs_dest_set_eraseall(&en->set);
        kfree(en);
}

                
/*
 *	Returns hash value for IPVS LBLCR entry
 */
static inline unsigned ip_vs_lblcr_hashkey(__u32 addr)
{
        return (ntohl(addr)*2654435761UL) & IP_VS_LBLCR_TAB_MASK;
}


/*
 *	Hash an entry in the ip_vs_lblcr_table.
 *	returns bool success.
 */
static int
ip_vs_lblcr_hash(struct ip_vs_lblcr_table *tbl, struct ip_vs_lblcr_entry *en)
{
        unsigned hash;

        if (!list_empty(&en->list)) {
                IP_VS_ERR("ip_vs_lblcr_hash(): request for already hashed, "
                          "called from %p\n", __builtin_return_address(0));
                return 0;
        }

        /*
         *	Hash by destination IP address
         */
        hash = ip_vs_lblcr_hashkey(en->addr);

        write_lock(&tbl->lock);
        list_add(&en->list, &tbl->bucket[hash]);
        atomic_inc(&tbl->entries);
        write_unlock(&tbl->lock);
        
        return 1;
}


#if 0000
/*
 *	Unhash ip_vs_lblcr_entry from ip_vs_lblcr_table.
 *	returns bool success.
 */
static int ip_vs_lblcr_unhash(struct ip_vs_lblcr_table *tbl,
                             struct ip_vs_lblcr_entry *en)
{
        if (list_empty(&en->list)) {
                IP_VS_ERR("ip_vs_lblcr_unhash(): request for not hashed entry, "
                          "called from %p\n", __builtin_return_address(0));
                return 0;
        }

        /*
         * Remove it from the table
         */
        write_lock(&tbl->lock);
        list_del(&en->list);
        INIT_LIST_HEAD(&en->list);
        write_unlock(&tbl->lock);

        return 1;
}
#endif


/*
 *  Get ip_vs_lblcr_entry associated with supplied parameters.
 */
static inline struct ip_vs_lblcr_entry *
ip_vs_lblcr_get(struct ip_vs_lblcr_table *tbl, __u32 addr)
{
        unsigned hash;
        struct ip_vs_lblcr_entry *en;
        struct list_head *l,*e;

        hash = ip_vs_lblcr_hashkey(addr);
        l = &tbl->bucket[hash];

        read_lock(&tbl->lock);
        
        for (e=l->next; e!=l; e=e->next) {
                en = list_entry(e, struct ip_vs_lblcr_entry, list);
                if (en->addr == addr) {
                        /* HIT */
                        read_unlock(&tbl->lock);
                        return en;
                }
        }

        read_unlock(&tbl->lock);

        return NULL;
}


/*
 *      Flush all the entries of the specified table.
 */
static void ip_vs_lblcr_flush(struct ip_vs_lblcr_table *tbl)
{
        int i;
        struct list_head *l;
        struct ip_vs_lblcr_entry *en;
        
        for (i=0; i<IP_VS_LBLCR_TAB_SIZE; i++) {
                write_lock(&tbl->lock);
                for (l=&tbl->bucket[i]; l->next!=l; ) {
                        en = list_entry(l->next,
                                        struct ip_vs_lblcr_entry, list);
                        ip_vs_lblcr_free(en);
                        atomic_dec(&tbl->entries);
                }
                write_unlock(&tbl->lock);
        }
}


static inline void ip_vs_lblcr_full_check(struct ip_vs_lblcr_table *tbl)
{
        unsigned long now = jiffies;
        int i, j;
        struct list_head *l, *e;
        struct ip_vs_lblcr_entry *en;

        for (i=0, j=tbl->rover; i<IP_VS_LBLCR_TAB_SIZE; i++) {
                j = (j + 1) & IP_VS_LBLCR_TAB_MASK;
                e = l = &tbl->bucket[j];
                write_lock(&tbl->lock);
                while (e->next != l) {
                        en = list_entry(e->next,
                                        struct ip_vs_lblcr_entry, list);
                        if ((now - en->lastuse) <
                            sysctl_ip_vs_lblcr_expiration) {
                                e = e->next;
                                continue;
                        }
                        ip_vs_lblcr_free(en);
                        atomic_dec(&tbl->entries);
                }
                write_unlock(&tbl->lock);
        }
        tbl->rover = j;
}


/*
 *      Periodical timer handler for IPVS lblcr table
 *      It is used to collect stale entries when the number of entries
 *      exceeds the maximum size of the table.
 *
 *      Fixme: we probably need more complicated algorithm to collect
 *             entries that have not been used for a long time even
 *             if the number of entries doesn't exceed the maximum size
 *             of the table.
 *      The full expiration check is for this purpose now.
 */
static void ip_vs_lblcr_check_expire(unsigned long data)
{
        struct ip_vs_lblcr_table *tbl;
        unsigned long now = jiffies;
        int goal;
        int i, j;
        struct list_head *l, *e;
        struct ip_vs_lblcr_entry *en;
        
        tbl = (struct ip_vs_lblcr_table *)data;

        if ((tbl->counter % COUNT_FOR_FULL_EXPIRATION) == 0) {
                /* do full expiration check */
                ip_vs_lblcr_full_check(tbl);
                tbl->counter = 1;
                goto out;
        }
                
        if (atomic_read(&tbl->entries) < tbl->max_size) {
                tbl->counter++;
                goto out;
        }

        goal = (atomic_read(&tbl->entries) - tbl->max_size)*4/3;
        if (goal > tbl->max_size/2)
                goal = tbl->max_size/2;

        for (i=0, j=tbl->rover; i<IP_VS_LBLCR_TAB_SIZE; i++) {
                j = (j + 1) & IP_VS_LBLCR_TAB_MASK;
                e = l = &tbl->bucket[j];
                write_lock(&tbl->lock);
                while (e->next != l) {
                        en = list_entry(e->next,
                                        struct ip_vs_lblcr_entry, list);
                        if ((now - en->lastuse) < ENTRY_TIMEOUT) {
                                e = e->next;
                                continue;
                        }
                        ip_vs_lblcr_free(en);
                        atomic_dec(&tbl->entries);
                        goal--;
                }
                write_unlock(&tbl->lock);
                if (goal <= 0)
                        break;
        }
        tbl->rover = j;
        
  out:
        mod_timer(&tbl->periodic_timer, jiffies+CHECK_EXPIRE_INTERVAL);
}


static int ip_vs_lblcr_init_svc(struct ip_vs_service *svc)
{
	int i;
        struct ip_vs_lblcr_table *tbl;
        
        /*
         *    Allocate the ip_vs_lblcr_table for this service
         */
        tbl = kmalloc(sizeof(struct ip_vs_lblcr_table), GFP_ATOMIC);
        if (tbl == NULL) {
                IP_VS_ERR("ip_vs_lblcr_init_svc(): no memory\n");
		return -ENOMEM;
        }
        svc->sched_data = tbl;
        IP_VS_DBG(0, "LBLCR hash table (memory=%dbytes) allocated for "
                  "current service\n",
                  sizeof(struct ip_vs_lblcr_table));
        
        /*
         *    Initialize the hash buckets
         */
        for (i=0; i<IP_VS_LBLCR_TAB_SIZE; i++) {
                INIT_LIST_HEAD(&tbl->bucket[i]);
        }
        tbl->lock = RW_LOCK_UNLOCKED;
        tbl->max_size = IP_VS_LBLCR_TAB_SIZE*16;
        tbl->rover = 0;
        tbl->counter = 1;

        /*
         *    Hook periodic timer for garbage collection
         */
	init_timer(&tbl->periodic_timer);
        tbl->periodic_timer.data = (unsigned long)tbl;
        tbl->periodic_timer.function = ip_vs_lblcr_check_expire;
        tbl->periodic_timer.expires = jiffies+CHECK_EXPIRE_INTERVAL;
        add_timer(&tbl->periodic_timer);
        
        MOD_INC_USE_COUNT;
        return 0;
}


static int ip_vs_lblcr_done_svc(struct ip_vs_service *svc)
{
        struct ip_vs_lblcr_table *tbl = svc->sched_data;

        /* remove periodic timer */
        del_timer(&tbl->periodic_timer);

        /* got to clean up table entries here */
        ip_vs_lblcr_flush(tbl);

        /* release the table itself */
        kfree(svc->sched_data);
        IP_VS_DBG(0, "LBLCR hash table (memory=%dbytes) released\n",
                  sizeof(struct ip_vs_lblcr_table));

        MOD_DEC_USE_COUNT;
        return 0;
}


static int ip_vs_lblcr_update_svc(struct ip_vs_service *svc)
{
        return 0;
}


static inline struct ip_vs_dest *
__ip_vs_wlc_schedule(struct ip_vs_service *svc, struct iphdr *iph)
{
        register struct list_head *l, *e;
        struct ip_vs_dest *dest, *least;
        int loh, doh;

        l = &svc->destinations;
        if (l == l->next)
                return NULL;

        /*
         * We think the overhead of processing active connections is fifty
         * times than that of inactive conncetions in average. (This fifty
         * times might be not accurate, we will change it later.) We use
         * the following formula to estimate the overhead:
         *                dest->activeconns*50 + dest->inactconns
         * and the load:
         *                (dest overhead) / dest->weight
         *
         * Remember -- no floats in kernel mode!!!
         * The comparison of h1*w2 > h2*w1 is equivalent to that of
         *                h1/w1 > h2/w2
         * if every weight is larger than zero.
         *
         * The server with weight=0 is quiesced and will not receive any
         * new connection.
         */

        for (e=l->next; e!=l; e=e->next) {
                least = list_entry(e, struct ip_vs_dest, n_list);
                if (least->weight > 0) {
                        loh = atomic_read(&least->activeconns) * 50
                                + atomic_read(&least->inactconns);
                        goto nextstage;
                }
        }
        return NULL;
        
        /*
         *    Find the destination with the least load.
         */
  nextstage:
        for (e=e->next; e!=l; e=e->next) {
                dest = list_entry(e, struct ip_vs_dest, n_list);
                doh = atomic_read(&dest->activeconns) * 50
                        + atomic_read(&dest->inactconns);
                if (loh*dest->weight > doh*least->weight) {
                        least = dest;
                        loh = doh;
                }
        }

        IP_VS_DBG(6, "LBLCR: server %d.%d.%d.%d:%d "
                  "activeconns %d refcnt %d weight %d overhead %d\n",
                  NIPQUAD(least->addr), ntohs(least->port),
                  atomic_read(&least->activeconns),
                  atomic_read(&least->refcnt), least->weight, loh);

        return least;
}


/*
 *   If this destination server is overloaded and there is a less loaded
 *   server, then return true.
 */
static inline int
is_overloaded(struct ip_vs_dest *dest, struct ip_vs_service *svc)
{
        if (atomic_read(&dest->activeconns) > dest->weight) {
                register struct list_head *l, *e;
                struct ip_vs_dest *d;
        
                l = &svc->destinations;
                for (e=l->next; e!=l; e=e->next) {
                        d = list_entry(e, struct ip_vs_dest, n_list);
                        if (atomic_read(&d->activeconns)*2 < d->weight) {
                                return 1;
                        }
                }
        }
        return 0;
}


/*
 *    Locality-Based (weighted) Least-Connection scheduling
 */
static struct ip_vs_dest *
ip_vs_lblcr_schedule(struct ip_vs_service *svc, struct iphdr *iph)
{
        struct ip_vs_dest *dest;
        struct ip_vs_lblcr_table *tbl;
        struct ip_vs_lblcr_entry *en;

        IP_VS_DBG(6, "ip_vs_lblcr_schedule(): Scheduling...\n");

        tbl = (struct ip_vs_lblcr_table *)svc->sched_data;
        en = ip_vs_lblcr_get(tbl, iph->daddr);
        if (en == NULL) {
                dest = __ip_vs_wlc_schedule(svc, iph);
                if (dest == NULL) {
                        IP_VS_DBG(1, "no destination available\n");
                        return NULL;
                }
                en = ip_vs_lblcr_new(iph->daddr);
                if (en == NULL) {
                        return NULL;
                }
                ip_vs_dest_set_insert(&en->set, dest);
                ip_vs_lblcr_hash(tbl, en);
        } else {
                dest = ip_vs_dest_set_min(&en->set);
                if (!dest || is_overloaded(dest, svc)) {
                        dest = __ip_vs_wlc_schedule(svc, iph);
                        if (dest == NULL) {
                                IP_VS_DBG(1, "no destination available\n");
                                return NULL;
                        }
                        ip_vs_dest_set_insert(&en->set, dest);
                }
                if (atomic_read(&en->set.size) > 1 &&
                    jiffies-en->set.lastmod > sysctl_ip_vs_lblcr_expiration) {
                        struct ip_vs_dest *m;
                        m = ip_vs_dest_set_max(&en->set);
                        if (m) ip_vs_dest_set_erase(&en->set, m);
                }
        }
        en->lastuse = jiffies;

        IP_VS_DBG(6, "LBLCR: destination IP address %u.%u.%u.%u "
                  "--> server %u.%u.%u.%u:%d\n",
                  NIPQUAD(en->addr),
                  NIPQUAD(dest->addr),
                  ntohs(dest->port));
                        
        return dest;
}


/*
 *      IPVS LBLCR Scheduler structure
 */
static struct ip_vs_scheduler ip_vs_lblcr_scheduler =
{
        {0},                     /* n_list */
        "lblcr",                 /* name */
        ATOMIC_INIT(0),          /* refcnt */
        ip_vs_lblcr_init_svc,    /* service initializer */
        ip_vs_lblcr_done_svc,    /* service done */
        ip_vs_lblcr_update_svc,  /* service updater */
        ip_vs_lblcr_schedule,    /* select a server from the destination list */
};


__initfunc(int ip_vs_lblcr_init(void))
{
        IP_VS_INFO("Initializing LBLCR scheduling\n");
        INIT_LIST_HEAD(&ip_vs_lblcr_scheduler.n_list);
        lblcr_sysctl_table.sysctl_header =
                register_sysctl_table(lblcr_sysctl_table.root_dir, 0);
        return register_ip_vs_scheduler(&ip_vs_lblcr_scheduler);
}


#ifdef MODULE
EXPORT_NO_SYMBOLS;

int init_module(void)
{
        INIT_LIST_HEAD(&ip_vs_lblcr_scheduler.n_list);

        /* module initialization by 'request_module' */
        if (register_ip_vs_scheduler(&ip_vs_lblcr_scheduler) != 0)
                return -EIO;

        lblcr_sysctl_table.sysctl_header =
                register_sysctl_table(lblcr_sysctl_table.root_dir, 0);

        IP_VS_INFO("LBLCR scheduling module loaded.\n");

        return 0;
}

void cleanup_module(void)
{
        /* module cleanup by 'release_module' */
        if (unregister_ip_vs_scheduler(&ip_vs_lblcr_scheduler) != 0) {
                IP_VS_INFO("cannot remove LBLCR scheduling module\n");
         } else {
                IP_VS_INFO("LBLCR scheduling module unloaded.\n");
        }
        unregister_sysctl_table(lblcr_sysctl_table.sysctl_header);
}

#endif /* MODULE */
