/*
 * IPVS:        Least-Connection Scheduling module
 *
 * Version:     $Id: ip_vs_lc.c,v 1.5 2000/11/24 02:02:53 wensong Exp $
 *
 * Authors:     Wensong Zhang <wensong@iinchina.net>
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Changes:
 *     Wensong Zhang            :     added the ip_vs_lc_update_svc
 *     Wensong Zhang            :     added any dest with weight=0 is quiesced 
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <net/ip_masq.h>
#ifdef CONFIG_IP_MASQUERADE_MOD
#include <net/ip_masq_mod.h>
#endif
#include <linux/ip_fw.h>
#include <net/ip_vs.h>


static int ip_vs_lc_init_svc(struct ip_vs_service *svc)
{
        MOD_INC_USE_COUNT;
        return 0;
}


static int ip_vs_lc_done_svc(struct ip_vs_service *svc)
{
        MOD_DEC_USE_COUNT;
        return 0;
}


static int ip_vs_lc_update_svc(struct ip_vs_service *svc)
{
        return 0;
}


/*
 *    Least Connection scheduling
 */
static struct ip_vs_dest* ip_vs_lc_schedule(struct ip_vs_service *svc,
                                            struct iphdr *iph)
{
	struct list_head *l, *e;
        struct ip_vs_dest *dest, *least;
	int lac, dac;

	IP_VS_DBG(6, "ip_vs_lc_schedule(): Scheduling...\n");

	l = &svc->destinations;
        if (l == l->next)
                return NULL;

	/*
         * Simply select the server with the least number of
         *        (activeconns<<5) + inactconns
         * Except whose weight is equal to zero.
         * If the weight is equal to zero, it means that the server is
         * quiesced, the existing connections to the server still get
         * served, but no new connection is assigned to the server.
         */

        for (e=l->next; e!=l; e=e->next) {
                least = list_entry (e, struct ip_vs_dest, n_list);
                if (least->weight > 0) {
                        lac = (atomic_read(&least->activeconns) << 5)
                                + atomic_read(&least->inactconns);
                        goto nextstage;
                }
        }
        return NULL;
        
        /*
         *    Find the destination with the least load.
         */
  nextstage:
	for (e=e->next; e!=l; e=e->next) {
	        dest = list_entry(e, struct ip_vs_dest, n_list);
                if (dest->weight == 0)
                        continue;
                dac = (atomic_read(&dest->activeconns) << 5)
                        + atomic_read(&dest->inactconns);
	        if (dac < lac) {
		        least = dest;
			lac = dac;
		}
	}

        IP_VS_DBG(6, "LC: server %d.%d.%d.%d:%d activeconns %d inactconns %d\n",
                  NIPQUAD(least->addr), ntohs(least->port),
                  atomic_read(&least->activeconns),
                  atomic_read(&least->inactconns));

        return least;
}


static struct ip_vs_scheduler ip_vs_lc_scheduler = {
        {0},                    /* n_list */
        "lc",                   /* name */
        ATOMIC_INIT(0),         /* refcnt */
        ip_vs_lc_init_svc,      /* service initializer */
        ip_vs_lc_done_svc,      /* service done */
        ip_vs_lc_update_svc,    /* service updater */
        ip_vs_lc_schedule,      /* select a server from the destination list */
};


__initfunc(int ip_vs_lc_init(void))
{
        IP_VS_INFO("Initializing LC scheduling\n");
        INIT_LIST_HEAD(&ip_vs_lc_scheduler.n_list);
        return register_ip_vs_scheduler(&ip_vs_lc_scheduler) ;
}


#ifdef MODULE
EXPORT_NO_SYMBOLS;

int init_module(void)
{
        INIT_LIST_HEAD(&ip_vs_lc_scheduler.n_list);

        /* module initialization by 'request_module' */
        if(register_ip_vs_scheduler(&ip_vs_lc_scheduler) != 0)
                return -EIO;

        IP_VS_INFO("LC scheduling module loaded.\n");

        return 0;
}

void cleanup_module(void)
{
        /* module cleanup by 'release_module' */
        if(unregister_ip_vs_scheduler(&ip_vs_lc_scheduler) != 0)
                IP_VS_INFO("cannot remove LC scheduling module\n");
        else
                IP_VS_INFO("LC scheduling module unloaded.\n");
}

#endif /* MODULE */
