/*
 * IPVS:        Weighted Round-Robin Scheduling module
 *
 * Version:     $Id: ip_vs_wrr.c,v 1.9 2000/11/24 01:57:23 wensong Exp $
 *
 * Authors:     Wensong Zhang <wensong@iinchina.net>
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Changes:
 *     Wensong Zhang            :     changed the ip_vs_wrr_schedule to return dest
 *     Wensong Zhang            :     changed some comestics things for debugging
 *     Wensong Zhang            :     changed for the d-linked destination list
 *     Wensong Zhang            :     added the ip_vs_wrr_update_svc
 *     Julian Anastasov         :     return -ENOMEM instead of ENOMEM in the
 *                                    ip_vs_wrr_init_svc
 *     Julian Anastasov         :     fixed the bug of returning destination
 *                                    with weight 0 when all weights are zero
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <net/ip_masq.h>
#ifdef CONFIG_IP_MASQUERADE_MOD
#include <net/ip_masq_mod.h>
#endif
#include <linux/ip_fw.h>
#include <net/ip_vs.h>

/*
 * current destination pointer for weighted round-robin scheduling
 */
struct ip_vs_wrr_mark {
        struct list_head *cl;        /* current list head */
        int cw;                      /* current weight */
};


static int ip_vs_wrr_init_svc(struct ip_vs_service *svc)
{
	/*
         *    Allocate the mark variable for WRR scheduling
         */
        svc->sched_data = kmalloc(sizeof(struct ip_vs_wrr_mark), GFP_ATOMIC);

        if (svc->sched_data == NULL) {
                IP_VS_ERR("ip_vs_wrr_init_svc(): no memory\n");
		return -ENOMEM;
        }
        memset(svc->sched_data, 0, sizeof(struct ip_vs_wrr_mark));

        ((struct ip_vs_wrr_mark*)svc->sched_data)->cl = &svc->destinations;

        MOD_INC_USE_COUNT;
        return 0;
}


static int ip_vs_wrr_done_svc(struct ip_vs_service *svc)
{
        /*
         *    Release the mark variable
         */
        kfree_s(svc->sched_data, sizeof(struct ip_vs_wrr_mark));
        
        MOD_DEC_USE_COUNT;
        return 0;
}


static int ip_vs_wrr_update_svc(struct ip_vs_service *svc)
{
        ((struct ip_vs_wrr_mark*)svc->sched_data)->cl = &svc->destinations;
        return 0;
}


/*
 *    Get the maximum weight of the service destinations.
 */
int ip_vs_wrr_max_weight(struct ip_vs_service *svc)
{
        register struct list_head *l, *e;
        struct ip_vs_dest *dest;
        int weight = 0;

        l = &svc->destinations;
        for (e=l->next; e!=l; e=e->next) {
                dest = list_entry(e, struct ip_vs_dest, n_list);
                if (dest->weight > weight)
                        weight = dest->weight;
        }

        return weight;
}

        
/*
 *    Weighted Round-Robin Scheduling
 */
static struct ip_vs_dest* ip_vs_wrr_schedule(struct ip_vs_service *svc,
                                             struct iphdr *iph)
{
        struct ip_vs_dest *dest;
        struct ip_vs_wrr_mark *mark = svc->sched_data;

	IP_VS_DBG(6, "ip_vs_wrr_schedule(): Scheduling...\n");

        /*
         * This loop will always terminate, because 0<mark->cw<max_weight,
         * and at least one server has its weight equal to max_weight.
         */
        while (1) {
                if (mark->cl == &svc->destinations) {
                        /* it is at the head of the destination list */
                        
                        if (mark->cl == mark->cl->next)
                                /* no dest entry */
                                return NULL;

                        mark->cl = svc->destinations.next;
                        mark->cw--;
                        if (mark->cw <= 0) {
                                mark->cw = ip_vs_wrr_max_weight(svc);
                                /*
                                 * Still zero, which means no availabe servers.
                                 */
                                if (mark->cw == 0) {
                                        mark->cl = &svc->destinations;
                                        IP_VS_INFO("ip_vs_wrr_schedule(): "
                                                   "no available servers\n");
                                        return NULL;
                                }
                        }
                }
                else mark->cl = mark->cl->next;

                if (mark->cl != &svc->destinations) {
                        /* not at the head of the list */
                        dest = list_entry(mark->cl, struct ip_vs_dest, n_list);
                        if (dest->weight >= mark->cw)
                                break;
                }
        }

        IP_VS_DBG(6, "WRR: server %d.%d.%d.%d:%d "
                  "activeconns %d refcnt %d weight %d\n",
                  NIPQUAD(dest->addr), ntohs(dest->port),
                  atomic_read(&dest->activeconns),
                  atomic_read(&dest->refcnt), dest->weight);

        return  dest;
}


static struct ip_vs_scheduler ip_vs_wrr_scheduler = {
	{0},                    /* n_list */
	"wrr",                  /* name */
	ATOMIC_INIT(0),         /* refcnt */
	ip_vs_wrr_init_svc,     /* service initializer */
	ip_vs_wrr_done_svc,     /* service done */
	ip_vs_wrr_update_svc,   /* service updater */
	ip_vs_wrr_schedule,     /* select a server from the destination list */
};


__initfunc(int ip_vs_wrr_init(void))
{
	IP_VS_INFO("Initializing WRR scheduling\n");
        INIT_LIST_HEAD(&ip_vs_wrr_scheduler.n_list);
	return register_ip_vs_scheduler(&ip_vs_wrr_scheduler) ;
}

#ifdef MODULE
EXPORT_NO_SYMBOLS;

int init_module(void)
{
        INIT_LIST_HEAD(&ip_vs_wrr_scheduler.n_list);

	/* module initialization by 'request_module' */
	if(register_ip_vs_scheduler(&ip_vs_wrr_scheduler) != 0)
	        return -EIO;

	IP_VS_INFO("WRR scheduling module loaded.\n");
	
        return 0;
}

void cleanup_module(void)
{
	/* module cleanup by 'release_module' */
	if(unregister_ip_vs_scheduler(&ip_vs_wrr_scheduler) != 0)
	        IP_VS_INFO("cannot remove WRR scheduling module\n");
	else
	        IP_VS_INFO("WRR scheduling module unloaded.\n");
}

#endif /* MODULE */
